#include <iostream>

#include <casacore/casa/BasicSL/String.h>

#include <jive/ms2uvfitsimpl/Converters/FileFormatConverter.h>
#include <jive/ms2uvfitsimpl/Converters/ms2idifitsConverter.h>


#include <signal.h>
#include <libgen.h>

#include <sstream>

#ifndef BUILDTIME
#define BUILDTIME "<no buildtime info present>"
#endif

using namespace casacore;


// define built-in defaults
static const char*      default_conversion = "m2i";

void Usage( const String& prgname ) {
    cout << "Usage: " << prgname << " [option] <source> <destination>" << endl;
    cout << "       where option can be:" << endl;
    cout << "         -m2i  -> conversion from MS (source) to IDI FITSfile" << endl
         << "                  (VLBA format)" << endl;
    cout << "         -wm2i -> conversion from Westerbork MS (source)" << endl
         << "                  to IDI FITSfile (destination)" << endl;
    cout << "         -fuzz SECONDS -> Extend scans for up to SECONDS at each end" << endl;
    cout << endl
         << "Defaults are:   '" << default_conversion << "'" << endl;
    return;
}

void throwonsignal( int signalnr ) {
    ostringstream   buf;
    
    buf << endl << "**** Caught a signal #" << signalnr << "..." << endl;
    
    throw( AipsError(buf.str()) );
}



int main(int, char const*const* argv ) {
    String                infile;
    String                outfile;
    String                method( default_conversion );
    
#ifdef BUILDTIME
    char    argv0copy[ 4096 ];
    char*   thisexe;
   
    ::strncpy(argv0copy, argv[0], sizeof(argv0copy));
    if( (thisexe=::basename(argv0copy))==0 )
        thisexe = argv0copy;
    cout << "\n"
        << thisexe << ": Version of " << BUILDTIME << " begins..."
        << "\n\n" 
        << flush;
#endif
    
    //  Set the handler for various signals!
    sigset( SIGBUS, throwonsignal );
    sigset( SIGINT, throwonsignal );
    sigset( SIGKILL, throwonsignal );
    
    //  Get the name of this program
    if( !*(argv+1) ) {
        Usage( *argv );
        return 0;
    }

    while( *(++argv) ) {
        String curarg( *argv );
        
        if( curarg[0]!='-' ) {
            //  Must be a filename?
            if( !(infile.length()) )
                infile = curarg;
            else if( !(outfile.length()) )
                outfile = curarg;
            else {
                cout << "Only two file names are allowd to be entered" << endl;
                return -1;
            }
        } else {
            //  Check if valid option:
            String const    opt( curarg.after(0) );
            
            if( opt=="m2i" || opt=="wm2i" )
                method = opt;
            else if( opt=="fuzz" ) {
                if( *(++argv) )
                    MSConversionInfo::setFuzz( ::atof(*argv) );
                else {
                    cout << "Missing fuzz value after -fuzz option" << endl;
                    return -1;
                }
            } else {
                cout << "Unrecognized option: " << opt << "!" << endl;
                return -1;
            }
        }
    }
    
    if( !infile.length() ) {
        cout << "You specified no inputfile" << endl;
        return -1;
    }
    
    if( !outfile.length() ) {
        cout << "No outputfile given" << endl;
        return -1;
    }
    
    // convert MS to IDI fits
    generatorlist_t   hdustodo;

    hdustodo.push_back( "idihdr" );
    hdustodo.push_back( "arraygeometry" );
    hdustodo.push_back( "idifrequency" );
    hdustodo.push_back( "idisource" );
    hdustodo.push_back( "idiantenna" );

    hdustodo.push_back( "idiFlag" );
    hdustodo.push_back( "idiGainCurve" );
    hdustodo.push_back( "idiSysTemp" );

    // for westerbork, add the following
    if( method=="wm2i" ) {
        hdustodo.push_back( "idity" ); 
        hdustodo.push_back( "idigainc" ); 
    }

    hdustodo.push_back( "uvtable" );//MS2UVF::uvtable

    try {
        ms2idifitsConverter converter( hdustodo );

        //  Attempt the transformation....
        if( !converter.convert(outfile, infile) ) {
            cout << "Rats. Tranformation not succeeded" << endl;
            return -1;
        }
    }
    catch( std::exception const& e ) {
        cout << "Caught exception: " << e.what() << endl;
        return -1;
    }
    catch( ... ) {
        cout << "Caught unknown exception" << endl;
        return -1;
    }
    return 0;
}
