#ifndef STRINGUTIL_H
#define STRINGUTIL_H

#include <vector>
#include <string>
#include <sstream>
#include <casacore/casa/BasicSL/String.h>

//
// Split a string into substrings at char 'c'
// Note: the result may contain empty strings, eg if more than one separation
// characters follow each other.
//
// E.g.
//    res = ::split( "aa,bb,,dd,", ',' )
//
// would yield the following array:
//   res[0] = "aa"
//   res[1] = "bb"
//   res[2] = ""
//   res[3] = "dd"
//   res[4] = ""
//
//
std::vector<std::string> split( const std::string& str, char c );


// Simple 'any printable value' to string
template <typename T>
std::string repr(const T& t) {
    std::ostringstream oss;
    oss << t;
    return oss.str();
}

template <typename T>
casacore::String Repr(T const& t) {
    return casacore::String(repr(t));
}

// Display a container on a stream
// the container must support .begin(), .end() and ::const_iterator
// use as
//     std::ostream& << print_container(<container instance>);
template <typename T>
struct print_container_impl {
    print_container_impl(const T& container):
        container_ref( container )
    {}

    const T&  container_ref;
};

template <typename T>
std::ostream& operator<<(std::ostream& os, const print_container_impl<T>& pt) {
    os << '[';
    for(typename T::const_iterator ptr=pt.container_ref.begin(); ptr!=pt.container_ref.end(); ptr++)
        os << *ptr << ',';
    return os << ']';
}

template <typename T>
print_container_impl<T> print_container(const T& ct) {
    return print_container_impl<T>(ct);
}

struct lowercase_cmp {
    template <typename T>
    bool operator()(T const& l, T const& r) const {
        return ::tolower(l) == ::tolower(r);
    }
};

template <typename T>
bool case_insensitive_compare(T const& l, T const& r) {
    if( l.size() != r.size() )
        return false;
    return std::equal(l.begin(), l.end(), r.begin(), lowercase_cmp());
}

#endif
