//  Experiment - Abstract base-class, specifying the kind information
//  a user can get from an experiment
//
//   Author:  Harro Verkouter, 8-12-1998
//
//   $Id: Experiment.h,v 1.10.2.2 2018/08/21 12:26:29 verkout Exp $
//
#ifndef EXPERIMENT_H
#define EXPERIMENT_H

#include <casacore/casa/aips.h>
#include <casacore/casa/BasicSL/String.h>
#include <casacore/casa/Arrays/Vector.h>
#include <casacore/casa/Logging.h>

#include <jive/j2ms/Visibility.h>
#include <jive/j2ms/SourceDB.h>
#include <jive/j2ms/AntennaDB.h>
#include <jive/j2ms/FrequencyDB.h>
#include <jive/j2ms/Scan.h>
#include <jive/j2ms/scanlist.h>
#include <jive/j2ms/option_map.h>
#include <jive/labels/polarization.h>
#include <jive/labels/CorrelationCode.h>

#include <string>
#include <iostream>


//  The class itself
class Experiment
{
public:
    //  Create an Experiment. If the default is used, the Current
    //  Working directory is used as exp.root dir.
    Experiment( const casacore::String& opt="", const casacore::String& exprootdir="" );

	
	// For a specific option, what was the value?
	// If the option was not set, return empty string
	casacore::String                optionValue( const casacore::String& opt ) const;
    casacore::String const&         options( void ) const;
    // HV: Apr 2018 - In order to support (re)setting options whilst running
    //                some machinery has to be implemented.
    //                If someone wants to reset options, the non-virtual
    //                resetOptions() needs to be called.
    //                The base class will analyze the new options string
    //                and yank the pure virtual method if something's
    //                changed.
    //                Derived(concrete) classes may respond to this as they
    //                see fit
    void                            resetOptions( casacore::String const& opts );
    virtual void                    optionsReset( void ) = 0;

    //  Provide access to the properties. Note: these methods must be
    //  implemented by a derived class!
    virtual const casacore::String&          getCode( void ) const = 0;

    //  Hrmpff... design error (c) Harro V....
    //
    //  These functions should be located somewhere else but here;
    //  they control the output the JIVE MS-filler produces. May need
    //  a revision sometime... The MS-Filler should ask the visibility
    //  buffer to deliver the data in a certain domain *and* the
    //  MS-Filler should be capable of deciding on it's own where to
    //  store the data...
    virtual const casacore::String&    getMSname( void ) const = 0;
    virtual void                   setMSname( const casacore::String& newmsname ) = 0;

    virtual Visibility::domain     getDestinationDomain( void ) const = 0;
    virtual casacore::Bool             setDestinationDomain( Visibility::domain newoutputdomain ) = 0;


    //  Ok... these belong here!
    virtual const casacore::Vector<casacore::String>&  getAntennas( void ) const = 0;
    virtual const casacore::Vector<casacore::String>&  getSources( void ) const = 0;

    // The experiment MUST have its own frequency DB!
    virtual const FrequencyDB&     getFrequencyDB( void ) const = 0;

    // The experiment MUST have an antennadatabase.
    virtual const AntennaDB&       getAntennaDB( void ) const = 0;

    //  Also, the experiment MUST have a source DB
    virtual const SourceDB&        getSourceDB( void ) const = 0;

    //  VisibilityBuffers's are a part of the experiment, therefore we
    //  should ask the Experiment if it would be so kind as to create
    //  the indicated VisibilityBuffer. If it fails to create it, a
    //  null-pointer is returned, otherwise a pointer to the newly
    //  created Buffer is returned!
    //
    //  NOTE: the caller is responsible for deleting the
    //  resources!!!!!!
    virtual VisibilityBuffer*      getVisibilityBuffer( const casacore::String& visbufname ) = 0;


    // Experiment can give us a list of scans. VisibilityBuffers can do that too.
    // Default VisibilityBuffer 'getScanlist()' implementation will return
    // the experiment's getScanlist().
    // Specific VisBuffer objects may decide to return their own list (eg 
    // the COFVisibilityBuffer does that).
    // This one *may* be overridden. Default implementation gives an empty list
    virtual const j2ms::scaninfolist_t& getScanlist( void ) const;

    // Return a stationbased scaninfolist
    // (do it by const ref such that we dont
    //  have to copy the data over and over
    //  again)
    //  The default implementation is to
    //  return an empty list.... so only classes
    //  that really can fill it in have to implement
    //  their own vsn.
    virtual const j2ms::stationscanlist_t&
                                getStationScanlist( int station ) const;

    //  Tells us whether or not the Vernier delay was applied.
    //  Also, this shouldn't live here i guess
    virtual casacore::Bool             hasVernierDelay( void ) const = 0;

    // The array-code (eg. EVN, GLOBAL, VLBA).
    //
    //  Note: derived classes *may* override this one, but don't have
    //  to. The return value defaults to EVN....(since we're at JIVE
    //  here...)
    virtual const casacore::String&          getArrayCode( void ) const;

    //  These methods are NOT virtual!
    const casacore::String&           getRootDir( void ) const;

    //  Log a message (if logging is enabled); usr. need not worry
    //  whether or not logging is enabled, function will verify all
    //  that's necessary)
    void                    logLogMessage( const casacore::String& msg,
										   const casacore::LogOrigin& org=casacore::LogOrigin() ) const;

    //  Delete the resources used by this experiment
    virtual ~Experiment();

private:
    //  The experiment's private parts
	casacore::String     myOptions;
    casacore::String     myRootDir;
    std::ostream**       myLogStreamptr;

	// Map of string-to-string; the 'myOptions' split up;
	// <key>=<value> will be translated to map[<key>] = <value>
	optionmap_type       myOptionValues;

    //  Prohibit these
    Experiment( const Experiment& );
    const Experiment& operator=( const Experiment& );
};

//  Print some diagnostics on a stream
std::ostream& operator<<( std::ostream& os, const Experiment& expref );


#endif
