// implementation
//
// $Id: GhostBusterFilter.cc,v 1.3.2.1 2018/08/21 12:26:29 verkout Exp $
//
// $Log: GhostBusterFilter.cc,v $
// Revision 1.3.2.1  2018/08/21 12:26:29  verkout
// HV: * Added 'option_map' implementation so it can be reused to
//       support "key=value,key2=value2..." passing of settings
//     * On multi-phase-center (700-ish) j2ms2 become /SUPER/ slow:
//       the ghostbuster-filter (enabled by default) was effectively
//       disabled (on purpose) by SFXCVisBuffer implementation but did
//       that in the most time-consuming way ...
//     * GhostBusterFilter changed:
//         - now takes options (see j2ms2 help) [can honour station based
//           data_good from VEX file if desired]
//         - responsibility for implementing the filter moved out of
//           *VisBuffer classes to better suited logical level
//     * More descriptive SFXCVisBuffer error messages
//     * UNBVisBuffer: overlooked the fact that subbands now have to be
//       sorted by frequency AND sideband in stead of just freq.
//
// Revision 1.3  2016/06/23 13:02:25  jive_cc
// HV: * moved codebase to casacore 2.x compatibility - the namespace is now
//       casacore::  in stead of casa::. A lot of improvement ......
//     * now link against vex parser which understands $BITSTREAMS and $THREADS
//
// Revision 1.2  2006-03-02 14:21:39  verkout
// HV: * Removed remaining instances of strstream, replaced with strstream * Got rid of (home grown) Container class!!!! YES! Finally replaced with STLs vector/list, whichever is appropriate * Added pure virtual method to hduGenerator - class writers MUST return an ID which (uniquely) identifies their specific Generator * mapping of generator-ID to actual Generator is now built automagically rather than by ms2uvfitsConverter, class writers should look in Registrar.h for details
//
// Revision 1.1  2006/02/14 13:30:49  verkout
// HV: Implemented a filter which removes ghostdata based on individual scan-lists (scanlist per station). If any of the participating stations in a baseline has no scan defined the visibility is discarded
//
//
#include <jive/j2ms/GhostBusterFilter.h>
#include <jive/j2ms/scanlist.h>
#include <jive/j2ms/VisibilityBuffer.h>
#include <jive/j2ms/Experiment.h>
#include <jive/Utilities/stringutil.h>
#include <jive/Utilities/jexcept.h>

#include <casacore/casa/Quanta.h>

#include <algorithm>
#include <iostream>

using std::cerr;
using std::endl;

using namespace j2ms;
using namespace casacore;


// Use stl's algo stuff to facilitate finding stuff
// in containers...

struct _time_finder {

    // note: even though this *is* a struct I'll make 
    // an explicit public section to indicate that this
    // thing also has a *private* section...
    public:
        // look for a given time
        _time_finder(MVEpoch const& t, bool ds) :
            __m_vistime( t.get() ), __m_datastart(ds)
        {}
    
        // check if the searched time is whithin the range
        // [ssi.scanStart , ssi.scanEnd)
        bool operator()( const stationscaninfo_t& ssi ) const {
            static MVEpoch const zeroSeconds(0.);
            return (__m_vistime>=(ssi.scanStart.get() + (__m_datastart ? ssi.dataStart : zeroSeconds).get()) && __m_vistime<ssi.scanEnd.get());
        }

    private:
        double const __m_vistime;
        bool const   __m_datastart;

        // *no* default c'tor
        _time_finder();
};

// a nifty 'define' to make things readable...
// find a scan in container 'a' by time 'b'
#define find_scan_by_time(a,b,c) \
    std::find_if<>(a.begin(), a.end(), _time_finder(b,c))

casacore::String throwOnValue(casacore::String v) {
    THROW_JEXCEPT("GhostBusterFilter option(s) do not support setting a value like '" << v << "'");
}


GhostBusterFilter::GhostBusterFilter() :
    Filter( "stationscan" )
{}

GhostBusterFilter::GhostBusterFilter( casacore::String const& option ) :
    Filter( "stationscan" ), __m_options( ::mk_optionmap(option, ',', lower_case, throwOnValue) )
{
    static casacore::String const recognized_options[] = {"relaxed", "data_good"};
    static size_t const           nOptions(sizeof(recognized_options)/sizeof(recognized_options[0]));

    // Check that all options are recognized
    for(optionmap_type::const_iterator p = __m_options.begin(); p!=__m_options.end(); p++)
        if( std::find(recognized_options, recognized_options+nOptions, p->first)==(recognized_options+nOptions) )
            THROW_JEXCEPT("GhostBusterFilter/unsupported option '" << p->first << "'");
    // if not strict treat empty station based list as "pass"(True) otherwise as "drop"(False)
    __m_strict   = (__m_options.find(casacore::String("relaxed"))==__m_options.end());

    __m_datagood = (__m_options.find(casacore::String("data_good"))!=__m_options.end());
}

//
// Extract the X- and Y- station of the baseline and the
// time of the visibility
// and verify that *both* have a valid scan definition
// for that time
casacore::Bool GhostBusterFilter::letPass( const Visibility& vis ) {
    // if the vis. has NO visibilitybuffer, we cannot continue
    // since we can't get the station-based scanlist (that should
    // be available from the visbuffer). In this case we will
    // NOT let the visibility pass!
    if( !vis.hasVisibilityBuffer() )
        return False;

    // ok get sum refs
    MEpoch const&       vtime( vis.getTime() );
    Experiment const&   expref( vis.getVisibilityBuffer().getExperiment() );
    BaseLineIdx const&  bl( vis.getLabel().baseline );

    if( !bl.is_sane() ) {
        cerr << "GhostBusterFilter: Visibility baselineidx is NOT sane?\n"
             << "  " << vis.getLabel() << endl;
        return False;
    }

    // Now extract the X/Y antennas and the MVEpoch
    MVEpoch         t;
    unsigned int    x,y;

    x = bl[0];
    y = bl[1];
    t = vtime.getValue();

    // now (try) to find scan(s) in the stationbased scanlist
    const stationscanlist_t&          xlist( expref.getStationScanlist((int)x) );
    stationscanlist_t::const_iterator xscan;

    // How to deal with empty station scan list depends upon strictness
    // not strict => let visibility pass (strict == False => return True)
    // strict     => drop visibility     (strict == True  => return False)
    if( xlist.empty() )
        return !__m_strict;

    if( (xscan = find_scan_by_time(xlist, t, __m_datagood))==xlist.end() )
        return False;

    // only *really* look for yscan (and test) if not X and Y equal
    if( y==x )
        return True;

    const stationscanlist_t&          ylist( expref.getStationScanlist((int)y) );

    if( ylist.empty() )
        return !__m_strict;

    return find_scan_by_time(ylist, t, __m_datagood)!=ylist.end();
}

GhostBusterFilter::~GhostBusterFilter() {}
