//
//  Implementation of the JIVEMSFiller-class
//
//  Author:  Harro Verkouter,   10-12-1998
//
//  $Id: JIVEMSFiller.cc,v 1.34.2.2 2018/08/21 12:26:29 verkout Exp $
//
#include <jive/j2ms/JIVEMSFiller.h>
#include <jive/j2ms/CorrelatorSetup.h>
#include <jive/j2ms/History.h>
#include <jive/j2ms/Scan.h>
#include <jive/j2ms/FrequencyDB.h>
#include <jive/j2ms/AntennaDB.h>
#include <jive/j2ms/SourceDB.h>
#include <jive/j2ms/noUVW.h>
#include <jive/j2ms/aipspppol2j2mspol.h>
#include <jive/Utilities/FreqCounter.h>
#include <jive/Utilities/jexcept.h>
#include <jive/Utilities/stringutil.h>
#include <jive/Utilities/sciprint.h>
#include <jive/labels/CorrelationCode.h>
#include <jive/labels/VisLabel.h>

#include <jive/assertfail.h>



// Filters recognized by us
#include <jive/j2ms/SourceFilter.h>
#include <jive/j2ms/GhostBusterFilter.h>
#include <jive/j2ms/AntennaInfo.h>

#ifdef MARK4
// COF stuff
#include <data_handler/MediumStatsFile.h>
#include <data_handler/MappingsFile.h>
#include <data_handler/MediumStatDataRecord.h>
#include <data_handler/CountedPointer.h>
#include <data_handler/MappingEntry.h>
#include <data_handler/ModelFile.h>
#include <data_handler/ModelDataRecord.h>
#endif

// CASA stuff
#include <casacore/ms/MeasurementSets.h>
#include <casacore/ms/MeasurementSets/MSPolarization.h>
#include <casacore/ms/MeasurementSets/MSSpectralWindow.h>
#include <casacore/ms/MeasurementSets/MSDataDescription.h>
#include <casacore/ms/MeasurementSets/MSPolColumns.h>
#include <casacore/ms/MeasurementSets/MSSpWindowColumns.h>
#include <casacore/ms/MeasurementSets/MSDataDescColumns.h>
#include <casacore/ms/MeasurementSets/MSAntenna.h>
#include <casacore/ms/MeasurementSets/MSAntennaColumns.h>
#include <casacore/ms/MeasurementSets/MSFeed.h>
#include <casacore/ms/MeasurementSets/MSFeedColumns.h>
#include <casacore/ms/MeasurementSets/MSField.h>
#include <casacore/ms/MeasurementSets/MSFieldColumns.h>
#include <casacore/ms/MeasurementSets/MSSource.h>
#include <casacore/ms/MeasurementSets/MSSourceColumns.h>
#include <casacore/ms/MeasurementSets/MSProcessor.h>
#include <casacore/ms/MeasurementSets/MSProcessorColumns.h>
#include <casacore/ms/MeasurementSets/MSObservation.h>
#include <casacore/ms/MeasurementSets/MSObsColumns.h>
#include <casacore/ms/MeasurementSets/MSState.h>
#include <casacore/ms/MeasurementSets/MSStateColumns.h>

#include <casacore/casa/Quanta.h>
#include <casacore/casa/Quanta/MVAngle.h>
#include <casacore/casa/Quanta/Quantum.h>
#include <casacore/casa/Quanta/Unit.h>
#include <casacore/casa/Quanta/Euler.h>
#include <casacore/casa/Quanta/QMath.h>
#include <casacore/casa/Quanta/RotMatrix.h>
#include <casacore/casa/Quanta/MVTime.h>
#include <casacore/casa/Quanta/MVEpoch.h>

#include <casacore/casa/Arrays/ArrayMath.h>
#include <casacore/casa/Arrays/ArrayLogical.h>
#include <casacore/casa/Arrays/ArrayUtil.h>

#include <casacore/casa/Utilities/Regex.h>

#include <casacore/casa/Exceptions.h>
#include <casacore/casa/OS/Time.h>

#include <casacore/measures/Measures/MeasMath.h>
#include <casacore/measures/Measures/Stokes.h>
#include <casacore/measures/Measures/MPosition.h>
#include <casacore/measures/Measures/Muvw.h>

#include <casacore/scimath/Mathematics/FFTServer.h>

#include <casacore/tables/Tables.h>
#include <casacore/tables/TaQL/ExprNode.h>
#include <casacore/tables/Tables/TableColumn.h>
#include <casacore/tables/Tables/TableDesc.h>
#include <casacore/tables/Tables/TableIter.h>



// system headers

#include <iomanip>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <dirent.h>
#include <libgen.h>
#include <time.h>
#include <limits.h>

// C++ and STL stuff
#include <set>
#include <sstream>
#include <algorithm>
#include <vector>
#include <cassert>
#include <iomanip>

using namespace std;
using namespace casacore;


// Static stuff
ExperimentCache      JIVEMSFiller::theirCache;



// If the linker wont recognize our template, let's stuff'm in here!
// (and see if that works)
const char* translate2ascii( const String& arg ) {
	return arg.chars();
}


//  Define RotMatrix*Vector (how to do it...)
Vector<Double> operator*( const RotMatrix& lhside, const Vector<Double>& rhside ) {
    Vector<Double>   retval(3);
    
    retval = 0.0;
    
    for( int i=0; i<3; i++ ) {
		for( int j=0; j<3; j++ ) {
			retval(i) += lhside( i, j )*rhside(j);
		}
    }
    return retval;
}




JIVEMSFiller::JIVEMSFiller( const String& opt ):
	myOptions( opt )
{
	//	cout << "Created JIVEMSFiller with opt=" << myOptions << endl;
	//
	// Parse the options (if any)
	
	// Let's remove all whitespace stuff!
	string           optcopy;
	string::iterator curchar = myOptions.begin();

	while( curchar<myOptions.end() ) {
		if( !::strchr(" \t\n\f", *curchar) ) {
			optcopy.append( 1, *curchar );
		}
		curchar++;
	}

	// Step one: split the options at semicolons ';' -> they're the delimiters
	// between separate options.
	Regex                    rx_filterwopt( "^filter/([^=]+)=(.+)$" );
	Regex                    rx_filter( "^filter/(.+)$" );
	vector<string>           opts;
	vector<string>::iterator curopt;

	opts   = ::split( optcopy, ';' );
	curopt = opts.begin();
	while( curopt!=opts.end() ) {
        String  tmp( *curopt );
        string  fspec, fopt;

        if( tmp.matches(rx_filterwopt) ) {
            Int    s, l;
            // filterspec is the first matchgroup
            rx_filterwopt.match_info(s, l, 1);
            fspec = tmp(s, l);
            // and the actual options are the second matchgroup
            rx_filterwopt.match_info(s, l, 2);
            fopt  = tmp(s, l);
        } else if( tmp.matches(rx_filter) ) {
            Int    s, l;
            // filterspec is the first matchgroup
            rx_filter.match_info(s, l, 1);
            fspec = tmp(s, l);
        } else if( curopt->size() ) {
			cerr << "JIVEMSFiller: unsupported option `" << *curopt << "'" << endl;
		}

        if( !fspec.empty() ) {
            Filter*  fptr( 0 );

			// See if we can recognize the filter....
			if( fspec=="source" ) {
				fptr = new SourceFilter( fopt );
                cout << "JIVEMSFiller: Adding SourceFilter [" << fopt << "]" << endl;
			} else if ( fspec=="ghostbuster" ) {
				fptr = new GhostBusterFilter( fopt );
                cout << "JIVEMSFiller: Adding GhostBusterFilter [" << fopt << "]" << endl;
            } else {
				cout << "JIVEMSFiller: unrecognized filter `" << fspec << "'"
					 << endl;
			}
			if( fptr )
                myFilters.push_back(fptr);
		}

		// Do *NOT* forget to move on! (Guess (again) how I found this one out...)
		curopt++;
	}
}

//  Do the job of actually making a new MS or concatenating it to an
//  existing one for the indicated scan in the given experiment....
Bool JIVEMSFiller::fillDataToExperimentsMS( Experiment& expref,
											const subjob_t& sjdescr,
											Bool ask )
{
    Bool              retval( False );
    VisibilityBuffer* ptr2buffer( 0 );
    
    // The very first check: if the indicated visibility buffer is
    // present in the experiment
    ptr2buffer = expref.getVisibilityBuffer( sjdescr.jobName );
    
    if( !ptr2buffer ) {
        cerr << "\n"
             << "Something went wrong opening '" << sjdescr.jobName << "'. Please check any error messages above.\n\n";

		return retval;
    }

    //  Ok. Locate an MS where to stuff the new data....
    MS*             ptr2ms( this->makeMS(expref, ask, *ptr2buffer) );

    // If we see a different MS name, clear caches
    static String   prevMS;
    if( expref.getMSname()!=prevMS )
        this->rebuild_caches( *ptr2ms );
    prevMS = expref.getMSname();
    
    if( !ptr2ms ) {
	    cout << "Failed to create output MS for the experiment!" << endl;
    	delete ptr2buffer;
    	return retval;
    }
    
    try {
        // TODO: deal with potential new filler options in the subjob
        //       description!
        retval = this->fillMSfromVisibilityBuffer( *ptr2ms,
												   *ptr2buffer,
												   sjdescr );
    }
    catch( const std::exception& se ) {
        cout << "standard exception: " << se.what() << endl;
    }
    catch( const char* s ) {
        cout << "Caught a string: " << s << endl;
    }
    catch( ... ) {
        cout << "Caught unknown exception?" << endl;
    }

    //  We don't need the pointer to the MS anymore....
    delete ptr2ms;

    //  Nor do we need the pointer to the scan anymore
    delete ptr2buffer;
    
    //  Tell'm what happened!
    return retval;
}



JIVEMSFiller::~JIVEMSFiller()
{}


//  The private methods!!!
MS* JIVEMSFiller::makeMS( const Experiment& expref, Bool ask, const VisibilityBuffer& visbuf ) {
    //  First we follow the method of TMS/NFRA, see MSDefiner{.cc,.h}
    MS*                   retval( 0 );
    Table*                proctable( 0 );
    String                key;
    TableDesc             td( MS::requiredTableDesc() );
// Disable writing of the MODEL until such time that we either need it
// (again) or decide to do it rite
#if 0
    Table*                modeltable( 0 );
    TableDesc             modeldesc( "JIVE_MODEL", TableDesc::Scratch );
    SetupNewTable*        modeltablesetup( 0 );
#endif
    TableDesc             procdesc( "JIVE_PROCESSING", TableDesc::Scratch );
    String const          MSname( expref.getMSname() );
    SetupNewTable*        tablesetupptr( 0 );
    SetupNewTable*        proctablesetup( 0 );
	MS::PredefinedColumns datacolEnum = MS::DATA;

#if 0
    // And the defintion of the JIVE_MODEL for 
    // storing the model used whilst correlating
    modeldesc.addColumn( (ScalarColumnDesc<Int>("ANTENNA_ID", "Antenna ID")) );
    modeldesc.addColumn( (ScalarColumnDesc<Int>("FEED_ID", "Feed ID")) );
    modeldesc.addColumn( (ScalarColumnDesc<Int>("SPECTRAL_WINDOW_ID", "Spectral window ID")) );
    modeldesc.addColumn( (ScalarColumnDesc<Double>("TIME", "Starttime of interval for which polys are valid")) );
    modeldesc.addColumn( (ScalarColumnDesc<Double>("INTERVAL", "Duration of interval")) );
    modeldesc.addColumn( (ArrayColumnDesc<Double>("DRY_ATMOS_DELAY", "'dry' atmospheric delay")) );
    modeldesc.addColumn( (ArrayColumnDesc<Double>("WET_ATMOS_DELAY", "'wet' atmospheric delay")) );
    modeldesc.addColumn( (ArrayColumnDesc<Double>("CLOCK_OFFSET", "Clock offset")) );
    modeldesc.addColumn( (ArrayColumnDesc<Double>("GROUP_DELAY", "Group delay polynomials")) );
    modeldesc.addColumn( (ArrayColumnDesc<Double>("PHASE_DELAY", "Phase delay polynomials")) );
	modeldesc.addColumn( (ScalarColumnDesc<Int>("PROCESSOR_ID", "Job where this data applies to")) );
#endif

    // Table to hold processing steps that were applied to the data in this MS
    // processing steps == non-standard/optional JIVE-specific processing that
    // happened to the data.
    procdesc.addColumn( ScalarColumnDesc<Int>("JOBID", "JIVE specific JobId") );
    procdesc.addColumn( ScalarColumnDesc<Double>("TIME", "MJD when this task was run") );
    procdesc.addColumn( ScalarColumnDesc<String>("PROCESSING", "The processing step that was run") );
    procdesc.addColumn( ScalarColumnDesc<String>("PROCESSING_PARAMS", "(optional) parameter(s) of the task") );

    // Ok, since this is the JIVE-filler maybe it's safe to assume
    // that we ALWAYS have complex data?! And we're filling
    // Visibilities, who always have a matrix-shaped block of data,
    // therefore we add a complex array column of dimensionality 2.
    //
    //  HV: 19-07-2000  With the arrival of MSv2.0 time/frequency data
    //                  has to be stored in different columns....
    String    datacolumnname = "";
    
    switch( expref.getDestinationDomain() )
    {
		case Visibility::time:
		    datacolumnname = "LAG_DATA";
			datacolEnum = MS::LAG_DATA;
		    break;
	    
		case Visibility::frequency:
			datacolumnname = "DATA";
			datacolEnum = MS::DATA;
			break;
	    
		default:
			//  Yikes! Cannot do anything with this!
			//
			//  We just let the datacolumnname be empty so we can test
			//  on that after the switch statement and throw up after
			//  that!
			break;	    
    }
	
    if( datacolumnname=="" ) {
		cout << "Yikes! Experiment's destination datadomain is unknown!\n"
	    	 << "Hence we cannot decode on in what column to store the data...\n";	
		return 0;
    }
    
	// Add our datacolumn to the table!
	MS::addColumnToDesc( td, datacolEnum, 2 );

    //  Ok table descriptor set up. Now we first try to create the
    //  table as NewNoReplace, which will throw an exception if the
    //  table already exists. So if we catch an exception, we prompt the user
    try
    {
		//  Ok. We want to use the TiledShapeStorage manager.  This
		//  requires us to define a hypercube (yummie..), so let's do
		//  that now
		td.defineHypercolumn( String("JIVEData"), (uInt)3,
							 ::stringToVector(datacolumnname+",FLAG") );

		tablesetupptr  = new SetupNewTable( MSname, td, Table::NewNoReplace );	
#if 0
        modeltablesetup = new SetupNewTable( MSname+"/"+"JIVE_MODEL",
                                             modeldesc, Table::NewNoReplace );
#endif
        proctablesetup = new SetupNewTable( MSname+"/"+"JIVE_PROCESSING",
                                             procdesc, Table::NewNoReplace );

		//  If we get to here, we are creating the table!
		//  Continue on, make the storage manager!
		//
		//  The TiledShapeStorageManager will be acting on the
		//  (LAG_)DATA and FLAG columns. We specify a default shape of
		//  (4,64).. (during filling we will reset the shape or we do
		//  not and rely on the system do that for us?)
		//TiledShapeStMan  tsm( "JIVEData", IPosition(3,1,4,64) );
        //  HV: 16 feb 2015 (!) Only now we find out that the axes in our
        //                      tiles are the wrong way round! We wanted to
        //                      group 4 polarizations of ~64 channels together 
        //                      but we grouped one polarization and 4
        //                      channels for 64 rows together.
        //                      Our indices were wrong because the underlying storage is in
        //                      Fortran order, wooopsie!
        uInt             nPol, nFreq, nFreqPerTile, nRow;
        CorrelatorSetup  cs( visbuf.correlatorSetup() );

        // Advanced heuristics follow below ... Pfff.
        //
        // Always group two polarizations together, unless we seem to be
        // storing single pol data. (Typically we have 1, 2 or 4 pol
        // combinations.)
        nPol  = cs.getCorrelations().size();
        nPol  = (nPol>1) ? 2 : 1;

        nFreq = cs.nrOfSpectralPoints();
        if( expref.getDestinationDomain()==Visibility::time )
            nFreq *= 2;

        nRow         = 1;
        nFreqPerTile = 32;

        // Now start working out a tile size of 8-16kB minimum

        while( (nPol * nFreqPerTile * nRow * 8) < 16384 ) {
            // Tile not large enough. Step one: start incrementing
            // the number of spectral channels per tile - but we don't want
            // less than 8 chunks per spectrum, unless the spectrum is very
            // small
            if( nFreqPerTile<nFreq && (nFreq/nFreqPerTile)>8 ) {
                nFreqPerTile *= 2;
                continue;
            }
            // Ok, now we have as many freqs together as we want. Add more rows.
            nRow *= 2;
        }

        // Now we can create the shaped storage manager
		TiledShapeStMan  tsm( "JIVEData", IPosition(3, nPol, nFreqPerTile, nRow) );
	
		//  Let's bind the data/flag columns to the correct storage
		//  manager!
		tablesetupptr->bindColumn( datacolumnname, tsm );
		tablesetupptr->bindColumn( MS::columnName(MS::FLAG), tsm );

		//  Ok! Create the MS and JIVE specific tables
		retval     = new MeasurementSet( *tablesetupptr );
#if 0
        modeltable = new Table( *modeltablesetup );
#endif
        proctable  = new Table( *proctablesetup );
	
		//  Create default tables
		retval->createDefaultSubtables( Table::New );

		//  Attach our own table!
#if 0
		retval->rwKeywordSet().defineTable( "JIVE_MODEL", *modeltable );
#endif
		retval->rwKeywordSet().defineTable( "JIVE_PROCESSING", *proctable );
		retval->initRefs();

		//  Add a column to the DATADESCRIPTION subtable if we're
		//  writing LAG_DATA..?
		if( expref.getDestinationDomain()==Visibility::time )
		    retval->dataDescription().addColumn( ScalarColumnDesc<Int>("LAG_ID") );

		// Let's add a pass Id column to the processor-table
		retval->processor().addColumn( ScalarColumnDesc<Int>("PASS_ID") );

		//  And we're done!
    }
    catch( const std::exception& x ) {
		//  This call is to make sure that a) the compiler does not
		//  complain, and b) to make sure the compiler does not
		//  optimize it out?!
		cout << "*** " << x.what() << endl;
	
		//  Make really sure that tablesetupptr = 0, I don't trust it
		//  to leave the value intact in case of an exception...
		tablesetupptr = 0;
    }
    
    //  If tablesetupptr = 0, the table already existed. Ask the user
    //  for confirmation to concatenate
    if( !tablesetupptr ) {
		//  Table existed!

		//  HV: 21-07-2000  With the arrival of MS v 2.0 we have to be a little
		//                  more careful here... before we go on and ask the usr
		//                  if he/she wishes to concatenate the data we will have
		//                  to varify that we *can* concatenate (since the data
		//                  can be stored in an other column as the one we need
		//                  to be writing to). Depending on the desired output
		//                  domain, we check if the associated datacolumn indeed
		//                  is in the table. If it is not, we do not ask for
		//                  confirmation but return an error...
		Bool    precondition1;
		Bool    precondition2;
		Bool    precondition3;
	
	
		retval = new MeasurementSet( MSname );

		precondition1 = retval->tableDesc().isColumn(datacolumnname);
		precondition2 = true;
	 	precondition3 = precondition1 &&
				( (expref.getDestinationDomain()==Visibility::time && 
			       retval->dataDescription().tableDesc().isColumn("LAG_ID")) ||
				  (expref.getDestinationDomain()==Visibility::frequency) ); 

		//  At this point we can safely delete the MS pointed to by retval
		//  since if we are going to use it we will have to open
		//  the MS with Table::Update anyway....
		delete retval;
		retval = 0;

		if( precondition1 && precondition2 && precondition3 ) {
		    //  Ok.. datacolumn exists and is writable...
		    //  In that case it makes sense to go on and ask for confirmation!
		    Bool       concatenate( (ask==False) );
	
		    if( ask ) {
				String     ans;
		
				cin.tie( &cout );
		
				cout << "MS `" << MSname << "' already exists. Concatenate data? [yn] :";
				cin >> ans;
		
				cin.tie( 0 );
		
				if( ans=="y" || ans=="Y" )
				    concatenate = True;
	    	}
	
		    if( concatenate ) {
				// Ok... do concatenation!
				retval = new MeasurementSet( MSname, Table::Update );
		    }
		}
		else
		{
		    cout << "**** Ok.. something's clearly wrong here...\n";
		    cout << "**** trying to concatenate data to MS '" << MSname << '\n';
	    
		    if( !precondition1 )
				cout << "> MS already contains data from other datadomain!\n";
		    if( !precondition2 )
				cout << "> Datacolumn '" << datacolumnname << "' is not writable?!\n";
		    if( !precondition3 )
				cout << "> MS is internally inconsistent. Appears to contain LAG_DATA\n"
			 		<< "  but has no LAG_ID in DATA_DESCRIPTION subtable!\n";
	    
		    cout << flush;
		}
    }

    //  Clean up...
    delete tablesetupptr;
#if 0
    delete modeltablesetup;
    delete modeltable;
#endif
    delete proctablesetup;
    delete proctable;
   
    return retval;
}

typedef std::map<unsigned int, Int>                   sb2ddid_type;
typedef std::map<unsigned int, antennafeed_pair_type> ant2ant_type;


// Loop over all defined correlations in the correlator setup, forming the
// swapped correlation codes and finding the indices in the original setup
// suppose:
//    original order: LL, RR, LR, RL
//                    0   1   2   3
//  then for a swapped baseline, the polarizations are reversed so the data,
//  relabelling the MS without reshuffling the data, would *actually* be:
//
//                    LL, RR, RL, LR
//  so we recompute what the correct index of the products would be:
//                    0   1   3   2
//  by looking up the index of the *actual* polarization product in the
//  original order (because that's how it is to be written in the MS)
rowmap_type compute_reshuffled_rows( CorrelatorSetup const& setup ) {
    correlationset_t const& ms_order( setup.getCorrelations() );
    unsigned int            srcRow( 0 );
    rowmap_type             result( ms_order.size() );

    for(correlationset_t::const_iterator p=ms_order.begin(); p!=ms_order.end(); p++) {
        correlationset_t::const_iterator pNew = std::find(ms_order.begin(), ms_order.end(), p->mk_swap());
        if( pNew==ms_order.end() )
            THROW_JEXCEPT(casacore::String("Internal error: reversed polarization product ")+Repr(p->mk_swap())+" not present in CorrelatorSetup");
        result[ srcRow++ ] = std::distance(ms_order.begin(), pNew);
    }
    return result;
}


// Only fill data from the visbuf where
// the scans are in the range sjdescr.firstScanNo() -> sjdescr.lastScanNo()
// (note: other filters may apply as well!! (see expirent/filler options)
Bool JIVEMSFiller::fillMSfromVisibilityBuffer( MeasurementSet& msref,
											   VisibilityBuffer& visbuf,
											   const subjob_t& sjdescr )
{
    uInt                 currentrownr;
    double               integrationtime;
    String               lastScanId;
    Visibility           cv;
	sb2ddid_type         sb2ddid;
    ant2ant_type         ant2ant;
    const j2ms::Scan*    curScanptr;
    const Experiment&    expref( visbuf.getExperiment() );
    j2ms::FreqCounter    sourceidcounter;

    // Make a number of column objects, to speed up access
    ScalarColumn<Int>     antenna1( msref, MS::columnName(MS::ANTENNA1) );
    ScalarColumn<Int>     antenna2( msref, MS::columnName(MS::ANTENNA2) );
    ScalarColumn<Int>     arrayid( msref, MS::columnName(MS::ARRAY_ID) );
    ScalarColumn<Int>     feed1( msref, MS::columnName(MS::FEED1) );
    ScalarColumn<Int>     feed2( msref, MS::columnName(MS::FEED2) );
    ScalarColumn<Int>     fieldid( msref, MS::columnName(MS::FIELD_ID) );
    ScalarColumn<Int>     observationidcol( msref, MS::columnName(MS::OBSERVATION_ID) );
    ScalarColumn<Int>     scannumber( msref, MS::columnName(MS::SCAN_NUMBER) );
    ScalarColumn<Int>     datadescid( msref, MS::columnName(MS::DATA_DESC_ID) );
    ScalarColumn<Int>     processorid( msref, MS::columnName(MS::PROCESSOR_ID) );
    ScalarColumn<Double>  timecentroid( msref, MS::columnName(MS::TIME_CENTROID) );
    ScalarColumn<Int>     observationid( msref, MS::columnName(MS::OBSERVATION_ID) );
    ScalarColumn<Int>     stateid( msref, MS::columnName(MS::STATE_ID) );
    ArrayColumn<Bool>     flag( msref, MS::columnName(MS::FLAG) );
    ScalarColumn<Bool>    flagrow( msref, MS::columnName(MS::FLAG_ROW) );
    ArrayColumn<Double>   uvw( msref, MS::columnName(MS::UVW) );
    ArrayColumn<Float>    weight( msref, MS::columnName(MS::WEIGHT) );
    ArrayColumn<Float>    sigma( msref, MS::columnName(MS::SIGMA) );
    ScalarColumn<Double>  exposure( msref, MS::columnName(MS::EXPOSURE) );
    ScalarColumn<Double>  intervalcol( msref, MS::columnName(MS::INTERVAL) );
    ScalarColumn<Double>  time( msref, MS::columnName(MS::TIME) );

    //  Leave the datacolumn unbound; we will find out later what the
    //  *real* datacolumn will eb (LAG_DATA or DATA)
    ArrayColumn<Complex>  data;

    //  Source stuff...
    Int                       currsourceindex;

    //  Vector of Floats: the SIGMA-vector
    IPosition                 datashape;
    Vector<Float>             sigmavalues;

    //  Matrix<Bool> -> for FLAG stuff...
    Bool                     flagvalue;
    Matrix<Bool>             flagmatrix;
   
    // experiment cache
    ExperimentCacheElement*  ptr2element; 

    //  Ok. Get data from the cache
    ptr2element = theirCache.locateExperiment( expref.getCode() );
    if( !ptr2element ) 
		ptr2element = theirCache.addExperiment( expref );
    if( !ptr2element ) {
		cout << "fillMSFromVisibilityBufer: Failed to add experiment to cache?!" << endl;
        cout << "   Cannot continue translating this buffer!" << endl;
		return False;
    }

    //  If the exp. wants lag-data we need to attach the lag-data column from the 
    //  main table, otherwise the plain-old- data column
    if( expref.getDestinationDomain()==Visibility::time )
		data.attach( msref, MS::columnName(MS::LAG_DATA) );
    else if( expref.getDestinationDomain()==Visibility::frequency )
		data.attach( msref, MS::columnName(MS::DATA) );
    
    
    //  Okiedokie: read visibilities from the scan-buffer until there
    //  are no more left!
    currentrownr = msref.nrow();
    
    //  Take some variables out of the loop to prevent them being put
    //  on the stack over and over again
	Int              scanno;
	Int              fScan;
	Int              lScan;
    volatile casacore::Bool available;

	// Filter variables:
	// - whether or not the visibility passed all filters
	// - filteriterator
	// - number of filtered visibilities
	bool                      pass; 
	uLong                     numWritten;
	uLong                     numFiltered;
	uLong                     numOutsideScan;
    uLong                     numOutsideSubjob;
	vector<Filter*>::iterator curfilt;
    
    // Track translation speed
    uLong            nrbytes( 0 );
    struct timeval   starttime,endtime;
    
    //  tell the system to disp. DMY...
    MVTime::setFormat(MVTime::DMY);

    available = True;
    lastScanId = "No scan yet";

	// Let's read whatever's present in the processor table (in case 
	// we're appending to an existing MS...)
    //
	// Note: we only have to check the current pass here since it
	// *should* not vary across the visibilitybuffer
	Int                       curpassid;
	Int                       curprocid;
    Int const                 curstateid = JIVEMSFiller::updateStateTable(msref, expref);
	map<int,int>              pass2pid;
	map<int,int>::iterator    ptr;
    // when writing a swapped baseline the data rows must be swapped
    // (for the cross pols). "rowMap[ i ] = j" means that source row 'i'
    // must be copied to destination row 'j'
    rowmap_type                reshuffled_rows; 

	{
		const MSProcessor&      proc( msref.processor() );
		ROMSProcessorColumns    proccols( proc );

		for( uInt i=0; i<proccols.nrow(); i++ ) {
			pass2pid[ (int)(proccols.passId()(i)) ] = (int)i;
		}
	}

	curpassid = visbuf.getVisBufId();
	ptr       = pass2pid.find(curpassid);

	if( ptr==pass2pid.end() ) {
		// was not yet defined...

		// add to table
		curprocid = this->updateProcessorTable( msref, curpassid, visbuf.getProcessorSubType() );

		// and add to map
		pass2pid[ curpassid ] = (Int)curprocid;
	} else {
		curprocid = ptr->second;
	}
    //  Go!
	scanno           = -1;
	numWritten       = 0;
	numFiltered      = 0;
	numOutsideScan   = 0;
    numOutsideSubjob = 0;
	fScan            = sjdescr.firstScanNo();
	lScan            = sjdescr.lastScanNo();
	currsourceindex  = -1;

	::gettimeofday(&starttime, 0);
	try
	{
        while( available ) {
    		available = visbuf.getCurrentVisibility( cv );

            // no more visibilities?
    		if( !available )
    			break;

            // Check for new possible new scan
    		curScanptr = visbuf.getScanptr();
    		if( curScanptr && (curScanptr->getScanId()!=lastScanId) ) {
                String                 fcode( curScanptr->getFrequencyCode() );
                FrequencyConfig        fc;
                FrequencyDB const&     fdb( expref.getFrequencyDB() );
                CorrelatorSetup const& setup( visbuf.correlatorSetup() );
            
    			lastScanId = curScanptr->getScanId();

    			// Show usr. we detected a new scan...
                ::printf("\r%80s", " ");
    			cout << endl
                     << "===> Find: " << *curScanptr << endl
                     << "     Corr: " << setup << endl;
	    
    			//  Look up the source of this scan in the experiment's source-list
                if( (currsourceindex=JIVEMSFiller::updateFieldTable(msref, *curScanptr))<0 )
                    THROW_JEXCEPT(String("Source '")+curScanptr->getSourceCode()+"' not found in database?! Cannot go on!");

    			//  Try to get the frequencysetup for this scan....
    			if( !fdb.searchFrequencyConfig(fc, fcode) ) {
    				cout << "Failed to locate setup " << fcode << endl;
			    } else {
        			cout << "\n" << fc << endl;
                }
			    integrationtime = setup.getIntegrationTime();

    			// Re-initialize lookuptable for subband -> datadescriptionId.
    			sb2ddid.clear();
                ant2ant.clear();
			
		    	// save scannumber
    			scanno = curScanptr->getScanNo();

                // and precompute the reshuffled rowmap for swapped
                // baselines
                reshuffled_rows = compute_reshuffled_rows( setup );
            } else if( !curScanptr ) {
        	    if( lastScanId!="" ) {
			    	lastScanId = "";
                    // clear the line
                    ::printf("\r%80s", " ");
			    	cout << endl
			    		 << "===> Finding visibilities without a Scan [not writing 'm]"
                         << endl;
		
    				integrationtime  = 0.0;
    				currsourceindex  = -1;
    				scanno           = -1;
    				sb2ddid.clear();
                    reshuffled_rows.clear();
			    }
		    }

            // Print counter *before* we can get into the 'continue'
            // thingy (theoretically we could show nothing,
            // if *everything* would be filtered or without scan)
            if( !((numWritten+numFiltered+numOutsideScan+numOutsideSubjob)%100) ) {
		    	cout << "ToMS: " << setw(8) << numWritten << " "
                     << "Filt: " << setw(8) << numFiltered << " "
                     << "NoSCN: " << setw(8) << numOutsideScan << " "
                     << "OOSubjobRNG: " << setw(8) << numOutsideSubjob
                     << "     \r"
                     << flush;
	    	}

    		if( !curScanptr ) {
    			numOutsideScan++;
    			visbuf.next();
    			continue;
	    	}

            // Check #1: is the scan within 'subjob' range?
    		//           the fScan/lScan are cached from
    		//           sjdescr.firstScanNo()/lastScanNo()
    		//           where sjdescr is a calling parameter
    		//           of this function.
    		//           
    		//           NOTE NOTE NOTE NOTE NOTE NOTE NOTE
    		//           
    		//           Do not try to optimize the condition
    		//           below since the system relies on the fact
    		//           that if (backwards compatibility mode!)
    		//           *no* scan-number based filtering is to be
    		//           done, the fScan==INT_MIN (such that the 
    		//           test scanno>=fScan is always true) and 
    		//           that lScan==INT_MAX (hence scanno<=lScan
    		//           always true). If you re-order the condition
    		//           it might stop working...
    		if( !(scanno>=fScan && scanno<=lScan) ) {
                numOutsideSubjob++;
    			visbuf.next();
    			continue;
    		}
		
            // Run the visibility through our filters! See if it passes
        	// all filters. If it does, then continue on!
    		pass    = true;
    		curfilt = myFilters.begin();

    		while( curfilt!=myFilters.end() && pass ) {
    			pass = pass && (*curfilt)->letPass(cv);
    			curfilt++;
    		}
		
    		if( !pass ) {
    			numFiltered++;
    			visbuf.next();
    			continue;
    		}

            //  Cache a reference to the visibilitylabel
            const ExtVisLabel&     label( cv.getLabel() );
            sb2ddid_type::iterator cursb2dd;
	
            //  (try to) get the datadescriptor-id for the subband
            if( (cursb2dd=sb2ddid.find(label.subbandnr))==sb2ddid.end() ) {
                Int                                ddId;
                pair<sb2ddid_type::iterator, bool> insres;

                // Let the system come up with the correct ddId
                ddId = JIVEMSFiller::updateSpectralAndPolarizationTable(msref, cv, *curScanptr);

                // Warn if we couldn't find it
				if( ddId==-1 ) {
					cout << "JIVEMSFiller: AAAARGHHH? Failed to get a datadescid "
                         << "for subband#" << label.subbandnr << endl;
				}
                insres = sb2ddid.insert( make_pair(label.subbandnr, ddId) );
                if( !insres.second )
                    THROW_JEXCEPT("Failed to insert entry in subband->datadescId cache!");
                cursb2dd = insres.first;
            }

            // ant1, ant2 (experiment) must map to ant1, ant2 (MS)
            // NOTE: see the FEED note below -
            //       we will only write FEED1, FEED2 = 0 so we
            //       don't need to lookup/cache those
            // NOOOOOOOH No NO no NOOOOH
            // HV: May 2018 We know do not make assumptions about those
            //              FEEDs. Let the updateAntenna....() return
            //              (generate if necessary) us
            //              a proper FEED for the current antenna +
            //              polarization setup
            ant2ant_type::const_iterator ant1ptr = ant2ant.find(label.baseline.idx1());

            if( ant1ptr==ant2ant.end() )
                ant1ptr = ant2ant.insert( make_pair(label.baseline.idx1(),
                                                    JIVEMSFiller::updateAntennaAndFeedTable(msref, label.baseline.idx1(), *curScanptr)) ).first;

            ant2ant_type::const_iterator ant2ptr = ant2ant.find(label.baseline.idx2());
            if( ant2ptr==ant2ant.end() )
                ant2ptr = ant2ant.insert( make_pair(label.baseline.idx2(),
                                                    JIVEMSFiller::updateAntennaAndFeedTable(msref, label.baseline.idx2(), *curScanptr)) ).first;
            // Now that antennas may be delivered in any order and be
            // matched to any antenna in the MS we must make sure that the
            // data in the MS is in 'baseline major order': ant1 <= ant2
            // where ant1, ant2 are the *MS* antenna numbers!
            //
            // Before writing stuff the the MS we must swap the indices.
            // Later below we must also swap the data ...
            bool const is_swapped_baseline( ant1ptr->second.first>ant2ptr->second.first );

            if( is_swapped_baseline )
                std::swap(ant1ptr, ant2ptr);


            //  Ok. Dump the data in the MS!
            msref.addRow();
	
    		//  Write the columns!
    		time.put( currentrownr, (cv.getTime().get("s")).getValue() );
    		antenna1.put( currentrownr, ant1ptr->second.first /*label.baseline.idx1()*/ );
    		antenna2.put( currentrownr, ant2ptr->second.first /*label.baseline.idx2()*/ );
    	
    		//  Assume: our antennas have only 1 feed (with max. two pols) so default
    		//  to using feed 0 on the current antennae....
            //  HV: May 2018 - The FEEDs are now updated dynamically
            //                 (normalized) in the FEED table
    		feed1.put( currentrownr, ant1ptr->second.second );
    		feed2.put( currentrownr, ant2ptr->second.second );
    		datadescid.put( currentrownr, cursb2dd->second );
    	
    		// [TODO: document this]
    		processorid.put( currentrownr, curprocid );
    		
    		fieldid.put( currentrownr, currsourceindex );
    		intervalcol.put( currentrownr, integrationtime );
    		exposure.put( currentrownr, integrationtime );
    		timecentroid.put( currentrownr, integrationtime );
    		scannumber.put( currentrownr, scanno );
    		arrayid.put( currentrownr, 0 );
    		observationid.put( currentrownr, 0 );
            // Mar 2018: DesS finds that:
            //           non-empty STATE table and STATE_ID==-1
            //           in MAIN table leads to SIGSEGV in casa
            //           listobs(...)
            //           (Empty STATE table w/ STATE_ID==-1 works fine,
            //            as well as STATE_ID==0, effectively referencing
            //            the actual row in the STATE subtable)
            // May 2018: Decided to stop writing STATE table.
            //           But since we can *append* to an existing MS we'll
            //           fall always do the right thing. See
            //           "updateStateTable()". The 'curstateid' will be fixed
            //           to the return value of that function
    		stateid.put( currentrownr, curstateid );
    	
    		//  We need to calculate UVW....
            //  That is, unless the visibility has it's UVW
            //  part filled in
            if( MVuvw(cv.getUVW().get("m")).nearAbs(j2ms::noMVuvw) )
        		uvw.put( currentrownr, JIVEMSFiller::calcUVW(cv) );
            else
                uvw.put( currentrownr, cv.getUVW().get("m").getValue() );
		
    		//  The makeData() function makes sure that the visibilitydata
    		//  is conform the requested domain as indicated in the
    		//  experiment AND it will reorder the visibility data
            //  in the order the experiment dictates.
            rowmap_type const      noShuffle;
    		Matrix<Complex> const& dataref( JIVEMSFiller::makeData(expref, cv, is_swapped_baseline ? reshuffled_rows : noShuffle) );

            data.put( currentrownr, dataref );
		
    		//  Get the shape
    		datashape = dataref.shape();
	
    		//  Resize sigma...
    		sigmavalues.resize( datashape(0) );
    		sigmavalues = 0.0f;
	
    		sigma.put( currentrownr, sigmavalues );
    		weight.put( currentrownr, cv.getWeight() );	
	
            //  Resize FLAG-matrix
        	flagmatrix.resize( datashape );
	    
    		flagvalue  = cv.getFlag();
    		flagmatrix = flagvalue;
		
    		flag.put( currentrownr, flagmatrix );
    		flagrow.put( currentrownr, flagvalue );
	
            // Increment the counter that keeps track of how many bytes
        	// were translated
    		nrbytes += dataref.shape().product()*sizeof(Complex);
    		numWritten++;
	
    		//  Move on to next visibility
    		currentrownr++;

    		available = visbuf.next();
    	}
	}
	catch( AipsError const& z ) {
		cout << "*gasp* got an aips exception!" << z.getMesg() << endl;
	}
    catch( const std::exception& e ) {
        cout << "Caught deadly std::exception: " << e.what() << endl;
    }
    //  get time!
	::gettimeofday(&endtime, 0);

	//  Update the observation table (we do this postfix because the
	//  observation start/end need to be in here... shiiiit.. is this
	//  dumb or what?) We do not really know these so we just derive
	//  these values from the first/last time in the current MS...
	if( !this->updateObservationTable(msref, expref) ) {
		cout << "*** Huh? Failed to update OBSERVATION-table in MS " << expref.getMSname() << "\n"
			<< "    for experiment " << expref.getCode() << endl;
		cout << "*** Some info maybe be stored incorrectly..." << endl;
		cout << "*** NOTE: THIS DOES NOT AFFECT THE DATA!" << endl;
		cout << "*** This is merely auxiliary data...." << endl;
    }
    

	double   ssec, esec;

	ssec = starttime.tv_sec + starttime.tv_usec/1E6;
	esec = endtime.tv_sec + endtime.tv_usec/1E6;
    ::printf("\r%80s", " ");
    cout << endl << "Translated @ "
		 << j2ms::sciprint<long double>(((long double)nrbytes)/(esec-ssec) , "bytes/s")
         << endl;

    cout << endl
         << "--------====== Summary ======------" << endl;
    ::printf("ToMS:        %9lu\n", numWritten);
    ::printf("Filt:        %9lu\n", numFiltered);
    ::printf("NoSCN:       %9lu\n", numOutsideScan);
    ::printf("OOSubjobRNG: %9lu\n", numOutsideSubjob);
    ::printf("\n");

    // Update the JIVE_PROCESSING table
    try {
        // Only try to access the table if there really *is*
        // something to write [does not break backward compat;
        // old MSs do not have this table but usually their
        // jobs do not have sufficient info to do processing...]
        processing_steps  ps( visbuf.getProcessingSteps() );

        if( ps.size() ) {
            Int                              cjid;
            Table                            ptab( msref.rwKeywordSet().asTable("JIVE_PROCESSING") );
            Double                           now;
            unsigned int                     row( ptab.nrow() );
            ScalarColumn<Int>                jid( ptab, "JOBID" );
            ScalarColumn<Double>             tm( ptab, "TIME" );
            ScalarColumn<String>             task( ptab, "PROCESSING" );
            ScalarColumn<String>             parms( ptab, "PROCESSING_PARAMS" );
            processing_steps::const_iterator curstep;
            
            // already reserve space for the new rows
            ptab.addRow( ps.size() );
            now  = (Time()).modifiedJulianDay();
            cjid = visbuf.getVisBufId();
            // Append data for the steps
            for(curstep=ps.begin(); curstep!=ps.end(); curstep++, row++) {
                jid.put(row, cjid);
                tm.put(row, now);
                task.put(row, curstep->task);
                parms.put(row, curstep->parameters );
            }
            ptab.flush();
        }
    }
    catch( ... ) {
        // no JIVE-PROCESSING table => bummer
    }

    // append even more stuff...
#if 0
    JIVEMSFiller::writeModel( msref, visbuf );
#endif 
    return True;
}

//////////////////////////////////////////////////////////////////////////////////////////
//
//  Normalization mapping/lookup v2.0
//
//  The following caches stuff by frequency config name
//  It is seeded from an existing MS::SPECTRAL_WINDOW + MS::DATA_DESC_ID + MS::FEED
//  such that appending data will properly normalize
//
//////////////////////////////////////////////////////////////////////////////////////////

// May 2018: For reasons (...) it would be appreciated if we write
//           spectral windows in a "meaningful way", say by increasing
//           frequency.
//           So what we'll do is write the full frequency config in one go
//           such that successive subbands have successive spectral window
//           ids. Me I'm not 100% convinced this is really useful but we aim
//           to please.
//           It becomes important to know wether a specific frequency config
//           has already been written to the spectral window table or not


//////////////////////////////////////////////////////////////////////////////////////////
//                        The spectral_window 
//////////////////////////////////////////////////////////////////////////////////////////

struct spw_entry_type {
    public:
        spw_entry_type(casacore::Int spw, casacore::String const& mode, casacore::Int nr, casacore::uInt nSp,
                       casacore::Double f0, casacore::Double bw, casacore::Int side): 
            __m_spwId(spw), __m_FreqGroup(mode)   , __m_FreqGroupNr(nr), __m_nSpectralPoints(nSp),
            __m_f0(f0)    , __m_TotalBandwidth(bw), __m_NetSideband(side)
        {}
        // special 'copy constructor' for just replacing the spectralwindowid
        spw_entry_type(casacore::Int spw, spw_entry_type const& other):
            __m_spwId(spw), __m_FreqGroup(other.__m_FreqGroup),
            __m_FreqGroupNr(other.__m_FreqGroupNr),
            __m_nSpectralPoints(other.__m_nSpectralPoints),
            __m_f0(other.__m_f0), 
            __m_TotalBandwidth(other.__m_TotalBandwidth),
            __m_NetSideband(other.__m_NetSideband)
        {}

        casacore::Int const    __m_spwId;
        casacore::String const __m_FreqGroup;
        casacore::Int const    __m_FreqGroupNr;
        casacore::uInt const   __m_nSpectralPoints;
        casacore::Double const __m_f0;
        casacore::Double const __m_TotalBandwidth;
        casacore::Int const    __m_NetSideband;

    private:
        spw_entry_type();
};

std::ostream& operator<<(std::ostream& os, spw_entry_type const& spw) {
    return os << "SPW: #" << spw.__m_spwId << " nFreq:" << spw.__m_nSpectralPoints
              << " " << spw.__m_TotalBandwidth/1e6 << "MHz " << (spw.__m_NetSideband==0 ? 'L' : 'U') << "SB"
              << "@" << spw.__m_f0/1e6 << "MHz";
}

////////////////////////////////////////////////////////////////////////////
//
//  We're going to need to compare many fields in >>1 structs; specifically
//  we need to overload these boolean operators
//      x <  y
//      x == y
//  and typing all that out is going to be a boring job
//
//  So with some helper template functions + macro expansions we make it
//  maintainable and bearable
//
////////////////////////////////////////////////////////////////////////////

// Compilert will automatically inline this/optimize it away
template <typename T>
bool AlwaysFalse(T const&, T const& ) {
    return false;
}

// nearAbs() is overloaded for all casacore types ...
template <typename T>
bool NotEqual(T const& l, T const& r) {
    return !casacore::nearAbs(l, r);;
}
// ... but for casacore::String
template <>
bool NotEqual<casacore::String>(casacore::String const& l, casacore::String const& r) {
    return l!=r;
}
// ... and for casacore::Vector/Array?
//template <typename T>
//bool NotEqual(casacore::Array<T> const& l, casacore::Array<T> const& r) {
//    return !casacore::allNearAbs(l, r);
//}
template <typename T>
bool NotEqual(casacore::Vector<T> const& l, casacore::Vector<T> const& r) {
    return !allEQ(l, r);
}
template <typename T>
bool NotEqual(casacore::Matrix<T> const& l, casacore::Matrix<T> const& r) {
    return !allEQ(l, r);
}


// Lexicographical compare on CASA[CORE]/AIPS++ Array<T> is not quite
// defined ffs
//template <typename T>
//bool Less(casacore::Array<T> const& l, casacore::Array<T> const& r) {
//    return std::lexicographical_compare(l.cbegin(), l.cend(), r.cbegin(), r.cend());
//}
template <typename T>
bool Less(casacore::Vector<T> const& l, casacore::Vector<T> const& r) {
    return std::lexicographical_compare(l.cbegin(), l.cend(), r.cbegin(), r.cend());
}
template <typename T>
bool Less(casacore::Matrix<T> const& l, casacore::Matrix<T> const& r) {
    return std::lexicographical_compare(l.cbegin(), l.cend(), r.cbegin(), r.cend());
}
template <typename T>
bool Less(T const& l, T const& r) {
    return l<r;
}

#define IFOP1RETURNOP2(a, b, op1, op2) \
    if( op1(a, b) ) \
        return op2(a, b);

// The macro COMPARE() expands to this:
//    if( left.member!=right.member )
//        return left.member < right.member;
#define COMPARE(member, left, right) \
    IFOP1RETURNOP2(left.member, right.member, NotEqual, Less)
// The macro NOTEQ() expands to this:
//    if( left.member!=right.member )
//        return false;
#define NOTEQ(member, left, right) \
    IFOP1RETURNOP2(left.member, right.member, NotEqual,  AlwaysFalse)


// Implement    spw1 < spw2
bool operator<(spw_entry_type const& l, spw_entry_type const& r) {
    COMPARE(__m_FreqGroup, l, r);
    COMPARE(__m_FreqGroupNr, l, r);
    COMPARE(__m_nSpectralPoints, l, r);
    COMPARE(__m_f0, l, r);
    COMPARE(__m_TotalBandwidth, l, r);
    // only one attribute left ...
    return l.__m_NetSideband < r.__m_NetSideband;
}

// Implement    spw1 == spw2
bool operator==(spw_entry_type const& l, spw_entry_type const& r) {
    NOTEQ(__m_FreqGroup, l, r);
    NOTEQ(__m_FreqGroupNr, l, r);
    NOTEQ(__m_nSpectralPoints, l, r);
    NOTEQ(__m_f0, l, r);
    NOTEQ(__m_TotalBandwidth, l, r);
    // only one attribute left ...
    return l.__m_NetSideband == r.__m_NetSideband;
}
typedef std::list<spw_entry_type>   spectral_window_table_type;
static spectral_window_table_type   spectral_window_table = spectral_window_table_type();


//////////////////////////////////////////////////////////////////////////////////////////
//                        The polarization table
//////////////////////////////////////////////////////////////////////////////////////////
struct polarization_entry_type {
    public:
        polarization_entry_type(casacore::Int pid, correlations_t const& corr, j2ms::log2polmap const& l2p):
            __m_polarizationId(pid), __m_correlations(corr), __m_l2p(l2p)
        {}
        polarization_entry_type(casacore::Int pid, polarization_entry_type const& other):
            __m_polarizationId(pid), __m_correlations(other.__m_correlations), __m_l2p(other.__m_l2p)
        {}

        // We transform the MS CORR_TYPE enums into j2ms2 ones for internal
        // convenience
        casacore::Int const    __m_polarizationId;
        correlations_t const   __m_correlations;
        j2ms::log2polmap const __m_l2p;
    private:
        polarization_entry_type();
};

std::ostream& operator<<(std::ostream& os, polarization_entry_type const& pe) {
    return os << "POL: #" << pe.__m_polarizationId << " " << as_corrcode(pe.__m_correlations) << " " << pe.__m_l2p;
}

// Implement polarization_entry_type < other
bool operator<(polarization_entry_type const& l, polarization_entry_type const& r) {
    // Do not check __m_polarizationId, it is not part of the 'primary key'
    if( l.__m_correlations!=r.__m_correlations )
        return l.__m_correlations < r.__m_correlations;
    return l.__m_l2p<r.__m_l2p;
}
// Implement polarization_entry_type == other
bool operator==(polarization_entry_type const& l, polarization_entry_type const& r) {
    if( l.__m_correlations!=r.__m_correlations )
        return false;
    return l.__m_l2p==r.__m_l2p;
}

typedef std::list<polarization_entry_type> polarization_table_type;
static polarization_table_type             polarization_table;

//////////////////////////////////////////////////////////////////////////////////////////
//                        The data description table
//////////////////////////////////////////////////////////////////////////////////////////
struct datadescription_entry_type {
    public:
        datadescription_entry_type(Int ddid, Int spw, Int pol):
            __m_dataDescId(ddid), __m_SpectralWindowId(spw), __m_PolarizationId(pol)
        {}
        datadescription_entry_type(Int ddid, datadescription_entry_type const& other):
            __m_dataDescId(ddid), __m_SpectralWindowId(other.__m_SpectralWindowId),
            __m_PolarizationId(other.__m_PolarizationId)
        {}

        casacore::Int const __m_dataDescId;
        casacore::Int const __m_SpectralWindowId;
        casacore::Int const __m_PolarizationId;

    private:
        datadescription_entry_type();
};

std::ostream& operator<<(std::ostream& os, datadescription_entry_type const& dd) {
    return os << "DD: #" << dd.__m_dataDescId << " SPW=" << dd.__m_SpectralWindowId << " POL=" << dd.__m_PolarizationId;
}

bool operator<(datadescription_entry_type const& l, datadescription_entry_type const& r) {
    // do NOT compare on __m_dataDescId - that's not part of the 'primary key'
    COMPARE(__m_SpectralWindowId, l, r);
    return l.__m_PolarizationId < r.__m_PolarizationId;
}

bool operator==(datadescription_entry_type const& l, datadescription_entry_type const& r) {
    return (l.__m_SpectralWindowId == r.__m_SpectralWindowId && 
            l.__m_PolarizationId   == r.__m_PolarizationId);
}

typedef std::list<datadescription_entry_type> datadescription_table_type;
static datadescription_table_type             datadescription_table;

//////////////////////////////////////////////////////////////////////////////////////////
//                        The antenna table
//////////////////////////////////////////////////////////////////////////////////////////
struct antenna_entry_type {
    antenna_entry_type(casacore::Int ant, casacore::String const& nm, casacore::String const& st, casacore::String const& mount,
                       casacore::Double diam, casacore::Vector<casacore::Double> const& pos, casacore::Vector<casacore::Double> const& off):
        __m_AntennaId(ant), __m_AntennaName(nm), __m_Station(st), __m_Mount(mount),
        __m_Diameter(diam), __m_Position(pos), __m_Offset(off)
    {
        static const casacore::IPosition    shp(1, 3);
        // assert that position + offset are vectors of length 3
        // as per MS definition, empty is not an option
        if( __m_Position.shape()!=shp )
            THROW_JEXCEPT(string("Inconsistent ANTENNA #")+repr(__m_AntennaId)+": position is not a 3-element vector");
        if( __m_Offset.shape()!=shp )
            THROW_JEXCEPT(string("Inconsistent ANTENNA #")+repr(__m_AntennaId)+": offset is not a 3-element vector");
    }
    // This is a copy-like constructor where we copy values from 'other'
    // so we don't have to verify Position/Offset vectors because other's
    // constructor will have done that for us :D
    antenna_entry_type(casacore::Int ant, antenna_entry_type const& other):
        __m_AntennaId(ant), __m_AntennaName(other.__m_AntennaName), __m_Station(other.__m_Station),
        __m_Mount(other.__m_Mount), __m_Diameter(other.__m_Diameter),
        __m_Position(other.__m_Position), __m_Offset(other.__m_Offset)
    {}

    casacore::Int const                      __m_AntennaId;
    casacore::String const                   __m_AntennaName;
    casacore::String const                   __m_Station;
    casacore::String const                   __m_Mount;
    casacore::Double const                   __m_Diameter;
    casacore::Vector<casacore::Double> const __m_Position;
    casacore::Vector<casacore::Double> const __m_Offset;
    private:
        antenna_entry_type();
};

std::ostream& operator<<(std::ostream& os, antenna_entry_type const& ant) {
    return os << "ANT: #" << ant.__m_AntennaId << " " << ant.__m_AntennaName << "[" << ant.__m_Station << "] "
              << ant.__m_Diameter << "m " << ant.__m_Mount << " @" << ant.__m_Position;
}

bool operator<(antenna_entry_type const& l, antenna_entry_type const& r) {
    // don't compare antenna id
    COMPARE(__m_AntennaName, l, r);
    COMPARE(__m_Station, l, r);
    COMPARE(__m_Mount, l, r);
    COMPARE(__m_Diameter, l, r);
    COMPARE(__m_Position, l, r);
    COMPARE(__m_Offset, l, r);
    return false;
}

bool operator==(antenna_entry_type const& l, antenna_entry_type const& r) {
    // don't compare antenna id
    NOTEQ(__m_AntennaName, l, r);
    NOTEQ(__m_Station, l, r);
    NOTEQ(__m_Mount, l, r);
    NOTEQ(__m_Diameter, l, r);
    NOTEQ(__m_Position, l, r);
    NOTEQ(__m_Offset, l, r);
    return true;
}

typedef std::list<antenna_entry_type> antenna_table_type;
static antenna_table_type             antenna_table;

//////////////////////////////////////////////////////////////////////////////////////////
//                        The feed table
//////////////////////////////////////////////////////////////////////////////////////////

// Great. In their infinite wisdom, the FEEDn from MAIN is NOT a direct
// index! Good grief. (Almost all other *IDs are direct indices in the
// tables (aka "rownumber") and therefore not part of the primary key)
// but for FEEDn it is. Thx guys!
struct feed_entry_type {

    // For our application we need to find a FEED for an ANTENNA, valid for
    // all SPECTRAL_WINDOWS which is consistent with the current LOG2POLMAP
    // (logical polarization number to physical polarization mapping).
    //
    // Therefore we only keep those values for browsing in memory 
    // Special search functors are implemented below
    //
    // Time dependency: (May 2018) 
    //   our code does not currently deal with time dependency. We will use
    //   records with time<=0 and interval<=0, which is what, at this point,
    //   casacore seems to mean "no time dependency"
    feed_entry_type(casacore::Int ant, casacore::Int fid, casacore::Int spw, casacore::Double t0, casacore::Double dt,
                    j2ms::log2polmap const& l2p):
        __m_AntennaId(ant), __m_FeedId(fid), __m_SpectralWindowId(spw), __m_Time(t0), __m_Interval(dt), __m_l2p(l2p)
    {}

    casacore::Int const                       __m_AntennaId;
    casacore::Int const                       __m_FeedId;
    casacore::Int const                       __m_SpectralWindowId;
    casacore::Double const                    __m_Time;
    casacore::Double const                    __m_Interval;
    j2ms::log2polmap const                    __m_l2p;

    private:
        feed_entry_type();
};

bool operator<(feed_entry_type const& l, feed_entry_type const& r) {
    // for this one we must match on feed id as well as it IS part of the
    // primary key
    COMPARE(__m_AntennaId, l, r);
    COMPARE(__m_FeedId, l, r);
    COMPARE(__m_SpectralWindowId, l, r);
    COMPARE(__m_Time, l, r);
    COMPARE(__m_Interval, l, r);
    return l.__m_l2p < r.__m_l2p;
}

bool operator==(feed_entry_type const& l, feed_entry_type const& r) {
    NOTEQ(__m_AntennaId, l, r);
    NOTEQ(__m_FeedId, l, r);
    NOTEQ(__m_SpectralWindowId, l, r);
    NOTEQ(__m_Time, l, r);
    NOTEQ(__m_Interval, l, r);
    return l.__m_l2p == r.__m_l2p;
}

typedef std::list<feed_entry_type>  feed_table_type;
static feed_table_type              feed_table;

// Get a subset of all the feeds defined for antenna 'ant'
// We must manually check if there's one that matches the rest of our
// criteria (valid for all spectral windows + valid for all times).
// We need to find all FEEDs for the antenna because, if it turns out there
// isn't one meeting our criteria, we must add a new entry and assign a new
// FEEDid for that ANTENNA and therefore we must know which FEEDid's have
// already been defined for this ANTENNA. Sigh.
struct antenna_matcher {
    antenna_matcher(casacore::Int ant):
        __m_AntennaId(ant)
    {}
    casacore::Int const  __m_AntennaId;

    // Support matching anything that has __m_AntennaId
    template <typename T>
    bool operator()(T const& t) const {
        return t.__m_AntennaId == __m_AntennaId;
    }

    private:
        antenna_matcher();
};

// It took 'm until C++11 to realize that "filter" is a desirable algorithm.
// Of course it's not /called/ filter(), in STL it's called "copy_if()"
// Since filter() is quite simple we'll write the template ourselves
template <typename InputIt, typename OutputIt, typename UnaryPredicate>
OutputIt filter(InputIt first, InputIt last, OutputIt d_first, UnaryPredicate pred) {
    while( first!=last ) {
        if( pred(*first) )
            *d_first++ = *first;
        first++;
    }
    return d_first;
}

feed_table_type feeds_by_antenna(casacore::Int ant, feed_table_type const& feedtab) {
    feed_table_type     rv;
    filter(feedtab.begin(), feedtab.end(), std::back_inserter(rv), antenna_matcher(ant));
    return rv;
}


//////////////////////////////////////////////////////////////////////////////////////////
//                        The field table
//////////////////////////////////////////////////////////////////////////////////////////

typedef casacore::Matrix<casacore::Double>  fielddir_type;

struct field_entry_type {

    field_entry_type(casacore::Int fid, casacore::Int sid, casacore::String const& fn, casacore::Double tm,
                     fielddir_type const& deldir, fielddir_type const& phdir, fielddir_type const& refdir):
        __m_FieldId(fid), __m_SourceId(sid), __m_FieldName(fn), __m_Time(tm),
        __m_DelayDir(deldir), __m_PhaseDir(phdir), __m_ReferenceDir(refdir)
    {}

    // The half-copy constructor, basically to record a new entry with an
    // updated (actual) assigned field id
    field_entry_type(casacore::Int fid, field_entry_type const& other):
        __m_FieldId(fid), 
        __m_SourceId( other.__m_SourceId ), __m_FieldName( other.__m_FieldName ), __m_Time( other.__m_Time ),
        __m_DelayDir( other.__m_DelayDir ), __m_PhaseDir( other.__m_PhaseDir ), __m_ReferenceDir( other.__m_ReferenceDir )
    {}

    casacore::Int const     __m_FieldId; // not part of the 'primary key'
    casacore::Int const     __m_SourceId; // note: our code requires an entry -1 here
    casacore::String const  __m_FieldName;
    casacore::Double const  __m_Time;  // note: our code requires an entry valid for all times (0.0)
    // note: our code will look for an entry that has identical values in
    //       all these three
    fielddir_type const     __m_DelayDir;
    fielddir_type const     __m_PhaseDir;
    fielddir_type const     __m_ReferenceDir;

    private:
        field_entry_type();
};


bool operator<(field_entry_type const& l, field_entry_type const& r) {
    // fieldid is NOT part of the primary key
    COMPARE(__m_SourceId, l, r);
    COMPARE(__m_FieldName, l, r);
    COMPARE(__m_Time, l, r);
    COMPARE(__m_DelayDir, l, r);
    COMPARE(__m_PhaseDir, l, r);
    COMPARE(__m_ReferenceDir, l, r);
    // Only end up here if all member "<"-comparisons are false,
    // ie they're all equal, i.e. l is not less-than r ...
    return false;
}

bool operator==(field_entry_type const& l, field_entry_type const& r) {
    NOTEQ(__m_SourceId, l, r);
    NOTEQ(__m_FieldName, l, r);
    NOTEQ(__m_Time, l, r);
    NOTEQ(__m_DelayDir, l, r);
    NOTEQ(__m_PhaseDir, l, r);
    NOTEQ(__m_ReferenceDir, l, r);
    // Only end up here if all member "!=" comparisions are false
    // ie all members are equal, thus l == r!
    return true;
}

typedef std::list<field_entry_type> field_table_type;
static field_table_type             field_table;

//////////////////////////////////////////////////////////////////////////////////////////
//                        The state table
//////////////////////////////////////////////////////////////////////////////////////////

struct state_entry_type {

    state_entry_type(casacore::Int sid, casacore::Bool sig, casacore::Bool ref, casacore::Double cal,
                     casacore::Double load, casacore::Int ss, casacore::String const& obsm):
        __m_StateId(sid), __m_Sig(sig), __m_Ref(ref), __m_Cal(cal), __m_Load(load), __m_SubScan(ss), __m_ObsMode(obsm)
    {}
    // Semi-copy-ctor in order to overwrite the state id only
    state_entry_type(casacore::Int sid, state_entry_type const& other):
        __m_StateId(sid), __m_Sig(other.__m_Sig), __m_Ref(other.__m_Ref), __m_Cal(other.__m_Cal),
        __m_Load(other.__m_Load), __m_SubScan(other.__m_SubScan), __m_ObsMode(other.__m_ObsMode)
    {}

    casacore::Int const    __m_StateId;// not part of the primary key!
    casacore::Bool const   __m_Sig;
    casacore::Bool const   __m_Ref;
    casacore::Double const __m_Cal;
    casacore::Double const __m_Load;
    casacore::Int const    __m_SubScan;
    casacore::String const __m_ObsMode;

    private:
        state_entry_type();
};


bool operator<(state_entry_type const& l, state_entry_type const& r) {
    // stateid is NOT part of the primary key
    COMPARE(__m_Sig, l, r);
    COMPARE(__m_Ref, l, r);
    COMPARE(__m_Cal, l, r);
    COMPARE(__m_Load, l, r);
    COMPARE(__m_SubScan, l, r);
    COMPARE(__m_ObsMode, l, r);
    // Only end up here if all member "<"-comparisons are false,
    // ie they're all equal, i.e. l is not less-than r ...
    return false;
}
bool operator==(state_entry_type const& l, state_entry_type const& r) {
    // stateid is NOT part of the primary key
    NOTEQ(__m_Sig, l, r);
    NOTEQ(__m_Ref, l, r);
    NOTEQ(__m_Cal, l, r);
    NOTEQ(__m_Load, l, r);
    NOTEQ(__m_SubScan, l, r);
    NOTEQ(__m_ObsMode, l, r);
    // Only end up here if all member "<"-comparisons are false,
    // ie they're all equal, i.e. l is not less-than r ...
    return true;
}

typedef std::list<state_entry_type> state_table_type;
static state_table_type             state_table;



// Request to rebuild all caches
void JIVEMSFiller::rebuild_caches(MeasurementSet const& ms) {
    // Deal with the SPECTRAL_WINDOWs
    ROMSSpWindowColumns    spwcols( ms.spectralWindow() );

    spectral_window_table.clear();
    for(uInt spwid=0; spwid<spwcols.nrow(); spwid++) {
        // If this spectral window is flagged don't load it
        if( spwcols.flagRow()(spwid) )
            continue;
        // Need to deduce some properties of the spectral window
        // in order to know what f0 is we need to know what sideband it is
        Int const            sideband( spwcols.netSideband()(spwid) );
        Vector<Double> const chans( spwcols.chanFreq().get(spwid) );
        uInt const           nSp( chans.shape()[0] );

        if( sideband<0 || sideband>1 ) {
            cerr << "WARN: Skipping spectral window #" << spwid << " with unsupported NET_SIDEBAND=" << sideband << " (this code uses 0==LSB, 1==USB)" << endl;
            continue;
        }
        if( nSp!=(uInt)spwcols.numChan()(spwid) )
            THROW_JEXCEPT(string("Inconsistent Spectral Window Table Entry #")+repr(spwid)+": NUM_CHAN="+repr(spwcols.numChan()(spwid))+" len(CHAN_FREQ)="+repr(nSp));
        spectral_window_table.push_back( spw_entry_type((Int)spwid, spwcols.freqGroupName()(spwid), spwcols.freqGroup()(spwid),
                                                        nSp, (sideband == 0 ? chans[nSp-1] : chans[0]),
                                                        spwcols.totalBandwidth()(spwid), sideband) );
    }
    cout << "Loaded " << spectral_window_table.size() << " entries from SPECTRAL_WINDOW" << endl;

    // Deal with the POLARIZATIONs
    ROMSPolarizationColumns    polcols( ms.polarization() );

    polarization_table.clear();
    for(uInt polid=0; polid<polcols.nrow(); polid++) {
        // If this polarization is flagged don't load it
        if( polcols.flagRow()(polid) )
            continue;
        Array<Int> const     corrProds( polcols.corrProduct().get(polid) );
        Vector<Int> const    corrTypes( polcols.corrType().get(polid) );
        correlations_t const j2( aipspppol2j2mspol(corrTypes) );
        uInt const           nCorr( corrTypes.shape()[0] );
        IPosition const      nProds( corrProds.shape() );

        // Consistency checks
        if( nCorr!=(uInt)polcols.numCorr()(polid) )
            THROW_JEXCEPT(string("Inconsistent Polarization Table Entry #")+repr(polid)+": NUM_CORR="+repr(polcols.numCorr()(polid))+" != len(CORR_TYPE)="+repr(nCorr));


        // Our code currently only deals with XYLR so we only remember
        // polarization types that are strictly formed of those.
        // "aipspppol2j2mspol()" will insert default/empty CorrelationCode
        // in the list if any of the StokesTypes values not element of the
        // LRXY combos.
        if( std::find(j2.begin(), j2.end(), CorrelationCode())!=j2.end() ) {
            cerr << "WARN: Skipping polarization #" << polid << " with unsupported CORR_TYPEs - this code only supports LRXY combination(s)" << endl;
            continue; 
        }
        // The corrProds should be (2,nCorr)
        if( nProds.nelements()!=2 || (nProds.nelements()==2 && (nProds[0]!=2 || nProds[1]!=nCorr)) )
            THROW_JEXCEPT(string("Inconsistent Polarization Table Entry #")+repr(polid)+": CORR_PRODUCT shape is not 2,"+repr(nProds));

        // verify that those map consistently to the polarizations 
        // in the corrTypes
        Matrix<Int> const         cpMat( corrProds );
        j2ms::log2polmap::maptype tmpMap;
        for(uInt cp=0; cp<nProds; cp++) {
            // corrType[cp]    = (P0,   P1) 
            // cpMat[0,1][cp]  = (idx0, idx1)
            //  => insert into logigcal 2 physical mapping entries
            //   (idx0, P0), (idx1, P1)
            for(uInt n=0; n<2; n++) {
                j2ms::log2polmap::val_type          tmpVal( make_pair(cpMat(n, cp), j2[cp][n]) );
                j2ms::log2polmap::maptype::iterator p;

                if( (p=tmpMap.find(tmpVal.first))!=tmpMap.end() )
                    if( p->second!=tmpVal.second )
                        THROW_JEXCEPT(string("Inconsistent Polarization Table Entry #")+repr(polid)+": CORR_PRODUCT "+repr(cp)+" element "+repr(n)+" inconsistent with CORR_TYPE "+repr(j2[cp]));
                tmpMap.insert( tmpVal );
            }
        }
        // Now create a real log2polmap. Will throw if anything in the
        // recovered mapping seems fishy
        j2ms::log2polmap    l2p;
        for(j2ms::log2polmap::maptype::const_iterator p = tmpMap.begin(); p!=tmpMap.end(); p++)
            l2p.insert( *p );
        // Olreit!
        polarization_table.push_back( polarization_entry_type(polid, j2, l2p) );
    }
    cout << "Loaded " << polarization_table.size() << " entries from POLARIZATION" << endl;

    // Load DATA_DESCRIPTION table
    ROMSDataDescColumns   ddcols( ms.dataDescription() );
    datadescription_table.clear();

    for(uInt ddid=0; ddid<ddcols.nrow(); ddid++) {
        // Don't consider flagged entries
        if( ddcols.flagRow()(ddid) )
            continue;
        datadescription_table.push_back(datadescription_entry_type((Int)ddid, ddcols.spectralWindowId()(ddid), ddcols.polarizationId()(ddid)));
    }
    cout << "Loaded " << datadescription_table.size() << " entries from DATA_DESCRIPTION" << endl;

    // Load ANTENNA table ...
    ROMSAntennaColumns  antcols( ms.antenna() );
    antenna_table.clear();
    for(Int antid=0; antid<(Int)antcols.nrow(); antid++) {
        // Don't consider flagged entries
        if( antcols.flagRow()(antid) )
            continue;
        antenna_table.push_back(antenna_entry_type(antid, antcols.name()(antid), antcols.station()(antid), antcols.mount()(antid),
                                                   antcols.dishDiameter()(antid), antcols.position()(antid), antcols.offset()(antid)));
    }
    cout << "Loaded " << antenna_table.size() << " entries from ANTENNA" << endl;

    // Load FEED table
    ROMSFeedColumns     feedcols( ms.feed() );
    feed_table.clear();
    for(uInt feed=0; feed<feedcols.nrow(); feed++) {
        // Generate the logical-to-physical mapping for this feed
        j2ms::log2polmap                         l2p;
        casacore::Int const                      nReceptor( feedcols.numReceptors()(feed) );
        casacore::Vector<casacore::String> const polType( feedcols.polarizationType().get(feed) );

        // Assert that the length of the vector == what it sais in num_receptor
        if( polType.shape()!=casacore::IPosition(1, (casacore::uInt)nReceptor) )
            THROW_JEXCEPT(String("Inconsistent FEED table entry #")+Repr(feed)+": len(POL_TYPE) "+Repr(polType.shape())+" != NUM_RECEPTORS ["+Repr(nReceptor));
        // Loop over the polarization types
        for(casacore::Int pol=0; pol<nReceptor; pol++)
            if( polType(pol).length()==1 )
                l2p.insert( std::make_pair(pol, j2ms::char2polcode(polType(pol).at((size_t)0))) );
        // I think that's all we need
        feed_table.push_back(feed_entry_type(feedcols.antennaId()(feed), feedcols.feedId()(feed), feedcols.spectralWindowId()(feed),
                                             feedcols.time()(feed), feedcols.interval()(feed), l2p));
    }
    cout << "Loaded " << feed_table.size() << " entries from FEED" << endl;

    // Load FIELD table!
    ROMSFieldColumns    fieldcols( ms.field() ); 
    field_table.clear();
    for(uInt field=0; field<fieldcols.nrow(); field++) {
        // skip flagged entries
        if( fieldcols.flagRow()(field) )
            continue;
        // Extract the directions and assert their shapes are consistent:
        //      (2, NUM_POLY+1)
        casacore::IPosition const   dirShape(2, 2, fieldcols.numPoly()(field)+1);
        fielddir_type const         delayDir( fieldcols.delayDir().get(field) );
        fielddir_type const         phaseDir( fieldcols.phaseDir().get(field) );
        fielddir_type const         referenceDir( fieldcols.referenceDir().get(field) );

        if( delayDir.shape()!=dirShape )
            THROW_JEXCEPT(string("Field #")+repr(field)+" delayDir has inconsistent shape "+repr(delayDir.shape())+" (expect: "+repr(dirShape)+")");
        if( phaseDir.shape()!=dirShape )
            THROW_JEXCEPT(string("Field #")+repr(field)+" phaseDir has inconsistent shape "+repr(phaseDir.shape())+" (expect: "+repr(dirShape)+")");
        if( referenceDir.shape()!=dirShape )
            THROW_JEXCEPT(string("Field #")+repr(field)+" referenceDir has inconsistent shape "+repr(referenceDir.shape())+" (expect: "+repr(dirShape)+")");
        // OK that's enough checking I guess
        field_table.push_back(field_entry_type((Int)field, fieldcols.sourceId().get(field), fieldcols.name().get(field),
                                               fieldcols.time().get(field), delayDir, phaseDir, referenceDir));
    }
    cout << "Loaded " << field_table.size() << " entries from FIELD" << endl;

    // Load the STATE table
    ROMSStateColumns  statecols( ms.state() );
    state_table.clear();
    for(uInt state=0; state<statecols.nrow(); state++) {
        // skip flagged entries
        if( statecols.flagRow()(state) )
            continue;
        state_table.push_back( state_entry_type((Int)state, statecols.sig().get(state), statecols.ref().get(state),
                                                statecols.cal().get(state), statecols.load().get(state),
                                                statecols.subScan().get(state), statecols.obsMode().get(state)) );
    }
    cout << "Loaded " << state_table.size() << " entries from STATE" << endl;
}


// Objective: come up with data_desc_id for the current visibility.
// At the moment only freq_group_name ("mode" from the vex file)
// and subband number (int) were used to say "yup we got that"
//
// But that's all wrong
//
// We must link spectral_window_id + polarization_id to one data_desc_id
//
// spectral_window_id: 
//    primary key here is the chan_freqs, really
//    but that's expensive. instead make new primary key:
//      #spectral channels
//      total bandwidth
//      channel edge frequency (take care of USB/LSB)
//    those three can be easily matched 
//
//  Then look up correlations in the polarization table 
//
//  => bob == uncle
//
//  The REAL primary key here is:
//     mode, correlations, subbandnr, nSpectralPoints, bandwidth, freq0, sideband
//  and it should map to:
//     data_desc_id
//
//  Behind the scenes we must keep:
//   SPWID:
//      mode(==freqGroupName), nSpectralPoints, bandwidth, freq0, sideband
//   POLARIZATIONID:
//      nPol, polarization_products
//   DATADESCID:
//      SPWID,POLARIZATIONID
Int JIVEMSFiller::updateSpectralAndPolarizationTable( MS& msref,
						      const Visibility& vref,
						      const j2ms::Scan& scanref )
{
    //  Before we spend to much time, verify a couple of things...
    ExtVisLabel const&      label( vref.getLabel() );
    Experiment const&       expref( scanref.getExperiment() );
    FrequencyConfig const*  freqConf( scanref.getFrequencyConfig() );
    BaseBand const*         curBB( (freqConf==0 ? 0 : freqConf->getBaseBand(label.subbandnr)) );
    CorrelatorSetup const&  correlatorsetup( scanref.getVisibilityBuffer().correlatorSetup() );

    // If the scan's frequency config does not exist we're in trouble,
    // id. for the baseband
    if( freqConf==0 ) 
        THROW_JEXCEPT(String("No FrequencyConfig for mode ")+scanref.getFrequencyCode()+" in experiment "+expref.getCode());
    if( curBB==0 )
        THROW_JEXCEPT(String("Failed to locate subband #")+Repr(label.subbandnr)+" in mode "+scanref.getFrequencyCode()+" in experiment "+expref.getCode());

    //bbprops const      bbProps( *curBB );
    // Step 1) Locate the/an appropriate SpectralWindow, updating the table
    //         if necessary
    unsigned int const nSpectralPoints( correlatorsetup.nrOfSpectralPoints() );
    spw_entry_type     cur((Int)label.subbandnr, freqConf->getCode(), freqConf->getFreqGroupNr(), nSpectralPoints,
                           curBB->getStartfreq(), curBB->getBandwidth(),
                           ((curBB->getSideBand() == BaseBand::lower) ? 0 :(curBB->getSideBand() == BaseBand::upper ? 1 : -1)));

    spectral_window_table_type::const_iterator curSPW = std::find(spectral_window_table.begin(), spectral_window_table.end(), cur);

    if( curSPW==spectral_window_table.end() ) {
        // Must add ROW to Spectral Window Table
        // - all channels have the same width
        Double const            channelwidth( cur.__m_TotalBandwidth/cur.__m_nSpectralPoints );
        Array<Double>           centerfreqs( IPosition(1, cur.__m_nSpectralPoints) );
        MSSpectralWindow&       msspwintable( msref.spectralWindow() );
        MSSpWindowColumns       spwcol( msspwintable );
        uInt const              newSpwId( msspwintable.nrow() );
        Array<Double> const     chwidths(IPosition(1, cur.__m_nSpectralPoints), channelwidth);
        
        //  Compute the centerfrequencies. Take care of upper/lower side band
        //  - we rely on the fact that code befor has asserted that sideband
        //    is either "0" (lower) or "1" (upper)
        for( Int i = 0, offset = ((cur.__m_NetSideband==1) ? 0 : (1-(Int)cur.__m_nSpectralPoints)); i<(Int)cur.__m_nSpectralPoints; i++)
            centerfreqs( IPosition(1, (uInt)i) ) = cur.__m_f0 + (i + offset)*channelwidth;

        // I think we now have all relevant info
        // - Add to the MS Spectral Window first
        msspwintable.addRow();

        spwcol.numChan().put( newSpwId, (Int)cur.__m_nSpectralPoints );
        // we do not give names to the spectral windows
        spwcol.name().put( newSpwId, "" );
        spwcol.refFrequency().put( newSpwId, 0.0);
        spwcol.chanFreq().put( newSpwId, centerfreqs );	

        //  Frequency reference? Heck.. i don't know... just put in
        //  RESTfrequency..
        spwcol.measFreqRef().put( newSpwId, (Int)MFrequency::REST );
        spwcol.chanWidth().put( newSpwId, chwidths );	
        spwcol.effectiveBW().put( newSpwId, chwidths );
        spwcol.resolution().put( newSpwId, chwidths );	
        spwcol.totalBandwidth().put( newSpwId, cur.__m_TotalBandwidth );
        spwcol.netSideband().put( newSpwId, cur.__m_NetSideband );
        spwcol.ifConvChain().put(newSpwId, 0 );
        spwcol.freqGroup().put( newSpwId, cur.__m_FreqGroupNr );
        spwcol.freqGroupName().put( newSpwId, cur.__m_FreqGroup );
        spwcol.flagRow().put( newSpwId, False );

        msspwintable.flush();
        // Also add the entry to cached version of the spectral window table
        curSPW = spectral_window_table.insert(spectral_window_table.end(), spw_entry_type(newSpwId, cur));
    } 

    // Step 2) Locate the/an appropriate POLarization ID, updating
    //         the POLARIZATION table if necessary
    //   correlatorsetup.getCorrelations() = ordered set
    //   in the cache we keep
    //      correlations_t
    //   which is unordered
    correlations_t                          tblcor;
    correlationset_t const&                 corcor( correlatorsetup.getCorrelations() );
    std::copy(corcor.begin(), corcor.end(), std::back_inserter(tblcor));

    polarization_entry_type const           curpol(-1, tblcor, correlatorsetup.l2pmap());
    polarization_table_type::const_iterator curPOL = std::find(polarization_table.begin(), polarization_table.end(), curpol);

    if( curPOL==polarization_table.end() ) {
        // Must add entry to POLARIZATION table
        Vector<Int>              corrType( curpol.__m_correlations.size() );
        Matrix<Int>              corrProd( 2, curpol.__m_correlations.size() );
        MSPolarization&          mspoltable( msref.polarization() ); 
        MSPolarizationColumns    polcols( mspoltable );
        uInt const               newPolId( mspoltable.nrow() );

        // 1) Convert internal j2ms2 values to MS values
        //  [j2ms CorrelationCode to casacore::StokesType]
        //  The aips++ Stokes stuff support constructing from string
        //  and the j2ms stuff (the current correlation) supports
        //  printing as string...
        // 2) From the correlations + log2polmap form the array of array
        // indices (corrProd)
        for(uInt i=0; i<curpol.__m_correlations.size(); i++) {
            corrType[i] = (Int)Stokes::type( Repr(curpol.__m_correlations[i]) );
            for(uInt j=0; j<2; j++)
                corrProd(j, i) = curpol.__m_l2p.index_of( curpol.__m_correlations[i][j] );
        }
        // OK values are now prepared!
        mspoltable.addRow();
        polcols.flagRow().put( newPolId, False );
        polcols.numCorr().put( newPolId, (Int)curpol.__m_correlations.size() );
        polcols.corrType().put( newPolId, corrType);
        polcols.corrProduct().put( newPolId, corrProd );

        mspoltable.flush();
        // Also add into cache, of course
        curPOL = polarization_table.insert(polarization_table.end(),  polarization_entry_type((Int)newPolId, curpol));
    }

    // 3) Now all we need is to figure out of there already is a data
    //    description ID for this specific combination of SPW + POL
    datadescription_entry_type                 ddid(-1, curSPW->__m_spwId, curPOL->__m_polarizationId);
    datadescription_table_type::const_iterator curDD = std::find(datadescription_table.begin(), datadescription_table.end(), ddid);

    if( curDD==datadescription_table.end() ) {
        // Nope. Must add row to data description table
        MSDataDescription&  msddtable( msref.dataDescription() ); 
        MSDataDescColumns   ddcols( msddtable );
        uInt const          newDDId( msddtable.nrow() );

        // We don't need to do any pre-processing for this entry
        msddtable.addRow();
        ddcols.flagRow().put( newDDId, False );
        ddcols.polarizationId().put( newDDId, ddid.__m_PolarizationId );
        ddcols.spectralWindowId().put( newDDId, ddid.__m_SpectralWindowId );
        msddtable.flush();

        // And also add to cache, obv.
        curDD = datadescription_table.insert(datadescription_table.end(), datadescription_entry_type((Int)newDDId, ddid));
    }
    return curDD->__m_dataDescId;
}


// UPDATES the ANTENNA and FEED tables for the given experiment into the
// indicated MS.
//
// HV: 09/May/2018 There is no reason at all why every antenna in the MS should match
//                 like for like with the list in the current experiment.
//                 Also the handling of FEEDs was ... dodgy at best I think.
//                 At the moment we make sure that all POLARIZATION indices
//                 are consistent between POLARIZATION + FEED subtables.
//                 This could not be guaranteed before.
//                 The whole mess about how polarization is indexed and
//                 referenced is explained below.
//                 I'll summarize here:
//
//                  - there are two locations where polarization labels and
//                    indices are mentioned: the Polarization and Feed
//                    subtables. One is antenna independent and the other
//                    isn't.
//
//                  - In Polarization there is a list of CASA StokesType
//                    enum values, of length NUM_CORR which, in principle, map to e.g.:
//                      [5, 8, 6, 7] => ["RR", "LL", "RL" , "LR"]
//                          (note: the values may be wrong but you get the idea)
//                    also there is a matrix of 2 x NUM_CORR of 0's and 1's, 
//                    corresponding to the polarization product indices:
//                      [(0,0), (1,1), (0,1), (1,0)]
//                    From this we might infer: 
//                      polarization index 0 == 'R' and 1 == 'L'
//                    but this is /antenna independent/!
//
//                  - In Feed there is, per antenna, a list of receptors
//                    defined for a particular feed. The types of receptors are
//                    listed as string, say like this:
//                      ["L", "R"]
//                    From this we might infer: 
//                      polarization index 0 == 'L' and 1 == 'R'
//
//                  Clearly this FEED definition cannot be used together
//                  with the POLARIZATION example because it would lead to
//                  an inconsistent mapping of index to physical
//                  polarization.
//
//                  The new code therefore attempts to find a FEED
//                  definition for an ANTENNA which is consistent with the
//                  polarization mapping of the current correlation. [It is
//                  sufficient if the sought polarization mapping is a
//                  subset of the FEED definition!]
//
//                  If no such FEED can be found, a new entry is added to
//                  the Feed table such that *that* feed /is/ consistent
//                  with the current Polarization setup.
//
// HV: 24/08/2006  Polarization handling reworked. 
//                 In the MS 2.0 defn the combination
//                 between FEED/POLARIZATION/DATA_DESCRIPTION
//                 is somewhat ... non-orthogonal.
//                 It boils down to what's said in the comment above this one
antennafeed_pair_type JIVEMSFiller::updateAntennaAndFeedTable(MeasurementSet& msref, unsigned int ant, j2ms::Scan const& scanref) {
    AntennaInfo             ai;
    Experiment const&       expref( scanref.getExperiment() );
    AntennaDB const&        antdb( expref.getAntennaDB() );
    // Note: make a COPY of "scanref.getVis....Setup().l2pmap()"!
    //       even though ".l2pmap()" returns const ref to data member of the
    //       correlator setup it makes the thing in here sigsegv if we
    //       iterate over it. Weird.
    j2ms::log2polmap const  l2p( scanref.getVisibilityBuffer().correlatorSetup().l2pmap() );

    // Step 1. Find the antenna in the experiment's database
    if( !antdb.searchAntenna(ai, (const Int)ant) )
        THROW_JEXCEPT(string("Request for experiment antenna #")+repr(ant)+" fails, no such antenna!");

    // Now we can form the antenna properties that we need to look for in the MS
    antenna_entry_type const           antenna(-1, ai.getName(), ai.getStation(), ai.getMount(), ai.getDiameter(),
                                               ai.getPosition().get("m").getValue(), ai.getOffset());
    antenna_table_type::const_iterator curANT = std::find(antenna_table.begin(), antenna_table.end(), antenna);

    // No such antenna??? Need to add a new one
    if( curANT==antenna_table.end() ) {
        MSAntenna&              antab( msref.antenna() );
        MSAntennaColumns        antabcols( antab );
        casacore::uInt const    newAntId( antab.nrow() );

        antab.addRow();
		antabcols.name().put( newAntId, antenna.__m_AntennaName);
		antabcols.dishDiameter().put( newAntId, antenna.__m_Diameter );
		antabcols.mount().put( newAntId, antenna.__m_Mount );
		antabcols.offset().put( newAntId, antenna.__m_Offset );
		antabcols.position().put( newAntId, antenna.__m_Position );
		antabcols.station().put( newAntId, antenna.__m_Station );
		antabcols.flagRow().put( newAntId, False );
		antabcols.type().put( newAntId, "GROUND-BASED" );
        antab.flush();
        // Do not forget to add the entry to the in-memory cache
        curANT = antenna_table.insert(antenna_table.end(), antenna_entry_type((Int)newAntId, antenna));
    }

    // Step 2. Find a FEED for the current antenna which is consistent with
    //         the current polarization mapping [and valid for all SPW, all
    //         times]
    //         Note: we already compile the set of feed ids defined for this
    //         antenna (because we're already iterating over them) such that
    //         IF we need to add a new one we can fairly easy find a new ID
    typedef std::set<casacore::Int> feedidset_type;
    feedidset_type                  feedidset;
    feed_table_type const           feeds = feeds_by_antenna(curANT->__m_AntennaId, feed_table);
    feed_table_type::const_iterator curFEED;

    for(curFEED = feeds.begin(); curFEED!=feeds.end(); curFEED++) {
        // It is a feed defined for the requested antenna
        feedidset.insert(curFEED->__m_FeedId);

        // But is this feed compatible at all? For that to be true, the
        // log2polmap of the current scan must be a subset of the FEED
        j2ms::log2polmap const&  feedl2p( curFEED->__m_l2p );
        if( !std::includes(feedl2p.begin(), feedl2p.end(), l2p.begin(), l2p.end()) )
            continue;
        // OK. Is this feed valid for all SPW?
        if( curFEED->__m_SpectralWindowId!=-1 )
            continue;
        // Great. But is it valid for all time ranges?
        // [either time or interval > 0]
        if( curFEED->__m_Time>0 || curFEED->__m_Interval>0 )
            continue;
        // 'k found it
        break;
    }
    if( curFEED==feeds.end() ) {
        // Oh. Crap. No compatible feed!
        // Must write a new entry.
        // First do some analysis on the l2p map that we must write.
        // l2p contains entries like:
        //      pol<index> => physical pol 
        // so l2p with size 1 (1 entry) could contain:
        //      pol0       => [RLXY]
        // OR
        //      pol1       => [RLXY]
        // But in MS we must write entries that are indexed by, well, linear
        // index. So the "pol1" in the 2nd possibility must have the actual
        // [LRXY] in it's 1st index (or else the mapping would be broken)
        //
        // So we get the max over the polX indices, *that*'s the size of the
        // arrays we must write ...
        casacore::Int                    nReceptor( -1 );
        j2ms::log2polmap::const_iterator curl2p;

        // l2p map indices are 0-based, so the number of indices is n+1
        for(curl2p = l2p.begin(); curl2p!=l2p.end(); curl2p++)
            nReceptor = std::max(nReceptor, curl2p->first+1);
        if( nReceptor<=0 )
            THROW_JEXCEPT(String("log2polmap for scan ")+scanref.getScanId()+" in experiment "+expref.getCode()+" seems to be empty?!");

        // Now we can properly size our arrays 
        casacore::Vector<String>  polType( IPosition(1, nReceptor) );

        // only fill the entries we know about ...
        for(curl2p = l2p.begin(); curl2p!=l2p.end(); curl2p++)
            polType(curl2p->first) = Repr(curl2p->second);

        // We must conjure up a new feed id. Start with preferred value of 0
        casacore::Int                 newFeedId( 0 );

        // If there is no feed id of 0 defined, grab it!
        if( std::find(feedidset.begin(), feedidset.end(), 0)!=feedidset.end() ) {
            // Dang. feed id 0 already in use.
            // Check if there is a hole in the sequence or wether we must
            // take the next bigger id.
            // Note: because we HAVE found index 0, it means that the
            //       feedidset is not empty. Thus there will be a min+max
            //       element. Also, because it's a set, min == *begin(),
            feedidset_type::const_iterator ptr = std::max_element(feedidset.begin(), feedidset.end());
            if( (*ptr - *feedidset.begin())!=(casacore::Int)feedidset.size() ) {
                // Hah! There *must* be a hole in the sequence. Because we
                // know the sequence is not empty we can unconditionally do
                // *begin() and also ++begin() w/o checking for ==.end()!
                for(ptr=feedidset.begin(), newFeedId=*ptr; ptr!=feedidset.end() && *ptr==newFeedId; ptr++, newFeedId++) {}
            } else {
                // Bummer. Must alloc the next highest feed id
                newFeedId = (*ptr)+1;
            }
        }

        // Now we have enough information to define a new feed_entry_type
        // Contrary to at other tables, we now add this one to memory first
        // such that when writing the MSFeed table we can take consistent
        // values from that entry and write them out
        curFEED = feed_table.insert(feed_table.end(), feed_entry_type(curANT->__m_AntennaId, newFeedId,
                                                                      -1/*all spectral windows!*/, 0.0, 0.0 /*no time dependency*/,
                                                                      l2p));
        // Now write the new FEED definition out to MSFeed
        MSFeed&          feedref( msref.feed() );
        uInt const       feedrow( feedref.nrow() );
        MSFeedColumns    feedcols( feedref);

        feedref.addRow();

		feedcols.antennaId().put( feedrow, curFEED->__m_AntennaId );
		feedcols.feedId().put( feedrow, curFEED->__m_FeedId  );
		feedcols.spectralWindowId().put( feedrow, curFEED->__m_SpectralWindowId );
		feedcols.time().put( feedrow, curFEED->__m_Time );
		feedcols.interval().put( feedrow, curFEED->__m_Interval );
		feedcols.numReceptors().put( feedrow, nReceptor );
		feedcols.polarizationType().put( feedrow, polType );
		//  [[ASSUMPTIONS]] No BEAM associated, no beam offset, perfect
        //  polarization response
		feedcols.beamId().put( feedrow, -1 );
		feedcols.beamOffset().put( feedrow, casacore::Matrix<casacore::Double>((size_t)nReceptor, (size_t)nReceptor, casacore::Double(0.0)) );

        // DTermn init with all zeroes, diagonal 1
        casacore::Matrix<casacore::Complex> dTerms((size_t)nReceptor, (size_t)nReceptor, casacore::Complex(0.0, 0.0));
        dTerms.diagonal() = casacore::Complex(1.0, 0.0);
		feedcols.polResponse().put( feedrow, dTerms );

		feedcols.position().put( feedrow, casacore::Vector<casacore::Double>((uInt)3, 0.0) );
		feedcols.receptorAngle().put( feedrow, casacore::Vector<casacore::Double>((size_t)nReceptor, casacore::Double(0.0)) );

        feedref.flush();
    }

    if( curFEED->__m_FeedId!=0 )
        cout << "WARN: Using non-standard FEEDid#" << curFEED->__m_FeedId << " for ANTENNA #" << curFEED->__m_AntennaId << " [" << curANT->__m_AntennaName << "]!!!!" << endl;
    return antennafeed_pair_type(curFEED->__m_AntennaId, curFEED->__m_FeedId);
}

const Matrix<Complex>& JIVEMSFiller::makeData( const Experiment& expref,
                                               const Visibility& vref,
                                               rowmap_type const& rowmap/*empty means: no remap, otherwise rowmap[i]=j means src[i] => dest[j]*/ )
{    //  These are always there....
    static Matrix<Complex>          srcdata;
    static Matrix<Complex>          fftdata;
    static Matrix<Complex>          resdata;
    static FFTServer<float,Complex> transformator;

    //  Cache the visibility datadomains
    bool                    needfft;
    Visibility::domain      srcdomain( vref.getDatadomain() );
    Visibility::domain      dstdomain( expref.getDestinationDomain() );
    Matrix<Complex> const&  visdata( vref.getData() );

    // pre-empty the result data for 'error' return values
    resdata.resize(0, 0);

    //  Perform sanity checks
    if( srcdomain==Visibility::undefined || dstdomain==Visibility::undefined ) 
        return resdata;
  
    needfft = (dstdomain!=srcdomain);
    
    // If we have to do something but we
    // cannot do it because of lack of info, we return the empty
    // data as well.
    // There will be other cases of when we do not have enough
    // info to complete the request but we cannot do that in one
    // statement.
    if( needfft &&
        (!vref.hasVisibilityBuffer() || 
         (vref.hasVisibilityBuffer() && !vref.getVisibilityBuffer().getScanptr())) ) 
        return resdata;

    //  Locate the baseband we're looking at....
    //  If we cannot find it or it lacks certain
    const BaseBand*          bbptr( 0 );
    FrequencyConfig          fc;
    const j2ms::Scan&        scanref( *(vref.getVisibilityBuffer().getScanptr()) );
    const FrequencyDB&       freqdb( expref.getFrequencyDB() );
  
    // At this point we still may or may not need to do a FFT.
    // The following tests basically only apply when we need to
    // do a FFT
    if( needfft ) {
        if( !freqdb.searchFrequencyConfig(fc, scanref.getFrequencyCode()) )
            return resdata;
    
        if( (bbptr=fc.getBaseBand(vref.getLabel().subbandnr))==0 )
            return resdata;
    
        if( bbptr->getSideBand()==BaseBand::none )
            return resdata;
    }

    // NOW we have arrived at a point where we're certain that WHATEVER we
    // need to do (FFT or not) we have sufficient info to DO it.
    //
    // Whatever we need to do we MUST reorder (potentially...) the visibility
    // datamatrix in order to guarantee its rows are in the order the
    // experiment dictates.
    //  The simplest case is when the visibility data-domain and the
    //  desired output domain match...

    // HV: As of January 2007 decided to let the visibilitybuffer 
    //     deliver the data in a predefined order, we are not going
    //     to mess around with it; we just take the data as-is.
    //     We just copy it across to our working array...
    // HV: June 2018 - reverted back; we must potentially do some
    //     reshuffling of the rows. Assume that the 'rowmap'
    //     empty implies: no reshuffling, otherwise we do
    //     shuffling + conjugating
    if( rowmap.empty() ) {
        // no reshuffling necessary.
        // if there's also no fft'ing necessary, we're done quickly!
        if( !needfft )
            return visdata;
        // be sure to resize the 'srcdata' thingy otherwise the 
        // assignment will throw at some point...
        srcdata.resize( IPosition(2,0,0) );
        srcdata = visdata.copy();
    } else {
        // Irrespective of wether we need to FFT or not, we must make a copy
        // of the data because we need to shuffle the rows
        srcdata.resize( visdata.shape() );
        // step 1: reshuffle + conjugate the rows
        for(unsigned int i=0; i<rowmap.size(); i++)
            srcdata.row( rowmap[i] ) = casacore::conj(visdata.row(i));
        // step 2: reverse each row as per Examples in
        //         https://casacore.github.io/casacore/structcasacore_1_1ArrayUtil__global__functions__reverseArray.html
        //         Note: the documentation in there is inconsistent:
        //         """Reversing axis 1 of a Matrix means that its rows are
        //            reversed. <snip>. Reversing axis 1 of a Matrix means
        //            that its columns are reversed."""
        //         Since casacore Matrix is (row, column) I'm going with
        //         "axis 0 reverses the rows, axis 1 reverses the columns"
        //
        srcdata = casacore::reverseArray(srcdata, 1);

        // if no further processing required, we're done
        if( !needfft )
            return srcdata;
    }
    
    //  If we arrive at this point in the program we've established
    //  a couple of facts:
    //
    //  1) source and destination domain for the data are NOT the
    //  same, hence we *need* to do a FFT
    //
    //  2) neither source nor destination domain is unknown, so no
    //  surprises are to be expected from those
    //
    //  3) we have located the baseband this data belongs to so we
    //  know exactly which part of the data we need to cut out (*if*
    //  we have to cut, that is)
    uInt                     fftsize;
    uInt                     nrffts;
    Bool                     expand;
    Bool                     fftflag;
    const Matrix<Complex>*   retvalptr;
    IPosition                srcshape( srcdata.shape() );
    
    //  What we do is: check if we need to expand the src data-matrix,
    //  i.e. if the source dat is an autocorrelation and the data is
    //  in time-domain, we need to expand the data (because we have
    //  already verified that the destination domain and the source
    //  domain are not equal and since the source domain (in this
    //  case) is time, the other must be frequency).
    nrffts  = srcdata.nrow();
    fftsize = srcdata.ncolumn();
    
    //  We need to expand the data if either of the following
    //  conditions holds:
    //
    //  1) we're going from time->frequency AND this is an
    //  autocorrelation (Albert gives us only half the number of lags
    //  in this case)
    //
    //  2) we're going from frequency->time domain
    //
    //  NOTE: the following two boolean variables will tell which (if
    //  any) of the two aforementioned conditions holds, they can
    //  *never* hold both (we make use of this property later on)
    Bool       expandcondition1;
    Bool       expandcondition2;
    
    //  30-09-1999: HV, explicitly set expandcondition 1 to False
    //  since this is (from this date on) already covered by the
    //  COFVisibilityBuffer class. Remains to be done for the
    //  DTDVisibilityBuffer!!! We can expect problems if people start
    //  reading data from this format! Changes to that are slim so
    //  forget about it for now!
    //
    //
    //expandcondition1 = ( vref.isAutoCorrelation() && srcdomain==Visibility::time );
    expandcondition1 = False;
    
    expandcondition2 = ( srcdomain==Visibility::frequency );
    
    if( (expand=(expandcondition1 || expandcondition2))==True )
	    fftsize *= 2;
    
    //  Now we can properly resize and initialise the 'fftdata' matrix
    fftdata.resize( IPosition(2,nrffts,fftsize) );
    fftdata = Complex( 0.0f, 0.0f );
    
    //  If we need to expand, do it now!
    if( expand ) {
        uInt     nr2copy( srcdata.ncolumn() );
	
        //  Depending on which expandcondition holds we take action
        if( expandcondition1 ) {
	        //  Autocorrelation && time->frequency
	        for( uInt i=0; i<nrffts; i++ ) {
                //  Copy all channels but channel0
                for( uInt lagpoint=1; lagpoint<nr2copy; lagpoint++ ) {
                    fftdata( i, nr2copy+lagpoint ) = 
                    fftdata( i, nr2copy-lagpoint ) = srcdata( i, lagpoint );
                }
                //  Copy channel 0 into the middle of the fftarray
                fftdata( i, nr2copy ) = srcdata( i, 0 );
            }
        } else /* expandcondition2==true */ {
            //  frequency->time, need to check sideband to know where
            //  to copy the data to
            uInt    startpoint;
	    
            if( bbptr->getSideBand()==BaseBand::upper )
		        startpoint = nr2copy;
	        else
		        startpoint = 1;

            //  Ok. Copy stuff
            for( uInt i=0; i<nrffts; i++ )
                (fftdata.row(i))( Slice(startpoint, nr2copy) ) = srcdata.row(i);
	    }
	} else /* no expand needed */ {
	    fftdata = srcdata;
    }
    
    //  Ok. fftdata is now properly sized and filled in, now we
    //  determine the FFT-flag and do the FFT
    //
    //  fftflag (=toFrequency argument)
    //  
    //  True  implies forward transform == time -> frequency
    //  False implies backward transform == frequency -> time
    fftflag = ( srcdomain==Visibility::time );
    
    //  Do the fft
    //  Do this row by row. Because the casacore::Array (and derivatives)
    //  use reference semantics by default, taking out a
    //  row of a matrix and operating on that, modifies the
    //  storage underlying the matrix.
    for( uInt i=0; i<nrffts; i++ ) {
		Vector<Complex>   schijt( fftdata.row(i) );
		transformator.fft( schijt, fftflag );
    }
    
    //  Now we have to get out the data. If we were doing
    //  timetofrequency, we must keep only half of the resultant
    //  values. datastart tells us where to start in the fft'ed
    //  data. The value of datastart is determined by the sideband of
    //  the data: USB -> we take the upper half of the array, LSB ->
    //  the lower half.
    uInt     datastart( 0 );
    uInt     datasize( fftsize );
    
    if( fftflag ) {
        datasize /= 2;
	
        //  Which half to take depends on the sideband
        if( bbptr->getSideBand()==BaseBand::upper ) 
            datastart = datasize;
        else
            datastart = 1;
    }
    
    // datasize!=0 && datasize unequal to fftsize this means that we
    // have to <snip> the array...
    if( datasize && datasize!=fftsize ) {
        //  Now we can resize and fill the 'resdata'-matrix
        resdata.resize( nrffts, datasize );

        for( uInt i=0; i<nrffts; i++ ) {
            //resdata.row(i) = (fftdata.row(i))( Slice(datastart, datasize) );
            for( uInt j=0; j<datasize; j++ )
		        resdata( i, j ) = fftdata( i, datastart+j );
	    }
        retvalptr = &resdata;
    } else if( datasize ) {
        retvalptr = &fftdata;
    } else /* error condition?! */ {
        // make really sure resdata is empty
        resdata.resize(0, 0);
	    retvalptr = &resdata;
    }

    //  Return the result....
    return *retvalptr;
}

const Vector<Double>& JIVEMSFiller::calcUVW( const Visibility& vref  ) {
    //  These statics are here to speed up UVW-calculation; we don't need to 
    //  recreate these each call
    static Double         gastInRadians;
    static Double         currentHA;
    static MEpoch         currentGAST;
    static MEpoch         currenttime;
    static RotMatrix      currentRotMatrix;
    static Vector<Double> sourceRADEC;
    static Vector<Double> uvw( 3 );
    static Vector<Double> ant1xyz( 3 );
    static Vector<Double> ant2xyz( 3 );
    static Vector<Double> baseline( 3 );    

    //  Init retval to a neat zero-value
    uvw = j2ms::noUVWv;

    if( !vref.hasVisibilityBuffer() ||
    	(vref.hasVisibilityBuffer() &&
         !vref.getVisibilityBuffer().getScanptr()) ) {
    	return uvw;
    }
   
    // at least there is something known about the visibility, try to 
    // go on
    const j2ms::Scan&       scanref( *(vref.getVisibilityBuffer().getScanptr()) );
    const Experiment&       itsexpref( scanref.getExperiment() );
    ScanCacheElement*       scancacheptr;
    const ExtVisLabel&      labelref( vref.getLabel() );
    ExperimentCacheElement* expcacheptr( theirCache.addExperiment(itsexpref) );
    
    //  (attempt to)locate the Scan in the Scan-cache, since that entry contains
    //  all the conversion stuff needed for the coordinate
    //  transformations.
    if( !expcacheptr )
    	return uvw;
    scancacheptr = expcacheptr->addScan( scanref ); 

    if( !scancacheptr )
    	return uvw;

    //  Verify that we can find the antennainfo....
    const AntennaDB&      antdb( itsexpref.getAntennaDB() );
    const SourceInfo&     sourceinforef( scancacheptr->itsSourceInfo );
    AntennaInfo           ant1, ant2;
    
    if( !antdb.searchAntenna(ant1, labelref.baseline.idx1()) ||
        !antdb.searchAntenna(ant2, labelref.baseline.idx2()) )
        return uvw;

    //  Ok. Convert the current visibility time to GAST (Greenwich Sidereal Time)
    currenttime = vref.getTime();
    currentGAST = scancacheptr->timeCvt( currenttime );

    //  Get the source's RA in radians.
    //
    //  NOTE: the source position should be precessed to the current
    //  visibility epoch!! Leave that for later (we're using 2000
    //  coordinates, mostly, so the difference should be very
    //  little...
    sourceRADEC = ((sourceinforef.getDirection()).getValue()).get();

    //  Convert GAST to radians as well.... (in order to be able to
    //  derive the HA as seen from Greenwich)
    //
    //  1) start off with the fractional day in GAST
    //  2) multiply by 2pi to get at an angle in radians
    gastInRadians  = (currentGAST.getValue()).getDayFraction();
    gastInRadians *= ( 2*C::pi );

    //  Now we can form the HA
    currentHA = gastInRadians - sourceRADEC(0);
    
    // Ok. Now attempt to form the baseline
    //
    // 1) get the antenna XYZ-positions
    // 2) subtract them to get the baseline vector
    ant1xyz = ((ant1.getPosition()).getValue()).getValue();
    ant2xyz = ((ant2.getPosition()).getValue()).getValue();
    
    baseline = ant2xyz - ant1xyz;
    
    //  Ok. All that's left is a) to set up the rotation matrix and b)
    //  do the rotation!
    currentRotMatrix = RotMatrix( Euler(-(C::pi_2-sourceRADEC(1)), 1, (currentHA - C::pi_2), 3) );

    //  Finally: do the rotation!
    uvw = currentRotMatrix*baseline;

    return uvw;
}


//  HV:  21-07-2000  With the arrival of MS v 2.0 I decided to write the 
//                   field table in one go. 
//       29-01-2007  Throws up if a problem. Verifies that, _if_ there are
//                   rows in the field-table, they are the same as in the
//                   experiment. FIELD_ID column in the main table
//                   is a direct index by now so the same comments as
//                   under the antenna table apply.
//                   Also stopped writing the SOURCE table. It is not 
//                   required nor does it really add more information
//                   (for our purposes)
//       18-05-2018  Build the table dynamically again ...
//                   This helps in leaking meta data from a multi-experiment
//                   VEX file into the measurement sets because now only the
//                   fields whose data is actually written into the MS are
//                   exported into the field table
casacore::Int  JIVEMSFiller::updateFieldTable( MeasurementSet& msref, const j2ms::Scan& scanref ) {
    SourceInfo              si;
    Experiment const&       expref( scanref.getExperiment() );
    SourceDB const&         srcdb( expref.getSourceDB() );
    casacore::String const  srccode( scanref.getSourceCode() );


    // Step 1. Find the source's information
    if( !srcdb.searchSource(si, srccode) )
        THROW_JEXCEPT(String("Request for experiment source '")+srccode+"' fails, no such source?!");

    // Do some pre-processing on the listed source direction in order to be
    // able to look for a correctly matching entry.
    // MDirection should yield vector of length 2: RA, DEC
    Vector<Double> const      srcdir( (si.getDirection().getValue()).get() );
    if( srcdir.size()!=2 )
        THROW_JEXCEPT(String("The source '")+srccode+"' direction is not just RA, DEC?!");
    // In the Field table it is allowed to have a direction as a polynomial
    // series in time with NUM_POLY coefficients, written as a matrix of
    // 2xNUM_POLY (Ra, Dec) * [num_poly + 1] coefficients.
    // casacore Matrix<T> is (nrow, ncolumn) so the direction matrices look:
    //     RA , RA(t)_0 , RA(t)_1 , ...  [expansion coefficients of RA(t) ]
    //     DEC, DEC(t)_0, DEC(t)_2, ...  [expansion coefficients of DEC(t)]
    //
    // Our code assumes the sources have constant position, thus NUM_POLY==0
    // but we must list NUM_POLY+1 "coefficients", including the 0th order
    // (constant term) 
    //
    // Let's see if we can go from Vector => properly shaped Matrix in one
    // fell sweep
    fielddir_type const fielddir( srcdir.reform( IPosition(2, 2/*ra, dec*/, 1/*only 1 term in the polynomials*/) ) );

    // Now we can build the field table entry we're looking for.
    // Note that our code always looks for [and inserts if not found] an entry with:
    //  * time = 0.0  [must be valid for all times]
    //  * delay_dir == phase_dir == reference_dir == source position from SourceDB
    //  * source_id == -1 [no associated entry in the source table]
    field_entry_type const           field(-1 /*No field id yet!*/, -1 /*No associated source*/,
                                           srccode, casacore::Double(0.0) /*Valid for all times*/,
                                           fielddir, fielddir, fielddir /*All three directions equal*/);
    field_table_type::const_iterator curFIELD = std::find(field_table.begin(), field_table.end(), field);

    // Ah?
    if( curFIELD==field_table.end() ) {
        // FIELD_ID in main is a direct index
        MSField&             fieldref( msref.field() );
        MSFieldColumns       fieldcols( fieldref );
        casacore::uInt const newFieldId( fieldref.nrow() );

        // Add row, fill it in and flush to disk
        fieldref.addRow();

        fieldcols.name().put( newFieldId, field.__m_FieldName );
        fieldcols.sourceId().put( newFieldId, field.__m_SourceId );
        fieldcols.time().put( newFieldId, field.__m_Time );
        fieldcols.numPoly().put( newFieldId, 0 );
        fieldcols.delayDir().put( newFieldId, field.__m_DelayDir );
        fieldcols.phaseDir().put( newFieldId, field.__m_PhaseDir );
        fieldcols.referenceDir().put( newFieldId, field.__m_ReferenceDir );
        fieldcols.flagRow().put( newFieldId, False );

        fieldref.flush();

        // Now all that's needed is to update our /internal/ copy of the
        // field table with a slightly updated field entry; now we have
        // assigned a new field_id to it
        curFIELD = field_table.insert(field_table.end(), field_entry_type((Int)newFieldId, field));
    }
    return curFIELD->__m_FieldId;
}

// Update the processor table with a new pass_id
//
// * NO CHECKING FOR DUPLICATES IS DONE HERE * 
// The return value
// will be the newly added 'PROCESSOR_ID'.... the caller could, if he/she
// so desired, keep a map from 'pass_id' -> 'processor_id' and only call
// this function if they need to write a new 'pass_id'... (just a hint)
Int JIVEMSFiller::updateProcessorTable( MeasurementSet& msref,
									    Int newpassid,
									    const String& subtype)
{
    //  Test if we can get out ASAP (without any intelligent testing
    //  whatsoever..)
	Int                retval;
    MSProcessor&       processor( msref.processor() );
    MSProcessorColumns proccols( processor );

    //  Write a single row...
	retval = processor.nrow();

    processor.addRow();
    
    proccols.type().put( retval, "CORRELATOR" );
    proccols.subType().put( retval, subtype );
    proccols.typeId().put( retval, 0 );
    proccols.modeId().put( retval, 0 );
    proccols.flagRow().put( retval, False );
    proccols.passId().put( retval, newpassid );
    processor.flush();

    return retval;
}


Bool JIVEMSFiller::updateObservationTable( MeasurementSet& msref,
										   const Experiment& expref )
{
    MSObservation&        observation( msref.observation() );
    MSObservationColumns  obscols( observation );

    //  If Observation has no rows, make it so...
    if( observation.nrow()==0 ) {
		char          buf1[1024];
		char*         basenameptr;
		Array<String> as( IPosition(1,1) );
		
		as( IPosition(1, 0) ) = "???";
		observation.addRow();
	
		//  Fill in basic stuff once
		obscols.telescopeName().put( 0, expref.getArrayCode() );
		obscols.timeRange().put( 0, Vector<Double>(2, 0.0) );
		obscols.observer().put( 0, "JIVE" );
		obscols.log().put( 0, as );
		obscols.scheduleType().put( 0, "VEX" );
	
		//  Create the vexfilename....
		//
		//  NOTE: We do not create the absolute path since that would
		//  be useless...
		::strncpy( buf1, expref.getRootDir().c_str(), sizeof(buf1) );
		
		as( IPosition(1, 0) ) = "<error in getting basename>";
	
		if( (basenameptr=::basename(buf1))!=0 ) {
		    ostringstream strm;

			strm << basenameptr << ".vix";
			
			as( IPosition(1, 0) ) = strm.str();    
		}
		obscols.schedule().put( 0, as );
		
		obscols.project().put( 0, expref.getCode() );
	
		//  Set target release date in four weeks time!
		//
		//MVTime      today( (Time()) );
		Time        t;
		MVTime      today( t );
		MVTime      fourweeks( Quantity(28.0, "d") );
		MVTime      target = (today.get()+fourweeks.get());
		Double      targetvalue = (target.get()).getValue("s");
	
		//obscols.releaseDate().put( 0, ((today+fourweeks).get()).getValue("s") );
		obscols.releaseDate().put( 0, targetvalue );
	
		obscols.flagRow().put( 0, False );
	}
    // Ok. The table already got a row filled in.
    // Now all that we do is updating the TIMERANGE column but we can
    // only do that if there are rows in the MS
    if( msref.nrow() ) {
        MSMainColumns    mscols( msref ); 
        Vector<Double>   times( mscols.time().getColumn() );
        Vector<Double>   tr( 2 );

        minMax( tr(0), tr(1), times );

        obscols.timeRange().put( 0, tr );
    }
    //  [[TODO]] Lets find out the min/max time? Consider this done as of 08/02/2002
    return True;
}

// May 2018: Decided to stop writing STATE table.
//           But since we can *append* to an existing MS we'll
//           fall back to:
//                    -1 if state empty
//              new row  if state non-empty ?
casacore::Int JIVEMSFiller::updateStateTable( MeasurementSet& msref, const Experiment& ) {
    MSState&             state( msref.state() );

    // If table empty, keep it that way
    if( state.nrow()==0 )
        return -1;

    // Look for specific state table entry
    state_entry_type const  theState(-1, casacore::True /*sig*/, casacore::False /*ref*/,
                                     casacore::Double(0.0) /*cal*/, casacore::Double(0.0) /*load*/,
                                     0 /*subscan*/, casacore::String("ON-SOURCE") /*obsmode*/);

    state_table_type::const_iterator curState = std::find(state_table.begin(), state_table.end(), theState);

    if( curState==state_table.end() ) {
        // Buggrit. Must add entry!
        casacore::uInt const newStateId( state.nrow() ); 
        MSStateColumns       statecols( state );

        state.addRow();

        statecols.sig().put( 0, theState.__m_Sig );
        statecols.ref().put( 0, theState.__m_Ref );
        statecols.cal().put( 0, theState.__m_Cal );
        statecols.load().put( 0, theState.__m_Load );
        statecols.subScan().put( 0, theState.__m_SubScan );
        statecols.obsMode().put( 0, theState.__m_ObsMode );
        statecols.flagRow().put( 0, casacore::False );

        state.flush();
        // And append to internal cache of the state table
        curState = state_table.insert(state_table.end(), state_entry_type((Int)newStateId, theState));
    }
    return curState->__m_StateId;
}

// entries in the modeltable have a unique key
// (explained further down below)
struct modelkey {
    casacore::Int    antenna_id;
    casacore::Double time;
	casacore::Int    processor_id;

    modelkey( casacore::Int ant, casacore::Double tm, casacore::Int proc ):
        antenna_id( ant ), time( tm ), processor_id( proc )
    {}
    
    private:
        // do not allow default modelkeys.
        modelkey();
};

// Less-than functor which implements
// the < operator for _modelkey thingies
// First order by antenna-id, then by time
struct _lt_modelkey {
    bool operator()( const modelkey& l, const modelkey& r ) const {
        return ( l.antenna_id<r.antenna_id ||
                 (l.antenna_id==r.antenna_id && l.time<r.time) ||
				 (l.antenna_id==r.antenna_id && l.time==r.time && 
				    l.processor_id<r.processor_id) );
    }
};

// this defines the modelcache. It is sufficient to
// remember if we have an entry for the modelkey
// since if that's granted, all info for all spectral windows
// is taken to be there
typedef std::set<modelkey,_lt_modelkey> modelcache_t;

// and an insert-result of inserting something into the cache
// which is pair<iterator,bool>, where the bool indicates if
// the insert was ok, if yes, then iterator points at the
// inserted element
typedef std::pair<modelcache_t::iterator,bool>  modelcacheinsertres_t;

// extract all the modeldata from the VisibilityBuffer and write it
// to the modeltable (if we didn't do that earlier)
#if 0
void JIVEMSFiller::writeModel( MeasurementSet& ms,  const VisibilityBuffer& vb ) {
    try {
        // if the VisibilityBuffer does not offer a model - don't do anything!
        unsigned int              nrWritten;
        unsigned int              alreadyInMS;
        modelcache_t              modelcache;
        const Experiment&         eref( vb.getExperiment() );
        ExperimentCacheElement*   elptr;
        const j2ms::modeldata_t&  model( vb.getModelData() );

        if( model.size()==0 ) {
            cout << "writeModel/No modelrecords available" << endl;
            return;
        }
        if( (elptr=JIVEMSFiller::theirCache.locateExperiment(eref.getCode()))==0 ) {
            cout << "writeModel/No experimentinfo available in cache?!" << endl;
            return;
        }
        // Create a ro-reference to the frequency-cache for the experiment
        const fqcache_type&          fqcache( elptr->freqCache );
        fqcache_type::const_iterator curfq;

        // Read the entries in the JIVE_MODEL table; build up a cache
        // to see if we already got the data for a given modelentry or not.
        // The modelpolynomials are uniquely keyed on (TIME,ANTENNA) [that's
        // not totally true - eg if we ever do multiple field centers, this
        // assumption DOES NOT HOLD]. However, when Mark K. decided to define
        // the JIVE_MODEL table format this was deemed as being unnecessary
        // [the unique key is (TIME,ANTENNA,SOURCE)] since we do not do it now
        // so we won't do it in the future.
        // Just so you know why the code over here doesn't deal with that just
        // now.
		//
		// 06/07/2009 - HV: In order to support telling data from reruns 
		//              apart - they may have been run with different MODELs -
		//              we'll add the PROCESSOR_ID to the unique key.
		
		// First, let's find the PROCESSOR_ID (row) for the
		// current JIVE JOB ID (visbuf.getVisBufId()/passID column)
		int            curpassid = -1;

		{
			const MSProcessor&      proc( ms.processor() );
			ROMSProcessorColumns    proccols( proc );

			for( uInt i=0; curpassid<0 && i<proccols.nrow(); i++ )
				if( (int)proccols.passId()(i)==vb.getVisBufId() )
					curpassid = (int)i;
		}
		// If the passid not found .. then what??
		if( curpassid<0 ) {
			cerr << "Error: JobId #" << vb.getVisBufId() << " could not "
				 << "be found in the PROCESSOR table. This is an internal "
				 << "inconsitency. Report to the author verkouter@jive.nl.\n"
				 << "THE MODEL IS NOT WRITTEN FOR THIS JOBID!\n";
			return;
		}
        // Anyway - whatever the unique key, we need to build up a cache of
        // entries (unique keys suffice) that we already DID write to the MS.
	    Table                      modeltable( ms.rwKeywordSet().asTable("JIVE_MODEL") );
        Table                      spwintab( ms.rwKeywordSet().asTable("SPECTRAL_WINDOW") );
    	ScalarColumn<Int>          antennaid( modeltable, "ANTENNA_ID" );
    	ScalarColumn<Int>          feedid( modeltable, "FEED_ID" );
    	ScalarColumn<Int>          spwid( modeltable, "SPECTRAL_WINDOW_ID" );
    	ScalarColumn<Double>       time( modeltable, "TIME" );
    	ScalarColumn<Double>       interval( modeltable, "INTERVAL" );
		ScalarColumn<Int>          processor( modeltable, "PROCESSOR_ID" );
    	ArrayColumn<Double>        dryatmdel( modeltable, "DRY_ATMOS_DELAY" );
    	ArrayColumn<Double>        wetatmdel( modeltable, "WET_ATMOS_DELAY" );
    	ArrayColumn<Double>        clockoffs( modeltable, "CLOCK_OFFSET" );
    	ArrayColumn<Double>        grpdel( modeltable, "GROUP_DELAY" );
    	ArrayColumn<Double>        phsdel( modeltable, "PHASE_DELAY" );

        // Well, we do not have to get the whole table, a strategically
        // formulated TaQL query might save us a lot of trouble.
        // Note: we may leave out the SPECTRAL_WINDOW_ID since if a record
        // is written for a specific ANTENNA,TIME combination,
        // all the SPECTRAL_WINDOW_IDs get written to this table.
        // I.e. the *unique* key is just the (ANTENNA,TIME) tuple (take
        // heed, read above...)
		//
		// 06/07/2009 - HV: (ANTENNA,TIME) => (ANTENNA,TIME,PROCESSOR_ID)
        Table                  qres;
        TaQLResult             taqlres;
        ROScalarColumn<Int>    antidcol;
		ROScalarColumn<Int>    proccol;
        ROScalarColumn<Double> timecol; 

        // Just get the distinct values in the two columns we need; no 
        // need to sort. The std::set() + our key_compare will do that.
        taqlres = ::tableCommand("SELECT DISTINCT ANTENNA_ID, TIME, PROCESSOR_ID FROM $1", modeltable);
        qres    = taqlres.table();

        antidcol.attach( qres, "ANTENNA_ID" );
        timecol.attach( qres, "TIME" );
		proccol.attach( qres, "PROCESSOR_ID" );
        for( uInt row=0; row<qres.nrow(); row++ ) {
            modelcacheinsertres_t   insres;
        
            insres = modelcache.insert( modelkey(antidcol(row),
												 timecol(row),
												 proccol(row)) );
            if( !insres.second ) {
                cerr << "writeModel()/Error building modelcache - " << endl
                     << " failed to insert (antenna_id,time)=("
                     << antidcol(row) << "," << timecol(row)
					 << "," << proccol(row) << ") " << endl
                     << "Usually this means a duplicate but.... "
                     << "this should NOT happen; it means that the casa "
                     << "Table Query Language command 'SELECT DISTINCT <...>' "
                     << "did not work as expected." << endl
                     << "Visit http://casa.nrao.edu and complain!"
                     << endl;
                THROW_JEXCEPT("Internal error inserting entry in modelcache");
            }
        }
        // done building modelcache
        // Now go over the entries from the visbuf
        Matrix<Double>                       grp_delay;
        Matrix<Double>                       phs_delay;
        // shapes of the arrays (vectors of lengts 2 && coefficients_type::nrElements)
        // are constant
        const IPosition                      taylor1(1, 2);
        const IPosition                      rowshp(1, j2ms::coefficients_type::nrElements);
        j2ms::log2polmap                     l2p( vb.correlatorSetup().l2pmap() );
        const j2ms::scaninfolist_t&          scanlist( vb.getScanlist() );
        modelcache_t::const_iterator         cacheptr;
        j2ms::modeldata_t::const_iterator    curmodel;
        j2ms::scaninfolist_t::const_iterator curscan;

        // the shape of the coeffient matrices is known now 
        // [and will not change during execution of this fn]
        grp_delay.resize( IPosition(2, l2p.size(), j2ms::coefficients_type::nrElements) );
        phs_delay.resize( IPosition(2, l2p.size(), j2ms::coefficients_type::nrElements) );
        
        curfq       = fqcache.end();
        curscan     = scanlist.end();
        nrWritten   = 0;
        alreadyInMS = 0;
        
        for( curmodel=model.begin(); curmodel!=model.end(); curmodel++ ) {
            MVEpoch            curMVepoch( Quantity(curmodel->startMJDInSeconds, "s") );
            modelkey           curkey( curmodel->stationID,
									   curmodel->startMJDInSeconds,
									   curpassid );
            mepoch_in_scaninfo condition( curMVepoch );

            // if in cache - discard
            if( (cacheptr=modelcache.find(curkey))!=modelcache.end() ) {
                alreadyInMS++;
                continue;
            }

            // Find out the scan so's we can unmap every experiment based value
            // (subbands in the modeldatarecord are with respect to the frequency-setup
            //  and have no relations whatsoever to the spectral-window-id in the MS).

            // if time of current modelrecord outside current scan, start looking
            // for new scan [includes triggering of search if no current scan yet]
            if( curscan!=scanlist.end() && !condition(*curscan) )
                curscan=scanlist.end();
            // if no current scan yet/anymore go find the/a new one
            if( curscan==scanlist.end() ) {
                curscan = find_scaninfo_by_time(scanlist, curMVepoch);
                // if curscan, by now, is _still_ scanlist.end(), we're up a certain
				// creek. w/o paddle, even.
                if( curscan==scanlist.end() ) {
                    cerr << "writeModel: No scan found for time "
                         << MVTime::setFormat(MVTime::DMY) << MVTime(curMVepoch) << "/Skipping modelrecord" 
                         << endl;
                    continue;
                }
                // reset curfq to trigger looking up a new one!
                curfq = fqcache.end();
            }
            // woohoo, got a scan. See if we can find the associated freqcache stuff etc
            // NOTE: curscan is assumed to point at something sensible from here on
            if( curfq==fqcache.end() ) {
                if( (curfq=fqcache.find(curscan->frequencyCode))==fqcache.end() ) {
                    cerr << "writeModel: No frequencyinfo found in cache for mode "
                         << curscan->frequencyCode << "/Skipping modelrecord" << endl;
                    continue;
                }
            }
            // Both curscan/curfq are valid. Now we can find the subbands and map 'm to
            // MS spectral-window-IDs
            
            // Iterator over the entries in the modeldatarecord
            const fqprops_type&                       fqprops( curfq->second );
            j2ms::sb2polymap_type::const_iterator     cursb;
            j2ms::src2sbpolymap_type::const_iterator  cursrc;
            for( cursrc=curmodel->src2sbpolymap.begin();
                 cursrc!=curmodel->src2sbpolymap.end();
                 cursrc++ ) {
                // cursrc->first == experiments sourceId == MSSourceId [there should be
                //                  1:1 mapping there]
                // cursrc->second contains records for all subbands that we have modeldata for
                for( cursb=cursrc->second.begin(); 
                     cursb!=cursrc->second.end();
                     cursb++ ) {
                    // cursb->first == baseband# in the frequencyconfig for the scan
                    //                 this record applies to. Must be translated to
                    //                 a MS Spectral Window ID. no 1:1 mapping here...
                    // cursb->second contains the polys, organized by polarization
                    fqprops_type::const_iterator           mssbptr; 
                    const j2ms::pol2polymap_type&          p2pmap( cursb->second );
                    j2ms::log2polmap::const_iterator       curpol;

                    // Try to look up the current subband in the
                    // current frequencysetup-cache-element
                    if( (mssbptr=fqprops.find(cursb->first))==fqprops.end() ) {
                        cerr << "writeModel/No info in MS for subband#" << cursb->first
                             << " in mode " << curfq->first << "/Skipping modeldatarecord" << endl;
                        continue;
                    }
                    // Loop over the polarizations && put the coefficients in the right place
                    for( curpol=l2p.begin(); curpol!=l2p.end(); curpol++ ) {
                        int                                    polidx = l2p.index_of( curpol->second );
                        j2ms::pol2polymap_type::const_iterator curpolypair;

                        // locate the arrayindex of the current polarization
                        if( polidx<0 ) 
                            THROW_JEXCEPT("writeModel/Internal inconsistency. log2polmap entry " << *curpol
                                          << " returns index_of(" << curpol->second << ") = " << polidx);
                        // locate the polynomial-pair for current polarization. If not found, there's
                        // just no data for this src/subband/polarization
                        if( (curpolypair=p2pmap.find(curpol->second))==p2pmap.end() )
                            break;
                        // Data for polarization curpol->second should be put in row #polidx in the
                        // array
                        grp_delay.row( polidx ) = Vector<Double>(rowshp, curpolypair->second.delay.data);
                        phs_delay.row( polidx ) = Vector<Double>(rowshp, curpolypair->second.phase.data);
                    }
                    // if the data for not all of the polarizations was found, don't write the
                    // polys
                    if( curpol!=l2p.end() ) {
                        cerr << "writeModel/Data for polarization " << curpol->second << " missing in "
                             << "subband#" << cursb->first
                             << " @" << MVTime::setFormat(MVTime::DMY) << curMVepoch << "/Skipping record" << endl;
                        continue;
                    }

                    // Phew. Now we can safely add a row to the modeltable
                    uInt        nwrow( modeltable.nrow() );
                    
                    modeltable.addRow();
                    antennaid.put(nwrow, curmodel->stationID);
                    spwid.put(nwrow, mssbptr->second.__spwId);
                    feedid.put(nwrow, 0);
					processor.put(nwrow, curpassid);
                    time.put(nwrow, curmodel->startMJDInSeconds);
                    interval.put(nwrow, curmodel->durationInSeconds);
                    dryatmdel.put(nwrow, Vector<Double>(taylor1, curmodel->dryAtmosphere.data));
                    wetatmdel.put(nwrow, Vector<Double>(taylor1, curmodel->wetAtmosphere.data));
                    clockoffs.put(nwrow, Vector<Double>(taylor1, curmodel->clockOffset.data));
                    grpdel.put(nwrow, grp_delay);
                    phsdel.put(nwrow, phs_delay);
                }
            }
            nrWritten++;
            // if succesfull, insert into cache as well
            modelcacheinsertres_t   ires;
            
            ires = modelcache.insert( modelkey(curmodel->stationID,
                                               curmodel->startMJDInSeconds,
											   curpassid) );
            if( !ires.second )
                THROW_JEXCEPT("Failed to insert new entry into modelcache?!");
        }
        cout << "writeModel/Wrote " << nrWritten << " modelentries out of " << model.size()
             << "  [" << alreadyInMS << " already in MS]"
             << endl;
    }
    catch( const std::exception& e ) {
        cerr << "writeModel: " << e.what() << endl;
    }
    return;
}
#endif
