//  JIVEMSFiller - a MS filler, dedicated to deal with the situation @
//  JIVE
//
//  i.e. Must be able to cope with 'Bos-format (=raw binary dump)' +
//  COF (Correlator Output Format)
//
//  Author:   Harro Verkouter,   10-12-1998
//
//  $Id: JIVEMSFiller.h,v 1.12.2.1 2018/06/07 15:27:12 verkout Exp $
//
#ifndef JIVE_J2MS_JIVEMSFILLER_H
#define JIVE_J2MS_JIVEMSFILLER_H

#include <casacore/casa/aips.h>

#include <casacore/casa/BasicSL/String.h>
#include <casacore/casa/Arrays/Vector.h>

#include <casacore/ms/MeasurementSets.h>
#include <casacore/casa/Arrays/Matrix.h>
#include <jive/j2ms/Experiment.h>
#include <jive/j2ms/ExperimentCache.h>
#include <jive/j2ms/subjob.h>
#include <jive/j2ms/Visibility.h>
#include <jive/j2ms/FrequencyConfig.h>
#include <jive/j2ms/VisibilityBuffer.h>
#include <jive/j2ms/Filter.h>
#include <vector>

typedef std::pair<casacore::Int, casacore::Int> antennafeed_pair_type;
typedef std::vector<unsigned int>               rowmap_type;

//  The Filler-class 
class JIVEMSFiller
{
public:
    //  Create a JIVEMSFiller. The string argument is optional and
    //  may contain filler options.
    //  Currently only filter options recognized
    JIVEMSFiller( const casacore::String& opt = "" );

    //  Do the job of adding the indicated scan to the MS of the
    //  indicated experiment. The boolean argument controls whether or
    //  not the user is asked to concatenate (if the MS already
    //  exists). true -> user is prompted; false -> data is
    //  concatenated blindly
    casacore::Bool       fillDataToExperimentsMS( Experiment& expref,
										const subjob_t& sjdescr,
										casacore::Bool ask=casacore::True );

    //  Delete all the resources used by the JIVEMSFiller...
    ~JIVEMSFiller();

private:
    //  The filler's private parts

	// Option passed when the Filler was created....
	casacore::String           myOptions;

	// Optionally, if the usr wanted filters to be applied:
	// let's keep a list of'm!
	std::vector<Filter*>  myFilters;

    //  The set of global databases:
    static ExperimentCache    theirCache;

    //  The private methods

    //  Attempt to create/open the MS for the indicated Experiment
    //  and set it up according to JIVE specs
    //  HV: 16 Feb 2015: we need to pass the visibility buffer such that
    //                   makeMS can use info from the data (e.g. correlator
    //                   setup ...) to size the tiles for the Tiled storage
    //                   manager
	casacore::MeasurementSet* makeMS( const Experiment& expref, casacore::Bool ask, const VisibilityBuffer& visbuf );
    void                      rebuild_caches(casacore::MeasurementSet const& ms);

    //  This function makes sure the visibility data is in the domain
    //  requested by the experiment. 
    //
    //  NOTE: this is ONLY for the datamatrix, nothing else!
    //
    //  NOTE: The contents of the matrix that is returned may or may
    //  not vary, since a reference to a static object is returned and
    //  the contents of the matrix may be changed after a call to
    //  makeData(). Therefore, copy the contents of the matrix to your
    //  own local variable to ensure proper operation.
    static const casacore::Matrix<casacore::Complex>& 
			makeData( Experiment const& expref,
					  Visibility const& vref,
                      rowmap_type const& rowmap );


    //  Fill the MS with the data from the scan
    casacore::Bool fillMSfromVisibilityBuffer( casacore::MeasurementSet& msref,
							       		   VisibilityBuffer& visbuf,
								           const subjob_t& sjdescr );

    //  Write the indicated frequency setup of the scan in the spectralwindow
    //  table of the given MS
    casacore::Int updateSpectralAndPolarizationTable( 
				casacore::MeasurementSet& msref,
				const Visibility& vref,
				const j2ms::Scan& scanref );

    //  Write all antennas in the experiment to the antennatable in
    //  the indicated MS.
    //  19/01/2007 HV - Updated. Does not return anymore. Throws.
    //                  If antenna's are listed in the MS's Antenna
    //                  subtable it verifies they are the same as in
    //                  the experiment. 
    //                  If they're not, we're up the creek. Equally,
    //                  we're also up the creek if we fail to write/
    //                  append to this table at all. So throwing up
    //                  seems like the right thing to do.
    //  18/05/2018 HV - Find rows in the ANTENNA and FEED table that
    //                  correctly and consistently describe the properties
    //                  of antenna 'ant'. Rows are added if needed
    antennafeed_pair_type updateAntennaAndFeedTable(casacore::MeasurementSet& msref,
                                                    unsigned int ant, j2ms::Scan const& scanref); 

    //  Return the FIELD_ID of the source (field) of the indicated scan.
    //  Matches a row in the field table or adds a new row if no such row
    //  was found
    casacore::Int  updateFieldTable(casacore::MeasurementSet& msref, j2ms::Scan const& scanref);

	// Updates the Processor subtable with a new row for a new passid
	casacore::Int  updateProcessorTable( casacore::MeasurementSet& msref,
									 casacore::Int newpassid,
					 				 const casacore::String& subtype );

    //  update the Observation table (for the min/max time in the MS?)
    casacore::Bool updateObservationTable( casacore::MeasurementSet& msref,
							           const Experiment& expref );
    
    casacore::Int updateStateTable( casacore::MeasurementSet& msref,
								    const Experiment& );
    //  Calculate the UVW-value for the Scan/Visibility combination
    //  (NOTE: data is taken from the experiment/scan cache)
    //
    //  IMPORTANT NOTE: this function returns a reference to static
    //  Vector-object. The vector is modified everytime the 'calcUVW'
    //  method is invoked. Therefore, don't rely on the fact that the
    //  values found in the vector are always the same and if you want
    //  to keep the values, you should make a local copy (not the
    //  copy-c'tor, because that uses reference semantics. Great isn't
    //  it?! Marvellous!)
    //
    //  NOTE: It may return JIVEMSFiller::noUVW in case of error
    static const casacore::Vector<casacore::Double>& calcUVW( const Visibility& vref );

#if 0
    static void writeModel( casacore::MeasurementSet& ms,
                            const VisibilityBuffer& vb );
#endif
    // For now, calculate the time of the visibility in the scan/frequencyconfig
    /*
    const casacore::MEpoch&         calcVisibilityTime( const Scan& scanref, const Visibility& vref,
					      const FrequencyConfig& fcref );
					      */

    //  Do not allow these object to be copied/assigned to!
    JIVEMSFiller( const JIVEMSFiller& );
    const JIVEMSFiller& operator=( const JIVEMSFiller& );
};


#endif
