#include <jive/j2ms/SFXCVisBuffer.h>
#include <jive/j2ms/Experiment.h>
#include <jive/j2ms/COFScan.h>
#include <jive/j2ms/Scan.h>
#include <jive/j2ms/DayConversion.h>
#include <jive/labels/CorrelationCode.h>
#include <jive/Utilities/jexcept.h>
#include <jive/Utilities/stringutil.h>

#include <casacore/casa/Quanta/MVTime.h>
#include <casacore/casa/Utilities/Regex.h>

#include <sys/types.h>
#include <dirent.h>
#include <fcntl.h>
#include <libgen.h>
#include <limits.h>

#include <cmath>
#include <cstdlib>
#include <list>
#include <string>
#include <iostream>
#include <algorithm>

#include <jive/j2ms/sfxc/output_header.h>

bool operator<(const SFXCKey &a, const SFXCKey &b)
{
  if (a.antenna == b.antenna)
    {
      if (a.polarisation == b.polarisation)
	return a.subband < b.subband;
      else
	return a.polarisation < b.polarisation;
    }

  return a.antenna < b.antenna;
}

bool operator!=(const SFXCKey &a, const SFXCKey &b)
{
  return (a.antenna != b.antenna || a.polarisation != b.polarisation ||
	  a.subband != b.subband);
}

bool operator<(const SFXCBaselineKey &a, const SFXCBaselineKey &b)
{
  if (a.baseline.first != b.baseline.first)
    return a.baseline.first < b.baseline.first;

  if (a.baseline.second != b.baseline.second)
    return a.baseline.second < b.baseline.second;

  if (a.polarisation.first != b.polarisation.first)
    return a.polarisation.first < b.polarisation.first;

  if (a.polarisation.second != b.polarisation.second)
    return a.polarisation.second < b.polarisation.second;

  if (a.subband.first != b.subband.first)
    return a.subband.first < b.subband.first;

  return a.subband.second < b.subband.second;
}

SFXCVisBuffer::SFXCVisBuffer(const casacore::String &vb,
			     const Experiment *exp,
			     const string& /*vbo*/) :
  VisibilityBuffer(exp),
  __rootdir(exp->getRootDir())
{
    // Transform the experiment's scanlist into a list of scanpointers
    const j2ms::scaninfolist_t& scanlist(exp->getScanlist());
    j2ms::scaninfolist_t::const_iterator scan;
    
    for (scan = scanlist.begin(); scan != scanlist.end(); scan++)
      __scanptrlist.push_back(new j2ms::COFScan(this, *scan));


    char buf[PATH_MAX];
    if (::realpath(__rootdir.c_str(), buf) == NULL)
      THROW_JEXCEPT("Unresolvable path" << __rootdir);

    const string path = __rootdir + "/" + vb;
    fd = ::open(path.c_str(), O_RDONLY | O_LARGEFILE);
    if (fd == -1)
      THROW_JEXCEPT("Cannot open " << path);

    Output_header_global base_hdr;
    std::memset(&base_hdr, 0, sizeof(base_hdr));

    // Read the header size
    ssize_t nbytes = ::read(fd, &base_hdr, sizeof(base_hdr.header_size));
    if (nbytes != sizeof(base_hdr.header_size))
      THROW_JEXCEPT("Cannot read global header size");

    // Check that at least the basic information is there
    if (base_hdr.header_size < 64)
      THROW_JEXCEPT("Global header size invalid (is " << base_hdr.header_size
		    << " expected at least " << 64 << "). Maybe this is not SFXC data.");

    // Allocate space for full header
    size_t size = std::max(sizeof(base_hdr), (size_t)base_hdr.header_size) + 1;
    Output_header_global *hdr = (Output_header_global *)std::malloc(size);
    std::memset(hdr, 0, size);

    // Now read the complete header
    ::lseek(fd, 0, SEEK_SET);
    nbytes = ::read(fd, hdr, base_hdr.header_size);
    if (nbytes != (ssize_t)base_hdr.header_size)
      THROW_JEXCEPT("Cannot read global header, maybe this is not SFXC data.");

    if (hdr->output_format_version != OUTPUT_FORMAT_VERSION)
      THROW_JEXCEPT("Global header version invalid (is " << hdr->output_format_version
		    << " expected " << OUTPUT_FORMAT_VERSION << "). Maybe this is not SFXC data.");

    correlator_version = hdr->correlator_version;
    integration_time = hdr->integration_time * 1e-6;
    job_nr = hdr->job_nr;

    // OK from here we assume we're looking at SFXC data
    // Now it's time to parse the file name for extra bits of info.
    // SFXC output file names are typically:
    //      <some base name>.cor{_<stuff>}*
    // i.e. any number of "_" separated bits of info can be appended to the
    // ".cor" extension.
    //
    // Known so far:
    //      _bin[0-9]+      pulsar bin ouput for bin [0-9]+
    //      _XXXX           might be source name for multiple phase center
    //                      but could also be something else
    //                      So if we find this we check if it is a source id
    //                      and otherwise (silently) ignore
    const string::size_type ext = vb.rfind(".cor");
    const casacore::Regex   rxPulsarBin("^bin[0-9]+$");
    SourceDB const&         sourceDB( exp->getSourceDB() );
    vector<string>          extensions = split( ((ext==string::npos) ? string() : vb.substr(ext+4)), '_' );
    SourceInfo              maybeSource;

    for(vector<string>::const_iterator extension = extensions.begin(); extension!=extensions.end(); extension++) {
        if( extension->empty() )
            continue;
        if( rxPulsarBin.match(extension->c_str(), extension->length())!=casacore::String::npos )
            continue;
        // Not a pulsar bin number, check if it's a valid source id
        if( sourceDB.searchSource(maybeSource, *extension)==false )
            // nope. ignore silently
            continue;
        // If we already have set an apparently valid source id, we cry
        // out loud!
        if( !__maybeSourceId.empty() )
          THROW_JEXCEPT("Multiple valid source-id extension strings found in file name: " << std::endl
                        << "  " << vb << std::endl
                        << "     found: " << __maybeSourceId << " and " << *extension << std::endl);
        // Allrighty!
        std::cout << "SFXCVisBuffer / setting multiphasecenter source id to " << *extension << std::endl;
        __maybeSourceId = *extension;
    }

    // Start reading after the header
    ::lseek(fd, hdr->header_size, SEEK_SET);

    unsigned month, day;
    DayConversion::dayNrToMonthDay(month, day, hdr->start_day, hdr->start_year);
    start_year = hdr->start_year;
    start_month = month + 1;
    start_day = day;
    start_sec = hdr->start_time;
    casacore::Double dayfraction = (start_sec + 0.5 * integration_time) / DayConversion::secondsPerDay;
    casacore::MVEpoch start_time = casacore::MVEpoch(casacore::MVTime(hdr->start_year, month + 1, day), dayfraction);

    if (hdr->number_stations > 0)
      createAntennaMapping(hdr);
    else
      createAntennaMapping();

    if (hdr->number_sources > 0)
      createSourceMapping(hdr);
    else
      createSourceMapping();

    baseline = baselines.end();
    subband = subbands.end();

    num_channels = hdr->number_channels;

    correlationset_t pols;
    switch (hdr->polarisation_type)
      {
      case Output_header_global::LEFT_POLARISATION:
	pols.insert(CorrelationCode(j2ms::l, j2ms::l));
	break;
      case Output_header_global::RIGHT_POLARISATION:
	pols.insert(CorrelationCode(j2ms::r, j2ms::r));
	break;
      case Output_header_global::LEFT_RIGHT_POLARISATION:
	pols.insert(CorrelationCode(j2ms::r, j2ms::r));
	pols.insert(CorrelationCode(j2ms::l, j2ms::l));
	break;
      case Output_header_global::LEFT_RIGHT_POLARISATION_WITH_CROSSES:
	pols.insert(CorrelationCode(j2ms::r, j2ms::r));
	pols.insert(CorrelationCode(j2ms::r, j2ms::l));
	pols.insert(CorrelationCode(j2ms::l, j2ms::r));
	pols.insert(CorrelationCode(j2ms::l, j2ms::l));
	break;
      default:
	THROW_JEXCEPT("Unknown polarisations in header");
    }

    j2ms::log2polmap l2p;
    switch (hdr->polarisation_type)
      {
      case Output_header_global::LEFT_POLARISATION:
	l2p.insert(std::make_pair(0, j2ms::l));
	break;
      case Output_header_global::RIGHT_POLARISATION:
	l2p.insert(std::make_pair(0, j2ms::r));
	break;
      default:
	l2p.insert(std::make_pair(0, j2ms::r));
	l2p.insert(std::make_pair(1, j2ms::l));
	break;
      }

    __setup = CorrelatorSetup(num_channels, pols, integration_time,
			      CorrelatorSetup::bypassVersion, l2p, -1);

    // Now we can pre-allocate the temporary storage.
    cplxbuf.resize(__setup.nrOfSpectralPoints());
    weight.resize(__setup.getCorrelations().size());
    data.resize(__setup.getCorrelations().size(),  __setup.nrOfSpectralPoints());
    cplxnullen = complexvector_type(__setup.nrOfSpectralPoints(), casacore::Complex(0.0f, 0.0f));

    // Move forward to first visibility.
    next();

    std::free(hdr);
}

void
SFXCVisBuffer::createAntennaMapping(void)
{
  const Experiment& exp(getExperiment());
  const AntennaDB& adb(exp.getAntennaDB());
  std::vector<casacore::String> antennas;
  std::vector<casacore::String>::const_iterator antenna;
  AntennaInfo ai;

  // Convert from casacore::Vector to std::vector.
  exp.getAntennas().tovector(antennas);

  // Sort antennas in alphabetical order.
  std::sort(antennas.begin(), antennas.end());

  for (antenna = antennas.begin(); antenna != antennas.end(); antenna++)
    {
      adb.searchAntenna(ai, *antenna);
      antenna_map.push_back(ai.getId());
    }
}

void
SFXCVisBuffer::createAntennaMapping(struct Output_header_global *hdr)
{
  const Experiment& exp(getExperiment());
  const AntennaDB& adb(exp.getAntennaDB());
  AntennaInfo ai;

  char *data = (char *)hdr + hdr->stations_offset;
  for (int i = 0; i < hdr->number_stations; i++) {
    if (data - (char *)hdr >= hdr->header_size)
      THROW_JEXCEPT("Corrupt antenna table");

    adb.searchAntenna(ai, data);
    antenna_map.push_back(ai.getId());
    data += strlen(data) + 1;
  }
}

void
SFXCVisBuffer::createSourceMapping(void)
{
  source_map.push_back(__maybeSourceId);
}

void
SFXCVisBuffer::createSourceMapping(struct Output_header_global *hdr)
{
  char *data = (char *)hdr + hdr->sources_offset;
  for (int i = 0; i < hdr->number_sources; i++) {
    if (data - (char *)hdr >= hdr->header_size)
      THROW_JEXCEPT("Corrupt source table");

    source_map.push_back(std::string(data));
    data += strlen(data) + 1;
  }
}

// Helper stuff for finding frequencies/sidebands
template <typename T>
T advancer(T t, size_t n) {
    std::advance(t, n);
    return t;
}
struct bbfinder_type {
    bbfinder_type(casacore::Double const& f0, BaseBand::SideBand side):
        __m_f0(f0), __m_sideband(side)
    {}
    bool operator()(BaseBand const& bb) const {
        return casacore::nearAbs(bb.getStartfreq(), __m_f0, 1.0e-6) && bb.getSideBand()==__m_sideband;
    }
    casacore::Double const   __m_f0;
    BaseBand::SideBand const __m_sideband;
};

void
SFXCVisBuffer::createSubbandMapping(void)
{
  const Experiment& exp(getExperiment());
  const FrequencyDB& fdb(exp.getFrequencyDB());
  FrequencyConfig fc;

  fdb.searchFrequencyConfig(fc, __mode);
  const basebandlist_t& bbs(fc.getBaseBands());

  // Sort + uniquefy frequencies.
  std::set<casacore::Double> freqs;

  std::transform(bbs.begin(), bbs.end(), std::inserter(freqs, freqs.end()), std::mem_fun_ref(&BaseBand::getStartfreq));

  // Start with a clean slate.
  // subbands = set[ pair<frequency_nr, sideband> ]
  // so we must look up, for each subband, which BaseBand nr matches,
  // based on frequency+sideband
  subband_map.clear();

  typedef std::set<std::pair<int, int> > spi_type;
  for(spi_type::const_iterator cursb = subbands.begin(); cursb!=subbands.end(); cursb++) {
      // Make sure we're not adressing frequencies outside defined range
      if( cursb->first<0 || cursb->first>=(int)freqs.size() )
          THROW_JEXCEPT("frequency_nr " << cursb->first << " out of range (min: 0, max: " << freqs.size()-1 << ")");
      // From frequency_nr => frequency and from int => BaseBand::sideband
      // so we can go look for that baseband
      bbfinder_type const            bbfinder(*advancer(freqs.begin(), cursb->first), (cursb->second==0 ? BaseBand::lower : BaseBand::upper));
      basebandlist_t::const_iterator curBB = std::find_if(bbs.begin(), bbs.end(), bbfinder);

      if( curBB==bbs.end() )
          THROW_JEXCEPT("frequency_nr " << cursb->first << "/" << bbfinder.__m_sideband << " not found in FrequencyConfig " << __mode);
      // Now we can fill in the mapping
      subband_map[ *cursb ] = curBB->getBaseBandNr();
  }
}

casacore::Bool
SFXCVisBuffer::getCurrentVisibility(Visibility &v)
{
  if (subbands.empty() || baseline == baselines.end())
    return casacore::False;

  if ((*baseline).first >= antenna_map.size()
      || (*baseline).second >= antenna_map.size())
    THROW_JEXCEPT("Unobserved station " << ((*baseline).first) << " or " << ((*baseline).second));
  if (subband_map.count(*subband) == 0)
    THROW_JEXCEPT("Unobserved subband " << ((*subband).first) << " pol=" << ((*subband).second));

  const correlationset_t &pols(__setup.getCorrelations());
  correlationset_t::const_iterator pol;

  bool swap = true;
  casacore::Int ant1 = antenna_map[(*baseline).first];
  casacore::Int ant2 = antenna_map[(*baseline).second];
  if (ant1 > ant2)
    {
      swap = false;
      ant1 = antenna_map[(*baseline).second];
      ant2 = antenna_map[(*baseline).first];
    }

  int i;
  for (pol = pols.begin(), i = 0; pol != pols.end(); pol++, i++)
    {
      std::vector<SFXCBaselineKey>::const_iterator key;

      std::pair<int, int> polarisation = std::make_pair((*pol)[0] == j2ms::l, (*pol)[1] == j2ms::l);
      if (!swap)
	polarisation = std::make_pair((*pol)[1] == j2ms::l, (*pol)[0] == j2ms::l);

      int j;
      for (key = keys.begin(), j = 0; key != keys.end(); key++, j++)
	{
	  if ((*key).baseline != (*baseline) || (*key).subband != (*subband))
	    continue;
	  
	  if ((*key).polarisation == polarisation)
	    break;
	}

      if (key == keys.end())
	{
	  data.row(i) = cplxnullen;
	  weight(i) = 0.0;
	  continue;
	}

      if ((*key).subband.second == 0)
	for (int channel = 0; channel < num_channels; channel++)
	  cplxbuf[num_channels - channel - 1] = casacore::Complex(visibilities[j][channel * 2], -visibilities[j][channel * 2 + 1]);
      else
	for (int channel = 0; channel < num_channels; channel++)
	  cplxbuf[channel] = casacore::Complex(visibilities[j][channel * 2], visibilities[j][channel * 2 + 1]);

      if (swap)
	{
	  for (int channel = 0; channel < num_channels; channel++)
	    cplxbuf[channel] = std::conj(cplxbuf[channel]);
	}

      SFXCKey key1, key2;
      key1.antenna = (*key).baseline.first;
      key1.polarisation = (*key).polarisation.first;
      key1.subband = (*key).subband;
      key2.antenna = (*key).baseline.second;
      key2.polarisation = (*key).polarisation.second;
      key2.subband = (*key).subband;

      if (key1 != key2) {
	// Cross-correlations need to have a quantization (Van Vleck)
	// correction applied.  The factor here is taken from equation
	// A8.13 in Thompson, Moran and Swenson, "Interferometry and
	// Synthesis in Radio Astronomy", 2nd Edition and assumes
	// 2-bit sampling and "ideal" sampler statistics.
	for (int channel = 0; channel < num_channels; channel++)
	  cplxbuf[channel] *= 1.1329552;
      }

      data.row(i) = cplxbuf;

      if (ant1 == ant2)
	weight(i) = weight_map[key1];
      else {
	// Correlator versions before SVN r1129 did not write out the
	// number of correlated valid samples, so we have to estimate
	// the baseline weight based on the weights of the
	// autocorrelations.  This assumes invalid samples in each
	// datastreams are uncorrelated.
	if (correlator_version < 1119)
	  weight(i) = weight_map[key1] * weight_map[key2];
	else
	  weight(i) = baseline_weight_map[(*key)];
      }
    }

  ExtVisLabel lab;

  lab.correlations.clear();
  for (pol = pols.begin(); pol != pols.end(); pol++)
    lab.correlations.push_back(*pol);
  lab.subbandnr = subband_map[*subband];
  lab.baseline = BaseLineIdx(ant1, ant2);

  if (uvw_map.count(ant1) == 0 || uvw_map.count(ant2) == 0)
    THROW_JEXCEPT("Missing uvws? ant#" << ant1 << " has " << (uvw_map.count(ant1)) << ", ant#" << ant2 << " has " << uvw_map.count(ant2));

  casacore::Muvw::Ref ref(casacore::Muvw::J2000);
  casacore::Muvw uvw = casacore::Muvw(uvw_map[ant1] - uvw_map[ant2], ref);
  v = Visibility(lab, data, Visibility::frequency, casacore::False, weight, slice_time, uvw, this);

  return casacore::True;
}

void
SFXCVisBuffer::read_timeslice_header(void)
{
  struct Output_header_timeslice hdr;
  ssize_t nbytes = ::read(fd, &hdr, sizeof(hdr));
  if (nbytes != sizeof(hdr))
    {
      if (nbytes == 0)
	{
	  next_slice = -1;
	  return;
	}
      THROW_JEXCEPT("Cannot read timeslice header");
    }

  next_slice = hdr.integration_slice;
  next_baselines = hdr.number_baselines;
  next_uvws = hdr.number_uvw_coordinates;
  next_stats = hdr.number_statistics;
}

void
SFXCVisBuffer::read_uvws(void)
{
  for (int i = 0; i < next_uvws; i++)
    {
      struct Output_uvw_coordinates uvw;
      ssize_t nbytes = ::read(fd, &uvw, sizeof(uvw));
      if (nbytes != sizeof(uvw))
	THROW_JEXCEPT("Cannot read uvw");

      if ((unsigned int)uvw.station_nr >= antenna_map.size())
	THROW_JEXCEPT("read_uvws: Unobserved station #" << uvw.station_nr);
      uvw_map[antenna_map[uvw.station_nr]] = casacore::MVuvw(uvw.u, uvw.v, uvw.w);
      if ((unsigned int)uvw.source_nr >= source_map.size())
	THROW_JEXCEPT("read_uvws: Unobserved source #" << uvw.source_nr);
      __maybeSourceId = source_map[uvw.source_nr];
    }
}

void
SFXCVisBuffer::read_stats(void)
{
  for (int i = 0; i < next_stats; i++)
    {
      struct Output_header_bitstatistics stats;
      ssize_t nbytes = ::read(fd, &stats, sizeof(stats));
      if (nbytes != sizeof(stats))
	THROW_JEXCEPT("Cannot read stats");

      int32_t nvalid = 0;
      for (int i = 0; i < 4; i++)
	nvalid += stats.levels[i];

      if ((unsigned int)stats.station_nr >= antenna_map.size())
	THROW_JEXCEPT("read_stats: Unobserved station " << stats.station_nr);
      SFXCKey key;
      key.antenna = stats.station_nr;
      key.polarisation = stats.polarisation;
      key.subband = std::make_pair(stats.frequency_nr, stats.sideband);
      ntotal_map[key] = nvalid + stats.n_invalid;
      weight_map[key] = (double)nvalid / ntotal_map[key];
    }
}

void
SFXCVisBuffer::read_baseline(void)
{
  struct Output_header_baseline hdr;
  ssize_t nbytes = ::read(fd, &hdr, sizeof(hdr));
  if (nbytes != sizeof(hdr))
    THROW_JEXCEPT("Cannot read baseline header");

  if (current_baseline >= visibilities.size())
      visibilities.push_back(new float[(num_channels + 1) * 2]);

  nbytes = ::read(fd, visibilities[current_baseline], (num_channels + 1) * 2 * sizeof(float));
  if (nbytes != (ssize_t)((num_channels + 1) * 2 * sizeof(float)))
    THROW_JEXCEPT("Cannot read baseline visibilities");

  SFXCBaselineKey key;
  key.baseline = std::make_pair(hdr.station_nr1, hdr.station_nr2);
  key.polarisation = std::make_pair(hdr.polarisation1, hdr.polarisation2);
  key.subband = std::make_pair(hdr.frequency_nr, hdr.sideband);
  keys.push_back(key);

  baselines.insert(key.baseline);
  subbands.insert(key.subband);

  SFXCKey key1, key2;
  key1.antenna = key.baseline.first;
  key1.polarisation = key.polarisation.first;
  key1.subband = key.subband;
  key2.antenna = key.baseline.second;
  key2.polarisation = key.polarisation.second;
  key2.subband = key.subband;
  uint32_t ntotal = std::max(ntotal_map[key1], ntotal_map[key2]);
  baseline_weight_map[key] = (double)hdr.weight / ntotal;

  baseline = baselines.begin();
  subband = subbands.begin();

  current_baseline++;
}

void
SFXCVisBuffer::read_timeslice(void)
{
  current_baseline = 0;

  uvw_map.clear();
  ntotal_map.clear();
  weight_map.clear();
  baseline_weight_map.clear();

  do {
    read_uvws();
    read_stats();
    while (next_baselines--)
      read_baseline();
    read_timeslice_header();
  } while (current_slice == next_slice);
}

casacore::Bool
SFXCVisBuffer::next(void)
{
  if (subbands.empty())
    {
      read_timeslice_header();
    }
  else
    {
      subband++;
      if (subband == subbands.end())
	{
	  baseline++;
	  subband = subbands.begin();
	}
    }

  if (baseline == baselines.end())
    {
      keys.clear();
      baselines.clear();
      subbands.clear();

      if (next_slice == -1)
	return casacore::False;

      current_slice = next_slice;
      casacore::Double dayfraction = (start_sec + (current_slice + 0.5) * integration_time) / DayConversion::secondsPerDay;
      slice_time = casacore::MVEpoch(casacore::MVTime(start_year, start_month, start_day), dayfraction);
      read_timeslice();

      j2ms::Scan const*                cur_scan = this->getScanptr();
      multiphasecenter_in_info         cond(slice_time, __maybeSourceId);

      // If current scan is null or time stamp / source not matching
      // we needs to search for new scan
      if( !cur_scan || !cond(*cur_scan) ) {
        scanptrlist_type::const_iterator scan = find_scaninfo_by_predicate(__scanptrlist, cond);

        if (scan != __scanptrlist.end()) {
            if (__mode != (*scan)->getFrequencyCode()) {
                __mode = (*scan)->getFrequencyCode();
                createSubbandMapping();
            }
            setCurrentScanptr(*scan);
        } else {
            THROW_JEXCEPT("Failed to find scan for " << __maybeSourceId << " at time " << casacore::MVTime(slice_time.get()).string(casacore::MVTime::DMY,7));
        }
      }
    }

  return casacore::True;
}

const casacore::String&
SFXCVisBuffer::getRootDir() const
{
  return __rootdir;
}

casacore::Int
SFXCVisBuffer::getVisBufId() const
{
  return job_nr;
}

casacore::String
SFXCVisBuffer::getProcessorSubType() const
{
  return "SFXC";
}

CorrelatorSetup
SFXCVisBuffer::correlatorSetup() const
{
  return __setup;
}

SFXCVisBuffer::~SFXCVisBuffer()
{
  scanptrlist_type::iterator scan;

  for(scan = __scanptrlist.begin(); scan != __scanptrlist.end(); scan++) 
    delete *scan;

  ::close(fd);
}
