// VisBuffer object for SFXC/SoftwareCorrelator output datafiles
//
// HV 13/2/2007   initial version
//
#ifndef JIVE_J2MS_SFXCVISBUFFER_H
#define JIVE_J2MS_SFXCVISBUFFER_H

#include <jive/j2ms/VisibilityBuffer.h>
#include <jive/j2ms/CorrelatorSetup.h>
#include <jive/j2ms/scanlist.h>

#include <casacore/casa/aips.h>
#include <casacore/casa/BasicSL/String.h>
#include <casacore/casa/Arrays/Vector.h>
#include <casacore/casa/Arrays/Matrix.h>
#include <casacore/measures/Measures/Muvw.h>

#include <string>
#include <stdint.h>

// have to forward declare this, otherwise we get
// circular include-probs
namespace j2ms {
    class COFScan;
}

struct SFXCKey
{
  uint8_t antenna;
  int polarisation;
  std::pair<int, int> subband;
};

struct SFXCBaselineKey
{
  std::pair<uint8_t, uint8_t> baseline;
  std::pair<int, int> polarisation;
  std::pair<int, int> subband;
};

class SFXCVisBuffer: public VisibilityBuffer
{
 public: 
  // we support this construction signature
  SFXCVisBuffer(const casacore::String& vb, const Experiment* exp,
		const std::string& vbopt = std::string());

  // We *must* supply these [the VisBuffer interface]
  virtual casacore::Bool                   getCurrentVisibility( Visibility& v );
  virtual casacore::Bool                   next( void );
  virtual const casacore::String&          getRootDir( void ) const;
  virtual casacore::Int                    getVisBufId( void ) const;
  virtual casacore::String                 getProcessorSubType( void ) const;
  virtual CorrelatorSetup              correlatorSetup( void ) const;
  virtual ~SFXCVisBuffer();

 private:
  // List of pointers to scans. Memory management is left up to user
  // [you keep track wether or not you 0wn the pointer :)]
  typedef std::vector<j2ms::COFScan*>  scanptrlist_type;

  // CorrelatorSetup is fixed per visbuffer
  casacore::String           __mode;
  // In order to support multiple phase centers the SFXC visbuffer
  // must select the correct scan entry. 
  // The list of scans is now enriched with one entry per scan per phase
  // center so we not only need to pick the right entry based on time [which is
  // what happened before] but also on source id.
  // The source id is not (yet) in the SFXC output format but is encoded in
  // the correlator output file name. So currently we parse it out of there,
  // if present.
  // If a better way of propagating this metadata comes along we can adapt
  // accordingly quite easy; the infrastructure is already there.
  // We call it maybeSourceId because it will default to empty - to provide
  // backwards compatibility.
  std::string                __maybeSourceId;
  casacore::String           __rootdir;
  CorrelatorSetup            __setup;
  scanptrlist_type           __scanptrlist;
  // The mapping of station-id to stationbased info
  j2ms::station2listmap_t    station2listmap;

  // We have same temporary storage that we can
  // pre-allocate. We do not want to alloc this
  // over and over again. Once we know the
  // nr-of-channels we can alloc these
  typedef casacore::Vector<casacore::Complex>  complexvector_type;
  typedef casacore::Vector<casacore::Float>    floatvector_type;
  typedef casacore::Matrix<casacore::Complex>  complexmatrix_type;

  complexvector_type            cplxbuf; 
  floatvector_type              weight;
  complexmatrix_type            data;
  complexvector_type            cplxnullen;

  // Current status

  // Helper functions
  void createAntennaMapping(void);
  void createAntennaMapping(struct Output_header_global *);
  std::vector<casacore::Int> antenna_map;

  void createSourceMapping(void);
  void createSourceMapping(struct Output_header_global *);
  std::vector<std::string> source_map;

  void createSubbandMapping(void);
  std::map<std::pair<int, int>, casacore::uInt> subband_map;

  std::map<casacore::Int, casacore::MVuvw> uvw_map;
  int next_uvws;

  std::map<SFXCKey, double> weight_map;
  std::map<SFXCKey, int32_t> ntotal_map;
  std::map<SFXCBaselineKey, double> baseline_weight_map;
  int next_stats;

  int fd;
  std::vector<float *> visibilities;
  std::vector<SFXCBaselineKey> keys;
  int num_channels;

  unsigned int current_baseline, next_baselines;
  int current_slice, next_slice;

  std::set<std::pair<uint8_t, uint8_t> > baselines;
  std::set<std::pair<int, int> > subbands;

  std::set<std::pair<uint8_t, uint8_t> >::const_iterator baseline;
  std::set<std::pair<int, int> >::const_iterator subband;

  casacore::Int start_year, start_month, start_day, start_sec;
  casacore::Double integration_time;
  casacore::MVEpoch slice_time;

  int32_t correlator_version;
  int32_t job_nr;

  void read_timeslice(void);
  void read_timeslice_header(void);
  void read_uvws(void);
  void read_stats(void);
  void read_baseline(void);

  // Prohibit these
  SFXCVisBuffer();
  SFXCVisBuffer(const SFXCVisBuffer&);
  const SFXCVisBuffer& operator=(const SFXCVisBuffer&);
};

#endif
