#include <jive/j2ms/UNBVisBuffer.h>
#include <jive/j2ms/Experiment.h>
#include <jive/j2ms/COFScan.h>
#include <jive/j2ms/Scan.h>
#include <jive/j2ms/DayConversion.h>
#include <jive/labels/CorrelationCode.h>
#include <jive/Utilities/jexcept.h>
#include <jive/Utilities/stringutil.h>
#include <jive/Utilities/streamutil.h>

#include <casacore/casa/Quanta/MVTime.h>
#include <casacore/casa/Utilities/Regex.h>
#include <casacore/casa/BasicMath/Math.h>
#include <casacore/casa/Arrays/ArrayMath.h>

#include <sys/types.h>
#include <dirent.h>
#include <fcntl.h>
#include <libgen.h>
#include <limits.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

#include <cmath>
#include <list>
#include <limits>
#include <string>
#include <fstream>
#include <algorithm>
#include <iterator>
#include <stdexcept>

using namespace std;

const casacore::Float   unb_magic( 3.14 );
const casacore::Complex unb_nul(0.0f, 0.0f);

///////////////////////////////////////////////
//
//  Data structure describing the packets
//  coming out of the Jive Uniboard Correlator
//
///////////////////////////////////////////////


struct unb_32bit {
    int32_t     real;
    int32_t     imag;
    uint32_t    validity;
};
struct unb_64bit {
    int64_t     real;
    int64_t     imag;
    uint32_t    validity;
};

// This defines the header of a uniboard data packet
struct unb_header {
    // DON'T EVEN THINK OF CHANGING THE TYPES
    // OF THE FIELDS TO OTHER uint*_t TYPES
    // (ESPECIALLY THOSE FIELDS WHERE YOU THINK
    //  'HEY, THIS WOULD FIT INNA 8-bit QUANTITY')
    // BECAUSE IT WILL BREAK THE BIT-FIELDS!
    //
    // Fields in word 0
    uint32_t    freqbin:12;
    uint32_t    reserved1:6;
    uint32_t    ce:2;
    uint32_t    node:8;
    uint32_t    datasize:1;
    uint32_t    version:3;
    // Fields in word 1
    uint32_t    firstproduct:12;
    uint32_t    payloadsize:12;
    uint32_t    reserved2:8;
    // Words 2 + 3 are the timestamp
    uint32_t    seconds:32;
    uint32_t    framenr:32;

    // Convert bitfields in the header to
    // standard entities
    unb_ce_id    ce_id( void ) const {
        return unb_ce_id(this->node, ce);
    }

    unb_node     node_nr( void ) const {
        return (unb_node)(node & 0x7);
    }

    unb_backplane backplane_nr( void ) const {
        return (unb_backplane)((node >> 3) & 0x1f);
    }

    unb_timestamp timestamp( void ) const {
        return unb_timestamp(seconds, framenr);
    }

    unsigned int nproduct( void ) const {
        // the payloadsize is the number of
        // 32-bit words in the packet, including
        // the header but exclusive padding -
        // the CE will pad to an even number of
        // products
        // datasize==0 => 32bit data =>
        //      3x 32bit word / product
        // datasize==1 => 64bit data =>
        //      2x 64bit (re/im) + 32bit validity
        //      = 5 x 32bit word / product
        const unsigned int nword_per_product = (datasize?5:3);
        const unsigned int nproduct = (payloadsize - 4) / nword_per_product;

        // make sure payloadsize is multiple of the nword_per_product to
        // signal errors
        if( (nproduct * nword_per_product + 4)!=payloadsize )
            cerr << "unb_header: inconsistent header "
                 << "- ize != (nproduct*productsize + headersize)!" << endl; 
        return nproduct;
    }


    unsigned int packet_size( void ) const {
        // We already know the payloadsize. We must determine if it's being
        // padded or not: the CE outputs an even number of products.
        unsigned int       pktsz( payloadsize );
        const unsigned int nword_per_product = (datasize?5:3);
       
        // compute the modulo of the payloadsize with twice the
        // nword-per-product - if non-zero there's an odd amount
        // of products
        // NB: the number of products is to be derived from the actual
        //     payloadsize - i.e. the paktsize minus the header size!!!!!
        if( (payloadsize-4) % (nword_per_product+nword_per_product) )
            pktsz += nword_per_product;
        return 4*pktsz;
    }
};

ostream& operator<<(ostream& os, const unb_header& hdr) {
    return os << hdr.nproduct() << " products, "
              << "bin#" << hdr.freqbin << " "
              << "from CE " << hdr.backplane_nr() << "/" << hdr.ce_id();
}

/////////////////////////////////////////////
//
// All of the mapping entities
//
/////////////////////////////////////////////



///////////////////////////////
//
//    unb_subband
//
///////////////////////////////
unb_subband_key::unb_subband_key(unb_backplane ub, unb_subband sb, unsigned int offs):
    uniboard( ub ), subband( sb ), offset( offs )
{}

ostream& operator<<(ostream& os, const unb_subband_key& unb_sb) {
    return os << "SB#" << unb_sb.subband << " @" << unb_sb.offset;
}

// NOTE: we IGNORE the offset member here such that all unb_subband_key's
// that address the same uniboard, subband map to the same key.
bool operator<(unb_subband_key const& l, unb_subband_key const& r) {
    if( l.uniboard==r.uniboard )
        return l.subband<r.subband;
    return l.uniboard<r.uniboard;
}

///////////////////////////////
//
//    unb_ce_id
//
///////////////////////////////
unb_ce_id::unb_ce_id(unb_node n, unb_ce engine):
    node(n), ce( engine )
{}

bool operator<(unb_ce_id const& l, unb_ce_id const& r) {
    if( l.node!=r.node )
        return l.node<r.node;
    return l.ce<r.ce;
}

bool operator==(const unb_ce_id& l, const unb_ce_id& r) {
    return l.node==r.node && l.ce==r.ce;
}

ostream& operator<<(ostream& os, const unb_ce_id& ce_id) {
    return os << "CE[unb:" << board_number(ce_id.node) << " node:" << node_number(ce_id.node) << " ce:" << ce_id.ce << "]";
}

///////////////////////////////
//
//    unb_baseline
//
///////////////////////////////
unb_baseline::unb_baseline(unb_station a, unb_station b):
    x( (a<b)?(a):(b) ), y( (b>a)?(b):(a) ), swapped( a>b )
{}

bool operator==(unb_baseline const& l, unb_baseline const& r) {
    return l.x==r.x && l.y==r.y;
}
bool operator<(unb_baseline const& l, unb_baseline const& r) {
    if( l.x==r.x )
        return l.y<r.y;
    return l.x<r.x;
}
ostream& operator<<(ostream& os, unb_baseline const& ubl) {
   return os << ubl.x << "-" << ubl.y << (ubl.swapped?"*":"") << " (unb)";
}


///////////////////////////////
//
//    unb_product
//
///////////////////////////////
unb_product::unb_product( unb_baseline const& ubl, CorrelationCode const& cc):
    bl( ubl ), polcode( (ubl.swapped?(CorrelationCode(cc[1],cc[0])):cc) )
{}

bool operator==(const unb_product& l, const unb_product& r) {
    return l.bl==r.bl && l.polcode==r.polcode;
}
bool operator<(const unb_product& l, const unb_product& r) {
    if( l.bl==r.bl )
        return l.polcode<r.polcode;
    return l.bl<r.bl;
}

ostream& operator<<(ostream& os, unb_product const& prod) {
    return os << prod.bl << " " << prod.polcode;
}

///////////////////////////////
//
//    mapping_result_type
//
///////////////////////////////
mappingresult_type::mappingresult_type(const unb_product_map& pm, 
                                       const unb_enabled_products_map& epm):
    product_map( pm ), enabled_products_map( epm )
{}

///////////////////////////////
//
//    vex_output_type
//
///////////////////////////////
vex_output_type::vex_output_type(station_id st, vex_subband sb):
    vexstation( st ), vexsubband( sb )
{}

ostream& operator<<(ostream& os, const vex_output_type& vexo) {
    return os << "vex/out " << vexo.vexstation << "/sb#" << vexo.vexsubband;
}

///////////////////////////////
//
//    unb_input_type
//
///////////////////////////////
unb_input_type::unb_input_type(unb_backplane board, unb_station st, unb_subband sb):
    unbboard( board ), unbstation( st ), unbsubband( sb )
{}

bool operator<(const unb_input_type& l, const unb_input_type& r) {
    if( l.unbboard!=r.unbboard )
        return l.unbboard<r.unbboard;
    if( l.unbstation!=r.unbstation )
        return l.unbstation<r.unbstation;
    return l.unbsubband<r.unbsubband;
}

ostream& operator<<(ostream& os, const unb_input_type& unbi) {
    return os << "unb/in unb#" << unbi.unbboard << "/sta#" << unbi.unbstation << "/sb#" << unbi.unbsubband;
}

///////////////////////////////
//
//    unb_timestamp
//
///////////////////////////////
unb_timestamp::unb_timestamp():
    t_seconds( 0 ), t_framenr( 0 )
{}

unb_timestamp::unb_timestamp(uint32_t s, uint32_t f):
    t_seconds( s ), t_framenr( f )
{}

bool operator<(const unb_timestamp& l, const unb_timestamp& r) {
    if( l.t_seconds==r.t_seconds )
        return l.t_framenr<r.t_framenr;
    return l.t_seconds<r.t_seconds;
}

bool operator==(const unb_timestamp& l, const unb_timestamp& r) {
    return (l.t_seconds==r.t_seconds && l.t_framenr==r.t_framenr);
}
bool operator!=(const unb_timestamp& l, const unb_timestamp& r) {
    return !(operator==(l, r));
}

ostream& operator<<(ostream& os, const unb_timestamp& ts) {
    return os << "T" << ts.t_seconds << ":" << ts.t_framenr;
}

///////////////////////////////
//
//    unb_key
//
///////////////////////////////
unb_key::unb_key(unb_baseline b, unb_subband s, const CorrelationCode& cc):
    baseline( b ), subband( s ), polcode( cc )
{}

bool operator<(const unb_key& l, const unb_key& r) {
    if( l.baseline==r.baseline ) {
        if( l.subband==r.subband )
            return l.polcode<r.polcode;
        return l.subband<r.subband;
    }
    return l.baseline<r.baseline;
}

ostream& operator<<(ostream& os, const unb_key& key) {
    return os << "unb_key[bl=" << key.baseline << ", sb=" << key.subband << " pol=" << key.polcode << "]";
}


///////////////////////////////
//
//    vex_baseline
//
///////////////////////////////
vex_baseline::vex_baseline(vex_station a, vex_station b):
    x( (a<b)?(a):(b) ), y( (b>a)?(b):(a) ), swapped( a>b )
{}

ostream& operator<<(ostream& os, vex_baseline const& vbl) {
   return os << vbl.x << "-" << vbl.y << (vbl.swapped?"*":"") << " (vex)";
}
bool operator==(const vex_baseline& l, const vex_baseline& r) {
    return (l.x==r.x && l.y==r.y);
}
bool operator<(const vex_baseline& l, const vex_baseline& r) {
    if( l.x==r.x )
        return l.y<r.y;
    return l.x<r.x;
}


///////////////////////////////
//
//    vex_key
//
///////////////////////////////
vex_key::vex_key(vex_baseline b, vex_subband s):
    baseline( b ), subband( s )
{}

ostream& operator<<(ostream& os, const vex_key& key) {
    return os << "vex_key[bl=" << key.baseline << ", sb=" << key.subband << "]";
}

bool operator<(const vex_key& l, const vex_key& r) {
    if( l.baseline==r.baseline )
        return l.subband<r.subband;
    return l.baseline<r.baseline;
}

////////////////////////////////
//
//    vex_product
//
////////////////////////////////
vex_product::vex_product(vex_station x, vex_station y, vex_subband sb, const CorrelationCode& cc, unsigned int r, bool swp):
    need_swap( swp ), key(vex_baseline(x, y), sb), row( r ), polcode( cc )
{}

ostream& operator<<(ostream& os, const vex_product& vp) {
    return os <<  vp.key << (vp.need_swap ? ":SWP" : "") << " POL:" << vp.polcode << " row#" << vp.row;
}


///////////////////////////////
//
//    unb_data
//
///////////////////////////////
unb_data::unb_data(unsigned int nFreq, unsigned int nPol):
    // HV: 22-May-2014 If we're going to do spectral averaging, we have to
    //                 accumulate. And if we're going to accumulate
    //                 we'd better start from 0.0+0.0i rather than from
    //                 3.14+0.0i ....
    data(nPol, nFreq, casacore::Complex(0.0, 0.0)),
    weight(nPol, -unb_magic),
    freqbincount(nPol, 0),
    freqbinflag(nPol, nFreq, casacore::True)
{}

///////////////////////////////
//
//    unb_sb_data
//
///////////////////////////////
unb_sb_data::unb_sb_data(unsigned int nf, unsigned int nprod, unsigned int navg):
    nFreq( nf ), nAvg( navg ), nEnabledProd( nprod )
{
    const unsigned int nTotal = nFreq * nEnabledProd;

    if( nTotal == 0 )
        throw runtime_error(string("Either nFreq=0 ("+repr(nFreq)+") or nEnabledProd=0 ("+repr(nEnabledProd)));
    if( nAvg == 0 )
        throw runtime_error(string("nAvg=0, which is unacceptable"));

    // the data
    sb_data = new casacore::Complex[ nFreq*nEnabledProd ];
    ::memset(sb_data, 0, nTotal*sizeof(casacore::Complex));

    // the counter per product
    sb_fbincount = new casacore::uInt[ nEnabledProd ];
    ::memset(sb_fbincount, 0, nEnabledProd*sizeof(casacore::uInt));

    // the weight
    sb_weight = new casacore::Float[ nEnabledProd ];
    for(unsigned int i=0; i<nEnabledProd; i++)
        sb_weight[i] = -unb_magic;

    // and the boolean flags - initialized to .TRUE.
    // (nothing has arrived yet, so it is flagged. If a 
    //  spectral point comes in, the flag will be removed)
    sb_fbinflag = new casacore::Bool[ nTotal ];
    for(unsigned int i=0; i<nTotal; i++)
        sb_fbinflag[i] = casacore::True;
    // Allocate space for the two-dimensional unaveraged-freqbin flags
    sb_fbinflag_raw = new casacore::Bool[ nTotal * nAvg ];
    for(unsigned int i=0; i<(nTotal * nAvg); i++)
        sb_fbinflag_raw[i] = casacore::True;
}

unb_sb_data::~unb_sb_data() {
    delete [] sb_data;
    delete [] sb_fbincount;
    delete [] sb_weight;
    delete [] sb_fbinflag;
    delete [] sb_fbinflag_raw;
}

////////////////////////////////////////////////////
//
// Keeping packet info based on CE id/first product
//
////////////////////////////////////////////////////

ce_id_first_key::ce_id_first_key(const unb_ce_id& id, const unsigned int fp):
    ce_id( id ), first_prod( fp )
{}

bool operator<(const ce_id_first_key& l, const ce_id_first_key& r) {
    if( l.ce_id==r.ce_id )
        return l.first_prod < r.first_prod;
    return l.ce_id < r.ce_id;
}

fpoffset_type::fpoffset_type(unsigned int prod):
    product_offset( prod )
{}


///////////////////////////////////////
//
// The filestatus thingy
//
///////////////////////////////////////

filestatus_type::filestatus_type(const string& fn, int f):
    fd( f ), idx( 0 ), sz( 0 ), filename( fn ), eof( false )
{
    // pre-read data into the buffer
    do_read();
}


unb_header const* filestatus_type::current( void ) {
    unb_header const*  ptr = (unb_header const*)&buffer[idx];
    const unsigned int n = (sz - idx);

    return (n<16)?(0):((n>=ptr->packet_size())?ptr:0);
}

void filestatus_type::next( void ) {
    unsigned int      n;
    unb_header const* ptr = this->current();

    // if no current yet (or not anymore)
    // no point in doing anything
    if( !ptr )
        return;

    // move idx to next packet
    // as long as our buffer is > the maximum packet size (9000 bytes) this
    // can never go beyond the buffer end. What *could* happen is that the
    // buffer is *exactly* as large as one packet, in which case the idx
    // would be addressing one outside the buffer
    idx += ptr->packet_size();

    // There should be enough bytes for the new current frame in the buffer
    //   * less than 16 bytes in buffer? there isn't a full header 
    //     present at index 'idx' so we're stuft!
    //   * more than 16 bytes but not enough for the whole packet?
    //     stuft also
    //   NOTE: a do_read() action never repositions the 'read' pointer we
    //   call 'idx'!
    //   after the read action we must re-evaluate these conditions
    //   and possibly decide there aren't going to be sufficient bytes
    //   available.
    while(  // not enough bytes for the new header!
            (n=(sz-idx))<16 ||
            // there are enough bytes for a header but not the payload
            (n>=16 && n<((unb_header const*)&buffer[idx])->packet_size()) ) {
            // We may shift remaining bytes to front
            // to make as much space for new bytes
            ::memmove(&buffer[0], &buffer[idx], n);
            idx = 0;
            sz  = n;
            // now attempt to read
            if( !do_read() ) {
                ptr = 0;
                break;
            }
    }
    return;
}

bool filestatus_type::do_read( void ) {
    if( eof )
        return false;

    // Can we read moar bytes?
    int       nr;

    // always read a multiple of four bytes (because of swapping)
    // FIXME: may go as soon as the unb does the byteswap for us!
    // 
    // HV: 04oct2012 - yes! it can go! the UniBoard does the swap!
    if( (nr=::read(fd, &buffer[sz], (sizeof(buffer)-sz)))<=0 ) {
        eof = true;
        if( nr<0 ) {
            cerr << "[" << filename << "]/Failed to read bytes - "
                 << ::strerror(errno) << endl;
        }
        // transform to no-op + false return value
        nr = 0;
    }
    // And bump the endposition of the buffer
    sz += nr;
    return (nr>0);
}

filestatus_type::~filestatus_type() {
    if( fd>=0 )
        ::close( fd );
    //cout << "~filestatus: closed fd=" << fd << " [" << filename << "]" << endl;
}


////////////////////////////////////////////////////////////
//
//   Human readable form of the correlator config
//
////////////////////////////////////////////////////////////
std::ostream& operator<<(std::ostream& os, fw_config_t const& fwconfig) {
    return os << "nDDR: " << fwconfig.n_ddr_modules << " "
              << "nFFT: " << (1 << fwconfig.log2_n_fbin) << " "
              << "nSUB: " << fwconfig.n_band_p_station;
}



////////////////////////////////////////////////////////////
//
//   The actual visibility buffer for UniBoard data
//
////////////////////////////////////////////////////////////

struct stringfinder {
    stringfinder(const string& needle):
        __needle( needle )
    {}

    bool operator()(const string& haystack) const {
        return haystack==__needle;
    }

    private:
        string __needle;
};

UNBVisBuffer::UNBVisBuffer( casacore::String const& vb,
			     Experiment const* exp,
			     string const& vbo) :
  VisibilityBuffer(exp),
  __debug( 0 ),
  __doNormalization( true ),
  __fw_config( 0 ),
  __nChannelAvg( 1 ),
  __nFFTpIntg( 0 ),
  __weightScale( 1 ),
  __rootdir(exp->getRootDir())
{
    // Transform the experiment's scanlist into a list of scanpointers
    vector<string>                       opts = ::split(vbo, ',');
    const j2ms::scaninfolist_t&          scanlist(exp->getScanlist());
    j2ms::scaninfolist_t::const_iterator scan;
   
    // debug flag set? 
    if( std::find_if(opts.begin(), opts.end(), stringfinder("debug"))!=opts.end() )
        __debug = 1;

    // Option "normalize=<digit>" to influence normalization?
    unsigned int                         normalVal;
    vector<string>::const_iterator       curopt;

    for(curopt=opts.begin(); curopt!=opts.end(); curopt++) {
        int   tmp;
        if( ::sscanf(curopt->c_str(), "normalize=%u", &normalVal)==1 )
            __doNormalization = (normalVal>0);
        if( ::sscanf(curopt->c_str(), "debug=%d", &tmp)==1 )
            __debug = tmp;
    }
    if( !__doNormalization )
        cout << "WARN: normalization switched off - providing raw correlator counts" << endl;

    for( scan = scanlist.begin(); scan != scanlist.end(); scan++)
        __scanptrlist.push_back(new j2ms::COFScan(this, *scan));

    char buf[PATH_MAX];
    if (::realpath(__rootdir.c_str(), buf) == NULL)
      THROW_JEXCEPT("Unresolvable path" << __rootdir);

#if 0
    string path = __rootdir + "/" + vb;
    fd = ::open(path.c_str(), O_RDONLY | O_LARGEFILE);
    if (fd == -1)
      THROW_JEXCEPT("Cannot open " << path);
#endif
    // Step 1: check if this is UniBoard data
    //         If it is, the 'vb' will be a JSON file, containing the
    //         metadata for that subjob. The root of the datafiles will
    //         be the dirname of the the concatenation of the experiments
    //         rootdir + the json filename
    const string jsonfilename = __rootdir + "/" + vb + ((vb.find(".json")==string::npos)?".json":"");
    char*        dup          = ::strdup( jsonfilename.c_str() );
    const string vbroot       = ::dirname(dup);
    ::free( dup );

    if( __debug )
        cout << "UNBVisBuffer: vb='" << vb << "' =>" << endl
             << "              __rootdir=" << __rootdir << endl
             << "              jsonfilename=" << jsonfilename << endl
             << "              vbroot=" << vbroot << endl;

    // Check if we can read the file
    if( ::access(jsonfilename.c_str(), R_OK) )
        THROW_JEXCEPT(string("Cannot access '")+jsonfilename+"'");

    // Read other configuration info from the jsonfile
    ifstream           jsonfile( jsonfilename.c_str(), ios_base::in );
    Json::Value        root;
    Json::Reader       reader;

    if( !reader.parse(jsonfile, root) )
        THROW_JEXCEPT("UNBVisBuffer: Failed to parse json file '"+jsonfilename+"'\n"
                      +reader.getFormattedErrorMessages());

    // Read the correlator firmware config from the JSON. If we don't know what the
    // firmware did we cannot usefully process its data, now can we?
    const Json::Value& corrConfigStr = root["setup"]["correlatorconfig"];

    if( corrConfigStr.isNull() )
        THROW_JEXCEPT("UNBVisBuffer: Did not find setup.correlatorconfig key in the JSON file '"+jsonfilename+"'\n");
    __fw_config.correlatorconfig = corrConfigStr.asUInt();
    if( __debug )
        cout << "UNBVisBuffer: firmware setup = " << __fw_config << endl;

    // Step 2: load correlator config and build
    //         mappings etc
    __setup = CorrelatorSetup::correlatorSetupFromJSON(jsonfilename);

    // If the UNB correlator is asked to produce less frequency bins than it
    // natively produces it gives us *accumulated* values for weight.
    // So we must compensate for that!
    // Note: do NOT use the shape of the actual data array in normalization;
    //       we could be doing spectral averaging too ... so we must really
    //       look at the setup.spectralpoints from the JSON file to figure
    //       out what the UB has done to the data
    __weightScale = (1 << __fw_config.log2_n_fbin) / __setup.nrOfSpectralPoints();

    if( __debug && __doNormalization && __weightScale!=1 )
        cout << "UNBVisBuffer: using weight scaling of " << __weightScale << ", assuming UB did spectral accumulation" << endl;


    //     2a.1: Check if channel averaging is set. We can only do this 
    //         sensibly after the correlator setup has been read
    const string    avgChannelOpt( exp->optionValue("avgchannel") );
   
    if( avgChannelOpt.size() ) { 
        char*           eocptr;
        unsigned long   avgChan;

        avgChan = ::strtoul(avgChannelOpt.c_str(), &eocptr, 0);
        if( eocptr==avgChannelOpt.c_str() || *eocptr!='\0' )
            THROW_JEXCEPT("'avgchannel' option value '" << avgChannelOpt << "' is not a number")
        // Ok, So it *was* a number, now check if it's a _valid_ number
        // (power of 2). Our test fails if avgChan=0xfffffffffff (ie all 'f's)
        if( avgChan==0 || (avgChan & (avgChan-1))!=0 )
            THROW_JEXCEPT("'avgchannel' value of " << avgChan << " is NOT a power of 2 that's >0 !");
        // Finally, verify that after averaging, we're left with at least 16
        // spectral points
        if( (__setup.nrOfSpectralPoints()/avgChan)<16 )
            THROW_JEXCEPT("averaging " << avgChan << " channels in "
                          << __setup.nrOfSpectralPoints() << " results in <16 output spectral points");
        __nChannelAvg = avgChan;

        if( __nChannelAvg>1 ) {
            cout << "##############################################################" << endl
                 << "#### Averaging " << __nChannelAvg << " channels, input " << __setup.nrOfSpectralPoints()
                 << ", producing " << __setup.nrOfSpectralPoints()/__nChannelAvg << endl
                 << "##############################################################" << endl;

            // But now we have to provide a new correlator setup, one with less
            // spectral points ...
            __setup = CorrelatorSetup( __setup.nrOfSpectralPoints()/__nChannelAvg,
                                       __setup.getCorrelations(),
                                       __setup.getIntegrationTime(),
                                       __setup.getSoftwareVersion(),
                                       __setup.l2pmap(),
                                       __setup.getBocfRate() );
        }
    }
    if( __nChannelAvg>1 )
        cout << "##############################################################" << endl
         << "#### Averaging " << __nChannelAvg << " channels, input " << __setup.nrOfSpectralPoints()
         << ", producing " << __setup.nrOfSpectralPoints()/__nChannelAvg << endl
         << "##############################################################" << endl;

    //     2a: from configured polarizations build the polarization-to-row
    //         mapping so we always know where data needs to end up
    //     2b: also build the vector of correlations - the Visibilities
    //         get a vector rather than a set [in the vector *we* can
    //         define the order of the polarization products, in the set
    //         we can't]
    unsigned int                      row;
    const correlationset_t&           correlationset = __setup.getCorrelations();
    correlationset_t::const_iterator  curpol;

    __correlations.resize( correlationset.size() );

    for( row=0, curpol=correlationset.begin();
         curpol!=correlationset.end(); curpol++, row++) {
            if( __pol2rowmap.insert( make_pair(*curpol, row) ).second==false )
                throw runtime_error("Failed to insert entry into polarization->row mapping?!");
            __correlations[ row ] = *curpol;
    }

    if( __debug ) 
        cout << "UNBVisBuffer: got correlator setup" << std::endl
             << "  " << __setup << std::endl;



    // Step 3: read start time of the correlator. Timestamps are counted in
    //         'integrations since the start of correlation' so multiplying
    //         by the integration time will give us the actual timestamp
    casacore::Quantity zulu;
    const Json::Value& zulustr = root["setup"]["starttime"];

    if( zulustr.isNull() )
        throw runtime_error("Correlator start time not found in the setup!");
    if( !zulustr.isString() )
        throw runtime_error("Correlator start time is NOT a string!");


    if( casacore::MVTime::read(zulu, casacore::String(zulustr.asString()), casacore::True)==false )
        throw runtime_error(string("Failed to convert correlator start time '")+zulustr.asString()+"'");

    __zulu            = casacore::MVEpoch(zulu);
    casacore::MVTime::setFormat(casacore::MVTime::FITS, 9);

    if( __debug )
        cout << "UNBVisBuffer: found start time "
             << casacore::MVTime(zulu)
             << " (" << zulustr << ")"
             << endl;

    // We should be reading the integration time in FFTs out of the JSON
    const Json::Value& nFFTstr = root["setup"]["integrationtimefft"];

    if( !nFFTstr.isNull() ) {
        // If the key is there, it better be a uInt orelse ...
        // (also, do not allow setting to value of 0
        if( !nFFTstr.isUInt() || (__nFFTpIntg=nFFTstr.asUInt())==0 )
            throw runtime_error(string("'integrationtimefft' value '")+nFFTstr.asString()+"' is not an unsigned integer > 0");
    }

    // If normalizing requested but __nFFTpIntg<=6 we cannot do that
    // because in normalizing we assume the first six FFT periods are
    // invalid (because they are). Thus we always subtract '6' from
    // the 'number of FFTs per integration' for the maximum weight.
    if( __doNormalization && __nFFTpIntg<=6 ) 
        throw runtime_error(
             string("\n"
                    "-----------========== ERROR ==========-------------\n" 
                    "---                                             ---\n"
                    "---   Normalization of data was requested but   ---\n"
                    "---   'nFFTpIntg' is <= 6.                      ---\n"
                    "---   You have two options:                     ---\n"
                    "---   1. Fix the JSON file:                     ---\n"
                    "---     \"setup\" : { \"integrationtimefft\" : X, } ---\n"
                    "---   2. Turn off normalization:                ---\n"
                    "---      append %normalize=0 to VisBuffer name  ---\n"
                    "---                                             ---\n"
                    "-----------========== ERROR ==========-------------\n"));

    // The input mapping defines which VEX output was connected to which
    // uniboard input 
    __inputmap        = UNBVisBuffer::readInputMap(root);

    if( __debug )
        for(inputmap_type::const_iterator curinp=__inputmap.begin();
            curinp!=__inputmap.end(); curinp++)
                cout << "Inputmap entry " << curinp->first << " -> " << curinp->second << endl;

    // The correlator engine to subband mapping
    __ce_sb_map       = UNBVisBuffer::readCESubbandMap(root);

    // Which products were enabled? This is per correlation engine.
    // (and correlation engine id is a combination of (back)node id and 
    //  correlation-engine-within-node
    mappingresult_type mappingresults = UNBVisBuffer::readEnabledProductsMap(root);

    __productmap      = mappingresults.product_map;
    __enabledproducts = mappingresults.enabled_products_map;

    // For each key in the enabled products list (ie the correlation
    // engines) we check there is an entry in the CE->subband map
    // Update: Jun 2015: Each CE can compute just part of a spectrum.
    //         For each entry that maps to the same unb_subband_key,
    //         verify that the enabled products match up
    typedef std::map<unb_subband_key, unb_enabled_products> unbce_map_type;
    unbce_map_type  unbcemap;

    for(unb_enabled_products_map::const_iterator curce=__enabledproducts.begin();
        curce!=__enabledproducts.end(); curce++) {
            unb_ce_sb_map::iterator cesbptr = __ce_sb_map.find(curce->first);
            if( cesbptr==__ce_sb_map.end() )
                throw logic_error(string("The correlation engine '")+repr(curce->first)+
                                  "'  mentioned in the enabled products "
                                  "does not appear in the 'ce_sb' mapping!");
            // Either insert an entry into unbcemap with the current enabled
            // products or verify that what's mapped to is identical
            unbce_map_type::iterator unbceptr = unbcemap.find( cesbptr->second );

            if( unbceptr==unbcemap.end() ) {
                unbcemap.insert( make_pair(cesbptr->second, curce->second) );
                continue;
            }
            if( unbceptr->second!=curce->second )
                throw logic_error(string("The enabled products for correlation engine '")+repr(curce->first)+
                                  "', which maps to "+repr(cesbptr->second)+", are different from previous enabled "+
                                  "products for this subband (computed by a different CE)!");
    }

    // Now we go over all enabled products, find all 'unb_key' values and if
    // it wasn't unmapped to vex values yet, do that now.
    // (Need to check if all stations/subbands mentioned in the input map
    // exist in the VEX file)
    AntennaInfo                ai;
    const AntennaDB&           antdb = exp->getAntennaDB();
    map<string, vex_station>   sta2id;

    // 1. build mapping of vex station name (string) to vex station id (int)
    for(size_t i=0; i<antdb.nrAntennas() && antdb.searchAntenna(ai, (casacore::Int)i); i++)
        if( sta2id.insert( make_pair(ai.getTLC().c_str(), (vex_station)i) ).second==false )
            throw runtime_error(string("Failed to insert VEX station ")+ai.getTLC().c_str()+" into mapping?!");

    // 2. Loop over all enabled products
    for(unb_enabled_products_map::const_iterator curce=__enabledproducts.begin();
        curce!=__enabledproducts.end(); curce++) {
            // We know which CE we're dealing with so we can already deduce
            // the logical subband and the actual uniboard.
            // It's been already verified that all entries exist.
            const unb_subband_key  unb_sb    = (__ce_sb_map.find(curce->first))->second;

            // It's also already verified that all ce_id's that map to the
            //      same unb_subband_key share the same enabled products, so
            //      if an entry in the enabled vex products map already
            //      exists for the indicated unb_subband_key, we can skip
            //      checking it
            if( __enabled_vex_products.find(unb_sb)!=__enabled_vex_products.end() )
                continue;

            // Now we can inspect all enabled products
            for(unb_enabled_products::const_iterator prodnum=curce->second.begin();
                prodnum!=curce->second.end(); prodnum++) {
                    // from product number -> product details
                    unb_product_map::const_iterator prodptr = __productmap.find(*prodnum);
                    // in product details we find logical unb_stations and
                    // logical inputs. We look those up in the input mapping
                    // to see if something is sent to this enabled product.
                    inputmap_type::const_iterator   xptr, yptr;

                    xptr = __inputmap.find( unb_input_type(unb_sb.uniboard, prodptr->second.bl.x, unb_sb.subband) );
                    yptr = __inputmap.find( unb_input_type(unb_sb.uniboard, prodptr->second.bl.y, unb_sb.subband) );

                    if( xptr==__inputmap.end() )
                        throw logic_error(string("Failed to find entry in input mapping for X-part of product #")+
                                            repr(*prodnum)+" = "+repr(prodptr->second)+", unb_subband "+repr(unb_sb));
                    if( yptr==__inputmap.end() )
                        throw logic_error(string("Failed to find entry in input mapping for Y-part of product #")+
                                            repr(*prodnum)+" = "+repr(prodptr->second)+", unb_subband "+repr(unb_sb));

                    // Now we have vex station name + vex subband
                    // check if vex station name is in the vex antenna db
                    map<string, vex_station>::const_iterator  xstaptr, ystaptr;

                    xstaptr = sta2id.find( xptr->second.vexstation );
                    ystaptr = sta2id.find( yptr->second.vexstation );

                    if( xstaptr==sta2id.end() )
                        throw logic_error(string("Failed to find station '")+xptr->second.vexstation+"' "+
                                          " (in input mapping) in the AntennaDB for this experiment");
                    if( ystaptr==sta2id.end() )
                        throw logic_error(string("Failed to find station '")+yptr->second.vexstation+"' "+
                                          " (in input mapping) in the AntennaDB for this experiment");

                    // Verify that both x+y map to the same vex subband
                    if( xptr->second.vexsubband!=yptr->second.vexsubband )
                        throw logic_error(string("The unmapped VEX subbands do not match for ")+
                                          "product #"+repr(*prodnum)+" = "+repr(*prodnum)+", unb_subband #"+
                                          repr(unb_sb)+". X = "+repr(xptr->first)+", Y = "+repr(yptr->first));

                    // Now we have everything:
                    //  logic: unb_station x, y  + subband
                    //  phys:  vex station id x, y + vex subband
                    // Do determine the final amount of swap to be applied;
                    // only need to swap the phase if unb/vex baselines are
                    // computed in the 'wrong' order; i.e. if one of them is
                    // swapped.
                    const unb_key          unb(prodptr->second.bl, unb_sb.subband, prodptr->second.polcode);
                    const vex_baseline     vbl(xstaptr->second, ystaptr->second);
                    const bool             do_swap( unb.baseline.swapped!=vbl.swapped );
                    const CorrelationCode& unbpol( prodptr->second.polcode );
                    const CorrelationCode  vexpol( do_swap ? CorrelationCode(unbpol[1], unbpol[0]) : unbpol );
                    // Look up the row number for the unmapped polarization
                    // combination
                    const pol2rowmap_type::iterator currow = __pol2rowmap.find( vexpol );

                    if( currow==__pol2rowmap.end() )
                        throw runtime_error(string("Failed to find destination row number for polarization product ")+
                                            repr(vexpol)+" "+repr(vbl)+" [UNB: "+repr(unb)+"]");
                    // Now that we have all vital unmapped info, store it in
                    // the mapping
                    const vex_product      vex(xstaptr->second, ystaptr->second, xptr->second.vexsubband, vexpol, currow->second, do_swap);

                    // Append to the products list for the current ce_id
                    __enabled_vex_products[ unb_sb ].push_back( vex );
            }
    }

    if( __debug ) {
        cout << "Productmap:" << endl;
        for(unb_product_map::const_iterator pp=__productmap.begin(); 
            pp!=__productmap.end(); pp++)
                cout << "#" << pp->first << " -> " << pp->second << endl;

        cout << "Enabled products:" << endl;
        for(unb_enabled_products_map::const_iterator p=__enabledproducts.begin();
            p!=__enabledproducts.end(); p++)
                cout << p->first << ": " << print_container(p->second) << endl;
        cout << "UNB_SUBBAND => VEX product" << endl;
        for(enabled_vex_products_map::const_iterator curce=__enabled_vex_products.begin();
            curce!=__enabled_vex_products.end(); curce++)
                cout << curce->first << ": " << print_container(curce->second) << endl;
    }

    // What is there left to do?
    // * open all datafiles
    this->openDataFiles(root, vbroot);
    

    // trigger reading of the first visibilities. We insert a dummy
    // integration in the cache and position the baseline iterator at the
    // end of the baselines (there aren't any).
    // .next() will try to advance and trigger reading of the next bit of
    // data
    __integrationcache.insert( make_pair(unb_timestamp(), unb_integration()) );
    __currentintegration = __integrationcache.begin();
    __currentbaseline    = __currentintegration->second.end();
    next();
}


casacore::Bool
UNBVisBuffer::getCurrentVisibility(Visibility &v) {
    casacore::MVEpoch                        t0;
    static ExtVisLabel                   label;
#if 0
    vsb2sidebandmap_type::const_iterator vsb2sbptr;
#endif
    if( __currentintegration==__integrationcache.end() )
        return false;

    unb_data&       data = __currentbaseline->second;
    const vex_key&  key  = __currentbaseline->first;

    // If we find lower side band data, reverse the frequency
    // axis and take the complex conjugate
    //
    // NOTE NOTE NOTE NOTE NOTE NOTE
    //
    // HV: 22-May-2014 Sergei P. informed us that the UniBoard
    //                 multiplies lower side band data by the
    //                 Nyquist frequency. This has the effect
    //                 of _shifting_ the lower side band data
    //                 to the same part of the spectrum and 
    //                 NOT flip it.
    //                 This is turn means that we don't have 
    //                 to flip it either.
#if 0
    if( (vsb2sbptr=__vsb2sidebandmap.find(key.subband))!=__vsb2sidebandmap.end() &&
        vsb2sbptr->second==BaseBand::lower ) {
            // Matrix = nrow x ncol, ncol == n spectral channels, always even
            casacore::Complex       cplx;
            unb_complexmatrix&  m( data.data );

            for( casacore::uInt nr = 0; nr<m.nrow(); nr++ ) {
                const casacore::uInt ncol = m.ncolumn();
                for( casacore::uInt nc = 0; nc<ncol/2; nc++ ) {
                    cplx = conj(m(nr, nc));
                    m(nr, nc) = conj(m(nr, ncol - nc - 1));
                    m(nr, ncol - nc -1) = cplx;
                }
            }
    }
#endif
    label.subbandnr    = key.subband;
    label.baseline     = BaseLineIdx(key.baseline.x, key.baseline.y);
    label.correlations = __correlations;

#if 0
    cout << key.baseline << "/" << key.subband << " @ " 
         << casacore::MVTime::setFormat(casacore::MVTime::DMY, 9)
         << casacore::MVTime(t0) << endl;
#endif
    v = Visibility(label, data.data, Visibility::frequency,
                   casacore::False, data.weight, 
                   casacore::MEpoch( __t0, casacore::MEpoch::UTC ),
                   j2ms::noMuvw,
                   this);
    return true;
}


// Helper stuff for sorting a mode's subbands by frequency AND sideband
struct basebandsorter {
    bool operator()(const BaseBand* l, const BaseBand* r) const {
        // Not same start frequency? Sort by it then
        if( !casacore::nearAbs(l->getStartfreq(), r->getStartfreq(), 1.0e-6) )
            return l->getStartfreq() < r->getStartfreq();
        // Not same bandwidth? Sort by that then
        if( !casacore::nearAbs(l->getBandwidth(), r->getBandwidth(), 1.0e-6) )
            return l->getBandwidth() < r->getBandwidth();
        // Not same side band? Use that as final sort key
        return l->getSideBand() < r->getSideBand();
    }
};

typedef set<const BaseBand*, basebandsorter>    sortedsb_type;


casacore::Bool
UNBVisBuffer::next(void)
{
    // If we've come to the end of the integrations we're don
    if( __currentintegration==__integrationcache.end() )
        return casacore::False;

    // Now we can try to advance the baseline iterator
    if( __currentbaseline!=__currentintegration->second.end() )
        __currentbaseline++;

    // haven't reached the end of the integration yet?
    if( __currentbaseline!=__currentintegration->second.end() )
        return casacore::True;

    // Ok so we *did* reach the end-of-an-integration so it can go
    __integrationcache.erase( __currentintegration );

    // Read moar data (if any)
    readNext();

    // Next integration is at the beginning of the cache
    // (which may be empty if there is no moar data)
    __currentintegration = __integrationcache.begin();

    // Position baselineiterator at begin of next integration, if any
    if( __currentintegration!=__integrationcache.end() ) {
        __currentbaseline = __currentintegration->second.begin();

        // And attempt to locate the actual scan the new integration is in;
        // set timestamp + current scan pointer
        const double                     t_int = __setup.getIntegrationTime();
        const unb_timestamp&             ts    = __currentintegration->first;
        scanptrlist_type::const_iterator scanptrptr;

        // compute current timestamp
        if( __debug>2 )
            cout << "UNBVisBuffer::next()/found new integration with time stamp " << ts.t_framenr << endl;
        __t0 = __zulu + casacore::MVEpoch( casacore::Quantity(ts.t_framenr*t_int+t_int/2.0, "s") );

        // If we find which scan the current integration came from, we
        // can/must re-initialize the VEX subband-to-sidebandedness mapping
        if((scanptrptr=find_scaninfo_by_time(__scanptrlist, __t0))==__scanptrlist.end()) {
            this->setCurrentScanptr( 0 );
            __vsb2sidebandmap.clear();
        } else {
            string        mode;
            j2ms::Scan*   scanptr = *scanptrptr;

            this->setCurrentScanptr( scanptr );

            // From the current scan, find the mode. If it's != current
            // mode, re-initialize the vsb2sideband mapping
            if( (mode=scanptr->getFrequencyCode())!=__mode ) {
                ostringstream     dbg;
                FrequencyConfig   fc;

                if( !this->getExperiment().getFrequencyDB().searchFrequencyConfig(fc, mode) ) {
                    THROW_JEXCEPT("Find data from scan '" << scanptr->getScanNo() << "' " <<
                                  " with mode '" << mode << "' which isn't found in the frequency database");
                }
                // Build new mapping
                __vsb2sidebandmap.clear();

                dbg << "  Initializing VEX subband => sideband mapping for mode " << mode << endl;

                // Sort basebands by frequency
                unsigned int                  i;
                sortedsb_type                 sortedsb;
                sortedsb_type::const_iterator sbptrptr;

                for( i=0; i<fc.getNrBaseBands(); i++ )
                    if( !sortedsb.insert(fc.getBaseBand(i)).second )
                        THROW_JEXCEPT("Duplicate subband inserted - " <<
                                      fc.getBaseBand(i)->getStartfreq() <<
                                      " mode=" << mode);
                // Now that we have them sorted, we can create a mapping
                // of subband => sideband; we relabel them 0..(N-1)
                for(sbptrptr=sortedsb.begin(), i=0; sbptrptr!=sortedsb.end(); sbptrptr++, i++) {
                    const BaseBand*  bbptr = (*sbptrptr);
                    __vsb2sidebandmap[ i ] = bbptr->getSideBand();
                    dbg << "(" << i << "=>" << bbptr->getSideBand() << ") ";
                }
                if( __debug )
                    cout << dbg.str() << endl;

                __mode = mode;
            }
        }

        // If requested, do normalization
        if( __doNormalization )
            this->doNormalize( __currentintegration );
    }

    return (__currentintegration!=__integrationcache.end());
}

const casacore::String&
UNBVisBuffer::getRootDir() const
{
    return __rootdir;
}

casacore::Int
UNBVisBuffer::getVisBufId() const
{
    return 0;
}

casacore::String
UNBVisBuffer::getProcessorSubType() const
{
    return "UniBoard v1";
}

CorrelatorSetup
UNBVisBuffer::correlatorSetup() const
{
    return __setup;
}

UNBVisBuffer::~UNBVisBuffer()
{
    scanptrlist_type::iterator scan;

    for(scan = __scanptrlist.begin(); scan != __scanptrlist.end(); scan++) 
        delete *scan;
}

// Open all datafiles listed in the "datafiles" list
void
UNBVisBuffer::openDataFiles(const Json::Value& json,
                            const string& visbufrootdir )
{
    // <root>."datafiles" = [ "<filename1>" , ... ]
    const Json::Value&                   datafiles = json["datafiles"];

    if( datafiles.isNull() || !datafiles.isArray() )
        throw runtime_error("The datafiles key is not present/not an array in the JSON file");

    for(Json::ArrayIndex idx=0; idx<datafiles.size(); idx++) {
        // Each entry in the array should be a string
        Json::Value        fname_j = datafiles[idx];

        if( fname_j.isNull() || !fname_j.isString() )
            throw runtime_error("An antry in the datafiles key in the JSON file is NULL or not a string");

        // if filename starts with "." we take it to mean a path relative to 
        // the visibilitybuffer rootdir (ie where the JSON file itself is
        // located).
        int                fd    = -1;
        const string       fname = fname_j.asString();

        if( fname.empty() )
            throw runtime_error("There is an empty datafile in the 'datafiles' list");

        if( fname[0]=='.' ) {
            // now we open the file
            if( (fd=::open((visbufrootdir+"/"+fname).c_str(), O_LARGEFILE|O_RDONLY))<0 )
                throw runtime_error(string("Failed to open file '")+visbufrootdir+"/"+fname+"'");
        } else {
            if( (fd=::open(fname.c_str(), O_LARGEFILE|O_RDONLY))<0 )
                throw runtime_error(string("Failed to open file '")+fname+"'");
        }
        if( fd<0 )
            throw domain_error(string("datafile '")+fname+"' is not a supported protocol");

        __filemap.push_back( new filestatus_type(fname, fd) );
    }
    return;
}

// locally overload operator< for two iterators to entries in the filemap
bool operator<(const filemap_type::iterator& l, const filemap_type::iterator& r) {
    return (*l)->fd < (*r)->fd;
}

typedef std::map<unb_ce_id, unsigned int>   statmap_type;

// Read the next bit of data
bool UNBVisBuffer::readNext( void ) {
    typedef std::set<filemap_type::iterator>       fileset_type;
    typedef std::list<filemap_type::iterator>      filelist_type;
    typedef std::map<unb_timestamp, filelist_type> timelist_type;

    // We can short-circuit this: if there are no more files left than we're
    // quickly done!
    // Note: we may have run out of files but that doesn't mean the cache
    //       is empty.
    if( __filemap.empty() && __sb_integrationcache.size()==0 )
        return false;

    // we must find the minimum timestamp in the datafiles.
    // once we start at the lowest timestamp the algorithm should work fine
    unsigned int       packets_read = 0;
    const unsigned int nFreq        = __setup.nrOfSpectralPoints();
    const unsigned int nPol         = __setup.getCorrelations().size();
    // the idea is to keep on reading until the integration cache contains
    // three entries - then we're quite sure the first integration has all
    // the data it can have. The second may still miss data [the files
    // contain raw UDP packet dumps - they may have been reordered]
    while( __sb_integrationcache.size()<3 ) {
        statmap_type     prodcnt;
        fileset_type     eraselist;
        timelist_type    filebytime;

        // Find the minimum timestamp across all the files. Mark the
        // exhausted ones for erasion.
        for(filemap_type::iterator fptr=__filemap.begin();
                fptr!=__filemap.end(); fptr++)
            if( (*fptr)->current() )
                filebytime[(*fptr)->current()->timestamp()].push_back(fptr);
            else
                eraselist.insert( fptr );

        // Immediately deal with exhausted files so we don't have to deal
        // with them anymore
        for(fileset_type::const_iterator curdel=eraselist.begin();
                curdel!=eraselist.end(); curdel++) {
            delete (**curdel);
            __filemap.erase( *curdel );
        }

        // No more entries in the filebytime mapping => all files exhausted
        if( filebytime.empty() )
            break;

        // For each time stamp we keep a mapping of unb_subband to
        // its data. All we need to find right now is the map itself
        timelist_type::iterator     curtime = filebytime.begin();
        unb_sb_integration&         curint  = __sb_integrationcache[ curtime->first ];

        // Now we can loop over all iterators for the current lowest timestamp
        // and extract all the data until we hit a different timestamp
        // The entries for each timestamp are iterators into the filelist
        for(filelist_type::iterator fptr=curtime->second.begin();
                fptr!=curtime->second.end(); fptr++) {
            // Keep on reading packets from the current file until the
            // timestamp is different from what we expect
            filestatus_type&  fs( ***fptr );
            unb_header const* hdr;

            while( (hdr=fs.current())!=0 &&
                   hdr->timestamp()==curtime->first ) {
                // In the map, the CE's data is pointed to by an iterator,
                // indexed by the unb_subband which is pointed to by the
                // CEid
                const unb_ce_id                          ce_id( hdr->ce_id() );
                unb_ce_sb_map::iterator                  cesbptr   = __ce_sb_map.find( ce_id );
                unb_sb_integration::iterator             sbdataptr = (cesbptr!=__ce_sb_map.end() ?
                                                                      curint.find(cesbptr->second) :
                                                                      curint.end());
                unb_enabled_products_map::const_iterator enabledptr = __enabledproducts.find( ce_id );

                if( __debug>2 )
                    cout << "PACKET: " << ce_id << " FBIN=" << hdr->freqbin << " TIME=" << hdr->framenr << endl;
                if( __debug>3 && cesbptr!=__ce_sb_map.end() )
                    cout << "        unb_subband_key = " << cesbptr->second << endl;

                // Verify sanity - do we expect data from this CE at all?
                if( enabledptr==__enabledproducts.end() )
                    throw runtime_error(string("Found data from correlation engine ")+
                                        repr(ce_id)+" in file '"+fs.filename+"' "+
                                        "which has no entry in the enabled products "
                                        "mapping. Check the json file for this job.");

                if( cesbptr==__ce_sb_map.end() )
                    throw runtime_error(string("Found data from correlation engine ")+
                                        repr(ce_id)+" in file '"+fs.filename+"' "+
                                        "which has no entry in the CE => subband "
                                        "mapping. Check the json file for this job.");

                // 08 Jan 2016: Now that we know that cesbptr is valid, we
                //              can cache the unb_subband key that it maps to
                const unb_subband_key&    unb_sb_key( cesbptr->second );

                // 11 Aug 2015: Des Small found data where in the data file
                //              there were more freqbins than were
                //              configured in the json. (Or alternatively:
                //              the json underspecified the actual number of
                //              freqbins). 
                //              This resulted in quite a rude coredump with
                //              no useful information at all.
                const unsigned int rawfbin    = hdr->freqbin + cesbptr->second.offset;
                const unsigned int actualfbin = rawfbin / __nChannelAvg;

                if( actualfbin>=nFreq )
                    throw runtime_error(string("Found data from correlation engine ")+
                                        repr(ce_id)+" in file '"+fs.filename+"' "+
                                        "for frequency bin "+repr(actualfbin)+" (after spectral "+
                                        "averaging of "+repr(__nChannelAvg)+") but the number of "+
                                        "frequency bins configured in the json file is only "+repr(nFreq)+
                                        ", so maximum acceptable frequency bin number is "+repr(nFreq-1)+".");

                // Let the outer scope know we did read some data!
                packets_read++;

                // We've got another nproduct numbers in
                prodcnt[ ce_id ] += hdr->nproduct();

                // Check if it was the first time we got data for this CE
                // because memory allocation is involved, we cannot simply
                // say something like:
                //   cedataptr = cache.insert( make_pair(..) ).second;
                // because that would mean that the c'tor inside make_pair()
                // would be called a bazillion times!
                if( sbdataptr==curint.end() ) {
                    pair<unb_sb_integration::iterator, bool> insres;

                    insres = curint.insert( make_pair(unb_sb_key, new unb_sb_data(nFreq, enabledptr->second.size(), __nChannelAvg)) );
                    if( insres.second==false )
                        throw runtime_error(string("Failed to add subband data to cache for unb_subband=")+
                                            repr(unb_sb_key)+" in file '"+fs.filename+"'");

                    sbdataptr = insres.first;
                }

                // The unb_subband data is now directly available
                unb_sb_data&                          sbdata = *sbdataptr->second;

                // From the first product nr in the header, we should be
                // able to find the offset. Check if we found this
                // combination of ce/first-product before
                ce_id_first_key                            ce_id_first(ce_id, hdr->firstproduct);
                first_prod_offset_map_type::const_iterator fp_offset_ptr = __first_prod_offset_map.find( ce_id_first );

                if( fp_offset_ptr==__first_prod_offset_map.end() ) {
                    unb_enabled_products::const_iterator  productnr;

                    // If the first product in the packet is not in the
                    // enabled list - we are in a LOT of trouble
                    if( (productnr=enabledptr->second.find(hdr->firstproduct))==enabledptr->second.end() ) 
                        throw runtime_error(string("First product #")+repr(hdr->firstproduct)+
                                            " in data from correlation engine "+repr(hdr->backplane_nr())+"/"+repr(ce_id)+
                                            " in file '"+fs.filename+"' is not in the "
                                            "enabled products list. Check unb_config.json for this job.");

                    // If there are more products in the packet than we have
                    // entries left from the 'first-product-in-packet' to the
                    // end of the enabled products list for this CE then we're
                    // in trouble! Check this only once so we can go over the
                    // individual products without checking

                    // Seriously, C++, wtf were you thinking? You should be able to get the distance type between
                    // iterators slightly easier than this yes?
                    typedef std::iterator_traits<unb_enabled_products::const_iterator>::difference_type product_distance; 
                    product_distance product_offset;

                    if( std::distance(productnr, enabledptr->second.end())<(product_distance)hdr->nproduct() )
                        throw runtime_error(string("There are more products in the datafile following first product #")+
                                            repr(hdr->firstproduct)+" for "+
                                            "correlation engine "+repr(hdr->backplane_nr())+"/"+repr(ce_id)+
                                            " in file '"+fs.filename+"' than entries in the "
                                            "enabled products list for that CE. "+
                                            "Check unb_config.json for this job.");

                    // We must know the index of hdr->firstproduct in the enabled product list 
                    product_offset = std::distance(enabledptr->second.begin(), productnr);

                    // Store the checked offset in the map
                    pair<first_prod_offset_map_type::iterator, bool> insres;

                    // The current ce_id packet has an offset because of being
                    // a combination of (ce_id, first product) but also
                    // because this ce_id may be computing only part of the
                    // spectrum. Via the ce_sb mapping we know which part.
                    // Account for that offset too. This offset is also in
                    // units of freqbins - if >1 CEs compute the spectrum,
                    // each delivers an amount of fbins, each numbered
                    // 0..N-1
                    // Take channel averaging into account!
                    insres = __first_prod_offset_map.insert( make_pair(ce_id_first,
                                                                       fpoffset_type((unsigned int)product_offset)) );

                    if( !insres.second )
                        throw runtime_error(string("Failed to insert first product offset into mapping for first product #")+
                                            repr(hdr->firstproduct)+" for "+
                                            "correlation engine "+repr(hdr->backplane_nr())+"/"+repr(ce_id)+
                                            " in file '"+fs.filename+"'");
                    fp_offset_ptr = insres.first; 
                }

                // Excellent - we can now iterate over the products
                // Create raw pointers to the source and destination data
                // and 1- and 2-Dimensional start offsets
                unsigned int       oneDoffset = fp_offset_ptr->second.product_offset;
                unsigned int       twoDoffset = fp_offset_ptr->second.product_offset * sbdata.nFreq +
                                                actualfbin;
                unsigned int       twoDrawoffset = fp_offset_ptr->second.product_offset * sbdata.nFreq * __nChannelAvg +
                                                   rawfbin;
                unb_32bit*         srcdataptr = (unb_32bit*)((unsigned char*)hdr+16);

                for(unsigned int n=0;
                    n<hdr->nproduct();
                    n++, oneDoffset++, twoDoffset += sbdata.nFreq, twoDrawoffset += (sbdata.nFreq * __nChannelAvg), srcdataptr++) {
                    // Lots of bookkeeping to do ...
                    // HV: 08 Jan 2016 - We have a flag wether or not a complex number
                    //                   (frequency bin) has been written to, so why
                    //                   don't we protect against overwriting?
                    //                   Note: a point is flagged (True) if it did NOT
                    //                         contain data yet. After writing to it
                    //                         it is 'unflagged'
                    if( sbdata.sb_fbinflag_raw[twoDrawoffset]==casa::False ) {
                        unb_enabled_products::iterator curprod = enabledptr->second.begin();
                        std::advance( curprod, fp_offset_ptr->second.product_offset + n );
                        throw runtime_error(string("About to overwrite frequency bin data for fbin=")+
                                             repr(actualfbin)+
                                             " (fbin in packet was "+repr(hdr->freqbin)+")"+
                                             " of product "+repr(*curprod)+
                                             " of "+repr(ce_id)+
                                             " (=> unb_subband="+repr(unb_sb_key)+") in file '"+fs.filename+"'"
                                           );
                    }
                    // HV: 07 Jan 2015 - We must ACCUMULATE into the ce_data
                    //                   because in case someone has done
                    //                   caused spectral averaging
                    //                 - Also, we found that ALL our phases
                    //                   have the opposite sign compared to
                    //                   SFXC so we introduce an implicit
                    //                   sign change here
                    sbdata.sb_data[twoDoffset]     += casacore::Complex(srcdataptr->real, -srcdataptr->imag);
                    // Flag this bin twice - once in the averaged bins and once in the un-averaged bins
                    sbdata.sb_fbinflag[twoDoffset]  = 
                        sbdata.sb_fbinflag_raw[twoDrawoffset ] =
                            casacore::False;
                    sbdata.sb_fbincount[oneDoffset]++;
                    if( sbdata.sb_weight[oneDoffset]==-unb_magic )
                        sbdata.sb_weight[oneDoffset] = srcdataptr->validity;
                    if( sbdata.sb_weight[oneDoffset]!=srcdataptr->validity )
                        cout << "WARN: weight varies across spectrum: get " << srcdataptr->validity
                             << " (bin #" << hdr->freqbin << "), was " << sbdata.sb_weight[oneDoffset] << endl;
                }

                // We've processed the full packet!
                fs.next();
            }
            // Done one file, move on to the next
        }
        // Now's the time to do some loss statistics
        unsigned int    expect = 0, got = 0;

        for(unb_enabled_products_map::const_iterator p=__enabledproducts.begin(); p!=__enabledproducts.begin(); p++) {
            statmap_type::const_iterator  cntptr = prodcnt.find(p->first);

            expect += p->second.size()*nFreq;
            if( cntptr!=prodcnt.end() )
                got += cntptr->second;
        }

        if( got!=expect )
            cout << "[" << curtime->first << "] loss " 
                 << j2ms::printeffer("%.2lf", ((double)(expect-got)/(double)expect) * 100.0) << "%" << endl;
    }

    ////// Part one of the data decoding is done - now it's time to assemble
    ////// the correlation-engine products into VEX-labelled baseline
    ////// matrices!
    unb_sb_integration_cache::iterator __current_sb_integration = __sb_integrationcache.begin();

    // If there is no pending data in the CE integration cache ... there is
    // not data left anymore
    if( __current_sb_integration==__sb_integrationcache.end() )
        return false;
   
    // Ok, __current_ce_integration points at some data for an integration
    // We collect all of it into a new unb_integration. Start by creating a
    // new entry with the same key (the time stamp!)
    pair<unb_integration_cache::iterator, bool> insres;
  
    insres = __integrationcache.insert( make_pair(__current_sb_integration->first, unb_integration()) );
    if( !insres.second )
        throw runtime_error(string("Failed to add unb_integration entry for time stamp ")+repr(__current_sb_integration->first));

    // Get a reference to the newly created unb_integration
    unb_integration&      unb_intgref = insres.first->second;
    const casacore::IPosition vecshp(1, nFreq);

    // Great. Now loop over all data products we found in the data file(s):
    for(unb_sb_integration::const_iterator p=__current_sb_integration->second.begin();
        p!=__current_sb_integration->second.end(); p++) {
            // Process data from unb_subband p->first
            // Get access to the VEX products from that unb_subband. Note that we
            // do not have to verify wether this exists! The fact that this
            // unb_subband exists in the current integration means that it existed in
            // the original 'enabled_products' mapping and THUS it existst
            // in the 'enabled_vex_products'!
            enabled_vex_products_map::const_iterator  enabledvexprods = __enabled_vex_products.find(p->first);
            enabled_vex_products_type::const_iterator vexprod;
            // Get access to the actual data itself
            const unb_sb_data&                       sbdata  = *p->second;
            unsigned int                             n, oneDoffset, twoDoffset; 

            // Loop over all vexproducts
            for(vexprod=enabledvexprods->second.begin(), n=0, oneDoffset=0, twoDoffset=0; vexprod!=enabledvexprods->second.end();
                vexprod++, n++, oneDoffset++, twoDoffset+=nFreq) {
                // Look up the destination vex product by key, creating it
                // if necessary
                unb_integration::iterator  pptr = unb_intgref.find(vexprod->key);

                if( pptr==unb_intgref.end() ) {
                    pair<unb_integration::iterator, bool> ires 
                        = unb_intgref.insert( make_pair(vexprod->key, unb_data(nFreq, nPol)) );
                    if( ires.second==false )
                        throw runtime_error(string("Failed to add data for baseline ")+
                                            repr(vexprod->key)+", time="+repr(__current_sb_integration->first));
                    pptr = ires.first;
                }
                // Get all relevant info from the current product into the
                // destination integration.
                unb_data&   u_data( pptr->second );

                // Copy data to appropriate row in the matrix, taking care
                // of swapping the phase if necessary
                if( vexprod->need_swap )
                    u_data.data.row(vexprod->row) = casacore::conj(casacore::Vector<casa::Complex>(vecshp, &sbdata.sb_data[twoDoffset], casacore::SHARE));
                else
                    u_data.data.row(vexprod->row) = casacore::Vector<casacore::Complex>(vecshp, &sbdata.sb_data[twoDoffset], casacore::SHARE);

                // Copy received frequency bin flags
                u_data.freqbinflag.row(vexprod->row) = casacore::Vector<casacore::Bool>(vecshp, &sbdata.sb_fbinflag[twoDoffset], casacore::SHARE);

                // Weight and received freqbin count
                u_data.weight[vexprod->row]       = sbdata.sb_weight[oneDoffset];
                u_data.freqbincount[vexprod->row] = sbdata.sb_fbincount[oneDoffset];
            }

            // Delete all data for this one
            delete p->second;
    }
    // Processed this integration, time to move on to the next one
    __sb_integrationcache.erase(__current_sb_integration);
    return true;
}


// Read the enabled products per backnode/correlation engine as well as the
// product mapping
mappingresult_type
UNBVisBuffer::readEnabledProductsMap(const Json::Value& json) {
    // <root>."enabled_products" = [
    //      [<prodnr>, <unb_station1>, <pol1>, <station2>, <pol2>, [CEids] ],
    //       ....
    // ]
    //
    // where [CEids] =
    //  [ "<unb>, <node>, <ce>", ... ]
    //
    // the list of CEs where this product is active on
    const Json::Value&                   enabled = json["enabled_products"];
    unb_product_map                      unb_products;
    unb_enabled_products_map             unb_ce_enabled_products;

    if( enabled.isNull() || !enabled.isArray() )
        throw string("The enabled_products key is not present/not an array in the JSON file");

    for(Json::ArrayIndex idx=0; idx<enabled.size(); idx++) {
        // Each entry in the array should be an array itself
        string             polarization1, polarization2;
        Json::Value        entry = enabled[idx];
        unb_station        x, y;
        unsigned int       product;
        const string       entry_s = entry.toStyledString();
        j2ms::polarization p0, p1;

        // ".toStyledString() always adds "\n"! (strip it)
        //entry_s.erase( products.size()-1 );

        if( entry.isNull() || !entry.isArray() || entry.size()!=6 )
            throw string("Entry '")+entry_s+"' in enabled products is malformed - not an array of length 6";

        // extract basic info
        product       = entry[0].asUInt();
        polarization1 = entry[2].asString();
        polarization2 = entry[4].asString();

        Json::Value&       cearray = entry[5];

        if( cearray.isNull() || !cearray.isArray() )
            throw string("The correlation engine array in entry '")+entry_s+"' is not an array (the sixth element)";

        // verification of the unb_station indices
        if( !entry[1].isUInt() )
            throw string("X-antenna in product '"+entry_s+"' is not a number");
        if( !entry[3].isUInt() )
            throw string("Y-antenna in product '"+entry_s+"' is not a number");
        if( entry[1].asUInt()>255 )
            throw string("X-antenna in product '"+entry_s+"' is out of range (max 255)");
        if( entry[3].asUInt()>255 )
            throw string("Y-antenna in product '"+entry_s+"' is out of range (max 255)");
        x = (unb_station)entry[1].asUInt();
        y = (unb_station)entry[3].asUInt();

        // verification of the polarizations
        if( entry[2].isNull() || polarization1.size()!=1 )
            throw string("X-polarization in entry '")+entry_s+"' is not a single character";
        if( entry[4].isNull() || polarization2.size()!=1 )
            throw string("Y-polarization in entry '")+entry_s+"' is not a single character";

        p0 = j2ms::char2polcode( polarization1[0] );
        p1 = j2ms::char2polcode( polarization2[0] );

        if( !ACCEPTABLE_POLARIZATION(p0) )
            throw string("X-polarization in entry '")+entry_s+"' is not a valid polarization identifier";
        if( !ACCEPTABLE_POLARIZATION(p1) )
            throw string("Y-polarization in entry '")+entry_s+"' is not a valid polarization identifier";

        // Here we can insert the product in the mapping
        unb_products.insert( make_pair(product,
                                       unb_product( unb_baseline(x,y), CorrelationCode(p0, p1) )) );

        // Now process the correlation engine list and for each CE add
        // the current product to the list of enabled products
        for(Json::ArrayIndex ceidx=0; ceidx<cearray.size(); ceidx++) {
            const string  ce_s = cearray[ceidx].toStyledString();
            unsigned int  unb, nodeid, cenr;

            //ce_s.erase( ce_s.size()-1 );
            if( cearray[ceidx].isNull() || !cearray[ceidx].isString() )
                throw string("Correlation engine key '")+ce_s+"' is not a non-empty string in entry '"+entry_s+"'";

            // expect each element to be a string
            // "<unb>, <node number>, <correlation engine number>"
            if( ::sscanf(cearray[ceidx].asString().c_str(), "%u,%u,%u", &unb, &nodeid, &cenr)!=3 )
                throw string("Correlation engine key '"+ce_s+"' is not a valid correlation engine identifier in '"+entry_s+"'");

            // too large?
            if( unb>31 )
                throw range_error(string("uniboard number in ")+ce_s+" is larger than 31!");
            if( nodeid>7 )
                throw range_error(string("FPGA node number in ")+ce_s+" is larger than 7!");
            if( cenr>255 )
                throw range_error(string("CE number in ")+ce_s+" is larger than 255!");

            const unb_ce_id   ceid( mk_node(unb, nodeid), (unb_ce)cenr );

            // Ok now we have everything!
            unb_enabled_products&  ce_set = unb_ce_enabled_products[ceid];

            if( ce_set.insert(product).second == false )
                throw string("Duplicate enabled product '")+entry_s+"' for correlation engine '"+ce_s+"'";
        }
    }
    return mappingresult_type(unb_products, unb_ce_enabled_products);
}

// Firmware version => limits mapping
struct firmware_limits {
    const unsigned int    nSta;
    const unsigned int    nSub;

    firmware_limits(unsigned int sta, unsigned int sub):
        nSta( sta ), nSub( sub )
    {}
};
typedef std::map<unsigned int, firmware_limits> firmwarelimitsmap_type;

// Read the enabled products per backnode/correlation engine as well as the
// product mapping
inputmap_type
UNBVisBuffer::readInputMap(const Json::Value& json) {
    // <root>."input_mapping" : {
    //      <vexoutputkey> : [<unb>, <unb_station>, <unb_subband>],
    //                ...
    // }
    //
    // where
    //   <vexoutputkey> = "<Station's VEX name>, <subband number>"
    //   <unb>          = the uniboard backplane number (=="which uniboard")
    //   <unb_station>  = the virtual station number (0..31 inclusive) on
    //                    the indicated uniboard
    //   <unb_subband>  = the virtual subband (0..n) number of the virtual station
    //                    on the indicated uniboard
    //
    //   and vex name of a station is its two-letter code and the subband
    //   number is the linear index of the vexchannel after sorting the
    //   frequency bands in increasing frequency 
    inputmap_type           inputmap;
    const Json::Value&      input_mapping = json["input_mapping"];
    const Json::Value&      correlatorconfig = json["correlatorconfig"];
    Json::Value::Members    members;
    firmware_limits const   unknown_firmware(255, 255);
    firmware_limits const*  firmwarelimits = &unknown_firmware;
    firmwarelimitsmap_type  firmwarelimits_map;

    if( input_mapping.isNull() || !input_mapping.isObject() )
        throw string("The input_mapping key is not present/not an object in the JSON file");

    firmwarelimits_map.insert( make_pair(5966340, firmware_limits(32, 4)) /* 32 Stations x 4 subbands */ );
    
    if( !correlatorconfig.isNull() && correlatorconfig.isUInt() &&
        firmwarelimits_map.find(correlatorconfig.asUInt())!=firmwarelimits_map.end() )
            firmwarelimits = &firmwarelimits_map.find(correlatorconfig.asUInt())->second;

    if( firmwarelimits==&unknown_firmware && !correlatorconfig.isNull())
        throw string("readInputMap: no firmware limits defined for correlatorconfig: ")+correlatorconfig.toStyledString();

    // Loop over all keys
    members = input_mapping.getMemberNames();

    for(Json::Value::Members::const_iterator curkey=members.begin();
        curkey!=members.end(); curkey++) {
            // Each entry in the array should be an array itself
            const Json::Value  unbinp_j( input_mapping[*curkey] );
            const string       entry_s = unbinp_j.toStyledString();

            if( unbinp_j.isNull() || !unbinp_j.isArray() || unbinp_j.size()!=3 )
                throw invalid_argument(string("input_mapping entry '")+entry_s+"' is invalid, not array with three values!");

            // the key is "<vex station>,<vex subband>",
            char*            eocptr;
            unsigned int     vsb;
            vector<string>   parts = ::split(*curkey, ',');
            
            if( parts.size()!=2 )
                throw runtime_error(string("input_mapping entry key '")+*curkey+"' does not have two comma separated parts");

            if( parts[0].empty() )
                throw runtime_error(string("input_mapping entry key '")+*curkey+"' does not have a non-empty first part");

            vsb = ::strtoul(parts[1].c_str(), &eocptr, 0);
            if( parts[1].empty() || *eocptr!='\0' )
                throw invalid_argument(string("input_mapping entry key '")+*curkey+"' does not have a numerical second part");
            if( vsb>255 )
                throw range_error(string("input_mapping entry key '")+*curkey+"' subband value is larger than 255!");

            const vex_output_type vexoutput(parts[0], (vex_subband)vsb);
            
            // the value is [<unb>, <unb_station>, <unb_subband>]
            unsigned int    unb, unb_st, unb_sb;
            
            if( !unbinp_j[0].isUInt() || !unbinp_j[1].isUInt() || !unbinp_j[2].isUInt() )
                throw runtime_error(string("input_mapping key '")+*curkey+"''s values '"+entry_s+"' are not all integers!");

            // max 32 uniboards
            if( (unb=unbinp_j[0].asUInt())>31 )
                throw range_error(string("input_mapping key '")+*curkey+"''s uniboard value >31 in '"+entry_s+"'");

            // [[TODO]] These constraints should really be taken from the firmware - such
            // that we can warn if they fall outside the correlator
            // firmware's range
            if( (unb_st=unbinp_j[1].asUInt())>firmwarelimits->nSta )
                throw range_error(string("input_mapping key '")+*curkey+"''s station value >"+repr(firmwarelimits->nSta)+" in '"+entry_s+"'");
            if( (unb_sb=unbinp_j[2].asUInt())>firmwarelimits->nSub )
                throw range_error(string("input_mapping key '")+*curkey+"''s subband value >"+repr(firmwarelimits->nSub)+" in '"+entry_s+"'");

            const unb_input_type    unbinput(unb, unb_st, unb_sb);

            // Now we have the vexoutput key and unbinput as value
            // but we reverse the mapping. Also make sure that not 
            // two vex outputs are sent to the same unb_input.
            if( (inputmap.insert(make_pair(unbinput, vexoutput))).second==false )
                throw logic_error(string("duplicate key: >1 vex output sent to unb_input '")+entry_s+"'");
    }
    return inputmap;
}


// Read the mapping of which correlation engine computes which 'logical'
// uniboard subband
unb_ce_sb_map
UNBVisBuffer::readCESubbandMap(const Json::Value& json) {
    // <root>."ce_sb" : {
    //      "<unb>" : [["<node>,<CE>", unb_subband], ... ],
    //                ...
    //  }
    //  Update Jun 2015: unb_subband can now be:
    //      unb_subband || "subband,offset"
    //
    // where
    //   "<unb>"   the uniboard backplane number as a string
    //             (numbers cannot be used as keys)
    //      maps to a list of node/ce keys mapping to a uniboard logical
    //      subband:
    //
    //  ["<node>,<CE>", unb_subband]
    //      "<node>,<CE>"    the FPGA node number on the board (0..7) 
    //                       combined with the correlation engine instance
    //                       on that node [there could be >1 CE/fpga node]
    //      unb_subband      the logical uniboard subband (0..3)
    //                       maybe other firmware will be able to process
    //                       more than 4 subbands at once
    //      "subband,offset" the firmware can now compute a single spectrum
    //                       (subband) across different CEs. Each CE
    //                       computes a part of the spectrum and the
    //                       'offset' value indicates where in the total
    //                       spectrum this CE's output should go
    //
    unb_ce_sb_map              cesb;
    const Json::Value&         cesb_data = json["ce_sb"];
    Json::Value::Members       members;

    if( cesb_data.isNull() || !cesb_data.isObject() )
        throw string("The 'ce_sb' key is not present/not an object in the JSON file");

    // Loop over all keys
    members = cesb_data.getMemberNames();

    for(Json::Value::Members::const_iterator curkey=members.begin();
        curkey!=members.end(); curkey++) {
            // Each entry in the array should be an array itself
            char*              eocptr;
            unsigned int       unb_nr;
            const Json::Value  cesb_j( cesb_data[*curkey] );
            const string       cdsb_s = cesb_j.toStyledString();

            // Decode the uniboard number from the key
            unb_nr = ::strtoul(curkey->c_str(), &eocptr, 0);
            if( eocptr==curkey->c_str() || *eocptr!=0 )
                throw invalid_argument(string("Entry '")+*curkey+"' in ce_sb map is not numerical");
            if( unb_nr>31 )
                throw range_error(string("uniboard number out of range in ce_sb mapping, key '")
                                  + *curkey + " (>31)");
            // Assert the key maps to an array
            if( !cesb_j.isArray() )
                throw invalid_argument(string("key '")+*curkey+"' in ce_sb map does not map to an array");

            // Loop over the entries in the array
            for(unsigned int i=0; i<cesb_j.size(); i++) {
                // Each element in the list should be a two-element list
                // itself
                unsigned int      sb, offs = 0;
                const Json::Value elem   = cesb_j[i];
                const string      elem_s = elem.toStyledString();

                // The value is ["<node_ce_key>", <unb_subband>]
                // Jun 2015: We must support <unb_subband> as well as "<unb_subband>,<offset>"
                //           as second member in the array

                // Assert assertions about the value
                if( elem.isNull() || !elem.isArray() || elem.size()!=2 )
                    throw invalid_argument(string("ce_sb element '")+elem_s+"' for key '"+*curkey+
                                           " is invalid, not array with two values!");
                if( !elem[0].isString() )
                    throw invalid_argument(string("first value in ce_sb element '")+elem_s+"' in key '"+
                                                  *curkey + "' is not a string");

                // Check element 1 of the array. 
                // If it's uInt it's old-style json with just subband number
                if( !elem[1].isUInt() ) {
                    // Ok, then it must be string
                    if( !elem[0].isString() )
                        throw invalid_argument(string("second value in ce_sb element '")+elem_s+"' in key '"+
                                                      *curkey + "' is not a number or string");
                    // Verify the string can be parsed as <int>,<int>
                    char*            eocptr;
                    vector<string>   parts = ::split(elem[1].asString(), ',');

                    if( parts.size()!=2 )
                        throw invalid_argument(string("second value in ce_sb element '")+elem_s+"' in key '"+
                                               *curkey + "' does not have two comma separated parts");

                    // Check subband part
                    sb = ::strtoul(parts[0].c_str(), &eocptr, 0);
                    if( parts[0].empty() || *eocptr!='\0' )
                        throw invalid_argument(string("second value in ce_sb element  '")+elem_s+"' in key '"+
                                               *curkey + "' does not have a numeric first part before the comma");

                    // subband may not be >uint16_t max
                    if( sb>std::numeric_limits<uint16_t>::max() )
                        throw invalid_argument(string("subband value in ce_sb element '")+elem_s+"' in key '"+
                                                      *curkey + "' is > uint16_t max value");

                    // Read offset (second part)
                    offs  = ::strtoul(parts[1].c_str(), &eocptr, 0);
                    if( parts[1].empty() || *eocptr!='\0' )
                        throw invalid_argument(string("second value in ce_sb element '")+elem_s+"' in key '"+
                                               *curkey + "' does not have a numeric second part after the comma");
                } else {
                    sb = elem[1].asUInt();

                    // subband may not be >uint16_t max
                    if( sb>std::numeric_limits<uint16_t>::max() )
                        throw invalid_argument(string("subband value in ce_sb element '")+elem_s+"' in key '"+
                                                      *curkey + "' is > uint16_t max value");
                }

                // Parse the string in the first element, it has the
                // "node,ce" pair
                char*            eocptr;
                unsigned int     node, ce;
                vector<string>   parts = ::split(elem[0].asString(), ',');

                if( parts.size()!=2 )
                    throw invalid_argument(string("first value in ce_sb element '")+elem_s+"' in key '"+
                                           *curkey + "' does not have two comma separated parts");

                node = ::strtoul(parts[0].c_str(), &eocptr, 0);
                if( parts[0].empty() || *eocptr!='\0' )
                    throw invalid_argument(string("first value in ce_sb element '")+elem_s+"' in key '"+
                                           *curkey + "' does not have a numeric first part before the comma");
                if( node>7 )
                    throw range_error(string("first value in ce_sb element '")+elem_s+"' in key '"+
                                           *curkey + "' node number is out of range (>7)");

                ce  = ::strtoul(parts[1].c_str(), &eocptr, 0);
                if( parts[1].empty() || *eocptr!='\0' )
                    throw invalid_argument(string("first value in ce_sb element '")+elem_s+"' in key '"+
                                           *curkey + "' does not have a numeric second part after the comma");

                // Second value in the list is UInt (and has been checked)
                // so we can now safely build the map
                cesb.insert( make_pair(unb_ce_id(mk_node(unb_nr, node), unb_ce(ce)), unb_subband_key(unb_nr, sb, offs)) );
            }
    }
    return cesb;
}




/////////////////////////////////////////////////////////////////////////////////////////////
//
//            Normalize all the data in an integration
//
/////////////////////////////////////////////////////////////////////////////////////////////


// Need a mapping of a station's id to its autocorrelation square root of the power
// Such that if we find a cross-product, we can divide by the power proper
// Of course, it must be done per frequency and per polarization. 

struct vex_np_key {
    vex_station        station;
    vex_subband        subband;
    j2ms::polarization polarization;

    vex_np_key(vex_station vst, vex_subband vsb, j2ms::polarization p):
        station( vst ), subband( vsb ), polarization( p )
    {}
};

bool operator<(const vex_np_key& l, const vex_np_key& r) {
    if( l.station!=r.station )
        return l.station < r.station;
    if( l.subband!=r.subband )
        return l.subband < r.subband;
    return l.polarization < r.polarization;
}

ostream& operator<<(ostream& os, const vex_np_key& vnpkey) {
    return os << "VEX Sta#" << vnpkey.station << "/SB#" << vnpkey.subband << "/POL#" << vnpkey.polarization;
}

// A normalization quantum - a power and a weight, this is per
//  station/subband/polarization
struct normal_quant {
    // Per normalization quantum this is the information we compute & use
    casacore::Float   power;
    casacore::Float   weight;

    // Initialize power to 0.0 to mean "not computed" so we can
    // catch errors where e.g. for the cross correlations there are more
    // polarization products than for the auto correlations ...
    // (hopefully unlikely but in our UNB correlator that could happen)
    normal_quant() :
        power( casacore::Float(0.0) ), weight( casacore::Float(0.0) )
    {}
};
typedef std::map<vex_np_key, normal_quant>    normalquantmap_type;

// Normalize the correlation coefficients in the integration.
// We follow the same method - or at least as close as possible -
// that is used by SFXC
void UNBVisBuffer::doNormalize( unb_integration_cache::iterator intgptr ) {
    // 'intgptr' is iterator to pair [unb_timestamp, unb_integration]
    // 'intg' is reference to unb_intgration:
    //
    // 'intg' is MAP [unb_key] => unb_data
    //   unb_data     = (vex_key, data, weight, freqbincnt, freqbinflag)
    //   vex_key      = (vex_baseline, vex_subband)
    //   data         = unb_complexmatrix ( == casacore::Matrix<casacore::Complex>)
    //                  dims: nPol x nFreq
    //   freqbincnt   = unb_uintvector, length: nPol
    //   freqbinflag  = unb_boolmatrix, dims: nPol x nFreq
    //   vex_baseline = (vex_station, vex_station, swapped)
    //
    // We must do the normalization based on vex_station because multiple
    // unb_stations [labels in unb_key] could map to the same vex station
    // that is why we ignore the 'unb_key' in the MAP [unb_key] => unb_data
    unb_integration&      intg( intgptr->second );

    // The UNB correlator has as property that the first 6 FFT periods are
    // invalid - this has to do with filling up the buffers of the polyphase
    // filter bank. It takes six FFT periods before it starts outputting 
    // sensible results.
    normalquantmap_type   normalquantmap;
    const unb_timestamp&  ts( intgptr->first );
    const casacore::Float maxFFTCount = (casacore::Float)__nFFTpIntg - 6;

    // Step one - find all the auto correlations, compute their power
    //            and normalize them
    for( unb_integration::iterator p=intg.begin(); p!=intg.end(); p++) {
        const vex_key&      vk( p->first );
        const vex_baseline& vb( vk.baseline );

        // not auto correlation?
        if( vb.x!=vb.y )
            continue;

        // yup!
        unb_floatvector&                          weight( p->second.weight );
        unb_uintvector&                           count( p->second.freqbincount );
        unb_complexmatrix&                        data( p->second.data );
        unb_boolmatrix&                           flag( p->second.freqbinflag );
        const casacore::IPosition                 shp = data.shape();


        // Ok, now compute total power
        // We can use the
        // If real part == unb_magic, this means no data was available
        // for that spectral point!
        const casacore::uInt  nPol = shp[0], nFreq = shp[1];

        for(casacore::uInt i=0; i<nPol; i++) {
            // Skip cross pols!
            if( __correlations[i].is_crosspol() )
                continue;

            // Now we can form the real normalization key, consisting of STA/SB/POL
            // [we can easily take polarization[0] from the the current
            // polarization combination because it is parallel hands of
            // polarization, i.e. polarization[0]==polarization[1]]
            const vex_np_key                          vnpk(vb.x, vk.subband, __correlations[i][0]);
            pair<normalquantmap_type::iterator, bool> insres = normalquantmap.insert( make_pair(vnpk, normal_quant()) );
            normal_quant&                             norm( insres.first->second );

            // assert it's not a duplicate
            if( insres.second==false )
                throw runtime_error(string("Found duplicate vex key for auto correlation ")+
                                    repr(vnpk));

            // No use in counting power if we didn't receive any bins at all
            for(casacore::uInt j=0; count[i]>0 && j<nFreq; j++)
                if( !flag(i, j) )
                    norm.power += casacore::real( data(i, j) );

            // The power must add up to nFreq, such that, on average, each
            // spectral point gets an amplitude of 1. Thus we must multiply
            // by a factor of nFreq
            //
            // Then we have to take into account any bins lost; multiply
            // by the fraction of bins we actually received. count[i] is the
            // amount received, with a maximum of nFreq * __nChannelAvg
            //
            // Combining the two factors:
            //      nFreq * ( count[i] / (nFreq * __nChannelAvg) ) 
            // this reduces to:
            //      count[ i ] / __nChannelAvg
            if( count[i] )
                norm.power = norm.power / (count[i] / __nChannelAvg);

            const bool    havenorm = (norm.power>1.0e-3);

            // SFXC normalizes the autocorrelations differently than the
            // crosses. So let's follow that principle.
            // Also let's put the value of zero there rather than 3.14?
            for(casacore::uInt j=0; j<nFreq; j++)
                if( havenorm && !flag(i, j) )
                    data(i, j) = casacore::real(data(i, j)) / norm.power;
                else
                    data(i, j) = unb_nul;

            // If we did not have normalization, do not store it in the
            // normalquantmap and continue
            if( havenorm==false && __debug>0 )
                cout << "WARN: " << ts << ": AC Power low (" << norm.power << ") "
                     << "for VEX Sta#" << vb.x << "/SB#" << vk.subband << "/" << __correlations[i] << " "
                     << " (" << ((float)count[i])/(nFreq * __nChannelAvg)*100.0 << "% received)"
                     << endl;

            // Let's scale the weigth from a count to a number between 0 and 1
            // and cache it it the normalization information.
            // Andalso incorporate the packet loss into it; we scale the
            // weight by the fraction of packets received.
            // This way, spectra that have no packets or very little packets
            // will automatically get a weight of ~ 0 so they be easy to
            // flagr!
            //
            // Normalization weight does not take into account number of
            // bins received because that's already accounted for in the
            // norm.power[i]. (norm.power[i] is the average power per
            // spectral bin, irrespective of how many we received).
            // 
            // The normalization weight that comes out of the correlator
            // is how many valid fft frames were put into this integration
            // so we divide by the maximum number of fft frames to get at
            // the correlated normalization weight for this station. If e.g.
            // only 80% of the frames were valid, the correlations to this
            // station will be affected by that factor. The fact that we
            // lose packets (thus: frequency bins) has nothing to do with
            // the correlated power.
            //
            // The baseline weight we DO scale by the amount of packets
            // received - it's a good approximation of how much valid data
            // is in this spectrum ... 'nFreq' (the length of this spectrum)
            // _used_ to tell how many frequency bins were potentially
            // available but it's now down by a factor of __nChannelAvg so
            // we must compensate for that.
            norm.weight = ((casacore::Float)weight[i] / maxFFTCount) / __weightScale;
            weight[ i ] = norm.weight * ((casacore::Float)count[i])/(nFreq * __nChannelAvg);
        }
    }

    ////////////////////////////////////////////////////////////////////
    //
    // Ok - we got a scaling factor for all the individual telescopes
    //
    // Step 2: Now go through all data points and apply the scaling
    //         For each baseline, find the factors for station X and Y.
    //
    ////////////////////////////////////////////////////////////////////

    // Keep an error map - we don't want to display the same message a
    // bazillion times!
    typedef std::set<vex_np_key>                   no_autocorr_type;
    typedef pair<no_autocorr_type::iterator, bool> acinsres_type;

    no_autocorr_type    no_autocorr;

    // Loop over all baselines
    for( unb_integration::iterator p=intg.begin(); p!=intg.end(); p++) {
        unb_data&                           u_data( p->second );
        const vex_key&                      key( p->first );
        unb_uintvector&                     u_count( u_data.freqbincount );
        unb_floatvector&                    u_weight( u_data.weight );
        unb_complexmatrix&                  u_matrix( u_data.data );
        unb_boolmatrix&                     u_flag( u_data.freqbinflag );
        const casacore::IPosition           shp = u_matrix.shape();
        // Still, shp[0] = nPol, shp[1] = nFreq
        const casacore::uInt                nPol = shp[0], nFreq = shp[1];

        // Loop over all polarizations
        for(casacore::uInt i=0; i<nPol; i++) {
            // Form the STA/SB/POL keys for both the X and Y part of the baseline
            const vex_np_key                    xnpkey(key.baseline.x, key.subband, __correlations[i][0]);
            const vex_np_key                    ynpkey(key.baseline.y, key.subband, __correlations[i][1]);
            normalquantmap_type::const_iterator xptr, yptr;

            // If this is an autocorrelation, we can skip it [cross pol on
            // auto baseline is not an autocorrelation ... (HV, 19 dec 2014) *sigh*]
            if( xnpkey.station==ynpkey.station && !__correlations[i].is_crosspol() )
                continue;

            xptr = normalquantmap.find( xnpkey );
            yptr = normalquantmap.find( ynpkey );

            // What to do if not found? Throw? Continue with warning?
            if( xptr==normalquantmap.end() ) {
                acinsres_type   acinsres = no_autocorr.insert( xnpkey );
                // "insert" only succeeds the first time
                if( acinsres.second )
                    cout << "normalize: " << ts << " no autocorrelation power for " << xnpkey << endl;
                continue;
            }
            if( yptr==normalquantmap.end() ) {
                acinsres_type   acinsres = no_autocorr.insert( ynpkey );
                // "insert" only succeeds the first time
                if( acinsres.second )
                    cout << "normalize: " << ts << " no autocorrelation power for " << ynpkey << endl;
                continue;
            }

            // Depending on the polarization product we'll have to get X and
            // Y's respective polarization power
            const normal_quant&                  xnorm( xptr->second );
            const normal_quant&                  ynorm( yptr->second );
            const bool                           nox    = casacore::near(xnorm.power, 0.0);
            const bool                           noy    = casacore::near(ynorm.power, 0.0);
            const casacore::Float                factor = xnorm.power * ynorm.power;
            casacore::Vector<casacore::Complex>  row    = u_matrix.row( i );

            // Cf. SFXCVisBuffer:
            // if (key1 != key2) {
            //     // Cross-correlations need to have a quantization (Van Vleck)
            //     // correction applied.  The factor here is taken from equation
            //     // A8.13 in Thompson, Moran and Swenson, "Interferometry and
            //     // Synthesis in Radio Astronomy", 2nd Edition and assumes
            //     // 2-bit sampling and "ideal" sampler statistics.
            //     for (int channel = 0; channel < num_channels; channel++)
            //         cplxbuf[channel] *= 1.1329552;
            // }
            //
            // So for UniBoard the same restrictions apply ... 
            // (ideal sampler stats, 2-bit sampling)
            row *= casacore::Complex(1.1329552, 0);

            // First convert cross correlation count to a weight
            // NOTE: Do not forget to take into account the scaling factor
            //       to compensate for UNB giving us accumulated values
            //       (in stead of averaged) in case we're asking it to
            //       produce less frequency bins than its native amount
            u_weight[ i ] = (u_weight[i]/maxFFTCount) / __weightScale;

            // Only do the normalization if we have sensible normalization
            // for _both_ auto correlations
            if( !(nox || noy) ) { 
                // Ok. Now we must account for the weights - the amount of 
                // uncorrelated data loss.
                // We have:
                //     ACx  = [0, 1]   (autocorrelation weight for station  X)
                //     ACy  = [0, 1]   (          "       "         "       Y)
                //     CCxy = [0, 1]   (cross-correlation "        baseline XY)
                // CCxy gives the amount of overlap between ACx and ACy and the
                // amount of power that went into the CC out of each constituent
                // AC is the overlap divided by the AC weight. (The AC weight 
                // also accounts for loss of power in the AC itself)
                //
                // Normalize taking care flagged spectral bins (set them to 0)
                const casacore::Float normval = ::sqrt(factor) * (u_weight[i]/xnorm.weight) * (u_weight[i]/ynorm.weight);
                for(casacore::uInt j=0; j<nFreq; j++) {
                    if( !u_flag(i, j) )
                        row(j) /= normval;
                    else
                        row(j) = unb_nul;
                }
            } else {
                row = unb_nul;
            }

            // take into account packet loss on the cross baseline!
            // we cannot do this at the beginning because the *actual*
            // weight is used during the normalization
            u_weight[i] = u_weight[i] * ((casacore::Float)u_count[i]/(nFreq * __nChannelAvg));
        }
    }
    return; 
}

