// VisBuffer object for UniBoard output datafiles
//
// HV 04/May/2012   initial version
//
#ifndef JIVE_J2MS_UNBVISBUFFER_H
#define JIVE_J2MS_UNBVISBUFFER_H

#include <string>
#include <map>
#include <set>
#include <list>
#include <iostream>

#include <jive/j2ms/VisibilityBuffer.h>
#include <jive/j2ms/CorrelatorSetup.h>
#include <jive/j2ms/BaseBand.h>
#include <jive/j2ms/scanlist.h>
#include <jive/labels/CorrelationCode.h>
#include <jive/j2ms/json.h>

#include <casacore/casa/aips.h>
#include <casacore/casa/BasicSL/String.h>
#include <casacore/casa/Arrays/Vector.h>
#include <casacore/casa/Arrays/Matrix.h>
#include <casacore/measures/Measures/Muvw.h>

#include <stdint.h>

// have to forward declare this, otherwise we get
// circular include-probs
namespace j2ms {
    class COFScan;
}

// Entitties in the UniBoard correlator system and mappings between them
//   * updated to uint16_t. Way too big. uint8_t would be MORE than enough
//     but then the FUCKING C++ COMPILER prints them as characters GVD!
//     (so an index of '0' or '1' would be printed as the non-printable
//      ASCII character \0 or \1 ... great eh?)
typedef uint16_t unb_backplane;// uniboard backplane nr 0..31
typedef uint16_t unb_station;  // uniboard numeric station 0..31
typedef uint16_t unb_subband;  // uniboard numeric subband 0..3 [could get higher if more CE's avail]

// Jun 2015: The UniBoard can now correlate wider bandwidths by splitting
//           input band over >1 back nodes; each back node computing 1/N'th
//           part of the spectrum. We have to stitch those parts back into
//           one whole spectrum. We rely on the control software to know
//           and tell us in which part of the spectrum a particular's
//           CorrelationEngine's output should go ('offset').
struct unb_subband_key {
    unb_backplane uniboard;
    unb_subband   subband;
    unsigned int  offset;

    unb_subband_key( unb_backplane ub, unb_subband sb, unsigned int offs );
    // do NOT allow default c'tor
    private:
        unb_subband_key();
};
// This comparator will IGNORE the 'offset' member! Therefore, if >1 CE
// refer to a the same subband but with different offset, the key will
// map to the same data and enabled vex products
bool operator<(const unb_subband_key& l, const unb_subband_key& r);
std::ostream& operator<<(std::ostream& os, const unb_subband_key& unb_sb);

typedef uint16_t unb_node;     // uniboard node id: backplane/node number
typedef uint16_t unb_ce;       // uniboard correlation engine (CE) number

inline unb_node mk_node(const unb_backplane board, const unb_node node) {
    return (unb_node)( ((board & 0x1f)<<3) + (node&0x7) );
}
inline unb_backplane board_number(const unb_node node) {
    return (unb_backplane)((node >> 3) & 0x1f);
}
inline unb_node node_number(const unb_node node) {
    return (unb_node)(node & 0x7);
}

// A node+ CE number fully identifies a correlation engine [we get our data
// from individual CEs]
struct unb_ce_id {
    const unb_node    node;
    const unb_ce      ce;
    unb_ce_id(unb_node n, unb_ce engine);
};
bool operator<(const unb_ce_id& l, const unb_ce_id& r);
bool operator==(const unb_ce_id& l, const unb_ce_id& r);
std::ostream& operator<<(std::ostream& os, const unb_ce_id& ce_id);

// The baseline
struct unb_baseline {
    const unb_station x;
    const unb_station y;
    const bool        swapped; // if a>b on creation
    unb_baseline(unb_station a, unb_station b);
};
bool operator==(const unb_baseline& l, const unb_baseline& r);
bool operator<(const unb_baseline& l, const unb_baseline& r);
std::ostream& operator<<(std::ostream& os, const unb_baseline& ubl);

// Each CE computes a number of correlation products.
// The product table describes for each product which unb_stations
// are correlated and which polarization product [use CorrelationCode]
struct unb_product {
    const unb_baseline    bl;
    const CorrelationCode polcode;
    unb_product(const unb_baseline& ubl, const CorrelationCode& cc);
};
bool operator==(const unb_product& l, const unb_product& r);
bool operator<(const unb_product& l, const unb_product& r);
std::ostream& operator<<(std::ostream& os, const unb_product& prod);

typedef std::map<unsigned int, unb_product>       unb_product_map;

// Per CE we have a set of enabled products [the CCS programs which
// products the CE should output].
// Also there is a per CE mapping defined telling which logical
// uniboard subband it produces
typedef std::set<unsigned int>                    unb_enabled_products;
typedef std::map<unb_ce_id, unb_enabled_products> unb_enabled_products_map;
typedef std::map<unb_ce_id, unb_subband_key>      unb_ce_sb_map;

// If you combine a unb_product [by number] combin from a particular
// Correlation Engine, we know exactly which subband and which
// polarization it was.


// One of the functions returns both the product map as well as the enabled
// products mapping
struct mappingresult_type {
    unb_product_map          product_map;
    unb_enabled_products_map enabled_products_map;

    mappingresult_type(const unb_product_map& pm,
                       const unb_enabled_products_map& epm);
};

// We need to know which subband from which station mapped to
// a particular uniboard input. The JSON file has a mapping
// from vex output -> uniboard input. We immediately reverse
// the mapping because we know the unb_input (from the datafile)
// and need the vex_output.
typedef std::string                               station_id;
typedef uint16_t                                  vex_subband;
typedef uint16_t                                  vex_station;
struct vex_output_type {
    station_id  vexstation;
    vex_subband vexsubband;
    vex_output_type(station_id st, vex_subband sb);
};
struct unb_input_type {
    unb_backplane unbboard;
    unb_station   unbstation;
    unb_subband   unbsubband;
    unb_input_type(unb_backplane board, unb_station st, unb_subband sb);
};
// implement this operator so we can use the unb_input_type
// as key in a map
bool operator<(const unb_input_type& l, const unb_input_type& r);
std::ostream& operator<<(std::ostream& os, const unb_input_type& unbi);
std::ostream& operator<<(std::ostream& os, const vex_output_type& vexo);

typedef std::map<unb_input_type, vex_output_type> inputmap_type;


typedef std::map<CorrelationCode, unsigned int> pol2rowmap_type;

// Timestamp of the uniboard
struct unb_timestamp {
    uint32_t    t_seconds;
    uint32_t    t_framenr;

    // initializes to {0, 0}
    unb_timestamp();
    unb_timestamp(uint32_t s, uint32_t f);
};
bool operator<(const unb_timestamp& l, const unb_timestamp& r);
bool operator==(const unb_timestamp& l, const unb_timestamp& r);
bool operator!=(const unb_timestamp& l, const unb_timestamp& r);
std::ostream& operator<<(std::ostream& os, const unb_timestamp& ts);

// Forward declare that there will, eventually, be a struct called
// unb_header
struct unb_header;

struct filestatus_type {
    const int          fd;
    unsigned int       idx, sz;
    unsigned char      buffer[1024*1024];  // make sure this is > maximum packet size (9000 bytes)!
    const std::string  filename;

    filestatus_type(const std::string& fn, int f);

    // returns pointer to current unb_header 
    unb_header const* current( void );

    // move one unb_header ahead, you can't go back
    void              next( void );

    ~filestatus_type();

    private:
        bool   eof;
        bool   do_read( void );
};

// All datafiles kept in a map of filedescriptor->string (filename)
// so that errors etc can be printed with the appropriate filename attached
//typedef std::map<int, filestatus_type>    filemap_type;
typedef std::list<filestatus_type*>    filemap_type;


// Onwards to integration. The products from uniboard are sent
// one frequency bin at a time so we must collect all the freqbins
// first before we have a complete integration
struct unb_key {
    const unb_baseline    baseline;
    const unb_subband     subband;
    const CorrelationCode polcode;

    unb_key(unb_baseline b, unb_subband s, const CorrelationCode& cc);
};
bool operator<(const unb_key& l, const unb_key& r);
std::ostream& operator<<(std::ostream& os, const unb_key& key);

// We keep the fully unmapped keys as well (ie unmapped from logical
// uniboard values to physical VEX values)
struct vex_baseline {
    const vex_station x;
    const vex_station y;
    const bool        swapped; // if a>b on creation
    vex_baseline(vex_station a, vex_station b);
};
std::ostream& operator<<(std::ostream& os, const vex_baseline& vbl);
bool operator==(const vex_baseline& l, const vex_baseline& r);
bool operator<(const vex_baseline& l, const vex_baseline& r);

struct vex_key {
    vex_baseline  baseline;
    vex_subband   subband;
    vex_key(vex_baseline b, vex_subband s);
};
std::ostream& operator<<(std::ostream& os, const vex_key& key);
bool operator<(const vex_key& l, const vex_key& r);

struct vex_product {
    const bool            need_swap;
    const vex_key         key;
    const unsigned int    row;
    const CorrelationCode polcode;

    vex_product(vex_station x, vex_station y, vex_subband sb, const CorrelationCode& cc, unsigned int r, bool swp);
};
std::ostream& operator<<(std::ostream& os, const vex_product& vp);

// For convenience purposes we keep a mapping of logical uniboard key to
// fully unmapped vex key

typedef casacore::Vector<casacore::Complex>  unb_complexvector;
typedef casacore::Vector<casacore::Float>    unb_floatvector;
typedef casacore::Vector<casacore::uInt>     unb_uintvector;
typedef casacore::Matrix<casacore::Complex>  unb_complexmatrix;
typedef casacore::Matrix<casacore::Float>    unb_floatmatrix;
typedef casacore::Matrix<casacore::Bool>     unb_boolmatrix;


struct unb_data {
    unb_complexmatrix  data;
    unb_floatvector    weight;
    // In order to ease normalization we keep a 
    // count of how many frequency bins arrived
    // for every polarization and a matrix of
    // booleans such that we know *which*
    // frequency bins came in. Note: we make the
    // flag matrix such that, if needed, it can
    // be put directly into the FLAG_SPECTRUM column
    // i.e. a value of 'True' in the matrix means the
    // point is _flagged_ (i.e. didn't arrive from UNB)!
    unb_uintvector     freqbincount;
    unb_boolmatrix     freqbinflag;

    unb_data(unsigned int nf, unsigned int np);
};

// Hoorah! Even more mappings and data types and shit!
// Update Jun 2015: take care of >1 CE computing the spectrum of one subband

// We want a mapping of unb_subband => list<vex_product>
// The reason is this:
//  * data packets are sent from a correlation engine (ie: board/node/ce);
//    the unb_ce_id
//  * the unb_ce_id maps to a specific unb_subband
//  * the unb_subband->enabled unb products should map to a list
//    of vex products (rather: row in a vex product matrix), depending on the
//    actual polarization combi
//
// By having this map pre-computed, we hope to speed up data conversion
// Because then, for each unb_ce_id, we can just iterate over the list of
// enabled *vex* products and move the spectrum to the appropriate row in
// the appropriate matrix. The matrix is located via the vex key, which is
// contained in the enabled vex products list entry.

typedef std::list<vex_product>                               enabled_vex_products_type;
typedef std::map<unb_subband_key, enabled_vex_products_type> enabled_vex_products_map;

// In a protointegration, we aggregate a list of spectra (no matrices yet)
// per correlation engine; a cornerturned version of the output of one
// correlation engine. Later, we will aggregate the spectra into matrices -
// when we take the polarization etc. into account. 
// UPDDATE: Jun 2015 - because each CE may now compute only part of a
//          spectrum we add the offset of the product into the total
//          spectrum and write the CE data just there.
//
// The speed-up consists of being able to create the cornerturned data
// blindly.
// A correlation engine (CE) may output its data in multiple packets.
// Each packet specifies, apart from the frequency bin, the 
// "first product number" and the number of products the packet carries data
// for. So for each packet we find the start index in the enabled products
// list. The data for a proto-integration maps 1:1 to a linear list of
// spectra, one for each enabled product.
// So, find the index of the "first product" in the enabled products list
// for the CE at hand and then we can just fill the
// subsequent spectra with data from the packet.
//
// Example:
//    CE [0,4,0]  (Unb 0, FPGA 4, CE# 0):
//      Enabled unb products:   1536, 1620, 1672, 1991, 2014, 2100
//      enabled uniboard :      0-0   1-1   2-2   0-1   0-2   1-2
//      enabled vex :           8-8   2-2   5-5   8-2*  8-5*  2-5   (* = needs phase swap)
//      Product index:          0     1     2     3     4     5
//
//  proto-integration data layout in memory:
//
//  |  prod #0   |   prod #1  |   prod #2  |  prod #3   |  prod #4   |  prod #5   |
//  -------------------------------------------------------------------------------
//  | 1024 fbins | 1024 fbins | 1024 fbins | 1024 fbins | 1024 fbins | 1024 fbins |
//  -------------------------------------------------------------------------------
//
//  Packet header sais:
//
//      CE=[0,4,0], fbin=337, first product=1672, nprod=2
//
//      * Get enabled unb product list for CE=[0,4,0]
//          => 1536, 1620, 1672, 1991, 2014, 2100
//      * find first product (1672) in the list
//          => that is product #2
//      * compute the offset in the data array where this spectral point is:
//          => 1024 * product nr + current fbin
//          => 1024 * 2          + 337
//          => 2385
//      * now can blindly copy the contents of the packet:
//          for(i = 0; i<nprod; i++)
//              memory[2385 + 1024*i] = packet[ i ]
//
//  |  prod #0   |   prod #1  |   prod #2  |  prod #3   |  prod #4   |  prod #5   |
//  -------------------------------------------------------------------------------
//  | 1024 fbins | 1024 fbins | 1024 fbins | 1024 fbins | 1024 fbins | 1024 fbins |
//
//  corresponding memory offsets:
//  -------------------------------------------------------------------------------
//  0            1024         2048         3072         4092         5120         |
//  -------------------------------------------------------------------------------
//
//  and memory locations written to by this packet:
//  -------------------------------------------------------------------------------
//  |                              2385         3409                              |
//  |                              *            *                                 |
//  -------------------------------------------------------------------------------
//

// HV 08 Jan 2016: Changed name from unb_ce_data to unb_sb_data to
//                 indicate that it really is *subband* data and
//                 not strictly correlation engine data.
//                 JonathanEHargreaves found an issue with the stitching
//                 of multiple CE's computing one spectrum
struct unb_sb_data {
    // Actually a matrix, overlayed on a 1-D array.
    // The dimensions of the matrix are nFreq x nEnabledProds
   casacore::Complex*  sb_data;
    // These are 1-D array with lengh nEnabledProds
    casacore::uInt*     sb_fbincount;
    casacore::Float*    sb_weight;
    // Again a matrix overlaid on a 1-D array; keeping track of
    // which freqbins arrived [this is after applying __nChannelAvg]
    casacore::Bool*     sb_fbinflag; 
    // In order to keep track of wether some of the raw freqbins
    // were mixed [indicative of an error in the mapping / offsets / averaging
    // settings] we also keep flags for the unaveraged freqbins
    casacore::Bool*     sb_fbinflag_raw; 

    unb_sb_data(unsigned int nf, unsigned int nprod, unsigned int navg);
    ~unb_sb_data();

    const unsigned int nFreq;
    const unsigned int nAvg;
    const unsigned int nEnabledProd;

    private:
        unb_sb_data();
        unb_sb_data( const unb_sb_data& );
};

// a proto integration ...
typedef std::map<unb_subband_key, unb_sb_data*>     unb_sb_integration;
typedef std::map<unb_timestamp, unb_sb_integration> unb_sb_integration_cache;


// Cache search results based on ce_id/first-product-in-packet
// such that we verify it once
struct ce_id_first_key {
    const unb_ce_id    ce_id;
    const unsigned int first_prod;

    ce_id_first_key(const unb_ce_id& id, const unsigned int fp);
};
bool operator<(const ce_id_first_key& l, const ce_id_first_key& r);

// Check the first product in the packet once against the
// enabled products list and find its (linear) index
// Update Jun 2015: Each CE may compute part of the spectrum. So we must
//                  take the fbin offset into account as well as the product
//                  offset
//        Jan 2016: The frequency bin offset is not kept in here any more;
//                  it is computed on the fly in the code that does the
//                  stitching of the output of >= 1 CE into 1 unb subband
struct fpoffset_type {
    unsigned int product_offset;

    fpoffset_type(unsigned int prod);

    private:
        fpoffset_type();
};
typedef std::map<ce_id_first_key, fpoffset_type> first_prod_offset_map_type;


// 01 Dec 2017 When 'spectral averaging' is done by the UB it actually
//             performs 'spectral accumulation'. Which means that e.g. the
//             weights are higher by a factor of 'number of bins
//             accumulated'
//             Therefore it is j2ms2's task to unaccumulate, or,
//             effectively, do the averaging.
//
//             For this we must work out the factor. Of course no-one tells
//             us directly so we must /infer/ this.
//
//             In the JSON file the actual number of spectral points is
//             registered. But depending on which firmware produced that,
//             the /actual/ amount of frequency bins produced per spectrum
//             may be different:
//                  16MHz firmware = 1024 bins/spectrum
//                  32MHz firmware = 2048 bins/spectrum
//                  64MHz firmware = 4096 bins/spectrum
//
//             From the "correlatorconfig" number we should be able to infer
//             which firmware was used for the correlation.
//             According to JonathanH. the config is a 32 bit number layed
//             out as below:
//                  0x0000 0000 0000 0000 0000 0000 0000 0000
//                  0x.... .... DDWW WWWW NNNN NNNN BBBB BBBB
//
//              with:
//                  B = number of bands per station   8bit
//                  N = log2 number of freq bins      8bit
//                  W = DDR local address bus width   6bit
//                  D = number of DDR modules         2bit
//
//             For N we typically find 0x0A, 0x0B, 0x0C for
//             2^10, 2^11 and 2^12 bins per spectrum respectively
//
struct fwconfig_fields {
    uint32_t    unused:8;
    uint32_t    n_ddr_modules:2;
    uint32_t    ddr_bus_width:6;
    uint32_t    log2_n_fbin:8;
    uint32_t    n_band_p_station:8;
};
union fw_config_t {
    template <typename T>
    fw_config_t(T const& t):
        correlatorconfig( t )
    {}

    uint32_t        correlatorconfig;
    struct {
        uint32_t    n_band_p_station:8;
        uint32_t    log2_n_fbin:8;
        uint32_t    ddr_bus_width:6;
        uint32_t    n_ddr_modules:2;
        uint32_t    unused:8;
    };
};

// We should be able to analyze the enabled products into
// a per-integration struct
typedef std::map<vex_key,unb_data>                  unb_integration;
typedef std::map<unb_timestamp, unb_integration>    unb_integration_cache;


class UNBVisBuffer: public VisibilityBuffer
{
    public: 
        // we support this construction signature
        UNBVisBuffer(const casacore::String& vb, const Experiment* exp,
                const std::string& vbopt = std::string());

        // We *must* supply these [the VisBuffer interface]
        virtual casacore::Bool                   getCurrentVisibility( Visibility& v );
        
        virtual casacore::Bool                   next( void );
        virtual const casacore::String&          getRootDir( void ) const;
        virtual casacore::Int                    getVisBufId( void ) const;
        virtual casacore::String                 getProcessorSubType( void ) const;
        virtual CorrelatorSetup                  correlatorSetup( void ) const;
        virtual ~UNBVisBuffer();

    private:
        // List of pointers to scans. Memory management is left up to user
        // [you keep track wether or not you 0wn the pointer :)]
        typedef std::vector<j2ms::COFScan*>  scanptrlist_type;

        // For the current mode keep a mapping of vex_subband => sidebandedness
        typedef std::map<vex_subband, BaseBand::SideBand>  vsb2sidebandmap_type;

        // CorrelatorSetup is fixed per visbuffer
        int                        __debug;
        bool                       __doNormalization;
        fw_config_t                __fw_config;
        std::string                __mode;
        unsigned int               __nChannelAvg;
        unsigned int               __nFFTpIntg;
        unsigned int               __weightScale;
        casacore::String           __rootdir;
        filemap_type               __filemap;
        inputmap_type              __inputmap;
        correlations_t             __correlations;
        pol2rowmap_type            __pol2rowmap;
        CorrelatorSetup            __setup;
        scanptrlist_type           __scanptrlist;
        vsb2sidebandmap_type       __vsb2sidebandmap;

        // UNB specific mappings
        unb_ce_sb_map             __ce_sb_map;
        unb_product_map           __productmap;
        unb_enabled_products_map  __enabledproducts;

        // Mapping from unb_ce_id => enabled VEX products
        enabled_vex_products_map  __enabled_vex_products;

        // The CE-data integration cache
        unb_sb_integration_cache           __sb_integrationcache;
        first_prod_offset_map_type         __first_prod_offset_map;

        // The integration cache
        unb_integration_cache           __integrationcache;
        unb_integration::iterator       __currentbaseline;
        unb_integration_cache::iterator __currentintegration;

        // Time-keeping: the start of the data taking (__zulu)
        // and the time of the current integration (__t0)
        casacore::MVEpoch                   __zulu;
        casacore::MVEpoch                   __t0;


        // non-static fuction(s)
        void                          openDataFiles(const Json::Value& json,
                                                    const std::string& visbufrootdir );

        bool                          readNext( void );
        void                          doNormalize( unb_integration_cache::iterator intgptr );

        // Static helper functions
        static mappingresult_type     readEnabledProductsMap(const Json::Value& json);
        static inputmap_type          readInputMap(const Json::Value& json);
        static unb_ce_sb_map          readCESubbandMap(const Json::Value& json);


        // Prohibit these
        UNBVisBuffer();
        UNBVisBuffer(const UNBVisBuffer&);
        const UNBVisBuffer& operator=(const UNBVisBuffer&);
};


#endif
