//
//  VEXFrequencyDB - Specialization of the FrequencyDB class, reading
//  frequency configurations from a VEX-file....
//
//  Author:  Harro Verkouter,  6-7-1999
//
//  $Id: VEXFrequencyDB.h,v 1.8.2.1 2018/06/07 15:27:12 verkout Exp $
//
//  $Log: VEXFrequencyDB.h,v $
//  Revision 1.8.2.1  2018/06/07 15:27:12  verkout
//  HV: * fix a lot of things in j2ms implementation
//          - meta data (antenna, spectral window, source) only written
//            if data exists for it
//          - hopefully proper handling of polarization/feed/products
//          - delete writing obsolete JIVE_TAPESTATS table
//          - support changing of "eo:..." options running from lis-file
//
//  Revision 1.8  2016/06/23 13:02:26  jive_cc
//  HV: * moved codebase to casacore 2.x compatibility - the namespace is now
//        casacore::  in stead of casa::. A lot of improvement ......
//      * now link against vex parser which understands $BITSTREAMS and $THREADS
//
//  Revision 1.7  2013-04-26 11:22:08  jive_cc
//  HV: * It is now possible to pass a reference station for frequency setups
//        to the experiment - this is needed for mixed mode observations.
//        Add "eo:setup_ref_station=XXXX" to the j2ms2 commandline.
//        The default is unchanged: uses the first station in the VEX file for
//        the frequency setup definitions.
//
//  Revision 1.6  2007-02-26 15:02:18  jive_cc
//  HV: - adher to new interface functions
//      - cosmetic changes
//      - throw if something's fishy since fishyness at
//        this level is unacceptable
//
//  Revision 1.5  2006/03/02 14:21:40  verkout
//  HV: * Removed remaining instances of strstream, replaced with strstream * Got rid of (home grown) Container class!!!! YES! Finally replaced with STLs vector/list, whichever is appropriate * Added pure virtual method to hduGenerator - class writers MUST return an ID which (uniquely) identifies their specific Generator * mapping of generator-ID to actual Generator is now built automagically rather than by ms2uvfitsConverter, class writers should look in Registrar.h for details
//
//  Revision 1.4  2006/01/13 11:35:40  verkout
//  HV: CASAfied version of the implement
//
//  Revision 1.3  2004/01/05 15:02:18  verkout
//  HV: Probs. with templates. Moved Container to different namespace.
//
//  Revision 1.2  2001/05/30 11:48:20  verkout
//  HV: Made changes to accomodate MSv2.0 and moved all code from the 'vlbi' branch to the 'jive' branch.
//
//  Revision 1.1.1.1  2000/03/20 15:14:36  verkout
//  HV: imported aips++ implement stuff
//
//
#ifndef VEXFREQUENCYDB_H
#define VEXFREQUENCYDB_H

#include <casacore/casa/aips.h>
#include <jive/j2ms/FrequencyDB.h>
#include <casacore/casa/BasicSL/String.h>
#include <jive/j2ms/FrequencyConfig.h>

#include <vector>


class VEXFrequencyDB :
    public FrequencyDB
{
public:
    //  Create the database from the vexfile!
    VEXFrequencyDB( const casacore::String& vexfilename, const casacore::String& refStation="" );

    //  Provide these functions

    // lookup by frequency-code
    virtual FrequencyConfig  operator[]( const casacore::String& key ) const;
    virtual casacore::Bool       searchFrequencyConfig( FrequencyConfig& fcref,
                                                    const casacore::String& key ) const;
    // lookup by freq-group-number
    virtual casacore::Bool       searchFrequencyConfig( FrequencyConfig& fcref,
                                                    casacore::Int key ) const;

    // HV: Apr 2018 - Maybe it's time to start thinking of pointing at
    //                something!
    virtual FrequencyConfig const*    findFrequencyConfigPtr( casacore::String const& key   ) const; 
    virtual FrequencyConfig const*    findFrequencyConfigPtr( casacore::Int           fqgrp ) const; 

    //  Can access them by index 0..(nEntries()-1)
    //  Note: frequency-group-number for entry #n may or may not be equal to 'n'.
    //  Don't count on it.
    virtual casacore::uInt       nEntries( void ) const;
    virtual FrequencyConfig  operator[]( casacore::uInt idx) const;

    // the recorded polarizations
    const j2ms::polarizations_t& recordedPolarizations( void ) const;
    casacore::String const&      setupRefStation( void ) const;

    //  Delete the resources for this database
    virtual ~VEXFrequencyDB();

private:
    //  The private parts
    //  Stick the frequency configs in a vector
    //  (we must be able to adress the n-th entry...
    //  and map<> destroys the order we insert'm in)
    typedef std::vector<FrequencyConfig>  configurations_t;

    casacore::String      myRefStation;
    configurations_t      myConfigurations;
    j2ms::polarizations_t myRecordedPolarizations;
    
    //  Prohibit these
    VEXFrequencyDB();
    VEXFrequencyDB( const VEXFrequencyDB& );
    const VEXFrequencyDB& operator=( const VEXFrequencyDB& );
};


#endif
