// implementation of the non-virtual VisibilityBuffer class methods
//
// $Id: VisibilityBuffer.cc,v 1.7.2.1 2018/08/21 12:26:29 verkout Exp $
//
// $Log: VisibilityBuffer.cc,v $
// Revision 1.7.2.1  2018/08/21 12:26:29  verkout
// HV: * Added 'option_map' implementation so it can be reused to
//       support "key=value,key2=value2..." passing of settings
//     * On multi-phase-center (700-ish) j2ms2 become /SUPER/ slow:
//       the ghostbuster-filter (enabled by default) was effectively
//       disabled (on purpose) by SFXCVisBuffer implementation but did
//       that in the most time-consuming way ...
//     * GhostBusterFilter changed:
//         - now takes options (see j2ms2 help) [can honour station based
//           data_good from VEX file if desired]
//         - responsibility for implementing the filter moved out of
//           *VisBuffer classes to better suited logical level
//     * More descriptive SFXCVisBuffer error messages
//     * UNBVisBuffer: overlooked the fact that subbands now have to be
//       sorted by frequency AND sideband in stead of just freq.
//
// Revision 1.7  2016/06/23 13:02:27  jive_cc
// HV: * moved codebase to casacore 2.x compatibility - the namespace is now
//       casacore::  in stead of casa::. A lot of improvement ......
//     * now link against vex parser which understands $BITSTREAMS and $THREADS
//
// Revision 1.6  2007-05-25 09:42:41  jive_cc
// HV: - Added interface for requesting which processing steps have been
//     done for the given visibility buffer. Default returns empty
//     list of processing steps.
//
// Revision 1.5  2007/03/15 16:35:43  jive_cc
// VisibilityBuffer.cc
//
// Revision 1.4  2007/02/26 15:45:09  jive_cc
// HV: - cosmetic changes
//     - interface change: a visibilitybuffer may now
//       also generate modelrecords
//
// Revision 1.3  2006/08/17 13:41:14  verkout
// HV: * commentchanges * Added interface function to which should return the full list of scans that may be found in the VisBuf
//
// Revision 1.2  2006/03/02 14:21:40  verkout
// HV: * Removed remaining instances of strstream, replaced with strstream * Got rid of (home grown) Container class!!!! YES! Finally replaced with STLs vector/list, whichever is appropriate * Added pure virtual method to hduGenerator - class writers MUST return an ID which (uniquely) identifies their specific Generator * mapping of generator-ID to actual Generator is now built automagically rather than by ms2uvfitsConverter, class writers should look in Registrar.h for details
//
// Revision 1.1  2006/02/14 13:32:49  verkout
// HV: Really dunno why this'un wasn't there before. Anyway, it contains code which is common to all VisibilityBuffers
//
//
#include <jive/j2ms/VisibilityBuffer.h>
#include <jive/j2ms/Experiment.h>

#include <casacore/casa/Exceptions.h>

//  Visibility buffers must be attached to an Experiment
VisibilityBuffer::VisibilityBuffer( const Experiment* experiment ) :
	itsScanptr( 0 ),
	itsExperimentptr( experiment )
{
	//  If no experiment -> we give up!
	if( !itsExperimentptr )
		throw( casacore::AipsError("VisibilityBuffer: No experiment given!") );
}


//  Can do this because we MUST have an experiment
const Experiment&  VisibilityBuffer::getExperiment( void ) const {
	return *itsExperimentptr;
}
    

//  Tell the current scan (if any)
const j2ms::Scan*  VisibilityBuffer::getScanptr( void ) const {
	return itsScanptr;
}

//  Tells us about vernier delay
casacore::Bool VisibilityBuffer::getVernierDelay( void ) const {
	return itsExperimentptr->hasVernierDelay();
}

// HV: Changed. Rather than returning an empty one, return the list
// gotten from the experiment
const j2ms::scaninfolist_t& VisibilityBuffer::getScanlist( void ) const {
    return itsExperimentptr->getScanlist();
}

const j2ms::modeldata_t& VisibilityBuffer::getModelData( void ) const {
    static const j2ms::modeldata_t     empty;
    return empty;
}

processing_steps VisibilityBuffer::getProcessingSteps( void ) const {
    static const processing_steps     empty;
    return empty;
}

VisibilityBuffer::~VisibilityBuffer() {}
