#include <jive/j2ms/scanlist.h>
#include <jive/j2ms/DayConversion.h>
#include <jive/Utilities/jexcept.h>

#include <iostream>
#include <iomanip>
#include <sstream>
#include <set>
#include <vector>

#include <casacore/casa/Quanta/MVTime.h>

// need this for VexPlus class
#include <jive/j2ms/vexplus.h>

using namespace std;
using namespace j2ms;
using namespace casacore;


namespace j2ms {

    stationscaninfo_t::stationscaninfo_t() :
        scanStart(0.0), scanEnd(0.0), dataStart(0.)
    {}

    stationscaninfo_t::stationscaninfo_t(MVEpoch const& s, MVEpoch const& e, casacore::Quantity const& ds):
        scanStart( s ), scanEnd( e ), dataStart(ds)
    {}

    stationscaninfo_t::stationscaninfo_t(MVEpoch const& s, const Quantity& d, casacore::Quantity const& ds):
        scanStart( s ), scanEnd( s+d ), dataStart(ds)
    {}
}


void parseVEXScans(j2ms::scaninfolist_t& sil, j2ms::station2listmap_t& s2lm, std::string const& vexfilename, bool verbose) {
    int               result;

    if( vexfilename.empty() )
        THROW_JEXCEPT("Empty vexfilename!");

    // Read the full scanlist from the vexfile
    VexPlus  vexfile( vexfilename );

    if( (result=vexfile.parseVex())!=0 ) {
		switch( result ) {
			case -1:
				THROW_JEXCEPT("could not open file " << vexfilename);
				break;
		
			case -2:
                THROW_JEXCEPT("error parsing vex-file " << vexfilename << " to memory");
				break;
		
			default:
                THROW_JEXCEPT("unknown error code " << result << " from VexPlus::parseVex()");
				break;
		}
    }
    // make sure the return values are emptied
    sil.clear();
    s2lm.clear();

    // Find the scanlist by retaining the scans by 'scanName'
    // rather than by any other identifying piece-o'-info
    // Loop over all stations and for all scans defined for the station,
    // check if we already have a scan by that name. If not, add it to 
    // the list.
	set<string>   scanList;

    for(int stationcnt=0; stationcnt<vexfile.N_Stations(); stationcnt++ ) {
		string const      curstation = vexfile.Station( stationcnt );
		int const         nscans     = vexfile.N_Scans( curstation );
        stationscanlist_t station_scans;
	
		if( curstation.empty() ) 
			THROW_JEXCEPT("parseVEXScans/Failed to get Station name from VEXfile for station #" << stationcnt);

        if( nscans==0 ) {
    		cout << "parseVEXScans/WARN: station " << curstation << " has no scans defined at all" << endl;
            continue;
        }
        if( verbose )
    		cout << "parseVEXScans/reading " << nscans << " scans for "
    		   	 << curstation << " (" << stationcnt << "/" << vexfile.N_Stations() << ")"
    			 << endl;

		for( int i=0; i<nscans; i++ ) {
			int          year, daynr, hour, minute, second;
			double       dayfraction;
            MVTime       scanmvtime;
			ScanInfo     si;
			unsigned int month, day;
			string const scanlab      = vexfile.ScanName(curstation, i);
			string const scanstart    = vexfile.ScanStart(curstation, i);
            string const scanmode     = vexfile.ScanMode(curstation, i);
            double const datastart    = vexfile.DataStart(curstation, i);
            double const scanduration = vexfile.ScanDuration(curstation, i);
	
			if( scanlab.empty() ) 
				THROW_JEXCEPT("Failed to get ScanName from VEXfile " << vexfilename << " for station '"
					            << curstation << "', scan #" << i);
			if( scanstart.empty() )
				THROW_JEXCEPT("Failed to get ScanStart from VEXfile " << vexfilename << " for station '"
                              << curstation << "', scan #" << i);
            if( scanmode.empty() )
				THROW_JEXCEPT("Failed to get ScanMode from VEXfile " << vexfilename << " for station '"
                              << curstation << "', scan #" << i);
            if( scanduration<=0.0 )
				THROW_JEXCEPT("Invalid ScanDuration " << scanduration << " from VEXfile " << vexfilename << " for station '"
                              << curstation << "', scan #" << i);
            if( datastart<0.0 )
				THROW_JEXCEPT("Invalid DataStart " << datastart << " from VEXfile " << vexfilename << " for station '"
                              << curstation << "', scan #" << i);

            if( datastart>0 && verbose )
                cout << "   scan#" << i << " for station " << curstation << " has data_good of " << datastart << endl;

			// HV: 15/03/2005 - Now the key is "scanName" and not "scanstart"
			//                  anymore
            // HV: 07/12/2017 - In order to support multiple phase centers
            //                  we need to add the source name to the label
            //                  to check if we've already got this
            //                  *particular* scan
            //                  Also: change the location of this test to
            //                  inside the loop over all sources for this
            //                  scan
	
            // Already form all the scan constants; then we can loop over
            // all sources defined for this scan for this station and check
            // if we already got this particular one

			//  Transform the scanstart into a mvepoch
			if( ::sscanf(scanstart.c_str(), "%dy%dd%dh%dm%ds", &year, &daynr, &hour, &minute, &second)!=5 )
                THROW_JEXCEPT("Error in format, could not decode 'scanstart' " 
                              << "for station '" << curstation << "', scan #" << i);
	
			DayConversion::dayNrToMonthDay( month, day, daynr, year );
	
			dayfraction = ((hour*3600.0)+(minute*60.0) + (double)second)/(double)DayConversion::secondsPerDay;
			scanmvtime  = MVTime(year, month+1, day, dayfraction);

			//  Fill in the parts of the scan definition that are constant
			si.scanNo          = scanList.size()+1;
			si.scanStart       = MVEpoch( scanmvtime.day() );
			si.scanEnd         = si.scanStart + Quantity(scanduration, "s");
			si.scanId          = scanlab;
			si.frequencyCode   = scanmode;

            // We can already add this scan to the station-based scan list;
            // this scan is definitely defined for this station.
            // Need to compute the end time of data for this scan
            station_scans.push_back( stationscaninfo_t(si.scanStart, si.scanEnd, casacore::Quantity(datastart, "s")) );

            // now loop over all sources defined for this scan
            vector<string> sources = vexfile.ScanSourceAll(curstation, i);
            for( vector<string>::const_iterator curSource = sources.begin(); curSource!=sources.end(); curSource++ ) {
                string const sourceName = vexfile.SourceName(*curSource);

                if( sourceName.empty() )
                    THROW_JEXCEPT("parseVEXScans/Failed to get source name from VEX file " << vexfilename << " for source " << *curSource);

                // See: add source code to scan label to see if we already
                //      have this one
                std::string const   actualLabel( scanlab + sourceName );
                if( scanList.find(actualLabel)!=scanList.end() )
                    continue;
                if( verbose )
                    std::cout << "parseVEXScans/Adding scan " << scanlab << "/" << sourceName << std::endl;
                // brrrrm... nice to check if the thing appears in scanList, but.. errr..
                // if you forget to add a new one everytime you find one, the list
                // will stay empty, won't it??!
                scanList.insert( actualLabel );
            
                si.sourceCode = sourceName;
                sil.push_back(si);
            }
        }
        // Finished looping over all scans for this station - now stick the
        // station-based-scan-list into the mapping
        // note: .insert() returns pair<iterator, bool>
        if( !s2lm.insert(std::make_pair(stationcnt, station_scans)).second )
            THROW_JEXCEPT("parseVEXScans: failed to insert station based scan list for station " << curstation << " (#" << stationcnt << ") into map");
    }
}

