//  Implementation of the ms2idifitsConverter class
//
#include <jive/ms2uvfitsimpl/Converters/ms2idifitsConverter.h>



#include <jive/ms2uvfitsimpl/FITSUtils/HDUObject.h>

#include <jive/ms2uvfitsimpl/Generators/antabGenerator.h>
#include <jive/ms2uvfitsimpl/Generators/maingroupGenerator.h>
#include <jive/ms2uvfitsimpl/Generators/asciitabGenerator.h>
#include <jive/ms2uvfitsimpl/Generators/fqtabGenerator.h>
#include <jive/ms2uvfitsimpl/Generators/tytabGenerator.h>
#include <jive/ms2uvfitsimpl/Generators/gctabGenerator.h>
#include <jive/ms2uvfitsimpl/Generators/sutabGenerator.h>
#include <jive/ms2uvfitsimpl/Generators/idihdrGenerator.h>
#include <jive/ms2uvfitsimpl/Generators/idisourceGenerator.h>
#include <jive/ms2uvfitsimpl/Generators/idifrequencyGenerator.h>
#include <jive/ms2uvfitsimpl/Generators/idityGenerator.h>
#include <jive/ms2uvfitsimpl/Generators/idiantennaGenerator.h>
#include <jive/ms2uvfitsimpl/Generators/idigaincGenerator.h>

#include <jive/ms2uvfitsimpl/Generators/idiFlagGenerator.h>
#include <jive/ms2uvfitsimpl/Generators/idiGainCurveGenerator.h>
#include <jive/ms2uvfitsimpl/Generators/idiPhaseCalGenerator.h>
#include <jive/ms2uvfitsimpl/Generators/idiSysTempGenerator.h>

#include <jive/ms2uvfitsimpl/Generators/uvtableGenerator.h>
#include <jive/ms2uvfitsimpl/Generators/arraygeomGenerator.h>
#include <jive/ms2uvfitsimpl/Generators/Registrar.h>

#include <casacore/casa/IO/RegularFileIO.h>

#include <jive/Utilities/jexcept.h>
#include <casacore/casa/BasicSL/String.h>

#include <casacore/tables/TaQL/ExprNode.h>

#include <jive/Utilities/streamutil.h>

#include <sstream>

#include <errno.h>

using namespace std;
using namespace casacore;

// Note: 
uLong nrVisInChunk(MSConversionInfo const& base,
                   double const max_size_per_chunk = (double)(((unsigned)1)<<31) - 200*(1024*1024) /*allow for ~200MB overhead per chunk?*/ ) {
    double               size_per_row;
    double               overhead_per_row;
    double               size_per_visibility;
    // Ok. From the MSConversionInfo we can derive the
    // approx size of a row in the MS...
    // Then we can easily work out how many rows of
    // the MS can be put in a single FITS-file
    // 
    // Note: there is a factor "nr-of-IFs" between the
    // two row-sizes.
    // fits-row-size = nrIFs * ms-row-size
    //
    // Note: you can check the 'uvtableGenerator.cc' source
    //       file to see what will get written to the FITS
    //       file
    //
    // Data is written as 3 Floats (Re/Im/Wgt) (=3*4bytes) per datapoint +
    //   a number of indices per row.
    // Number of datapoints in a single row: 
    //   number of stokes parms (number of polarizations) *
    //   number of channels  * nr-of-IFs
    size_per_visibility = (double)base.nrPolarizations() *
                          (double)base.nrChannels() *
                          (double)base.nrIFs() *
                          (3 * sizeof(Float));

    // The per-row overhead is a number of extra indices and
    // some per-visibility constant numbers (UVW, date, time)
    //
    // We have 4*Float (for U,V,W, and inttim),
    //  3*(unsigned)Int (baseline, sourceid, freqid),
    //  and 2 doubles (date,time)
    overhead_per_row = 4*sizeof(Float) +
                       3*sizeof(Int) +
                       2*sizeof(Double);
    
    // This gives us the nominal size_per_row:
    size_per_row = size_per_visibility + overhead_per_row;

    // Now make an estimate how many row we can fit in the
    // max size/chunk. Note: these are *fits* rows
    // [so they take up nr-of-IFs MS-rows]
    return (uLong)(max_size_per_chunk/size_per_row);
}







ms2idifitsConverter::ms2idifitsConverter():
    FileFormatConverter()
{}

ms2idifitsConverter::ms2idifitsConverter( const generatorlist_t& hdus ):
    FileFormatConverter(),
    hdus2do( hdus )
{}

typedef std::list<hduGenerator*>    genpointers_type;
//  This function really converts the MS into a UVFITS file....
bool ms2idifitsConverter::convert( const String& fitsfilename,
                                   const String& MSfilename ) const {
    // Do some preprocessing on the HDU generator list of the HDUs we're
    // supposed to generate. Make sure there is an actual generator
    // registered for each type.
    genpointers_type                 genpointers;
    uvtableGenerator*                uvtable( 0 );
    hduGenerator::idtype_t const     uvtable_type = uvtableGenerator().getID();

    for(generatorlist_t::const_iterator curgen = hdus2do.begin(); curgen!=hdus2do.end(); curgen++) {
        hduGenerator*   actualgen = Registrar::find_ID( *curgen );

        if( actualgen==0 )
            throw std::runtime_error(std::string("ms2idifits: no actual HDU generator registered for type '")+*curgen+"'");
        // If it's the UV_TABLE generator one, don't add it to the list but keep it apart
        if( *curgen==uvtable_type ) {
            if( uvtable )
                throw std::runtime_error("ms2idifits: duplicate UV_TABLE generator in the list-of-HDUs-to-generate?!");
            uvtable = dynamic_cast<uvtableGenerator*>(actualgen);
            if( uvtable==0 )
                throw std::runtime_error("ms2idifits: UV_TABLE generator pointer does not point at uvtableGenerator? (dynamic_cast fail)");
        } else {
            // Append to list 
            genpointers.insert(genpointers.end(), actualgen);
        }
    }
    // Assert that, in fact, there IS a UV_TABLE generator
    if( uvtable==0 )
        throw std::runtime_error("ms2idifits: the uvtable generator is not present!");

    //  Now we face the task of converting the named MS to the named
    //  idifiTS-file...

    // We need to iterate over all the Visibilities in the MS
    // in order (otherwise AIPS starts nagging + having TIME/BASELINE
    // sort order is convenient for the users)
    hdulist_t           commonHDUs;
    Block<String>       sortNames( 4 );
    std::ostringstream  noHDU;

    sortNames[0]="TIME";
    sortNames[1]="ANTENNA1";
    sortNames[2]="ANTENNA2";
    sortNames[3]="FIELD_ID";

    cout << "ms2idi: Please wait whilst opening + sorting " << MSfilename << " by TIME, BL, SOURCE ... " << std::endl;
    MS const         sorted( MS(MSfilename, Table::Old).sort(sortNames) );

    // Right. If the MS is empty we don't even have to do nothing!
    if( sorted.nrow()==0 ) {
        cout << "ms2idi: the MS seems to be empty so no FITS file(s) generated" << endl;
        return true;
    }

    cout << "ms2idi: Start to analyze the MeasurementSet if it can be represented in FITS format ..." << std::endl;
    MSConversionInfo conversion_info( sorted );
    cout << std::endl
         << "ms2idi: Start generating HDUs" << std::endl;

    // Now we can already generate all the HDUs that are common between
    // all the pieces  [Note: the UV_TABLE generator has at this point
    // already been filtered out]
    for(genpointers_type::const_iterator genptr=genpointers.begin(); genptr!=genpointers.end(); genptr++) {
        hdulist_t   toappend( (*genptr)->generate(conversion_info) );
        if( toappend.empty() )
            noHDU << (*genptr)->getID() << " ";
        std::copy(toappend.begin(), toappend.end(), std::back_inserter(commonHDUs));
    }
    std::string const noHDUs( noHDU.str() );
    if( !noHDUs.empty() )
        cout << "ms2idi: no HDUs generated by generators: " << noHDUs << std::endl;

    // Need to work out how many visibilities per FITS file. At this
    // point we make FITS files <= 2GB
    uLong const         nVisPerChunk = nrVisInChunk( conversion_info );
    Block<String>       iternames( 4 );
    std::ostringstream  filenamestrm;

    iternames[0]="TIME";
    iternames[1]="ANTENNA1";
    iternames[2]="ANTENNA2";
    iternames[3]="FIELD_ID";

    // OK we're going to keep on converting until this iterator runs dry
    TableIterator     iter(conversion_info.getMS(), iternames, TableIterator::Ascending);

    //  Reset the counters for this run
    uvtable->visCount   = 0;
    uvtable->pieceCount = 0;

    while( !iter.pastEnd() ) {
        // OK need to generate a new FITS chunk
        hdulist_t            uvdata;
        filenamestrm.str( string() );
        filenamestrm << fitsfilename << (uvtable->pieceCount+1);

        RegularFileIO        fitsFile(RegularFile(filenamestrm.str()), ByteIO::NewNoReplace);

        // First flush all the common HDUs
        for(hdulist_t::const_iterator hdu=commonHDUs.begin(); hdu!=commonHDUs.end(); hdu++)
            fitsFile << **hdu;

        // Let the UV_TABLE generator do its thang
        uvdata = uvtable->generate(conversion_info, iter, nVisPerChunk);
        if( uvdata.empty() )
            cout << "ms2idi: no UV_TABLEs generated for file#" << uvtable->pieceCount << endl;
        for(hdulist_t::const_iterator hdu=uvdata.begin(); hdu!=uvdata.end(); hdu++) {
            fitsFile << **hdu;
            delete *hdu;
        }
        // Moving on to next piece
        uvtable->pieceCount++;
    }

    // Before leaving, delete the common HDUs
    for(hdulist_t::const_iterator hdu=commonHDUs.begin(); hdu!=commonHDUs.end(); hdu++)
        delete *hdu;

    // If only one piece was constructed, rename it back to the name w/o
    // the "1" suffix
    if( uvtable->pieceCount==1 )
        if( ::rename(filenamestrm.str().c_str(), fitsfilename.c_str())==-1 )
            throw std::runtime_error(string("Failed to rename fits file: ")+::strerror(errno));

    return true;
}

//  Configure ourselves such that the conversion is run according to
//  the user's wishes!
Bool ms2idifitsConverter::setConversion( const generatorlist_t& hdus ) {
    unsigned int                    gencount;
    generatorlist_t::const_iterator curgen;
    
     cout << endl;
     cout << "Next conversion will produce, in the following order:" << endl;
    
     hdus2do = hdus;
    
    for( curgen=hdus2do.begin(), gencount=0;
         curgen!=hdus2do.end();
         curgen++, gencount++ ) {
		cout << j2ms::printeffer("%3d: ",gencount) << *curgen << endl;
    }
    return True;
}



//  Destruct the conversion engine...
ms2idifitsConverter::~ms2idifitsConverter()
{}

