//
//  Class:
//
//        FITSColumn
//
//
//  Purpose:
//
//        This class will describe the properties of a
//        column in a FITS table. These properties encompass the
//        datatype ( ..FORM fields in the FITSdescription), the
//        description of the column and the width and the offset of
//        the column within a row in the FITS table.
//
//        I choose not to have this column templated bute use the
//        AIPS++ DataType enumeration instead, to prevent us from
//        having to upcast to templated type.
//
//
//
//
//  Author:
//
//        Harro Verkouter     27-3-1998
//
//
//      $Id: FITSColumn.h,v 1.7 2016/06/23 13:02:28 jive_cc Exp $
//
#ifndef FITSCOLUMN_H_INC
#define FITSCOLUMN_H_INC


#include <jive/ms2uvfitsimpl/FITSUtils/BasicFITSUtility.h>
//#include <jive/ms2uvfitsimpl/FITSUtils/FITSTableHDU.h>
#include <casacore/casa/aips.h>
#include <casacore/casa/BasicSL/String.h>
#include <casacore/casa/Arrays/IPosition.h>
#include <casacore/casa/Arrays/Slicer.h>
#include <casacore/casa/Utilities/DataType.h>
#include <casacore/casa/Arrays/Array.h>
#include <casacore/casa/Containers/RecordInterface.h>
#include <casacore/casa/Containers/RecordDesc.h>
#include <casacore/casa/Containers/RecordFieldId.h>

#include <casacore/fits/FITS.h>

#include <casacore/tables/Tables/ScalarColumn.h>
#include <casacore/tables/Tables/ArrayColumn.h>

#include <iostream>


class FITSTableHDU;

class FITSColumn :
    public BasicFITSUtility
{
public:
    //
    // Return the AIPS++ datatype that describes the nature of this
    // column
    //
    casacore::DataType                getColumnDataType( void ) const;
    casacore::RecordDesc              getColumnDescription( void ) const;

    //
    //  Translate this Column to Fitskeywords
    //
    virtual casacore::FitsKeywordList asFITSKeys( void ) const;

    //
    //  Get or put the value in a cell. Derived columns MUST implement these methods!
    //
    virtual casacore::Bool            get( casacore::uInt rownr, casacore::RecordInterface& recordref, const casacore::RecordFieldId& fieldidx ) const =0;
    virtual casacore::Bool            put( casacore::uInt rownr, const casacore::RecordInterface& recordref, const casacore::RecordFieldId& fieldidx ) =0;

    
    //
    //  Let the user have access (mostly read-only) to some of the
    //  private parts!
    //

    //
    // The width of this particular column (in bytes)
    //
    virtual casacore::uLong           getWidthInBytes( void ) const;
    

    //
    //  Accessors, use FITS nomenclature:
    //
    const casacore::String&           ttype( void ) const;
    const casacore::String&           tform( void ) const;
    const casacore::String&           tunit( void ) const;

    //
    //  Some administrative functions, I don't think users will be
    //  much using these, but anyway.
    //
    casacore::Int                     columnnr( void ) const;
    casacore::uLong                   columnoffset( void ) const;

    //
    //  Do we happen to be a binary column, by any chance??
    //
    casacore::Bool                    isBinaryColumn( void ) const;

    //
    //  Destruct this Column
    //
    virtual ~FITSColumn();

protected:
    //
    //  Three constructors. We make all of them protected, so only
    //  derived classes (actual specialization of the column class)
    //  are able to construct a column object.
    //
    //  1) Create a complete column -> user has control over all
    //  Column properties (but the Column number and the offset)
    //
    //  2) Create a column from a FitsKeywordlist -> intended use:
    //  attempt to read the column's settings from the list of
    //  keywords just read from file! The parent HDU tells us what our
    //  ordinal number and offset are.
    //
    //  3) A kind of copy-constructor -> the column settings of
    //  'other' are copied across apart from the columnnumber and
    //  offset. This constructor finds its use when a HeaderDataUnit
    //  needs to add a column. The user specifies a column that needs
    //  to be added to the HeaderdataUnit. The user does not know the
    //  actual number of the column. The HeaderdataUnit does know and
    //  uses this c'tor to create the correct column.
    //
    //
    //  NOTE: the number parameter (as well as the private datamember
    //  'itsColumnnumber' are Ints, because the value '-1' is used to
    //  indicate that the column has not been assigned to a valid HDU!
    //
    FITSColumn( const casacore::String& type, const casacore::String& comment,
                const casacore::String& form, const casacore::String& unit );
    
    FITSColumn( casacore::Int number, casacore::uLong offset,
                casacore::FitsKeywordList& kwsource, FITSTableHDU* ptr2parent );
    
    FITSColumn( casacore::Int number, casacore::uLong offset,
                const FITSColumn& other, FITSTableHDU* ptr2parent );
   

    // 
    // Because of gcc3.4 (and higher) ISO C++ standard compliance,
    // this copy c'tor must exists (bummer).
    // "temporaries, even those passed by const-reference *MUST*
    //  be copy constructable" <-- b*stards! ;)
    // And this construct we use in the 'addColumn()' methods...
    // (passing a temporary by const ref)
    //
    // NOTE NOTE NOTE NOTE
    //    IF the 'other' object has a Parent, it doesn't get copied
    //    across!
    //
    FITSColumn( const FITSColumn& other );

    //
    //  The derived classes must have acces to these members: they
    //  first have to parse the TFORM field and hence they can decide
    //  what column type/shape datatype etc. it is!
    //
    casacore::Bool      setDataType( const casacore::DataType& newdatatype );
    casacore::Bool      setColumnDescription( const casacore::RecordDesc& newdescription );
    casacore::Bool      setForm( const casacore::String& newform );

    //
    //  Sometimes (e.g. for an ASCII-column) we may have to change the
    //  offset for this column...
    //
    void      setColumnOffset( casacore::uLong newoffset );
    

protected:

    //
    //  Yuck! Do I not like this.... derived classes (and friends for
    //  that matter) can use this method to tell this Column
    //  Object that is really is a BinaryColumn... but.. since we must
    //  make decisions for keywords in another object (the
    //  BinaryTable) we must be able to find out if there are Matrix
    //  columns present and if so, we must even know which one it is!
    //
    //  Wait... you haven't heard the best part yet... it's not even
    //  part of the standard, these so-called matrix columns! They're
    //  an invention of the the Classic AIPS team... and since we
    //  *really* need classic AIPS we'll need to find a way to get our
    //  data through into classic AIPS...
    // 
    void        setBinaryColumnFlag( void );

private:
    //
    //  These are the variables that describe the Column
    //

    //
    //  The Column type
    //
    casacore::String         itsColumnType;
    
    //
    //  The comment
    //
    casacore::String         itsComment;
    
    //
    // This string describes the form of the column, according to
    // form= element_count+BASCIC_TYPE_CHARACTER,
    //
    //  e.g.  a column of single precision floats would have the
    //  following form: "1E"
    //
    //  a column with an array of 10 characters would have the form:
    //  "10A"
    //
    casacore::String         itsForm;
    
    //
    //  The unit of the data in the column
    //
    casacore::String         itsUnit;
    

    //
    //  If appropriate: which Columnnumber is this?
    //
    casacore::Int            itsColumnnumber;

    //
    //  For that matter, we keep the current datatype in this variable
    //  (this is the AIPS++ DataType enum). This enum also holds
    //  arraytypes etc. just what we need!
    //
    casacore::DataType       columnDataType;

    //
    //  How many data-type elements does this column take (form=
    //  <nrelements><datatype>)
    //
    casacore::Int            itsNumberElements;
    
    //
    //  We also need to have the total column width and the column
    //  offset in the row of the table (if known of course: this is
    //  only valid if the column is attached to a HeaderDataUnit)!
    //

    //
    //  This is the column width in bytes
    //
    casacore::uInt           itsTotalWidth;
    
    //
    //  The offset of this column in the row (also in bytes)
    //
    casacore::uLong          itsOffsetInRow;

    //
    //  The record description that describes this column in AIPS++
    //  terminology (like a combination of DataType and IPosition). It
    //  will be filled in by the 'CheckFormatAndInit()' private member
    //  function.
    //
    casacore::RecordDesc     itsColumnDescription;

    //
    //  The flag that tells us if we are a matrix column or not..
    //
    casacore::Bool           myBinaryColumnFlag;


    //
    //  The Column's private methods
    //

    
    //
    //  We do not allow copying and assigning to Columns, nor do we
    //  allow default Columns!
    //
    FITSColumn();
    FITSColumn& operator=( const FITSColumn& );
    
};

//
//  Display the columns properties on a stream....
//
std::ostream& operator<<( std::ostream& osref, const FITSColumn& columnref );

#endif
