//
//  Implemenmtation of the FITSDataSegment-class
//
//
//  Author:
//
//        Harro Verkouter,  27-04-1998
//
//
//     $Id: FITSDataSegment.cc,v 1.11.2.1 2018/06/08 11:44:05 verkout Exp $
//
#include <jive/ms2uvfitsimpl/FITSUtils/FITSDataSegment.h>
#include <jive/Utilities/stringutil.h> // for repr()
#include <casacore/casa/IO/ByteIO.h>
#include <casacore/casa/Exceptions.h>
#include <jive/ms2uvfitsimpl/FITSUtils/freefunctions.h>
#include <casacore/fits/FITS/fits.h>
#include <casacore/casa/Containers/Record.h>
#include <casacore/casa/Containers/RecordFieldId.h>

#include <cstdlib>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>

#include <cerrno>
#include <cstring>

using namespace std;
using namespace casacore;

//
//  Do the binary output into the stream!
//
ByteIO& operator<<( ByteIO& bioref, const FITSDataSegment& dsref )
{
    uLong    blocks2do( dsref.nrFitsBlocks() );
    
    for( uLong i=0; i<blocks2do; i++ )
    {
        bioref.write( 2880, dsref[i*2880] );
    }
    return bioref;
}




//
//  Create a DataSegment of 'size' bytes. Build some smartness in it!
//
FITSDataSegment::FITSDataSegment( uLong size, DataType basicfitstype ) :
    itsOrigin( FITSDataSegment::memory ),
    itsFitsSize( 0L ),
    itsOriginalSize( size ),
    itsOffset( 0L ),
    itsStreamptr( 0 ),
    itsDataptr( 0 ),
    itsTempfileDescriptor( -1 ),
    itsTempfileName( "" ),
    itsBasicFITSType( basicfitstype )
{
    //  Check the basic fitstype....
    if( !isBasicFitsType(itsBasicFITSType) )
        throw( AipsError("Invalid basic fits-type whilst trying to create a FITSDataSegment") );
    
    //  Transform itsSize into an integral number of FITS 2880-byte
    //  blocks...(if size != 0)
    itsFitsSize = (itsOriginalSize > 0) ? ((itsOriginalSize + 2880 - 1) / 2880) * 2880 : itsOriginalSize;
    
    //  If we end up here and itsDataptr is still NULL, we may assume
    //  that our 'intelligent' scheme has failed somewhere. Therfore
    //  we will just create a memory block of the requested size!
    //
    //  I assume that 'new' throws an exception when it cannot
    //  allocate the memory (it just closes the App. if I'm
    //  correct...)
    if( itsFitsSize ) {
        itsDataptr= ::malloc(itsFitsSize * sizeof(casacore::uChar)); //new uChar[ itsFitsSize ];
        if( itsDataptr==0 )
            throw AipsError( std::string("Failed to malloc ")+repr(itsFitsSize * sizeof(casacore::uChar))+" bytes of memory" );

        //  Correctly initialize the block!
        initDataSegment( itsDataptr, basicfitstype, itsFitsSize );
    }
}


//
//  Create a DataSegment of 'size' bytes out of a bytestream!
//
FITSDataSegment::FITSDataSegment( ByteIO* bytestream, uLong size, DataType basicfitstype ) :
    itsOrigin( FITSDataSegment::disk ),
    itsFitsSize( 0L ),
    itsOriginalSize( size ),
    itsOffset( 0L ),
    itsStreamptr( bytestream ),
    itsDataptr( 0 ),
    itsTempfileDescriptor( -1 ),
    itsTempfileName( "" ),
    itsBasicFITSType( basicfitstype )
{
    //  Incorrect basic fits type? 
    if( !isBasicFitsType(itsBasicFITSType) )
        throw( AipsError("Invalid basic fitstype whilst trying to create FITSDataSegment") );

    //  Well then: do some checking....
    if( !itsStreamptr )
        throw( AipsError("No streamptr whilst trying to attach to a stream?!") );
    
    //  Transform itsSize into an integral number of FITS 2880-byte
    //  blocks...
    itsFitsSize = (itsOriginalSize > 0) ? ((itsOriginalSize + 2880 - 1) / 2880) * 2880 : itsOriginalSize;

    //  Check if we're not crossing the 2**31-1 byte boundary
    if( itsFitsSize>=((1LL<<31)-1) )
        throw( AipsError("FITSDataSegment: Attempt to create segment with size >= 2**31-1 bytes. Unsupported by the OS") );
    
    //  If the stream is not seekable, we might just want to copy
    //  this to a tempfile on disk (which is seekable!)
    //  Note June 2018: This was basically always done, only based on
    //                  FITSSize being > 0
    if( itsFitsSize!=0 ) {
        //  The current filepointer in the original stream!
        itsOffset= itsStreamptr->seek( (long long int)0L, ByteIO::Current );
        
        //  Check if the segment does not extend beyond the size of
        //  the stream
        if( (itsOffset+itsFitsSize)>(uLong)itsStreamptr->length() )
            throw( AipsError("FITSDataSegment: segment extends beyond end-of-stream!") );

        //  Create a tempfile!
        itsDataptr= attachTempFile( itsTempfileName, itsTempfileDescriptor, itsFitsSize );

        //  Well then, if itsDataptr==0, something went
        //  wrong during creation. Now try to allocate memory in
        //  stead!
        if( itsDataptr==0 && (itsDataptr=::malloc(itsFitsSize * sizeof(casacore::uChar)))==0 )
            throw AipsError( std::string("Failed to malloc ")+repr(itsFitsSize * sizeof(casacore::uChar))+" bytes of memory" );

        //  Copy everything across....
        uLong         done( 0L );
        uLong         remain( itsFitsSize );
        uLong         maxcopy;

        while( remain ) {
            maxcopy=( (remain<2880)?(remain):(2880) );

            itsStreamptr->read( maxcopy, (void *)((uChar *)itsDataptr+done) );
            
            //  Update
            remain -= maxcopy;
            done   += maxcopy;
        }

        itsStreamptr->seek( (long long int)(itsOffset+itsFitsSize), ByteIO::Begin );
    }
}

//  Tell the user the size of this datasegment
uLong FITSDataSegment::getSegmentSize( void ) const {
    return itsFitsSize;
}


FITSDataSegment::source_t FITSDataSegment::dataSegmentOrigin( void ) const {
    return itsOrigin;
}


void FITSDataSegment::resize( casacore::uLong size ) {
    // request to resize ourselves ...
    if( itsOrigin!=FITSDataSegment::memory )
        throw AipsError("Cannot resize FITSDataSegment that's not originating from memory");

    // Must realloc the data if larger, otherwise just adjust the fits size
    casacore::uLong const newFitsSize = (size > 0) ? ((size + 2880 - 1) / 2880) * 2880 : size;

    if( newFitsSize>itsFitsSize ) {
        // need more data
        casacore::uLong const sz = newFitsSize * sizeof(casacore::uChar);

        if( itsDataptr==0 ) {
            if( (itsDataptr = ::malloc(sz)) == 0)
                throw AipsError( std::string("Failed to malloc ")+repr(sz)+" bytes of memory" );
            initDataSegment(itsDataptr, itsBasicFITSType, newFitsSize);
        } else {
            if( (itsDataptr = ::realloc(itsDataptr, sz)) == 0 )
                throw AipsError( std::string("Failed to realloc ")+repr(sz)+" bytes of memory" );
            // initialize the new bits of the segment
            initDataSegment(((casacore::uChar*)itsDataptr)+itsFitsSize, itsBasicFITSType, newFitsSize-itsFitsSize);
        }
    } 
    // adjust internal state
    itsOriginalSize = size;
    itsFitsSize     = newFitsSize;
}

//  Destruct the object
FITSDataSegment::~FITSDataSegment() {
    if( itsDataptr ) {
        // Unmap the datapointer!

        //  Try to figure out where the ptr came from (either use munmap or delete!)
        if( itsTempfileDescriptor>=0 ) {
            //  We may safely assume that the pointer was mapped to a
            //  tempfile! Now use munmap!
            ::munmap( (char*)itsDataptr, itsFitsSize );
        } else {
            //  ptr came from call to new[], use delete!
            //  nope, came from malloc now
            ::free( itsDataptr );
        }
    }
    
    // If a tempfile was associated, delete it!
    if( itsTempfileDescriptor>=0 ) {
        ::close( itsTempfileDescriptor );
        ::unlink( itsTempfileName.c_str() );
    }
}

/********************************************************************************

                           The protected section

 ********************************************************************************/
uLong  FITSDataSegment::nrFitsBlocks( void ) const {
    return itsFitsSize/(uLong)2880;
}

void* FITSDataSegment::operator[]( uLong offset ) {
    return (offset < itsFitsSize) ? (void*)(((casacore::uChar *)itsDataptr)+offset) : 0;
}


const void* FITSDataSegment::operator[]( uLong offset ) const {
    return (offset < itsFitsSize) ? (void*)(((casacore::uChar *)itsDataptr)+offset) : 0;
}



/********************************************************************************

                           The private section

 ********************************************************************************/


template <typename T>
void initFITSBlock(void* ptr, size_t sz) {
    T              fitsNul;
    T              localNul( 0 );
    size_t const   nItem( sz / sizeof(T) );
    unsigned char* dst( reinterpret_cast<unsigned char*>(ptr) );

    if( (nItem * sizeof(T))!=sz )
        throw AipsError( std::string("Initializing FITS block: item of size ")+repr(sizeof(T))+" does not fit integer amount of times in block of size "+repr(sz));
    // Convert one item of type T to FITS
    FITS::l2f((void*)&fitsNul, &localNul, 1);
    // Replicate it across the block
    for(size_t i=0; i<nItem; i++, dst+=sizeof(T))
        ::memcpy(dst, &fitsNul, sizeof(T));
}

// For Char it's more simple :-)
template <>
void initFITSBlock<casacore::Char>(void* ptr, size_t sz) {
    ::memset(ptr, 0x20, sz);
}

void FITSDataSegment::initDataSegment( void* ptr2block, DataType whichtype, uLong blocksize ) {
    //  If no datablock: nothing to do!
    if( !ptr2block || !blocksize )
        return;
    
    //  'Klad'block (prototype block filled with appropriately formatted
    //  zeroes of the indicated FITS type
    uChar         kladblok[2880];

    // Initialize the prototype block correctly based on type
    switch( whichtype ) {
        case TpUChar:
        case TpChar:
            ::initFITSBlock<casacore::Char>(kladblok, sizeof(kladblok));
            break;
            
        case TpShort:
            ::initFITSBlock<casacore::Short>(kladblok, sizeof(kladblok));
            break;

        case TpInt:
            ::initFITSBlock<casacore::Int>(kladblok, sizeof(kladblok));
            break;

        case TpFloat:
            ::initFITSBlock<casacore::Float>(kladblok, sizeof(kladblok));
            break;
        
        case TpDouble:
            ::initFITSBlock<casacore::Double>(kladblok, sizeof(kladblok));
            break;
            
        default:
            //  Nothing to do!
            break;
    }

    // All that's left is to replicate the kladblok
    unsigned char*  ptr = reinterpret_cast<unsigned char*>(ptr2block);
    // Note: do not care if user requests non-integer amount of fits blocks!
    for(size_t i=0; i<(blocksize/sizeof(kladblok)); i++, ptr+=sizeof(kladblok))
        ::memcpy(ptr, kladblok, sizeof(kladblok));
    return;
}


//
//  Attempt to create a tempfile and map the file to memory!
//
void* FITSDataSegment::attachTempFile( String& allocatedname, Int& allocatedfd, uLong size )
{
    //
    //  Attempt to make a tempfile and map it's contents to a pointer
    //
    char    templatename[]="datasegment.XXXXXX";
    void*   retval( 0 );
    
    //
    //  Step 0) Initialize the return values to unsuccesful indicators
    //
    allocatedname="";
    allocatedfd=-1;
    
    
    //
    //  Step 1) attempt to make a tempfilename 
    //
    allocatedfd = ::mkstemp(templatename);
   
    //
    //  Check if mktemp succeeded in creating a filename
    //
    if( allocatedfd==-1 )
    {
        cout << "Failed to make+open tempfile - " 
             << ::strerror(errno)
             << endl;
        return retval;
    }
    allocatedname = String(templatename);
    //
    //  Set the filepointer to the filesize and write a char, so the
    //  OS thinks the file is at least the amount of bytes we want to
    //  map!
    //
    Char    t( ' ' );
    
    ::lseek( allocatedfd, size, SEEK_SET );
    ::write( allocatedfd, &t, 1 );

    //
    //  Succesfully created the file!  Now attach the file
    //  to the program as 'virtual' memory!
    //
    retval = ::mmap( 0, size,
                     PROT_READ | PROT_WRITE, MAP_SHARED,
                     allocatedfd, 0 );

    if( !retval )
    {
        cout << "Failed to mmap - "
             << ::strerror(errno)
             << endl;
        
        //
        //  Close and clean up!
        //
        ::close( allocatedfd );
        allocatedfd=-1;

        ::unlink( allocatedname.chars() );
        allocatedname="";
    }
    return retval;
}


/********************************************************************************/
//
//
//               Some global functions here
//
//
/********************************************************************************/
//
//  Fill a recordfield from the datasegment (if possible!)
//
Bool fillRecordField( RecordInterface& recordref,
                      const RecordFieldId& fieldidx,
                      const FITSDataSegment& datasegmentref,
                      uLong offset )
{
    //
    //  The return value of the function
    //
    Bool         retval( False );
    DataType     fieldtype( recordref.dataType(fieldidx) );
    IPosition    fieldshape( recordref.shape(fieldidx) );

    //
    //  First: attempt to get the dataptr
    //
    const void*  dataptr( datasegmentref[offset] );
    
    if( !dataptr )
    {
        return retval;
    }
    
    //
    //  Ok. pointer is validated (though only the starting point is
    //  validated, the end point we might want to check later on, if
    //  we know how many bytes we really need...)
    //

    //
    //  Depending on the type of the recordfield, we do something
    //
    switch( fieldtype )
    {
        case TpChar:
        case TpUChar:
        case TpArrayChar:
        case TpArrayUChar:
            {
                if( !isArray(fieldtype) )
                {
                    const uChar*  ptr2uchar( (const uChar*)dataptr );

                    recordref.define( fieldidx, *ptr2uchar );
                    retval=True;
                }
                else
                {
                    //
                    //  We need to construct an array!
                    //
                    uLong         bytesneeded;
                    Array<uChar>  tmp( fieldshape );
                    
                    //
                    //  Verify that the datasegment we're trying to
                    //  read from is large enough....
                    //
                    bytesneeded=tmp.nelements();
                    
                    if( (offset+bytesneeded)<datasegmentref.getSegmentSize() )
                    {
                        uChar*   buffer( new uChar[tmp.nelements()] );
                        
                        //
                        //  FITS -> LOCAL conversion
                        //
                        FITS::f2l( buffer, (void *)dataptr, tmp.nelements() );

                        //
                        //  Now that's verified, we can safely take
                        //  the storage!
                        //
                        tmp.takeStorage( fieldshape, buffer, SHARE );
                        recordref.define( fieldidx, tmp );

                        //
                        //  Delete the buffer
                        //
                        delete [] buffer;

                        //
                        // And indicate success.
                        //
                        retval=True;
                    }
                }
            }
            break;
            
        case TpShort:
        case TpArrayShort:
            {
                if( !isArray(fieldtype) )
                {
                    Short  tmp;
                    
                    //
                    //  Do FITS -> LOCAL conversion
                    //
                    FITS::f2l( &tmp, (void *)dataptr, 1 );
                    
                    //
                    //  And stuff the value into the recordfield
                    //
                    recordref.define( fieldidx, tmp );
                    
                    retval=True;
                }
                else
                {
                    //
                    //  Array of short!
                    //
                    uLong        bytesneeded;
                    Array<Short> tmp( fieldshape );
                    
                    bytesneeded=tmp.nelements()*sizeof(Short);
                    
                    if( (offset+bytesneeded)<datasegmentref.getSegmentSize() )
                    {
                        Short*  buffer( new Short[tmp.nelements()] );
                        
                        //
                        //  FITS -> LOCAL
                        //
                        FITS::f2l( buffer, (void *)dataptr, tmp.nelements() );
                        
                        //
                        //  Let the array have the storage (for now)
                        //
                        tmp.takeStorage( fieldshape, buffer, SHARE );
                        
                        recordref.define( fieldidx, tmp );
                        
                        //
                        //  Buffer can go now!
                        //
                        delete [] buffer;

                        retval=True;
                    }
                }
                
            }
            break;

        case TpInt:
        case TpArrayInt:
            {
                if( !isArray(fieldtype) )
                {
                    Int  tmp;
                    
                    //
                    //  Do FITS -> LOCAL conversion
                    //
                    FITS::f2l( &tmp, (void *)dataptr, 1 );
                    
                    //
                    //  And stuff the value into the recordfield
                    //
                    recordref.define( fieldidx, tmp );
                    
                    retval=True;
                }
                else
                {
                    //
                    //  Array of ints!
                    //
                    uLong      bytesneeded;
                    Array<Int> tmp( fieldshape );
                    
                    bytesneeded=tmp.nelements()*sizeof(Int);
                    
                    if( (offset+bytesneeded)<datasegmentref.getSegmentSize() )
                    {
                        Int*  buffer( new Int[tmp.nelements()] );
                        
                        //
                        //  FITS -> LOCAL
                        //
                        FITS::f2l( buffer, (void *)dataptr, tmp.nelements() );
                        
                        //
                        //  Let the array have the storage (for now)
                        //
                        tmp.takeStorage( fieldshape, buffer, SHARE );
                        
                        recordref.define( fieldidx, tmp );
                        
                        //
                        //  Buffer can go now!
                        //
                        delete [] buffer;

                        retval=True;
                    }
                }
            }
            break;

        case TpFloat:
        case TpArrayFloat:
            {
                if( !isArray(fieldtype) )
                {
                    Float  tmp;
                    
                    //
                    //  Do FITS -> LOCAL conversion
                    //
                    FITS::f2l( &tmp, (void *)dataptr, 1 );
                    
                    //
                    //  And stuff the value into the recordfield
                    //
                    recordref.define( fieldidx, tmp );
                    
                    retval=True;
                }
                else
                {
                    //
                    //  Array of floats!
                    //
                    uLong        bytesneeded;
                    Array<Float> tmp( fieldshape );
                    
                    bytesneeded=tmp.nelements()*sizeof(Float);
                    
                    if( (offset+bytesneeded)<datasegmentref.getSegmentSize() )
                    {
                        Float*  buffer( new Float[tmp.nelements()] );
                        
                        //
                        //  FITS -> LOCAL
                        //
                        FITS::f2l( buffer, (void *)dataptr, tmp.nelements() );
                        
                        //
                        //  Let the array have the storage (for now)
                        //
                        tmp.takeStorage( fieldshape, buffer, SHARE );
                        
                        recordref.define( fieldidx, tmp );
                        
                        //
                        //  Buffer can go now!
                        //
                        delete [] buffer;
                        
                        retval=True;
                    }
                }
            }
            break;

        case TpDouble:
        case TpArrayDouble:
            {
                if( !isArray(fieldtype) )
                {
                    Double  tmp;
                    
                    //
                    //  Do FITS -> LOCAL conversion
                    //
                    FITS::f2l( &tmp, (void *)dataptr, 1 );
                    
                    //
                    //  And stuff the value into the recordfield
                    //
                    recordref.define( fieldidx, tmp );
                    
                    retval=True;
                }
                else
                {
                    //
                    //  Array of doubles!
                    //
                    uLong         bytesneeded;
                    Array<Double> tmp( fieldshape );
                    
                    bytesneeded=tmp.nelements()*sizeof(Double);
                    
                    if( (offset+bytesneeded)<datasegmentref.getSegmentSize() )
                    {
                        Double*  buffer( new Double[tmp.nelements()] );
                        
                        //
                        //  FITS -> LOCAL
                        //
                        FITS::f2l( buffer, (void *)dataptr, tmp.nelements() );
                        
                        //
                        //  Let the array have the storage (for now)
                        //
                        tmp.takeStorage( fieldshape, buffer, SHARE );
                        
                        recordref.define( fieldidx, tmp );
                        
                        //
                        //  Buffer can go now!
                        //
                        delete [] buffer;

                        retval=True;
                    }
                }
            }
            break;

        case TpComplex:
        case TpArrayComplex:
            {
                if( !isArray(fieldtype) )
                {
                    Complex  tmp;
                    
                    //
                    //  Do FITS -> LOCAL conversion
                    //
                    FITS::f2l( &tmp, (void *)dataptr, 1 );
                    
                    //
                    //  And stuff the value into the recordfield
                    //
                    recordref.define( fieldidx, tmp );
                    
                    retval=True;
                }
                else
                {
                    //
                    //  Array of complex!
                    //
                    uLong          bytesneeded;
                    Array<Complex> tmp( fieldshape );
                    
                    bytesneeded=tmp.nelements()*sizeof(Complex);
                    
                    if( (offset+bytesneeded)<datasegmentref.getSegmentSize() )
                    {
                        Complex*  buffer( new Complex[tmp.nelements()] );
                        
                        //
                        //  FITS -> LOCAL
                        //
                        FITS::f2l( buffer, (void *)dataptr, tmp.nelements() );
                        
                        //
                        //  Let the array have the storage (for now)
                        //
                        tmp.takeStorage( fieldshape, buffer, SHARE );
                        
                        recordref.define( fieldidx, tmp );
                        
                        //
                        //  Buffer can go now!
                        //
                        delete [] buffer;

                        retval=True;
                    }
                }
            }
            break;

        case TpDComplex:
        case TpArrayDComplex:
            {
                if( !isArray(fieldtype) )
                {
                    DComplex  tmp;
                    
                    //
                    //  Do FITS -> LOCAL conversion
                    //
                    FITS::f2l( &tmp, (void *)dataptr, 1 );
                    
                    //
                    //  And stuff the value into the recordfield
                    //
                    recordref.define( fieldidx, tmp );
                    
                    retval=True;
                }
                else
                {
                    //
                    //  Array of dcomplex!
                    //
                    uLong           bytesneeded;
                    Array<DComplex> tmp( fieldshape );
                    
                    bytesneeded=tmp.nelements()*sizeof(DComplex);
                    
                    if( (offset+bytesneeded)<datasegmentref.getSegmentSize() )
                    {
                        DComplex*  buffer( new DComplex[tmp.nelements()] );
                        
                        //
                        //  FITS -> LOCAL
                        //
                        FITS::f2l( buffer, (void *)dataptr, tmp.nelements() );
                        
                        //
                        //  Let the array have the storage (for now)
                        //
                        tmp.takeStorage( fieldshape, buffer, SHARE );
                        
                        recordref.define( fieldidx, tmp );
                        
                        //
                        //  Buffer can go now!
                        //
                        delete [] buffer;

                        retval=True;
                    }
                }
            }
            break;


        default:
            //
            //  Unhandled....
            //
            break;
    }
    return retval;
}

//
//  Flush the contents of the indicated recordfield to the datasegment
//  (if possible)
//
Bool flushRecordField( const RecordInterface& recordref,
                       const RecordFieldId& fieldidx,
                       FITSDataSegment& datasegmentref,
                       uLong offset )
{
    //
    //  The return value for this function
    //
    Bool       retval( False );

    //
    //  Elementary knowledge about the field we're adressing
    //
    DataType   fieldtype( recordref.dataType(fieldidx) );
    IPosition  fieldshape( recordref.shape(fieldidx) );

    //
    //  First: attempt to get the dataptr
    //
    void*      dataptr( datasegmentref[offset] );
    
    if( !dataptr )
    {
        return retval;
    }
    
    //
    // Ok. Starting pont validated. End point validation will be done
    // later
    //

    //
    //  Take action depending on the type of the field...
    //
    switch( fieldtype )
    {
        case TpChar:
        case TpUChar:
        case TpArrayChar:
        case TpArrayUChar:
            {
                if( !isArray(fieldtype) )
                {
                    *((uChar*)dataptr)=recordref.asuChar( fieldidx );
                    retval=True;
                }
                else
                {
                    //
                    //  It's an array. We havr to do some more work
                    //  now....
                    //
                    Bool                deleteit;
                    uLong               bytesneeded;
                    const uChar*        ptr2storage;
                    const Array<uChar>& tmp( recordref.asArrayuChar(fieldidx) );
                    
                    //
                    //  Check endpoint of the array....
                    //
                    bytesneeded=tmp.nelements();
                    
                    if( (offset+bytesneeded)<datasegmentref.getSegmentSize() )
                    {
                        //
                        //  Ok. We have enough space to write all the
                        //  databytes...
                        //
                        if( (ptr2storage=tmp.getStorage(deleteit))!=0 )
                        {
                            //
                            //  Do local -> fits conversion
                            //
                            FITS::l2f( dataptr, (uChar*)ptr2storage, tmp.nelements() );
                            retval=True;
                        }
                        
                        //
                        //  Release the storage
                        //
                        tmp.freeStorage( ptr2storage, deleteit );
                    }
                }
            }
            break;

        case TpShort:
        case TpArrayShort:
            {
                if( !isArray(fieldtype) )
                {
                    Short tmp( recordref.asShort(fieldidx) );
                    
                    //
                    //  Do local -> fits conversion
                    //
                    FITS::l2f( dataptr, &tmp, 1 );

                    retval=True;
                }
                else
                {
                    //
                    //  It's an array. We havr to do some more work
                    //  now....
                    //
                    Bool                deleteit;
                    uLong               bytesneeded;
                    const Short*        ptr2storage;
                    const Array<Short>& tmp( recordref.asArrayShort(fieldidx) );
                    
                    //
                    //  Check endpoint of the array....
                    //
                    bytesneeded=(uLong)tmp.nelements()*(uLong)sizeof(Short);
                    
                    if( (offset+bytesneeded)<datasegmentref.getSegmentSize() )
                    {
                        //
                        //  Ok. We have enough space to write all the
                        //  databytes...
                        //
                        if( (ptr2storage=tmp.getStorage(deleteit))!=0 )
                        {
                            //
                            //  Do local -> fits conversion
                            //
                            FITS::l2f( dataptr, (Short*)ptr2storage, tmp.nelements() );
                            retval=True;
                        }
                        
                        //
                        //  Release the storage
                        //
                        tmp.freeStorage( ptr2storage, deleteit );
                    }
                }
            }
            break;

        case TpInt:
        case TpArrayInt:
            {
                if( !isArray(fieldtype) )
                {
                    Int tmp( recordref.asInt(fieldidx) );
                    
                    //
                    //  Do local -> fits conversion
                    //
                    FITS::l2f( dataptr, &tmp, 1 );

                    retval=True;
                }
                else
                {
                    //
                    //  It's an array. We havr to do some more work
                    //  now....
                    //
                    Bool              deleteit;
                    uLong             bytesneeded;
                    const Int*        ptr2storage;
                    const Array<Int>& tmp( recordref.asArrayInt(fieldidx) );
                    
                    //
                    //  Check endpoint of the array....
                    //
                    bytesneeded=(uLong)tmp.nelements()*(uLong)sizeof(Int);
                    
                    if( (offset+bytesneeded)<datasegmentref.getSegmentSize() )
                    {
                        //
                        //  Ok. We have enough space to write all the
                        //  databytes...
                        //
                        if( (ptr2storage=tmp.getStorage(deleteit))!=0 )
                        {
                            //
                            //  Do local -> fits conversion
                            //
                            FITS::l2f( dataptr, (Int*)ptr2storage, tmp.nelements() );
                            retval=True;
                        }
                        
                        //
                        //  Release the storage
                        //
                        tmp.freeStorage( ptr2storage, deleteit );
                    }
                }
            }
            break;
            
        case TpFloat:
        case TpArrayFloat:
            {
                if( !isArray(fieldtype) )
                {
                    Float tmp( recordref.asFloat(fieldidx) );
                    
                    //
                    //  Do local -> fits conversion
                    //
                    FITS::l2f( dataptr, &tmp, 1 );

                    retval=True;
                }
                else
                {
                    //
                    //  It's an array. We havr to do some more work
                    //  now....
                    //
                    Bool                deleteit;
                    uLong               bytesneeded;
                    const Float*        ptr2storage;
                    const Array<Float>& tmp( recordref.asArrayFloat(fieldidx) );
                    
                    //
                    //  Check endpoint of the array....
                    //
                    bytesneeded=(uLong)tmp.nelements()*(uLong)sizeof(Float);
                    
                    if( (offset+bytesneeded)<datasegmentref.getSegmentSize() )
                    {
                        //
                        //  Ok. We have enough space to write all the
                        //  databytes...
                        //
                        if( (ptr2storage=tmp.getStorage(deleteit))!=0 )
                        {
                            //
                            //  Do local -> fits conversion
                            //
                            FITS::l2f( dataptr, (Float*)ptr2storage, tmp.nelements() );
                            retval=True;
                        }
                        
                        //
                        //  Release the storage
                        //
                        tmp.freeStorage( ptr2storage, deleteit );
                    }
                }
            }
            break;

        case TpDouble:
        case TpArrayDouble:
            {
                if( !isArray(fieldtype) )
                {
                    Double tmp( recordref.asDouble(fieldidx) );
                    
                    //
                    //  Do local -> fits conversion
                    //
                    FITS::l2f( dataptr, &tmp, 1 );

                    retval=True;
                }
                else
                {
                    //
                    //  It's an array. We havr to do some more work
                    //  now....
                    //
                    Bool                 deleteit;
                    uLong                bytesneeded;
                    const Double*        ptr2storage;
                    const Array<Double>& tmp( recordref.asArrayDouble(fieldidx) );
                    
                    //
                    //  Check endpoint of the array....
                    //
                    bytesneeded=(uLong)tmp.nelements()*(uLong)sizeof(Double);
                    
                    if( (offset+bytesneeded)<datasegmentref.getSegmentSize() )
                    {
                        //
                        //  Ok. We have enough space to write all the
                        //  databytes...
                        //
                        if( (ptr2storage=tmp.getStorage(deleteit))!=0 )
                        {
                            //
                            //  Do local -> fits conversion
                            //
                            FITS::l2f( dataptr, (Double*)ptr2storage, tmp.nelements() );
                            retval=True;
                        }
                        
                        //
                        //  Release the storage
                        //
                        tmp.freeStorage( ptr2storage, deleteit );
                    }
                }
            }
            break;

        case TpComplex:
        case TpArrayComplex:
            {
                if( !isArray(fieldtype) )
                {
                    Complex tmp( recordref.asComplex(fieldidx) );
                    
                    //
                    //  Do local -> fits conversion
                    //
                    FITS::l2f( dataptr, &tmp, 1 );

                    retval=True;
                }
                else
                {
                    //
                    //  It's an array. We havr to do some more work
                    //  now....
                    //
                    Bool                  deleteit;
                    uLong                 bytesneeded;
                    const Complex*        ptr2storage;
                    const Array<Complex>& tmp( recordref.asArrayComplex(fieldidx) );
                    
                    //
                    //  Check endpoint of the array....
                    //
                    bytesneeded=(uLong)tmp.nelements()*(uLong)sizeof(Complex);
                    
                    if( (offset+bytesneeded)<datasegmentref.getSegmentSize() )
                    {
                        //
                        //  Ok. We have enough space to write all the
                        //  databytes...
                        //
                        if( (ptr2storage=tmp.getStorage(deleteit))!=0 )
                        {
                            //
                            //  Do local -> fits conversion
                            //
                            FITS::l2f( dataptr, (Complex*)ptr2storage, tmp.nelements() );
                            retval=True;
                        }
                        
                        //
                        //  Release the storage
                        //
                        tmp.freeStorage( ptr2storage, deleteit );
                    }
                }
            }
            break;

        case TpDComplex:
        case TpArrayDComplex:
            {
                if( !isArray(fieldtype) )
                {
                    DComplex tmp( recordref.asDComplex(fieldidx) );
                    
                    //
                    //  Do local -> fits conversion
                    //
                    FITS::l2f( dataptr, &tmp, 1 );

                    retval=True;
                }
                else
                {
                    //
                    //  It's an array. We havr to do some more work
                    //  now....
                    //
                    Bool                   deleteit;
                    uLong                  bytesneeded;
                    const DComplex*        ptr2storage;
                    const Array<DComplex>& tmp( recordref.asArrayDComplex(fieldidx) );
                    
                    //
                    //  Check endpoint of the array....
                    //
                    bytesneeded=(uLong)tmp.nelements()*(uLong)sizeof(DComplex);
                    
                    if( (offset+bytesneeded)<datasegmentref.getSegmentSize() )
                    {
                        //
                        //  Ok. We have enough space to write all the
                        //  databytes...
                        //
                        if( (ptr2storage=tmp.getStorage(deleteit))!=0 )
                        {
                            //
                            //  Do local -> fits conversion
                            //
                            FITS::l2f( dataptr, (DComplex*)ptr2storage, tmp.nelements() );
                            retval=True;
                        }
                        
                        //
                        //  Release the storage
                        //
                        tmp.freeStorage( ptr2storage, deleteit );
                    }
                }
            }
            break;

        default:
            //
            //  Unhandled datatype
            //
            break;
    }

    return retval;
}








