//
//  Class:
//
//         FITSDataSegment
//
//  Purpose:
//
//         The FITSDataSegment class can be created in some ways. the
//         datasegment object knows where it was created and
//         therefore, read and write actions from the HDU objects can
//         be coordinated to make sure the get/put actions affect the
//         correct pieces of memory (or file). Whenever possible, the
//         FITSdatasegment will try to map a piece of a file to a
//         pointer to data, so access speeds up!
//
//
//
//  Author:
//
//         Harro Verkouter,    27-04-1998
//
//
//        $Id: FITSDataSegment.h,v 1.3.2.1 2018/06/08 11:44:05 verkout Exp $
//
#ifndef FITSDATASEGMENT_H_INC
#define FITSDATASEGMENT_H_INC

#include <casacore/casa/aips.h>
#include <casacore/casa/BasicSL/String.h>
#include <casacore/casa/Utilities/DataType.h>
#include <casacore/casa/IO/ByteIO.h>
#include <casacore/casa/Containers/RecordFieldId.h>
#include <casacore/casa/Containers/RecordInterface.h>

class HDUObject;

class FITSDataSegment
{
    //
    //  The output-function is a friend of this object....
    //
    friend casacore::ByteIO& operator<<( casacore::ByteIO& bioref,
									 const FITSDataSegment& dsref );

    friend HDUObject;

public:
    //
    //  An enum, telling where the object originated
    //
    enum source_t 
    {
        none, disk, memory
    };
    

    //
    //  Create the object
    //

    //
    //  This will create a datasegment of 'size' bytes long; a memory
    //  object. Pass the basic fitstype (either: TpChar, TpShort,
    //  TpInt, TpFloat or TpDouble), this parameter is used to pad the
    //  block with the zero representation of the correct type.
    //  
    //  NOTE: it is sufficient to pass the requested size, the object
    //  (since it's a FITSDataSegment, will calculate how many
    //  2880-byte block are necessary to grant the request.
    //
    //
    FITSDataSegment( casacore::uLong size, casacore::DataType basicfitstype );
    
    //
    //  Create the FITSDataSegment from a section in a ByteIO stream,
    //  starting at the current filepointer (will be determined by the
    //  FITSDataSegment class) and being 'size' bytes long
    //
    //  NOTE: The pointer is kept, but not taken over!
    //
    FITSDataSegment( casacore::ByteIO* bytestream,
					 casacore::uLong size,
					 casacore::DataType basicfitstype );
    

    //
    //  Return the actual size of the datasegment
    //
    virtual casacore::uLong        getSegmentSize( void ) const;

    //
    //  Allow the user access to the data......
    //
    //  NOTE: don't delete this pointer our from under us!
    //
    //
    //  These functions check that the index is valid and within the
    //  datasegment!
    //
    //  NOTE: if the index is outside the datasegment, a NULL-pointer
    //  is returned!
    //
    virtual void*        operator[]( casacore::uLong offset );
    virtual const void*  operator[]( casacore::uLong offset ) const;    
    
    //
    // Where did we originate from?
    //
    virtual FITSDataSegment::source_t dataSegmentOrigin( void ) const;

    //
    //  Destruct it
    //
    virtual ~FITSDataSegment();

protected:
    //
    //  Protected methods: a FITSDataSegment consists of a number of
    //  blocks of data. For the read/write-functions it may be useful
    //  to get access to whole blocks of data rather than having to
    //  compute that themselves!
    //
    //  NOTE: these functions are made virtual, so we can introduce a
    //  dummy FITSDataSegment with minimal overhead!
    //
    virtual casacore::uLong        nrFitsBlocks( void ) const;

    void                           resize( casacore::uLong size );

    //
    //  Initialize the datasegment with the 0-representation of the
    //  data type, as indicated by the functionparameter
    //
    static void          initDataSegment( void* ptr2block,
			 							  casacore::DataType whichtype,
										  casacore::uLong blocksize );

private:
    //
    //  The data segment's private parts
    //


    //
    //  The origin of the datasegment .....
    //
    source_t       itsOrigin;
    
    //
    //  The size of the segment
    //
    casacore::uLong          itsFitsSize;
    casacore::uLong          itsOriginalSize;
    
    //
    //  If appropriate: the offset within the stream it was created
    //  from
    //
    casacore::uLong          itsOffset;
    
    //
    //  And again (if appropriate) the stream whence it'd came
    //
    casacore::ByteIO*        itsStreamptr;
    
    //
    //  The pointer to the actual array of bytes!!!
    //
    void*          itsDataptr;

    //
    //  Filedescriptor of the (eventual) tmpfile!
    //
    casacore::Int            itsTempfileDescriptor;
    casacore::String         itsTempfileName;
    casacore::DataType const itsBasicFITSType;

    //
    //  Private methods
    //

    //
    //  This static function can be used to generate a temp-file. The
    //  assigned filedescriptor and generated filename will be
    //  returned in the function parameters. The unsigned char* return
    //  value indicates whether or not the function was succesful: if
    //  the pointer is non-null: everything OK and memory succesfully
    //  mapped to return value. ptr=0 -> something went wrong!
    //
    static void*   attachTempFile( casacore::String& allocatedname,
			                       casacore::Int& allocatedfd, casacore::uLong size );
    

    //
    //  No defaults, copying or assignment!
    //
    FITSDataSegment();
    FITSDataSegment( const FITSDataSegment& );
    FITSDataSegment& operator=( const FITSDataSegment& );
};


//
//  Easy access functions, read or write the indicated recordfield
//  at the offset 'offset' in the datasegment. These functions
//  take care of the conversion LOCAL -> FITS and vice
//  versa. Also, they check for a valid offset.
//
//  NOTE: these functions expect that the type of the field is set
//  correctly. It will take actions accordingly.
//
//  NOTE: No exceptions are thrown if anything goes wrong; if we
//  would, then the return value Bool would be not necessary....
//
casacore::Bool fillRecordField( casacore::RecordInterface& recordref,
                            const casacore::RecordFieldId& fieldidx,
                            const FITSDataSegment& datasegmentref,
                            casacore::uLong offset );

casacore::Bool flushRecordField( const casacore::RecordInterface& recordref,
                             const casacore::RecordFieldId& fieldidx,
                             FITSDataSegment& datasegmentref,
                             casacore::uLong offset );

#endif
