//
//  Class:
//
//          FITSPrimaryGroup
//
//  Purpose:
//
//          This class describes a PrimaryGroup in a FITS file!
//
//
//  Author:
//
//          Harro Verkouter,   26-05-1998
//
//
//     $Id: FITSPrimaryGroup.h,v 1.4 2016/06/23 13:02:29 jive_cc Exp $
//
#ifndef FITSPRIMARYGROUP_H_INC
#define FITSPRIMARYGROUP_H_INC


//
//  We need this include since it is our baseclass!
//
#include <jive/ms2uvfitsimpl/FITSUtils/FITSArrayHDU.h>
#include <jive/ms2uvfitsimpl/FITSUtils/FITSParameter.h>

#include <casacore/casa/Arrays/Vector.h>
#include <casacore/casa/Containers/RecordInterface.h>


//
//  The primary group hdu class. It isA FITSArrayHDU, therefore we
//  derive it from that class
//
class FITSPrimaryGroup :
    public FITSArrayHDU
{
    //
    //  The get/put functions of the FITSParameter class are friends
    //  of ours!
    //
    friend casacore::Bool FITSParameter::put( casacore::uLong idx, const casacore::RecordInterface& riref,
                                    const casacore::RecordFieldId& fieldidx );
    friend casacore::Bool FITSParameter::get( casacore::uLong idx, casacore::RecordInterface& riref,
                                    const casacore::RecordFieldId& fieldidx ) const;

public:
    //
    //  The public interface
    //

    //
    //  Create a primary group of a given type (basic fits type: the
    //  type of data items),
    //
    FITSPrimaryGroup( casacore::DataType basicfitstype );

    //
    //  Create the primary array from a fitskeywordlist, that's just
    //  been read from an existing fitsfile represented by
    //  'ptr2stream'
    //
    FITSPrimaryGroup( casacore::FitsKeywordList& kwlist, casacore::ByteIO* ptr2stream );

    //
    //  Primary groups have the possibility of adding parameters! The
    //  return value is the parameter number that was assigned to the
    //  parameter. If the return value is <0, the parameter was NOT
    //  added.
    //
    virtual casacore::Int       addParameter( const FITSParameter& parameter );

    //
    //  Set the number of groups (if possible)
    //
    casacore::Bool              setNumberOfGroups( casacore::uLong nrofgroups );

    //
    //  The type of this hdu
    //
    virtual HDUObject::hdu_t getHDUtype( void ) const;

    //
    // Define get/put actions on the primary group!
    //
    // Note: these functions are NOT defined on the baseclass, and
    // they are virtual in THIS class -> derived classes from this
    // object may overload the get/put actions.
    //
    virtual casacore::Bool      put( casacore::uLong group, const casacore::RecordInterface& recordref );
    
    virtual casacore::Bool      get( casacore::uLong group, casacore::RecordInterface& recordref ) const;

    //
    //  We have to implement our own version of the
    //  'getCurrentDataVolumeInBytes()' method, since there are two
    //  points we have to take into account:
    //
    //  1) axis 0 has length 0 -> product of all axes=0 which is
    //  usually quite wrong...
    //
    //  2) We have a number of groups, preceded by a number of
    //  parameters -> we have to take the parameters in account too.
    //
    virtual casacore::uLong     getCurrentDataVolumeInBytes( void ) const;

    //
    //  Users might be interested in the following info
    //

    //
    //  The shape of a group in the random groups structure
    //
    casacore::IPosition         groupShape( void ) const;
    casacore::uLong             getGroupSizeInBytes( void ) const;
    casacore::uLong             getNumberOfGroups( void ) const;
    casacore::RecordDesc        getGroupDescription( void ) const;

    //
    //  Check if the given parametr exists in this HDU. The number of
    //  occurrences is returned.
    //
    casacore::uInt              parameterExists( const casacore::String& parametertype ) const;
    casacore::uInt              getNumberOfParameters( void ) const;
    
    //
    //  Access parameters (by name or by number). Note: if the
    //  adressed parameter cannot be found, an aipsError exception is
    //  thrown!
    //
    FITSParameter&    getParameter( const casacore::String& parametername, casacore::uInt occurrence=0 ) const;
    FITSParameter&    getParameter( casacore::uInt parameternumber ) const;
    

    //
    //  Destruct the primary array
    //
    virtual ~FITSPrimaryGroup();
    
protected:
    //
    //  Protected functions
    //
    virtual casacore::FitsKeywordList   getDataDescriptionKeywords( void ) const;
    
private:
    //
    //  The primary groups private parts
    //
    
    //
    //  The number of groups represented by this HDU
    //
    casacore::uLong                   itsNumberOfGroups;
    
    //
    //  Primary groups have parameters! Keep a vector of pointers to
    //  the parameters
    //
    casacore::Vector<FITSParameter*>  itsParameterptrs;
    
    //
    //  The private methods
    //

    //
    //  These are undefined and inaccessible
    //
    FITSPrimaryGroup();
    FITSPrimaryGroup( const FITSPrimaryGroup& );
    FITSPrimaryGroup& operator=( const FITSPrimaryGroup& );
};



#endif
