//
//  DataTransferDescriptor Look-Up Table, to speed up LOGICAL indexing
//  into Albert's datatransfer descriptor, which is rather PHYSICALLY
//  orientated.
//
//  We might speed up things for those who want to adress in a
//  baseline based way, because in the datatransferdescriptor
//  everything is indexed by interferometer (one baseline can consist
//  of more than one interferometer, at least ONE per correlation,
//  when timemultiplexing is done even more!)
//
//
//  The key-issue here is to map logical adresses to physical
//  interferometers as fast as possible. What do we know?
//
//  1) A logical adress is baseline based. We want to have all info
//  about a certain baseline concentrated. Per baseline, we might have
//  multiple correlations (if we're correlating more than one
//  polarization/physical station).
//
//  2) Additional labels are job/subarray/subband.
//
//  In the DTD we have:
//
//  1) It is interferometer based. Per interferometer we can find out
//  what baseline it is attached to and which correlationcode it is
//  (correlationcode= RR/RL..XX,YX,etc..), so the correlationcode is
//  also a pair of indices.
//
//  2) As far as we're concerned, the interferometer order is quite
//  random.
//
//
//  Put all this info together and you'll find out that trying to look
//  for a specific baseline and a specific correlationcode is a time
//  consuming job.
//
//   So what we have to do is to map a logicakl address to a
//   interferometer number as fast as possible. If we have the
//   interferometer number, we may find out all info we need!
//
//
//
//  Author:     Harro Verkouter
//              17-07-1998
//
//
//   $Id: DtdLUT.h,v 1.1.1.1 2000/03/20 15:05:34 verkout Exp $
//
#ifndef DTDLUT_H
#define DTDLUT_H

//
//  datatype - typedefs
//
#ifndef HVTYPES_H
#include <vlbi/DtdLUT/hvtypes.h>
#endif

//
//  Include this for the DATATRANSFERDESCR-typedefs
//
#ifndef ALBERTTYPEDEFS_H
#include <vlbi/DtdLUT/AlbertTypedefs.h>
#endif

#ifndef CONTAINER_H
#include <vlbi/tmplsrc/Container.h>
#endif

#ifndef JOBDESCR_H
#include <vlbi/DtdLUT/JobDescr.h>
#endif

#include <iostream.h>

//
//  Forward declarations
//
class VisLabel;

//
//  The lookup table base-class. A lookup table implements the
//  'lookup' function to translate the logical 'VisLabel' into a
//  0-based inetrferometer number, so as we can directly jump in the
//  INTFR_DATADESCR_ARRAY to get at the interferometers internals!
//
class LookupTable
{
public:
    LookupTable()
    {
    }
    
    //
    //  Derived classes need to implememt this in order to be a real
    //  lookup-table!
    //
    virtual uint16   lookup( const VisLabel& idx ) const = 0;

    virtual ~LookupTable()
    {
    }
private:
    //
    //  No private parts!
    //
};





//
//  The lookuptable class itself.
//
//  NOTE: The LUT class does NOT store any pointer or references to
//  the DTD is is supposed to reference. It only stores indices into
//  the DTD. It is the user's responsibility to make sure that the DTD
//  and LUT stay in sync.
//
//
//  When the LUT-object is attached to a DTD, it scans the DTD once,
//  finding all stations, unique baselines and correlations in order
//  to map the logical adress to an interferometer number as fast as
//  possible!
//
//
class DtdLUT :
    public LookupTable
{
    friend ostream& operator<<( ostream& os, const DtdLUT& dref );

public:
    //
    //  The public interface:
    //

    //
    //  Default lookup table, not attached to any DTD
    //
    DtdLUT();
    
    //
    //  Create from an existing DTD
    //
    DtdLUT( const dtd& dtdref );
    
    //
    // Attach this LUT to another DTD
    //
    virtual bool    attach( const dtd& dtdref );


    //
    //  Lookup the adressed interferometer
    //
    virtual uint16  lookup( const VisLabel& idx ) const;

    //
    // Return how many jobs...
    //
    uint16          getNrJobs( void ) const
    {
	return jobs.elements();
    }

    //
    //  Get acces to a specific job. Empty job is returned if invalid job is adressed
    //
    const JobDescr& getJobDescr( uint16 jobnr ) const
    {
        static JobDescr    emptyjob;
        
        if( jobnr>=jobs.elements() )
        {
            return emptyjob;
        }
        return jobs( jobnr );
    }

    //
    //  Destruct the LUT, delete all memory resources. Don't be afraid
    //  we ruin the dtd, becuase we don't know what dtd we're
    //  referencing!
    //
    virtual ~DtdLUT();
    
protected:
    //
    //  Let them have access to the job array!
    //
    Container<JobDescr>&       jobarray( void );
    const Container<JobDescr>& jobarray( void ) const;

private:
    //
    //  All the private parts!
    //
    Container<JobDescr>     jobs;

    //
    // For now, we might forbid copying/assignment?
    //
    DtdLUT( const DtdLUT& );
    DtdLUT& operator=( const DtdLUT& );
};


//
//  This class attaches to a DTD allright, but implements another
//  'attach' method; it does the filling in based on the XSIGOUTP and
//  YSIGOUTP members rather than on XANTNR/YANTNR/XPOLNR/YPOLNR
//
class DtdHack :
    public DtdLUT
{
public:
    DtdHack();
    DtdHack( const dtd& dtdref );

    virtual bool   attach( const dtd& dtdref );

    virtual uint16 lookup( const VisLabel& idx ) const;

    ~DtdHack();
private:
    //
    //  Prevent copying/assignment
    //
    DtdHack( const DtdHack& );
    DtdHack& operator=( const DtdHack& );
};

//
//  Derive from DtdLUT to open up the internals?
//
class ExposedDtd :
    public DtdLUT
{
public:
    //
    //  nothing..
    //
    ExposedDtd();
    ExposedDtd( const dtd& dtdref );

    Container<JobDescr>&        jobs( void );
    const Container<JobDescr>&  jobs( void ) const;
    
    virtual ~ExposedDtd();
private:
    //
    //  Nothing!
    //
    ExposedDtd( const ExposedDtd& );
    ExposedDtd& operator=( const ExposedDtd& );
};


#endif
