//
//  A dual index keeps two indices together (like a baseline or a
//  correlationindex).
//
//  This is a base-class for specific indices that are a composite of
//  two numbers. An added feature is that we can map it to a linear
//  index.
//
//  The idea behind it is the following:
//
//  Suppose index1 ramges from 0-(n-1) and index2 ranges from 0-(m-1),
//  then we have at maximum n*m elements that we need to be able to
//  discriminate. If we map the index (x,y) to one number, say l (for
//  linear), we can jump into an array of length (n*m) to position l
//  to find the desired element directly. I think this may speed up
//  indexing quite a bit.
//
//
//
#ifndef DUALINDEX_H
#define DUALINDEX_H

#ifndef HVTYPES_H
#include <vlbi/DtdLUT/hvtypes.h>
#endif

//
//  Forward declarations
//
class ostream;


class DualIndex
{

public:
    //
    // Create it
    //
    DualIndex( uint16 max1 ) : idxval1( 0 ), idxval2( 0 ), maxidx( max1 )
    {
    }

    //
    //  This is what it's all about: translate from dual to linear index
    //
    inline uint16 operator()( void ) const
    {
        return (idxval1*maxidx)+idxval2;
    }
    
    //
    //  Do the same when we want to cast this one to uint16 (a linear index!)
    //
    inline operator uint16() const
    {
        return (idxval1*maxidx)+idxval2;
    }        
    inline operator int16() const
    {
        return (int16)((idxval1*maxidx)+idxval2);
    }        
    

    //
    //  Accessors
    //
    inline uint16&       idx1( void )
    {
        return idxval1;
    }
    inline uint16&       idx2( void )
    {
        return idxval2;
    }
    inline const uint16& idx1( void ) const
    {
        return idxval1;
    }
    inline const uint16& idx2( void ) const
    {
        return idxval2;
    }
    
    //
    //  Destruct the index
    //
    virtual ~DualIndex()
    {
    }

protected:
    //
    //  I know, i know.... it is VERY dumb to have protected
    //  datamebers..... but in this case (since we're trying to get
    //  the most speed out of this thing, we want to grant derived
    //  classes direct access to the index values!
    //

    //
    //  The two values, making up the complete index
    //
    uint16    idxval1;
    uint16    idxval2;
    

    //
    //  Maximum # of idx
    //
    uint16    maxidx;
private:
    //
    //  Nothing:
    //
};

//
// Dump the index nicely formatted on a stream
//
ostream& operator<<( ostream& os, const DualIndex& diref );

#endif
