/*********************************************************************************

  Class:
           VisLabel

  Purpose:
           To faciliate/unify indexing/labelling visibilities.

           One can set jobindex, subarrayindex,
           framenr(integrationnr), polarization, interferometer,
           baseline...

 *********************************************************************************/

#ifndef VISLABEL_H
#define VISLABEL_H



#ifndef HVTYPES_H
#include <vlbi/DtdLUT/hvtypes.h>
#endif

#ifndef BASELINEIDX_H
#include <vlbi/DtdLUT/BaseLineIdx.h>
#endif

#ifndef CONTAINER_H
#include <vlbi/tmplsrc/Container.h>
#endif

//
//  Forward declarations
//
class ostream;


class VisLabel
{
    //
    //  Send formatted output to the stream
    //
    friend ostream& operator<<( ostream& os, const VisLabel& vlref );
    
public:

    //
    //  Default ctor -> zero index
    //
    VisLabel( uint16 jobnr=0, uint16 sanr=0, uint16 sbnr=0, uint16 crnr=0, uint32 frnr=0,
               uint16 ifnr=0, const BaseLineIdx& bl=BaseLineIdx() );

    //
    //  Copy/assignment
    //
    VisLabel( const VisLabel& other );
    VisLabel& operator=( const VisLabel& rval );
    
    
    //
    //  Jobnr
    //
    uint16&         jobnr( void )
    {
	return itsJobnr;
    }
    
    const uint16&   jobnr( void ) const
    {
	return itsJobnr;
    }
    
    
    //
    //  The subarray under consideration
    //
    uint16&         subarraynr( void )
    {
	return itsSubarraynr;
    }
    
    const uint16&   subarraynr( void ) const
    {
	return itsSubarraynr;
    }
    

    //
    //  Subband (IF in classic AIPS)
    //
    uint16&         subbandnr( void )
    {
	return itsSubbandnr;
    }
    
    const uint16&   subbandnr( void ) const
    {
	return itsSubbandnr;
    }
    

    //
    //  The correlation index, to indicate which correlation is
    //  adressed.
    //
    uint16&         correlationnr( void )
    {
	return itsCorrelationnr;
    }
    
    const uint16&   correlationnr( void ) const
    {
	return itsCorrelationnr;
    }
    
    //
    // The 'framenr', the integration nr.
    //
    uint32&         framenr( void )
    {
	return itsFramenr;
    }
    
    const uint32&   framenr( void ) const
    {
	return itsFramenr;
    }
    

    //
    //  The interferometer number
    //
    uint16&         ifnr( void )
    {
	return itsInterferometernr;
    }
    
    const uint16&   ifnr( void ) const
    {
	return itsInterferometernr;
    }
    
    //
    //  Sometimes functions work on a complete baseline rather than on
    //  interferometer index.
    //
    BaseLineIdx&       baseline( void )
    {
	return itsBaseLine;
    }
    
    const BaseLineIdx& baseline( void ) const
    {
	return itsBaseLine;
    }

    //
    //  The destructor
    //
    virtual ~VisLabel();    

private:
    //
    //  The index of this job in the array of jobs
    //
    uint16   itsJobnr;

    //
    //  The current subarraynr
    //
    uint16   itsSubarraynr;

    //
    //  Which subband?
    //
    uint16   itsSubbandnr;

    //
    //  Which correlation?
    //
    uint16   itsCorrelationnr;

    //
    //  The current integration (of possibly multiple integrations)
    //
    uint32   itsFramenr;

    //
    //  The interferometer index
    //
    uint16   itsInterferometernr;

    //
    //  What is the baseline?
    //
    BaseLineIdx   itsBaseLine;
};


//
//  Oops... I forgot to include in VisLabel the fact that a visibility
//  does not necessarily has one correlation.... Therefore we need to
//  implement an extended VisLabel that holds an array of up to four
//  correlationnumbers!
//
//  Also, for us at JIVE, it might be interesting to keep track of the
//  SU where the data originally came from..
//
class ExtVisLabel :
    public VisLabel
{
    
	friend ostream& operator<<( ostream& os, const ExtVisLabel& visref );

public:
    //
    //
    //
    ExtVisLabel();    
    ExtVisLabel( const ExtVisLabel& other );
    
    ExtVisLabel& operator=( const ExtVisLabel& other );
    
    //
    //  The returned container contains the ORDINAL numbers of the
    //  correlations present int this visibility. Use the
    //  CorrelationCode::ordinal2code() (static!) function to get a
    //  real correlation code, telling you which one it really was.
    //
    Container<uint16>&       getCorrelations( void );
    const Container<uint16>& getCorrelations( void ) const;

    //
    //  De-allocate resources...
    //
    virtual ~ExtVisLabel();

private:
    //
    // The Extended VisLabel's private parts!
    //

    //
    //  The coded polarization combinations present in this visibility
    //
    Container<uint16>    itsCorrelationnrs;
};

//
//  The JIVE-visibility labels features stuff like station units and BOCFs...
//
class JIVEVisLabel :
    public ExtVisLabel
{
public:
    //
    //  Default c'tor
    //
    JIVEVisLabel();
    
    //
    //  Copy...
    //
    JIVEVisLabel( const JIVEVisLabel& other );

    //
    //  Assignment
    //
    JIVEVisLabel& operator=( const JIVEVisLabel& other );
    
    //
    // Jive specific labels
    //

    //
    //  For use with raw data -> begin/end BOCF nrs.
    //  
    //  set to 0xFFFFFFFF is invalid/not known
    //
    uint32&           getStartBOCF( void );
    const uint32&     getStartBOCF( void ) const;

    uint32&           getEndBOCF( void );
    const uint32&     getEndBOCF( void ) const;

    //
    //  And SU's
    //
    uint16&           getXStationUnit( void );
    const uint16&     getXStationUnit( void ) const;

    uint16&           getYStationUnit( void );
    const uint16&     getYStationUnit( void ) const;

    //
    //  Delete allocated resources
    //
    virtual ~JIVEVisLabel();
    
private:
    //
    //  Jive label private parts...
    //
    //
    //  The start/end BOCF nr.
    //
    uint32               itsStartBOCF;
    uint32               itsEndBOCF;

    //
    //  And the SU-nrs...
    //
    uint16               itsXStationUnit;
    uint16               itsYStationUnit;
};


#endif
