//
//  Implementation of the BOCF->time class
//
//  Author: Harro Verkouter, 11-03-1999
//
//  $Id: BOCFToTime.cc,v 1.1.1.1 2000/03/20 15:14:35 verkout Exp $
//
#ifndef BOCFTOTIME_H
#include <vlbi/j2ms/BOCFToTime.h>
#endif

#ifndef AIPSEXCEPTIONOBJECT_H
#include <vlbi/Utilities/AIPSExceptionObject.h>
#endif

#ifndef AIPS_QUANTA_H
#include <aips/Quanta.h>
#endif


BOCFToTime::BOCFToTime() :
    itsBOCFZeroPoint( MEpoch( Quantity( 0.0, "s"), MEpoch::UTC ) ),
    itsBOCFRate( 1.0 )
{
}

BOCFToTime::BOCFToTime( const BOCFToTime& other ) :
    itsBOCFZeroPoint( other.itsBOCFZeroPoint ),
    itsBOCFRate( other.itsBOCFRate )
{
}



BOCFToTime::BOCFToTime( const MEpoch& approxstart, Double bocfrate ) :
    itsBOCFRate( bocfrate )
{
    //
    //  Check if we do have a valid bocf rate...
    //
    if( itsBOCFRate==0.0 )
    {
	throw( AIPSExceptionObject("BOCF rate of zero not supported...") );
    }
    

    //
    //  Ok. Find out, from the approx. start of scan what the
    //  zeropoint of bocf should be.... it should be the last instance
    //  of time that was a multiple of 10min (or 600s)
    //
    uint32   tenminuteints;
    Double   kut;
    Quantity tmp;
    
    kut = approxstart.get("s").getValue();
    
    tenminuteints = (uint32)( kut/600.0 );
    tmp = Quantity( (Double)tenminuteints*600.0, "s" );
    
    itsBOCFZeroPoint = MEpoch( tmp, approxstart.getRef() );
}

//
//  Assignment....
//
BOCFToTime& BOCFToTime::operator=( const BOCFToTime& other )
{
    if( this!=&other )
    {
	itsBOCFZeroPoint = other.itsBOCFZeroPoint;
	itsBOCFRate = other.itsBOCFRate;
    }
    return *this;
}


uint32 BOCFToTime::getMaxBOCFnr( void ) const
{
    uint32   retval;
    
    retval = (uint32)(600.0*itsBOCFRate) - 1;
    return retval;
}

Double BOCFToTime::getBOCFRate( void ) const
{
    return itsBOCFRate;
}

const MEpoch& BOCFToTime::operator()( Double anybocf ) const
{
    Quantity       secs;
    static MEpoch  retval;

    secs = Quantity( itsBOCFZeroPoint.get("s") );
    secs += Quantity( anybocf/itsBOCFRate, "s" );
    
    retval = MEpoch( secs, itsBOCFZeroPoint.getRef() );
    
    return retval;
}

void BOCFToTime::advanceZeroPoint( void )
{
    Double     tmp;
    Quantity   ntime;

    tmp = itsBOCFZeroPoint.get("s").getValue();
    tmp += 600.0;
    
    ntime = Quantity( tmp, "s" );
    itsBOCFZeroPoint = MEpoch( ntime, itsBOCFZeroPoint.getRef() );
    return;
}

void BOCFToTime::rewindZeroPoint( void )
{
    Double     tmp;
    Quantity   ntime;

    tmp = itsBOCFZeroPoint.get("s").getValue();
    tmp -= 600.0;
    
    ntime = Quantity( tmp, "s" );
    itsBOCFZeroPoint = MEpoch( ntime, itsBOCFZeroPoint.getRef() );
    return;
}

void BOCFToTime::adjust( Int nrintervals )
{
    //
    //  Do clock adjustments
    //
    while( nrintervals )
    {
	if( nrintervals<0 )
	{
	    cout << "** Setting BOCF zeropoint back" << endl;
	    
	    this->rewindZeroPoint();
	    nrintervals++;
	}
	else
	{
	    cout << "** Advancing BOCF zeropoint" << endl;
	    this->advanceZeroPoint();
	    nrintervals--;
	}
    }
    return;
}



BOCFToTime::~BOCFToTime()
{
    //
    //  Nothing to do yet....
    //
}
