//
//  BaseBand -> a class describing the structure of a baseband:
//
//  bandwidth, startfreq, nr. of channels, polarization combinations
//
//  Assume startfreq is always the LOWER frequency edge of the band
//
//  Author:  Harro Verkouter,    2-12-1998
//
//  $Id: BaseBand.h,v 1.1.1.1 2000/03/20 15:14:35 verkout Exp $
//
#ifndef BASEBAND_H
#define BASEBAND_H

#ifndef AIPS_AIPS_H
#include <aips/aips.h>
#endif

#ifndef CONTAINER_H
#include <vlbi/tmplsrc/Container.h>
#endif


//
//  Forward decalarations
//
class String;
class ostream;
class CorrelationCode;
class FrequencyConfig;


class BaseBand
{
    //
    //  FrequencyConfig is our friend!
    //
    friend class FrequencyConfig;
    
public:
    //
    //  BaseBands are upper or lower sideband or not initialised!
    //
    typedef enum _SideBand
    {
	none, upper, lower
    } SideBand;
    
    //
    //  Map string to SideBand
    //
    static BaseBand::SideBand  stringToSideBand( const String& sbtxt );

    //
    //  Empty baseband
    //
    BaseBand();

    //
    //  Copy c'tor
    //
    BaseBand( const BaseBand& other );
    
    //
    //  Create a fully defined baseband
    //
    //  * All freq's are in Hz
    //
    BaseBand( Double startfreq, Double bandwidth, BaseBand::SideBand sideband );

    //
    // Assignment
    //
    BaseBand& operator=( const BaseBand& other );

    Double                 getStartfreq( void ) const;
    Double                 getBandwidth( void ) const;
    BaseBand::SideBand     getSideBand( void ) const;

    //
    //  Test if this BaseBand has a parent configuration...
    //
    Bool                   hasConfiguration( void ) const;

    //
    //  Get the parent configuration.
    //
    //  NOTE: throws AipsError if no parentconfiguration!!!!!
    //
    const FrequencyConfig& getConfiguration( void ) const;

    //
    //  Compare with another BaseBand
    //
    //  Return true id frequency/bandwidth/sideband are the same for
    //  both sides of the equation...
    //
    bool    operator==( const BaseBand& other );

    //
    //  Delete the baseband
    //
    ~BaseBand();


protected:
    //
    //  Only FrequencyConfig's can create a REALLY fully defined
    //  BaseBand. The frequencyconfig will fit the baseband with a) a
    //  subbandnr (or ID) and b) a pointer to itself, so the baseband
    //  knows who's his parent.
    //
    //  All the frequency settings/polarizations are COPIED from
    //  bbdata!
    //
    BaseBand( const FrequencyConfig* ptr2parent, uInt basebandnr, const BaseBand& bbdata );
    

private:    
    //
    //  The private parts
    //

    //
    //  These settings describe the BaseBand
    //
    Double             itsStartfreq;
    Double             itsBandwidth;
    BaseBand::SideBand itsSideBand;
    
    //
    //  These internals can only be set by a FrequencyConfig (i.e. not
    //  configurable by the user)
    //
    uInt                   itsBaseBandnr;
    const FrequencyConfig* itsParentConfigurationptr;
};

//
//  Output on a stream
//
ostream& operator<<( ostream& os, const BaseBand& bbref );
ostream& operator<<( ostream& os, const BaseBand::SideBand& sbref );

#endif
