//
//  Implementation of COFBaseline* stuff
//
//  Author:  Harro Verkouter 01-06-1999
//
//  $Id: COFBaselineIterator.cc,v 1.1.1.1 2000/03/20 15:14:35 verkout Exp $
//
#ifndef COFBASELINEITERATOR_H
#include <vlbi/j2ms/COFBaselineIterator.h>
#endif

#ifndef IFITERATOR_H
#include <data_handler/IFIterator.h>
#endif

#ifndef DATADESCRIPTOR_H
#include <data_handler/DataDescriptor.h>
#endif

#ifndef BASELINEIDX_H
#include <vlbi/DtdLUT/BaseLineIdx.h>
#endif

#ifndef CORRELATIONCODE_H
#include <vlbi/DtdLUT/CorrelationCode.h>
#endif

#ifndef INTERFEROMETERDESCRIPTOR_H
#include <data_handler/InterferometerDescriptor.h>
#endif



//
//  local function to map a polarization into a code
//
//  Officially, we should do that by using a polarization map for the
//  experiment. For know we <<ASSUME>> the following:
//
//  if not known, 0xFFFF is returned....
//
uint16 polarizationToCode( InterferometerDescriptor::Polarization p )
{
    uint16    retval( 0xFFFF );
    
    switch( p )
    {
	case InterferometerDescriptor::x:
	case InterferometerDescriptor::L:
	    retval = 0;
	    break;
	    
	case InterferometerDescriptor::y:
	case InterferometerDescriptor::R:
	    retval = 1;
	    break;
	    
	default:
	    break;
    }
    return retval;
}


//
//  Show the contents of a COFBaseline struct in HRF on a stream!
//
ostream& operator<<( ostream& os, const _COFBaseline& cbref )
{
    os << "Baseline: " << cbref.myIndex << "  SUBBAND=" << cbref.mySubbandIndex << '\n';
    for( uint32 i = 0; i<4; i++ )
    {
	os << "Correlation " << CorrelationCode::ordinal2code( (uint16)i ) << " is";

	if( cbref.myPresentFlag[i] )
	{
	    os << " present.";
	}
	else
	{
	    os << " not present.";
	}
	os << '\n';
    }
    os << flush;
    return os;
}


//
//  Show a complete baseline set on a stream (in HRF)
//
ostream& operator<<( ostream& os, const COFBaselineSet& cbsetref )
{
    COFBaselineIterator   blptr( cbsetref.begin() );
    
    os << "\n\nShowing contents of COFBaselineSet...\n";
    
    while( blptr )
    {
	os << *blptr;
	blptr++;
    }
    
    return os;
}


//
//
//  The COF-Baseline struct
//
//
_COFBaseline::_COFBaseline( const BaseLineIdx& blidx, uint32 subbandindex ) :
    mySubbandIndex( subbandindex ),
    myIndex( blidx ),
    next( 0 )
{
    InterferometerDescriptor   defaultIF;

    for( uint32 i=0; i<4; i++ )
    {
	myPresentFlag[ i ] = False;
	myFileOffsetInBytes[ i ] = 0;
	myInterferometers[ i ] = defaultIF;
    }
}

uint32 _COFBaseline::getSubbandIndex( void ) const
{
    return mySubbandIndex;
}


const BaseLineIdx& _COFBaseline::getIndex( void ) const
{
    return myIndex;
}

bool* _COFBaseline::getConfigured( void ) const
{
    bool*    retval( new bool[4] );
    
    for( uint32 i=0; i<4; i++ )
    {
	retval[ i ] = myPresentFlag[ i ];
    }
    return retval;
}

    
Container<uint16> _COFBaseline::getCorrelations( void ) const
{
    Container<uint16>    retval;

    for( uint16 i=0; i<4; i++ )
    {
	if( myPresentFlag[i] )
	{
	    retval.add( i );
	}
    }
    return retval;
}


const InterferometerDescriptor& _COFBaseline::operator[]( uint32 idx ) const
{
    static InterferometerDescriptor    dummy;
    const InterferometerDescriptor*    retvalptr( &dummy );

    if( idx<4 )
    {
	retvalptr = &myInterferometers[ idx ];
    }
    
    return (*retvalptr);
}



Bool _COFBaseline::addInterferometer( const InterferometerDescriptor& ifdescr,
				      const CorrelationCode& ccref )
{
    Bool      retval( False );
    uint32    linearidx( (uint32)CorrelationCode::code2ordinal(ccref) );

    //cout << "COFBaseline - idx=" << myIndex << ", adding correlation " 
    // << ccref << " (ordinal nr=" << linearidx << ")" << endl;
    
    if( linearidx>3 )
    {
	return retval;
    }
    
    if( !myPresentFlag[linearidx] )
    {
	myPresentFlag[ linearidx ] = True;
	myInterferometers[ linearidx ] = ifdescr;

	retval = True;
    }
    return retval;
}



//
//
//  The COF-baseline-set class
//
//
COFBaselineSet::COFBaselineSet() :
    myFirstBaselineptr( 0 )
{
}

COFBaselineSet::COFBaselineSet( const DataDescriptor& ddref ) :
    myFirstBaselineptr( 0 )
{
    this->attachDataDescriptor( ddref );
}

COFBaselineSet& COFBaselineSet::operator=( const DataDescriptor& ddref )
{
    this->attachDataDescriptor( ddref );
    return *this;
}


//
//  Clear out the linked list of baselines!
//
COFBaselineSet::~COFBaselineSet()
{
    this->removeBaselineList();
}

//
//
//                     Private method section
//
//

//
//  Parse the DataDescriptor and form all the baselines
//
Bool COFBaselineSet::attachDataDescriptor( const DataDescriptor& ddref )
{
    Bool            retval( False );
    IFIterator      ifit( ddref.begin() );
    BaseLineIdx     blidx;
    COFBaseline*    cofbaselineptr;
    CorrelationCode cc;

    //
    //  Clear the baseline list...
    //
    this->removeBaselineList();

    //
    //  Process all interferometers in the DataDescriptor
    //
    while( ifit )
    {
	Bool                             swapit( False );
	Bool                             autocorrelation;
	Bool                             crosspolarization;
	const InterferometerDescriptor&  ifref( *ifit );
	

	blidx = BaseLineIdx( ifref.getXAntenna(), ifref.getYAntenna() );


	if( blidx.idx1()>blidx.idx2() )
	{
	    blidx = BaseLineIdx( blidx.idx2(), blidx.idx1() );
	    swapit = True;
	}
	
	cc = CorrelationCode( polarizationToCode(ifref.getXPolarization()),
			      polarizationToCode(ifref.getYPolarization()) );
	
	autocorrelation   = ToBool( blidx.idx1()==blidx.idx2() );
	crosspolarization = ToBool( cc.idx1()!=cc.idx2() );

	if( swapit )
	{
	    cc = CorrelationCode( cc.idx2(), cc.idx1() );
	}
	

	if( (cofbaselineptr=this->locateBaseline(blidx, ifref.getSubband()))==0 )
	{
	    //
	    //  Not found. Add it...
	    //
	    cofbaselineptr = new COFBaseline( blidx, ifref.getSubband() );
	    
	    this->addBaselineptr( cofbaselineptr );
	}
	
	//
	//  Add the correlation to the baseline
	//
	cofbaselineptr->addInterferometer( ifref, cc );
	
	//
	//  For autocorrelations/crosspolarizations, Albert only
	//  calculates one of the products, therefore, iin order to
	//  keep the shape constant, when we find one of the
	//  cross-polarizations on a autocorrelation, we automatically
	//  add the other polarization with the same interferometer
	//
	if( autocorrelation && crosspolarization )
	{
	    cofbaselineptr->addInterferometer( ifref, CorrelationCode(cc.idx2(), cc.idx1()) );
	}

	//
	//  *sigh* .. do NOT forget to increment
	//  iterator.... otherwise you may end up in an infinite
	//  loop...... (guess what happend before I found THAT out...)
	//
	ifit++;
    }
    return retval;
}

COFBaselineIterator COFBaselineSet::begin( void ) const
{
    return COFBaselineIterator( myFirstBaselineptr );
}



//
//  Delete all Baseline entries
//
void COFBaselineSet::removeBaselineList( void )
{
    COFBaseline*   baselptr;
    
    while( myFirstBaselineptr )
    {
	baselptr = myFirstBaselineptr->next;
	
	delete myFirstBaselineptr;
	
	myFirstBaselineptr = baselptr;
    }
    return;
}

COFBaseline* COFBaselineSet::locateBaseline( const BaseLineIdx& blidx, uint32 subbandidx )
{
    COFBaseline*   retval( myFirstBaselineptr );
    
    while( retval )
    {
	const COFBaseline& cbref( *retval );
	
	if( cbref.getIndex()==blidx && cbref.getSubbandIndex()==subbandidx )
	{
	    break;
	}
	retval = retval->next;
    }
    return retval;
}

void COFBaselineSet::addBaselineptr( COFBaseline* baselptr )
{
    COFBaseline*    lastptr( myFirstBaselineptr );
    
    while( lastptr )
    {
	if( !lastptr->next )
	{
	    break;
	}
	lastptr= lastptr->next;
    }
    
    if( !lastptr )
    {
	//
	//  No entries yet
	//
	myFirstBaselineptr = baselptr;
    }
    else
    {
	lastptr->next = baselptr;
    }
    return;
}


//
//
//  The baseline iterator
//
//
COFBaselineIterator::COFBaselineIterator() :
    myBaselineptr( 0 )
{
}


COFBaselineIterator::COFBaselineIterator( const COFBaselineIterator& other ) :
    myBaselineptr( other.myBaselineptr )
{
}

COFBaselineIterator& COFBaselineIterator::operator=( const COFBaselineIterator& other )
{
    if( this!=&other )
    {
	myBaselineptr = other.myBaselineptr;
    }
    return *this;
}

COFBaselineIterator::operator bool() const
{
    return ( myBaselineptr!=0 );
}

bool COFBaselineIterator::operator!() const
{
    return ( myBaselineptr==0 );
}


COFBaselineIterator& COFBaselineIterator::operator++( void )
{
    if( myBaselineptr )
    {
	myBaselineptr = myBaselineptr->next;
    }
    return *this;
}

COFBaselineIterator& COFBaselineIterator::operator++( int )
{
    if( myBaselineptr )
    {
	myBaselineptr = myBaselineptr->next;
    }
    return *this;
}

COFBaseline& COFBaselineIterator::operator*( void )
{
    return *myBaselineptr;
}

const COFBaseline& COFBaselineIterator::operator*( void ) const
{
    return *myBaselineptr;
}

COFBaseline* COFBaselineIterator::operator->( void )
{
    return myBaselineptr;
}

const COFBaseline* COFBaselineIterator::operator->( void ) const
{
    return myBaselineptr;
}

//
//  Destruct -> nothing to do yet
//
COFBaselineIterator::~COFBaselineIterator()
{
}


COFBaselineIterator::COFBaselineIterator( COFBaseline* ptr ) :
    myBaselineptr( ptr )
{
}
