//
//  COFBaselineIterator - in order to be able to iterate over the
//  baselines in a CorrelatedDataRecord (part of the COF)
//
//  Since the 'shape' or 'layout' of a correlated data record is
//  determined by the last DataDescriptor encountered in the
//  CorrelatedDataFile, we attach the COFBaselineIterator to a
//  DataDescriptor. The data for the baseline is gotten by indexing
//  into the CorrelatedDataRecord of you choice!
//
//
#ifndef COFBASELINEITERATOR_H
#define COFBASELINEITERATOR_H

//
//  Bool etc.
//
#ifndef AIPS_AIPS_H
#include <aips/aips.h>
#endif

#ifndef BASELINEIDX_H
#include <vlbi/DtdLUT/BaseLineIdx.h>
#endif

#ifndef INTERFEROMETERDESCRIPTOR_H
#include <data_handler/InterferometerDescriptor.h>
#endif

#ifndef CONTAINER_H
#include <vlbi/tmplsrc/Container.h>
#endif


//
//  Forward declarations
//
class ostream;
class DataDescriptor;
class COFBaselineSet;
class CorrelationCode;
class COFBaselineIterator;



typedef struct _COFBaseline
{
    friend class COFBaselineSet;
    friend class COFBaselineIterator;
    friend ostream& operator<<( ostream& os, const _COFBaseline& cbref );
    
public:
    //
    //  do proper initialization
    //
    _COFBaseline( const BaseLineIdx& blidx, uint32 subbandindex );

    //
    //  Attempt to add a correlation to the Baseline
    //
    Bool     addInterferometer( const InterferometerDescriptor& ifdescr,
				const CorrelationCode& ccref );

    //
    // Read-only access to private parts
    //
    uint32              getSubbandIndex( void ) const;
    const BaseLineIdx&  getIndex( void ) const;

    //
    //  Return an array of 4 booleans telling which interferometers
    //  were set via the 'addInterferometr' method...
    //
    //  NOTE:  user is responsible for deleting the allocated memory!!!
    //
    //  NOTE2: allocation is done via new[] (so use delete [])!
    //
    bool*               getConfigured( void ) const;

    //
    //  Return the *ordinal* numbers of the configured correlations
    //  (which are quite different from the cast to uint16's of the
    //  configured correlations) in this baseline!
    //
    Container<uint16>   getCorrelations( void ) const;
    

    //
    //  Return the indexed InterferometerDescriptor (note:
    //  boundschecking is performed; NOT if the indexed
    //  InterferometerDescriptor was filled in!)
    //
    const InterferometerDescriptor&  operator[]( uint32 idx ) const;
    
    
private:
    //
    //  Our private parts
    //
    Bool                     myPresentFlag[ 4 ];
    uint32                   mySubbandIndex;
    uint32                   myFileOffsetInBytes[ 4 ];
    BaseLineIdx              myIndex;
    InterferometerDescriptor myInterferometers[ 4 ];
    struct _COFBaseline*     next;

    //
    //  Prohibit these
    //
    _COFBaseline();
    _COFBaseline( const _COFBaseline& );
    _COFBaseline& operator=( const _COFBaseline& );

} COFBaseline;


//
//
//  A Baseline ietrator -> we can iterate over a BaselineSet
//
//
class COFBaselineIterator
{
    friend class COFBaselineSet;
    
public:
    //
    //  An empty COFBaselineIterator -> points at nothing in
    //  particular
    //
    COFBaselineIterator();
    
    //
    //  copy-c'tor
    //
    COFBaselineIterator( const COFBaselineIterator& other );
    
    //
    //  assignment
    //
    COFBaselineIterator& operator=( const COFBaselineIterator& other );
    
    //
    //  operator bool tells us whether or not we are pointing at
    //  something! Useful in if() and while() constructions...
    //
    operator bool() const;

    bool operator!() const;
    
    //
    // pre- and postfix advancement of the iterator!
    //
    COFBaselineIterator&  operator++( void );
    COFBaselineIterator&  operator++( int );

    //
    //  De-refrence the pointer.
    //
    //  NOTE: NO CHECK FOR NULL pointer IS PERFORMED!!!
    //
    COFBaseline&          operator*( void );
    const COFBaseline&    operator*( void ) const;
    
    //
    //  Access to the pointer. No checking is performed. Use at own
    //  risk.
    //
    COFBaseline*          operator->( void );
    const COFBaseline*    operator->( void ) const;

    //
    //  de-allocate resources
    //
    ~COFBaselineIterator();
    
protected:
    //
    //  Only the COFBaselineSet class can create valid iterators!
    //
    COFBaselineIterator( COFBaseline* ptr );

private:
    //
    //  The private parts of the iterator
    //
    COFBaseline*   myBaselineptr;
};



//
//  Parse all the entries in the dataDescriptor to form a baseline set
//  rather than an interferometerset!
//
class COFBaselineSet
{
public:
    //
    //  An empty baseline set
    //
    COFBaselineSet();

    //
    //  create from a datadescriptor
    //
    COFBaselineSet( const DataDescriptor& ddref );
    
    //
    //  re-init, make a new set from the give datadescriptor
    //
    COFBaselineSet& operator=( const DataDescriptor& ddref );
    
    //
    //  return an iterator for the first baseline in the set!
    //
    COFBaselineIterator   begin( void ) const;

    //
    // De-allocate resources
    // 
    ~COFBaselineSet();
    

private:
    //
    //  Our private parts
    //
    COFBaseline* myFirstBaselineptr;
    
    //
    //  Private methods
    //
    //  This method alters the state of this object such that it
    //  represents all the baselines present in the mentiones
    //  datadescriptor
    //
    Bool         attachDataDescriptor( const DataDescriptor& ddref );

    //
    //  clear the linked list of baselines
    //
    void         removeBaselineList( void );
    
    //
    //  Locate a certain baseline in the list
    //
    COFBaseline* locateBaseline( const BaseLineIdx& blidx, uint32 subbandidx );

    //
    //  Addthe pointer to our list. NOTE: the object takes over the
    //  responsibility so don't delete the pointer out from under us!
    //
    void         addBaselineptr( COFBaseline* baselptr );

    //
    //  Prohibit these
    //
    COFBaselineSet( const COFBaselineSet& );
    COFBaselineSet& operator=( const COFBaselineSet& );    
};

//
//  Dump the contents of the set on a stream (in HRF)
//
ostream& operator<<( ostream& os, const COFBaselineSet& cbsetref );


#endif
