//
//  COF VisibilityBuffer - Correlator OutputFormat visibility buffer
//
//  Read visibilities from a file set known as the JIVE Correlator
//  Output Format.
//
//  This is a specialization of the VisibilityBuffer class
//
//
//  Author:   Harro Verkouter, 31-05-1999
//
//
//  $Id: COFVisibilityBuffer.h,v 1.2 2001/06/22 14:05:05 verkout Exp $
//
//
#ifndef COFVISIBILITYBUFFER_H
#define COFVISIBILITYBUFFER_H

//
//  Bool, Double etc...
//
#ifndef AIPS_AIPS_H
#include <aips/aips.h>
#endif

#ifndef AIPS_STRING_H
#include <aips/Utilities/String.h>
#endif

//
//  This is our base-class so we need it
//
#ifndef VISIBILITYBUFFER_H
#include <vlbi/j2ms/VisibilityBuffer.h>
#endif

//
//  We keep a baseline set and a baseline iterator since we iterate
//  through the set baseline by baseline
//
#ifndef COFBASELINEITERATOR_H
#include <vlbi/j2ms/COFBaselineIterator.h>
#endif

#ifndef AIPS_MATRIX_H
#include <aips/Arrays/Matrix.h>
#endif

#ifndef AIPS_VECTOR_H
#include <aips/Arrays/Vector.h>
#endif

#ifndef AIPS_COMPLEX_H
#include <aips/Mathematics/Complex.h>
#endif

#ifndef AIPS_MEPOCH_H
#include <aips/Measures/MEpoch.h>
#endif

#ifndef DATAFILE_H
#include <data_handler/DataFile.h>
#endif

#ifndef SCAN_H
#include <vlbi/j2ms/Scan.h>
#endif

#ifndef CONTAINER_H
#include <vlbi/tmplsrc/Container.h>
#endif

#ifndef CORRELATORSETUP_H
#include <vlbi/j2ms/CorrelatorSetup.h>
#endif

//
//  We need uint32....
//
#ifndef HVTYPES_H
#include <vlbi/DtdLUT/hvtypes.h>
#endif


//
//  Forward declarations
//
namespace j2ms
{
	class COFScan;
};
class VEXperiment;
class MappingsFile;
class MappingEntry;
class DataDescriptor;
class CorrelatedDataFile;
class CorrelatedDataRecord;



class COFVisibilityBuffer :
    public VisibilityBuffer
{
public:

    //
    //  Create a COF visibility buffer
    //
    COFVisibilityBuffer( uint32 jobid, uint32 subjobid, const VEXperiment* experimentptr=0 );
    
    //
    //  Get the current visibility from the buffer
    //
    virtual Bool          getCurrentVisibility( Visibility& vref );
    
    //
    // Advance to next visibility
    //
    virtual Bool          next( void );

    //
    //  What's the directory of the visibilitybuffer??
    //
    virtual const String& getRootDir( void ) const;

    //
    //  These are COFVisibilityBuffer specific
    //
    uint32                getJobID( void ) const;
    uint32                getSubJobID( void ) const;
    
    //
    //  Delete all allocated resources
    //
    virtual ~COFVisibilityBuffer();
    
private:
    //
    //  Our private parts
    //
    uint32                myJobID;
    uint32                mySubJobID;
    String                myRootDir;
    MappingsFile*         myMappingsFileptr;
	CorrelatorSetup       mySetup;
    CorrelatedDataFile*   myCorrelatedDataFileptr;

    //
    // Pointer to the current CorrelatedDataRecord!
    //
    CorrelatedDataRecord* myCurrentCorrelatedDataRecordptr;

    //
    //  Pointer to the last DataDescriptor found...
    //
    DataDescriptor*       myCurrentDataDescriptorptr;
    
    //
    //  The current mappingentry
    //
    MappingEntry*         myCurrentMappingptr;

    //
    //  Iteration stuff. The current baseline set and the iterator
    //  that iterates over the set.
    //
    COFBaselineSet        myBaselineSet;
    COFBaselineIterator   myBaselineIterator;

    //
    //  The list of scans found in the vexfile...
    //
    Container<ScanInfo>         myScanList;
    Container<j2ms::COFScan *>  myScanPtrs;
    
    //
    //  Private methods
    //
    //  Get all the data for the configured interferometers in the baseline
    //
    const Matrix<Complex>& getBaselineData( Vector<Double>& weight,
											const COFBaseline& cbref ) const;

    //
    //  A struct, used to pass decoded mapping information to the
    //  caller
    //
    struct _MapInfo
    {
		uint32     XTelescopeNr;
		uint32     YTelescopeNr;
		uint32     XDataSourceId;
		uint32     YDataSourceId;
		uint32     XLogicalChannel;
		uint32     YLogicalChannel;
	
		//
		//  default c'tor -> initialize properly; unitialized entries
		//  are set to 0xFFFFFFFF
		//
		_MapInfo();
    };
    typedef _MapInfo MapInfo;
    

    //
    //  Static method -> transfrom a FFROT (presumably DataDescriptor.getStartROT()) and
    //  two SFROTs (start/end SFROT) into an MEpoch-measure
    //
    //
    static const MEpoch&   calcMidPointAsMEpoch( const FFROT& ffrot,
												 const SFROT& startrot,
					   							 const SFROT& endrot );
    
    
    static COFVisibilityBuffer::MapInfo decodeMapping( const MappingEntry& mapping,
													   const COFBaseline& baseline );
    
    //
    //  The COFVisibilityBuffer is the one that knows how to read a
    //  correlatorsetup from a vexfile....
    //
    static bool            readCorrelatorSetup( CorrelatorSetup& result,
												const char* vexfilename );
    
    //
    //  Prohibit these
    //
    COFVisibilityBuffer();
    COFVisibilityBuffer( const COFVisibilityBuffer& );
    COFVisibilityBuffer& operator=( const COFVisibilityBuffer& );
};


#endif
