//
//  CorrelatorSetup - describes how a scan was correlated, holds info
//  like nr of lags, which polarizations were formed, integrationtime
//  etc.
//
//  In short: it is more an aggregation than an abstraction, therefore
//  I probably should have made it a struct rather than an object.
//
//  Author:  Harro Verkouter,   13-07-1999
//
//  $Id: CorrelatorSetup.h,v 1.2 2001/05/10 06:29:01 verkout Exp $
//
//  $Log: CorrelatorSetup.h,v $
//  Revision 1.2  2001/05/10 06:29:01  verkout
//  HV: Added support to j2ms2 to make decisions based on Albert's different software versions (ABMajor/ABMinor keywords in output VEXfile). Also a 'bypass' mode is present now (12345/67890)=>data is left unprocessed.
//
//  Revision 1.1.1.1  2000/03/20 15:14:36  verkout
//  HV: imported aips++ implement stuff
//
//
//
#ifndef CORRELATORSETUP_H
#define CORRELATORSETUP_H

#ifndef AIPS_AIPS_H
#include <aips/aips.h>
#endif

#ifndef HVTYPES_H
#include <vlbi/DtdLUT/hvtypes.h>
#endif

#ifndef CONTAINER_H
#include <vlbi/tmplsrc/Container.h>
#endif

//
//  Forward declarations
//
class ostream;




class CorrelatorSetup
{
    //
    //  Comparison
    //
    friend Bool operator==(const CorrelatorSetup& lhside, const CorrelatorSetup& rhside);


public:

	//
	//  Correlator software version
	//
	struct _CorSWVersion
	{
	public:
		_CorSWVersion( unsigned int major=0, unsigned int minor=0 ) :
			myMajor( major ),
			myMinor( minor )
		{
		}

		//
		//  Allow read-only access....
		//
		inline unsigned int major( void ) const
		{
			return myMajor;
		}

		inline unsigned int minor( void ) const
		{
			return myMinor;
		}

		//
		//  Comparison operator...
		//
		inline Bool operator==( const _CorSWVersion& rhs ) const
		{
			return ToBool(myMajor==rhs.myMajor && myMinor==rhs.myMinor);
		}

		inline Bool operator!=( const _CorSWVersion& rhs ) const
		{
			return ToBool(!(*this==rhs));
		}

		// must be able to compare versions :)
		Bool operator<( const _CorSWVersion& rhs ) const;
		Bool operator<=( const _CorSWVersion& rhs ) const;
		Bool operator>( const _CorSWVersion& rhs ) const;
		Bool operator>=( const _CorSWVersion& rhs ) const;


	private:
		unsigned int   myMajor;
		unsigned int   myMinor;
	};
	typedef struct _CorSWVersion    CorSWVersion;


    //
    //  An empty correlatorsetup can exist..
    //
    //  However, values will be set to nothing or ludicrous values to
    //  indicate that they are not set. Look at the accessor methods
    //  to find out specifics
    //
    CorrelatorSetup();
    
    CorrelatorSetup( uint32 nrspectralpoints, const Container<uint16>& correlations,
					 double integrationtime, 
					 const CorrelatorSetup::CorSWVersion& swver );

    //
    //  Copy stuff across/assignment
    //
    CorrelatorSetup( const CorrelatorSetup& other );
    CorrelatorSetup& operator=( const CorrelatorSetup& other );

    
    //
    //  How many spectral points?  NB: we're safe up to 0xFFFFFFFF-1
    //  lags... which is approx. 4.2billion lags...
    //
    //  If not initialized by usr, it is set to 0xFFFFFFFF
    //
    uint32                   nrOfSpectralPoints( void ) const;

    //
    //  Retrieve the correlations that were formed during this
    //  correlator setup..
    //
    const Container<uint16>& getCorrelations( void ) const;

    //
    //  Retrieve the integrationtime for this setup (units is seconds)
    //
    double                   getIntegrationTime( void ) const;

	const CorSWVersion&      getSoftwareVersion( void ) const;


    //
    //  De-allocate any allocated resources (if any)
    //
    ~CorrelatorSetup();


    /**********************************/
    //          STATIC METHOD(S)
    /**********************************/
    
    //
    //  From some predefined correlator modes (see Huib's document),
    //  return the accompanying Setup.
    //
    //  NOTE: you still need to supply the integrationtime, since that
    //  cannot be derived from the correlatormode...
    //
    static CorrelatorSetup     correlatorModeToSetup( int correlatormode,
													  double integrationtime );


	/**********************************/
	//       STATIC MEMBER(S)
	/**********************************/
	//
	//  Dedicated unique software version that (should) put the system
	//  in bypass mode: i.e. all corrections after reading the data 
	//  (e.g. vernier delay, rotating of the fourth polarizationproduct)
	//  should be discarded.....
	//
	//  Check CorrelatorSetup.cc for actual definition of the version number,
	//  I started off with: ABMajor=12345, ABMinor=67890.
	//
	static const CorSWVersion  bypassVersion;

    
private:
    //
    //  The setup's private parts
    //

    //
    //  How many spectral points were produced? (NOT the number of
    //  lags!!)
    //
    uint32            myNumberOfSpectralPoints;    

    //
    //  Which polarization combinations were formed by the correlator?
    //  This is an array of CorrelationCode-codes.
    //
    //  A CorrelationCode (see: CorrelationCode.h), consisting of 2
    //  indices (the polarization codes of the constituting
    //  polarizations) can be cast to an uint16 -> the mapping is
    //  unique, therefore, you can identify polarization combinations
    //  with just one index. So we just keep an array of these
    //  indices. When you want to really know what the combination
    //  was, do a CorrelationCode::ordinal2code( uint16 code ) to
    //  retrieve a full CorrelationCode.
    //
    Container<uint16> myCorrelations;

    //
    //  Eevery correlator setup has an integration time associated
    //  with it...
    //
    //  NOTE: Units is seconds!
    //
    double            myIntegrationTime;

	//
	//  Correlator software version -> can be used to make
	//  switches on...
	//
	CorSWVersion      mySoftwareVersion;
};

//
//  Show the correlator setup in Human Readable Format
//
ostream& operator<<( ostream& os, const CorrelatorSetup& csref );

//
//  This one doesn't need to be a friend!
//
Bool operator!=(const CorrelatorSetup& lhside, const CorrelatorSetup& rhside);

#endif
