//
//  Experiment - Abstract base-class, specifying the kind information
//  a user can get from an experiment
//
//   Author:  Harro Verkouter, 8-12-1998
//
//   $Id: Experiment.h,v 1.2 2001/06/22 14:05:12 verkout Exp $
//
#ifndef EXPERIMENT_H
#define EXPERIMENT_H

#ifndef AIPS_AIPS_H
#include <aips/aips.h>
#endif

#ifndef AIPS_STRING_H
#include <aips/Utilities/String.h>
#endif

#ifndef AIPS_VECTOR_H
#include <aips/Arrays/Vector.h>
#endif

#ifndef VISIBILITY_H
#include <vlbi/j2ms/Visibility.h>
#endif

#ifndef AIPS_LOGGING_H
#include <aips/Logging.h>
#endif


//
//  Forward declarations...
//
class ostream;
class SourceDB;
class AntennaDB;
class FrequencyDB;


//
//  The class itself
//
class Experiment
{
public:
    //
    //  Create an Experiment. If the default is used, the Current
    //  Working directory is used as exp.root dir.
    //
    Experiment( const String& exprootdir="" );

    //
    //  Provide access to the properties. Note: these methods must be
    //  implemented by a derived class!
    //
    virtual const String&          getCode( void ) const = 0;

    //
    //  Hrmpff... design error (c) Harro V....
    //
    //  These functions should be located somewhere else but here;
    //  they control the output the JIVE MS-filler produces. May need
    //  a revision sometime... The MS-Filler should ask the visibility
    //  buffer to deliver the data in a certain domain *and* the
    //  MS-Filler should be capable of deciding on it's own where to
    //  store the data...
    //
    virtual const String&          getMSname( void ) const = 0;
    virtual void                   setMSname( const String& newmsname ) = 0;

    virtual Visibility::domain     getDestinationDomain( void ) const = 0;
    virtual Bool                   setDestinationDomain( Visibility::domain newoutputdomain ) = 0;


    //
    //  Ok... these belong here!
    //
    virtual const Vector<String>&  getAntennas( void ) const = 0;
    virtual const Vector<String>&  getSources( void ) const = 0;

    //
    // The experiment must have its own frequency DB!
    //
    virtual const FrequencyDB&     getFrequencyDB( void ) const = 0;

    //
    // The user MAY have specified an antennadatabase.
    //
    // NOTE: attempt to get an AntennaDB whilst none was configured
    // will throw an AipsError!!!!
    //
    virtual Bool                   hasAntennaDB( void ) const = 0;
    virtual const AntennaDB&       getAntennaDB( void ) const = 0;

    //
    //  Also, the user MAY have specified a source DB
    //
    virtual Bool                   hasSourceDB( void ) const = 0;
    virtual const SourceDB&        getSourceDB( void ) const = 0;

    //
    //  VisibilityBuffers's are a part of the experiment, therefore we
    //  should ask the Experiment if it would be so kind as to create
    //  the indicated VisibilityBuffer. If it fails to create it, a
    //  null-pointer is returned, otherwise a pointer to the newly
    //  created Buffer is returned!
    //
    //  NOTE: the caller is responsible for deleting the
    //  resources!!!!!!
    //
    virtual VisibilityBuffer*      getVisibilityBuffer( const String& visbufname ) = 0;

    //
    //  Tells us whether or not the Vernier delay was applied.
    //
    virtual Bool                   hasVernierDelay( void ) const = 0;

    //
    // The array-code (eg. EVN, GLOBAL, VLBA).
    //
    //  Note: derived classes *may* override this one, but don't have
    //  to. The return value defaults to EVN....(since we're at JIVE
    //  here...)
    //
    virtual const String&          getArrayCode( void ) const;

    //
    //  These methods are NOT virtual!
    //
    const String&           getRootDir( void ) const;

    //  Log a message (if logging is enabled); usr. need not worry
    //  whether or not logging is enabled, function will verify all
    //  that's necessary)
    //
    void                    logLogMessage( const String& msg, const LogOrigin& org=LogOrigin() ) const;

    //
    //  Delete the resources used by this experiment
    //
    virtual ~Experiment();

private:
    //
    //  The experiment's private parts
    //
    String     myRootDir;
    ostream**  myLogStreamptr;

    //
    //  Prohibit these
    //
    Experiment( const Experiment& );
    Experiment& operator=( const Experiment& );
};

//
//  Print some diagnostics on a stream
//
ostream& operator<<( ostream& os, const Experiment& expref );


#endif
