//
//  Implementation of the ExperimentCache - classes...
//
//
#ifndef EXPERIMENTCACHE_H
#include <vlbi/j2ms/ExperimentCache.h>
#endif

#ifndef AIPS_VECTOR_H
#include <aips/Arrays/Vector.h>
#endif

#ifndef ANTENNADB_H
#include <vlbi/j2ms/AntennaDB.h>
#endif

#ifndef SOURCEDB_H
#include <vlbi/j2ms/SourceDB.h>
#endif

#ifndef EXPERIMENT_H
#include <vlbi/j2ms/Experiment.h>
#endif

#ifndef SCAN_H
#include <vlbi/j2ms/Scan.h>
#endif

#ifndef AIPS_EXCEPTIONS_H
#include <aips/Exceptions.h>
#endif

#include <stdlib.h>


//
//  Init ScanCacheElement.....
//
_ScanCacheElement::_ScanCacheElement() :
    outEpochRef( MEpoch::GAST ),
    prefixtimeadjust( 0 ),
    postfixtimeadjust( 0 ),
    samplecount( 0 ),
    itsLastIntegration( 0xFFFFFFFF ),
    itsLastendbocf( 0xFFFFFFFF ),
    nextScan( 0 )
{
    //
    // Check if the aipspath-environment variable is set. If it isn't,
    // we cant do the conversions (likely to crash..)
    //
    if( !getenv("AIPSPATH") )
    {
	throw( AipsError( String("Environment var. `AIPSPATH' not set. source an aipsinit.csh -")
			 +String(" script to set it for you.")) );
    }
    

    //
    //  Okiedokie, set references
    //
    timeCvt.setOut( outEpochRef );
}


_ScanCacheElement::~_ScanCacheElement()
{
}


//
//  Init ptr2next to NULL
//
_ExperimentCacheElement::_ExperimentCacheElement() :
    itsFirstScanptr( 0 ),
    nextExperiment( 0 )
{
}

ScanCacheElement* _ExperimentCacheElement::addScan( const j2ms::Scan& scanref,
													const SourceDB* alternatedbptr )
{
    ScanCacheElement*   scanptr;
    ScanCacheElement*   lastscanptr( 0 );
    
    //
    //  First: locate the scan?
    //
    scanptr = itsFirstScanptr;
    
    while( scanptr )
    {
	if( scanptr->itsScanCode==scanref.getScanId() )
	{
	    break;
	}
	lastscanptr = scanptr;
	scanptr = scanptr->nextScan;
    }

    if( !scanptr )
    {
	Bool               found;
	const Experiment&  expref( scanref.getExperiment() );
	
	//
	//  Add the scan!
	//
	scanptr = new ScanCacheElement;
	
	scanptr->itsScanCode = scanref.getScanId();
	
	//
	//  Attempt to locate the source-info
	//
	found = False;
	if( expref.hasSourceDB() )
	{
	    found = expref.getSourceDB().searchSource( scanptr->itsSourceInfo, scanref.getSourceCode() );
	}
	
	if( !found && alternatedbptr )
	{
	    found = alternatedbptr->searchSource( scanptr->itsSourceInfo, scanref.getSourceCode() );
	}
	
	if( !found )
	{
	    cout << "Failed to find SourceInfo for source `" << scanref.getSourceCode() << "'" << endl;
	}
	
	//
	//  Add the scan to the cache-list
	//
	if( lastscanptr )
	{
	    lastscanptr->nextScan = scanptr;
	}
	else
	{
	    itsFirstScanptr = scanptr;
	}
    }
    return scanptr;
}

ScanCacheElement* _ExperimentCacheElement::locateScan( const String& scancode )
{
    ScanCacheElement*    retval( itsFirstScanptr );
    
    while( retval )
    {
	if( retval->itsScanCode==scancode )
	{
	    break;
	}
	retval = retval->nextScan;
    }
    return retval;
}



//
//  Auto-delete the next element??! NO!!!!
//
_ExperimentCacheElement::~_ExperimentCacheElement()
{
    //
    // Clean the scan-cache
    //
    ScanCacheElement*   savedptr;
    
    while( itsFirstScanptr )
    {
	savedptr = itsFirstScanptr->nextScan;
	
	delete itsFirstScanptr;
	
	itsFirstScanptr = savedptr;
    }
}


//
//  The cache-class
//
ExperimentCache::ExperimentCache() :
    itsFirstElementptr( 0 )
{
}

//
//  Add an element into the cache (if it wasn't there already)
//
ExperimentCacheElement* ExperimentCache::addExperiment( const Experiment& expref, const AntennaDB* alternatedbptr )
{
    ExperimentCacheElement*   ptr2element( 0 );
    ExperimentCacheElement*   ptr2last( 0 );

    //
    //  First: look if exp. already in cache
    //
    ptr2element = itsFirstElementptr;
    while( ptr2element )
    {
	if( ptr2element->itsCode==expref.getCode() )
	{
	    //
	    //  Found it!
	    //
	    break;
	}
	ptr2last = ptr2element;
	ptr2element = ptr2element->nextExperiment;
    }
    
    //
    //  If element not yet found, add it!
    //
    if( !ptr2element )
    {
	//
	//  Get the antennacodes
	//
	Bool                   infofound;
	AntennaInfo            ai;
	const Vector<String>&  antennacodes( expref.getAntennas() );
	
	//
	//  Make room for a new element
	//
	ptr2element = new ExperimentCacheElement;
	
	ptr2element->itsCode = expref.getCode();
	
	//
	//  Attempt to fill in the AntennaTable
	//
	ptr2element->itsAntennaTable.resize( antennacodes.nelements() );
	
	for( uInt i=0; i<antennacodes.nelements(); i++ )
	{
	    infofound = False;
	    
	    if( expref.hasAntennaDB() )
	    {
		infofound = expref.getAntennaDB().searchAntenna( ai, antennacodes(i) );
	    }
	    
	    if( !infofound && alternatedbptr )
	    {
		infofound = alternatedbptr->searchAntenna( ai, antennacodes(i) );
	    }
	    
	    if( infofound )
	    {
		ptr2element->itsAntennaTable(i) = ai;

		//
		//  Help?
		//
		//cout << "Found antenna[" << i << "]:" << endl << ptr2element->itsAntennaTable(i) << endl;
	    }
	    else
	    {
		cout << "Cache: Failed to locate info for antenna `" << antennacodes(i) << "'" << endl;
	    }
	}
	
	//
	//  Don't forget to add the element to the list!
	//
	if( ptr2last )
	{
	    ptr2last->nextExperiment = ptr2element;
	}
	else
	{
	    itsFirstElementptr = ptr2element;
	}

	//
	//  Thats it for now...
	//
    }
    
    return ptr2element;
}

//
//  Locate an experiment based on its ExperimentCode
//
ExperimentCacheElement* ExperimentCache::locateExperiment( const String& expcode ) const
{
    ExperimentCacheElement*   retval( itsFirstElementptr );
    
    while( retval )
    {
	if( retval->itsCode==expcode )
	{
	    break;
	}
	retval = retval->nextExperiment;
    }
    return retval;
}


//
//  Delete all cached objects
//
ExperimentCache::~ExperimentCache()
{
    ExperimentCacheElement*    savedptr;
    
    while( itsFirstElementptr )
    {
	savedptr = itsFirstElementptr->nextExperiment;
	
	delete itsFirstElementptr;
	
	itsFirstElementptr = savedptr;
    }
    
    //
    // And we're done!!
    //
}


