//
//  Implementation
//
#ifndef IFBASEDDTDVISBUFFER_H
#include <vlbi/j2ms/IFBasedDTDVisBuffer.h>
#endif

#ifndef AIPS_EXCEPTIONS_H
#include <aips/Exceptions.h>
#endif

#ifndef INIFILE_H
#include <inifile/IniFile.h>
#endif

#ifndef CORRELATION_CODE_H
#include <vlbi/DtdLUT/CorrelationCode.h>
#endif

#ifndef EVNFRAFLAGS_H
#include <vlbi/evnfraflags.h>
#endif

#ifndef BASELINEIDX_H
#include <vlbi/DtdLUT/BaseLineIdx.h>
#endif

#include <iostream.h>
#include <stdlib.h>



IFBasedDTDVisBuffer::IFBasedDTDVisBuffer( const String& filename, const Experiment* experimentptr ) :
    VisibilityBuffer( experimentptr ),
    itsBaseFilename( filename ),
    itsRootDir( this->getExperiment().getRootDir()+"/"+itsBaseFilename ),
    itsDTDptr( new dtd ),
    itsDatafileptr( 0 ),
    itsCurrentjob( 0 ),
    itsCurrentsubarray( 0 ),
    itsCurrentIF( 0 ),
    itsCurrentintegration( 0 ),
    itsNrintegrations( 0L )
{
    Bool      dtdread( True );
    uLong     dtdoffset( 0L );
    uLong     dtdfilesize;
    uLong     filesize;
    uLong     jobdatasize;
    IniFile   theInifile( itsRootDir+"/"+itsBaseFilename+".ini" );
    ifstream  theDTDfile( itsRootDir+"/"+itsBaseFilename+".dtd", ios::in | ios::nocreate );
    
    //
    //  Check if DTD file was found
    //
    if( !theDTDfile.good() )
    {
	throw( AipsError("Sorry, DTD file `"+filename+".dtd' does not exist!") );
    }
    

    //
    //  Ok Set up the buffer, check for existance of both .cor and
    //  .dtd files....
    //
    itsDatafileptr = new ifstream( filename+".cor", ios::in | ios::nocreate );
    
    if( !itsDatafileptr->good() )
    {
	throw( AipsError("Sorry, binary datafile `"+filename+".cor' does not exist!") );
    }
    
    //
    //  Ok. Having verified that, we now try to read the DTD from the
    //  DTDfile...
    //
    theDTDfile.seekg( 0L, ios::end );
    dtdfilesize = theDTDfile.tellg();

    //
    //  Check if an offset was specified in the ini-file
    //
    String   tmp;
    
    theInifile.getValue( "DTD", "offset", tmp, "-1" );
    if( atol((const char*)tmp)>=0 )
    {
	dtdoffset = atol( (const char*)tmp );
    }

    //
    //  Do some checking....
    //
    if( (dtdoffset+(uLong)sizeof(dtd))>(uLong)dtdfilesize )
    {
	float   fraction( ((float)dtdfilesize-(float)dtdoffset)/(float)sizeof(dtd) );

	cout << "### Warning! DTD file `" << filename << ".dtd' is too small." << endl;
	cout << "### Cannot read complete DTD!" << endl;
	
	if( fraction>=0.9 )
	{
	    cout << "*** The system will read " << fraction*100.0 << "% of DTD." << endl;
	}
	else
	{
	    dtdread = False;
	    cout << "*** Less than 90% of datatransferdescriptor could be recovered!" << endl;
	    cout << "*** Stop!" << endl;
	}
    }

    if( !dtdread )
    {
	throw( AipsError("Blast. No DTD read.") );
    }
    
    theDTDfile.seekg( dtdoffset, ios::beg );
    theDTDfile.read( itsDTDptr, sizeof(dtd) );
    
    //
    //  Deduce the number of integrations?
    //

    //
    //  Get the total datafilesize
    //
    itsDatafileptr->seekg( 0L, ios::end );
    filesize = itsDatafileptr->tellg();
    
    itsDatafileptr->seekg( 0L, ios::beg );

    //
    //  The job datasize
    //
    jobdatasize=( itsDTDptr->TOTAL_DATASIZE*4L );

    itsNrintegrations = filesize/jobdatasize;

    //
    //  Check if we need 2 overwrite nrintegrations from the inifile
    //
    theInifile.getValue( "DTD", "nrintegrations", tmp, "<DEF>" );
    
    if( tmp!="<DEF>" )
    {
	itsNrintegrations = atol( (const char*)tmp );
    }
}


//Bool IFBasedDTDVisBuffer::getCurrentVisibility( Visibility& vis ) const
Bool IFBasedDTDVisBuffer::getCurrentVisibility( Visibility& vis )
{
    ExtVisLabel        lab;
    Matrix<Complex>    t;
    Visibility::domain d;

    //
    //  Init vis to be an empty visibility....
    //
    vis = Visibility();
    
    //
    //  Check adressing
    //
    if( itsCurrentjob>=(uInt)itsDTDptr->NJOB )
    {
	return False;
    }
    
    const jobdescr& curjobref( itsDTDptr->JOB_DATADESCR_ARRAY[itsCurrentjob] );
    
    if( itsCurrentsubarray>=(uInt)curjobref.NSUBARRAY )
    {
	return False;
    }
    
    const sadescr&  cursaref( itsDTDptr->SA_DATADESCR_ARRAY[curjobref.SA_DINDEX+itsCurrentsubarray] );
    
    if( itsCurrentIF>=(uInt)cursaref.NINTFR )
    {
	return False;
    }
    
    const ifdescr&  ifref( itsDTDptr->INTFR_DATADESCR_ARRAY[cursaref.IF_DINDEX+itsCurrentIF] );
    
    lab.jobnr() = itsCurrentjob;
    lab.ifnr() = itsCurrentIF;
    lab.subarraynr() = itsCurrentsubarray;
    lab.subbandnr() = ifref.SBANDNR;
    lab.framenr() = itsCurrentintegration;
    lab.baseline() = BaseLineIdx( ifref.XANTNR, ifref.YANTNR );

    lab.getCorrelations().resize( 1 );
    lab.getCorrelations()( 0 ) = CorrelationCode( ifref.XPOLNR, ifref.YPOLNR );

    //
    //  read the matrix
    //
    t = this->readData( lab, d );

    //
    //  Find out if the visibility has the spectral flag set or not
    //
    vis=Visibility( lab, t, d );

    return True;
}

Bool IFBasedDTDVisBuffer::next( void )
{
    Bool     retval( False );
    
    //
    //  Increase to next visibility??
    //
    //
    //  Check adressing
    //
    if( itsCurrentjob>=(uInt)itsDTDptr->NJOB )
    {
	return retval;
    }
    
    const jobdescr& curjobref( itsDTDptr->JOB_DATADESCR_ARRAY[itsCurrentjob] );
    
    if( itsCurrentsubarray>=(uInt)curjobref.NSUBARRAY )
    {
	return retval;
    }
    
    const sadescr&  cursaref( itsDTDptr->SA_DATADESCR_ARRAY[curjobref.SA_DINDEX+itsCurrentsubarray] );
    
    if( itsCurrentIF>=(uInt)cursaref.NINTFR )
    {
	return retval;
    }
    
    //
    // Attempt to increment basleline...
    //
    if( ++itsCurrentIF>=(uInt)cursaref.NINTFR )
    {
	//
	//  Ok. Start at zero...
	//
	itsCurrentIF=0;

	if( ++itsCurrentsubarray>=(uInt)curjobref.NSUBARRAY )
	{
	    //
	    //  Attempt to move on to next integration...
	    //
	    itsCurrentsubarray=0;

	    if( ++itsCurrentintegration>=itsNrintegrations )
	    {
		itsCurrentintegration=0;
		
		if( ++itsCurrentjob>=(uInt)itsDTDptr->NJOB )
		{
		    return retval;
		}
	    }
	}
    }
    return True;
}


IFBasedDTDVisBuffer::~IFBasedDTDVisBuffer()
{
    //
    //  Clean up!
    //
    delete itsDTDptr;
    
    delete itsDatafileptr;
}

//
//
//  Private methods!
// 
//
//
Matrix<Complex> IFBasedDTDVisBuffer::readData( const ExtVisLabel& labref,
					       Visibility::domain& domainref ) const
{
    static Float*          buffer( 0 );
    static Matrix<Complex> noans;


    //
    //  The return value
    //
    noans.resize( 0, 0 );

    //
    // Ok. Get the interferometer data
    //
    uInt         nrlags;
    uInt         nrfloats;
	uInt         ifindex( labref.ifnr() );
    //jobdescr*    jptr( &itsDTDptr->JOB_DATADESCR_ARRAY[labref.jobnr()] );
    //sadescr*     saptr( &itsDTDptr->SA_DATADESCR_ARRAY[jptr->SA_DINDEX+labref.subarraynr()] );
    ifdescr*     ifptr( &itsDTDptr->INTFR_DATADESCR_ARRAY[ifindex] );
	dbdescr*     dbptr( &itsDTDptr->DB_DATADESCR_ARRAY[ifptr->DB_DINDEX] );

    //
    // Let's fill in the domain of the data
    //
    if( ifptr->IF_DATATYPE&SPECTRUMFLAG )
    {
	domainref = Visibility::frequency;
    }
    else
    {
	domainref = Visibility::time;
    }

    //
    //  If this is LAG data and not an autocorrelation, we have twice
    //  the number of channels in the lag-domain, since 'nrchannels'
    //  now holds the value in the spectral domain.
    //
    nrfloats = nrlags = ifptr->NLAG;
    
    //
    //  If we're dealing with cmplex data, nlags is sum of real+imag,
    //  therefore, we only have half the number of complex numbers.
    //
    if( ifptr->IF_DATATYPE & CMPLXFLAG )
    {
	nrlags/=2;
    }

    //
    // Check if the IF_DATASIZE is at least the nrlags...
    //
    if( (uInt)ifptr->IF_DATASIZE<nrfloats )
    {
	cout << "IF_DATASIZE=" << ifptr->IF_DATASIZE << " whilst nrfloats=" << nrfloats << endl;
	return noans;
    }

    //
    // The next step is to locate the data for this interferometer in
    // Alberts datablock!
    //

    //
    //  Resize our matrix
    //
    noans.resize( 1, nrlags );

    //
    //  Get the buffer
    //
    buffer = new Float[ nrfloats ];

    //
    //  And read the data; first determine the total fileoffset in
    //  words, then multilply by sizeof(Float) to get the offset in
    //  bytes!
    //
    Long        offset( 0 );
    Long        datablockoffset( 0 );

    //
    // 1) -> the start of the whole block of data this dtd describes:
    //
    datablockoffset = itsDTDptr->TOTAL_DATASIZE*labref.framenr();

    //
    //  2) Now then, the ifptr->DB_DINDEX tells us what the ordinal
    //  number of the datablock attached to this interferometer
    //  is. This in itself doesn't tell us much; what we want is to
    //  locate the block. Albert uses equal-sized blocks, the size
    //  being the maximum datasize found in the subarray. In the
    //  subarraydescr this value is saved. Hence we multiply the
    //  blocknr with the blocksize and TADA! we have the offset of the
    //  datablock for this interferometer.
    //
    //datablockoffset += ( ifptr->DB_DINDEX * saptr->MDB_DATASIZE );
	datablockoffset  += dbptr->DB_DOFFS;
    
    //
    //  3) Now we find the INTERFEROMETERdata by adding to the
    //  block-offset the IF_DATAOFFS value.
    //
    offset = datablockoffset + ifptr->IF_DATAOFFS;

    //
    //  4) Since Albert works with units of words (=32-bit floating
    //  point values), we multiply the found value by 4 to get the
    //  offset in bytes!
    //
    offset *= 4;
    
    itsDatafileptr->seekg( offset, ios::beg );
    
    itsDatafileptr->read( buffer, ifptr->IF_DATASIZE*sizeof(Float) );

    //
    //  Now transform the buffer into complex numbers
    //  The organization in the data-blocks is:
    //
    //  _____ nrchannels _____ ____ nrchannels _______
    //  |                     ||                      |
    //  <real><real>.....<real><imag><imag>......<imag>
    //
    if( ifptr->IF_DATATYPE & ACFLAG )
    {
        for( uInt lagcnt=0; lagcnt<nrlags; lagcnt++ )
        {
            noans( IPosition(2,0,lagcnt) )=Complex( buffer[lagcnt], 0.0 );
        }
    }
    else
    {
        for( uInt lagcnt=0; lagcnt<nrlags; lagcnt++ )
        {
            noans( IPosition(2,0,lagcnt) )=Complex( buffer[lagcnt], buffer[lagcnt+nrlags] );
        }
    }

    //
    //  Cleanup
    //
    delete [] buffer;
    buffer = 0;

    return noans;
}


