//
//  Implementation of the experiment class
//
//  Author:  Harro Verkouter, 8-12-1998
//
//
#ifndef INIEXPERIMENT_H
#include <vlbi/j2ms/IniExperiment.h>
#endif

#ifndef INIFILE_H
#include <inifile/IniFile.h>
#endif

//
//  AIPS++ exceptions
//
#ifndef AIPSEXCEPTIONOBJECT_H
#include <vlbi/Utilities/AIPSExceptionObject.h>
#endif

//
//  COF exceptions
//
#ifndef EXCEPTIONOBJECT_H
#include <ExceptionObject/ExceptionObject.h>
#endif

#ifndef AIPS_ARRAYUTIL_H
#include <aips/Arrays/ArrayUtil.h>
#endif

#ifndef DTDSCAN_H
#include <vlbi/j2ms/DTDScan.h>
#endif

#ifndef INIFREQUENCYDB_H
#include <vlbi/j2ms/IniFrequencyDB.h>
#endif

#ifndef INIANTENNADB_H
#include <vlbi/j2ms/IniAntennaDB.h>
#endif

#ifndef INISOURCEDB_H
#include <vlbi/j2ms/IniSourceDB.h>
#endif

#ifndef DTDVISIBILITYBUFFER_H
#include <vlbi/j2ms/DTDVisibilityBuffer.h>
#endif


//
//  For JB def's of uint32 etc
//
#ifndef STD_TYPES_H
#include <fundamental/std_types.h>
#endif

#include <iostream.h>
#include <fstream.h>

#include <unistd.h>


ostream& operator<<( ostream& os, const IniExperiment& expref )
{
    os << "************ IniExperiment " << expref.getCode() << " ************" << endl;
    os << "Antennas: " << expref.getAntennas().ac() << endl;
    os << "Sources: " << expref.getSources().ac() << endl;
    os << "Output is placed in: " << expref.getMSname() << endl;
    os << "Data is written in the MS in the " << expref.getDestinationDomain() << " domain" << endl;
    return os;
}




IniExperiment::IniExperiment( const String& rootdir ) :
    Experiment( rootdir ),
    itsDestinationDomain( Visibility::undefined ),
    itsFrequencyDBptr( 0 ),
    itsAntennaDBptr( 0 ),
    itsSourceDBptr( 0 ),
    itsVernierDelayFlag( True ),
    itsArrayCode( "EVN" )
{
    String      tmpstring;
    IniFile     theInifile( this->getRootDir()+"/experiment.ini" );
    
    //
    //  Read settings from the inifile
    //
    theInifile.getValue( "Experiment", "code", itsCode, "<DEF>" );
    
    if( itsCode=="<DEF>" )
    {
	throw( AipsError("IniExperiment: no experiment code specified!") );
    }
    
    theInifile.getValue( "Experiment", "msout", itsMSname, "<DEF>" );
    
    if( itsMSname=="<DEF>" )
    {
	throw( AipsError("IniExperiment "+itsCode+": no output MS-name specified!") );
    }

    //
    //  In what domain do we write the data?
    //
    theInifile.getValue( "Experiment", "domain", tmpstring, "frequency" );

    if( tmpstring=="frequency" )
    {
	itsDestinationDomain = Visibility::frequency;
    }
    else if( tmpstring=="time" )
    {
	itsDestinationDomain = Visibility::time;
    }
    else
    {
	throw( AipsError("IniExperiment "+itsCode+": `"+tmpstring+"' is not recognized as a valid domain\n"+
			 "Use either time or frequency") );
    }
    
    //
    //  Read the antennas and sources!
    //
    theInifile.getValue( "Experiment", "antennas", tmpstring, "<DEF>" );
    
    if( tmpstring=="<DEF>" )
    {
	throw( AipsError("IniExperiment "+itsCode+": no antennas specified!") );
    }
    itsAntennas = stringToVector( tmpstring );


    theInifile.getValue( "Experiment", "sources", tmpstring, "<DEF>" );
    
    if( tmpstring=="<DEF>" )
    {
	throw( AipsError("IniExperiment "+itsCode+": no sources specified!") );
    }
    itsSources = stringToVector( tmpstring );
    

    theInifile.getValue( "Experiment", "arraycode", itsArrayCode, "EVN" );


    //
    // Attempt to make the IniFrequencyDatabase
    //
    itsFrequencyDBptr = new IniFrequencyDB( this->getRootDir()+"/frequency.ini" );

    if( !itsFrequencyDBptr )
    {
	throw( AipsError("IniExperiment "+itsCode+": no frequency-database...!") );
    }

    //
    //  Check if the user specified his/her own antenna-database
    //
    ifstream*    test( new ifstream(this->getRootDir()+"/antenna.ini") );
    
    if( test->good() )
    {
	test->close();
	
	itsAntennaDBptr = new IniAntennaDB( this->getRootDir()+"/antenna.ini" );
    }
    delete test;
    test = 0;

    //
    //  Check for existance of source db
    //
    test = new ifstream( this->getRootDir()+"/source.ini" ) ;
    
    if( test->good() )
    {
	test->close();
	
	itsSourceDBptr = new IniSourceDB( this->getRootDir()+"/source.ini" );
    }
    delete test;

    //
    //  See if Vernierdelay was specified
    //
    theInifile.getValue( "Experiment", "vernierdelay", tmpstring, "<DEF>" );

    if( tmpstring!="<DEF>" )
    {
	Regex     onstring( "^[ \t]*[oO][nN][ \t]*$" );
	Regex     offstring( "^[ \t]*[oO][fF]\\{2\\}[ \t]*$" );
	
	if( tmpstring.matches(onstring) )
	{
	    itsVernierDelayFlag = True;
	}
	else if( tmpstring.matches(offstring) )
	{
	    itsVernierDelayFlag = False;
	}
	else
	{
	    cout << "IniExperiment: invalid VernierDelay string `" << tmpstring << "'. Keeping default value." << endl;
	}
    }
    

    //
    //  Ok!
    //
}

const String& IniExperiment::getCode( void ) const
{
    return itsCode;
}

const String& IniExperiment::getMSname( void ) const
{
    return itsMSname;
}

void IniExperiment::setMSname( const String& newmsname )
{
    itsMSname = newmsname;
    return;
}

Visibility::domain IniExperiment::getDestinationDomain( void ) const
{
    return itsDestinationDomain;
}

Bool IniExperiment::setDestinationDomain( Visibility::domain newoutputdomain )
{
    itsDestinationDomain = newoutputdomain;
    cout << "IniExperiment " << itsCode << " - data will now be written in the "
	 << itsDestinationDomain << " domain.\n" << flush;
    return True;
}


const FrequencyDB& IniExperiment::getFrequencyDB( void ) const
{
    return *itsFrequencyDBptr;
}

Bool IniExperiment::hasAntennaDB( void ) const
{
    return ToBool( itsAntennaDBptr!=0 );
}

const AntennaDB& IniExperiment::getAntennaDB( void ) const
{
    if( !itsAntennaDBptr )
    {
	throw( AipsError("IniExperiment: attempt to get non-existant antennaDB!") );
    }
    return *itsAntennaDBptr;
}

Bool IniExperiment::hasSourceDB( void ) const
{
    return ToBool( itsSourceDBptr!=0 );
}

const SourceDB& IniExperiment::getSourceDB( void ) const
{
    if( !itsSourceDBptr )
    {
	throw( AipsError("IniExperiment: attempt to get non-existant sourceDB!") );
    }
    return *itsSourceDBptr;
}


/*
Scan* IniExperiment::getScan( const String& scanid )
{
    Scan*     retval( 0 );
    
    try
    {
	//
	//  Attempt to create a DTDScan.... AipsError will be thrown
	//  when not available
	//
	retval = new DTDScan( scanid, this );
    }
    catch( AipsError& x )
    {
	//
	//  Make sure that the retval is zero
	//
	cout << "IniExperiment::getScan: " << x.getMesg() << endl;
	retval = 0;
    }
    return retval;
}
*/


VisibilityBuffer* IniExperiment::getVisibilityBuffer( const String& visbufname )
{
    VisibilityBuffer*     retval( 0 );
    
    try
    {
	//
	//  Attempt to create a DTDScan.... AipsError will be thrown
	//  when not available
	//
	retval = new DTDVisibilityBuffer( visbufname, this );
    }
    catch( AipsError& x )
    {
	//
	//  Make sure that the retval is zero
	//
	cout << "IniExperiment::getVisibilityBuffer: " << x.getMesg() << endl;
	retval = 0;
    }
    return retval;
}



Bool IniExperiment::hasVernierDelay( void ) const
{
    return itsVernierDelayFlag;
}


const String& IniExperiment::getArrayCode( void ) const
{
    return itsArrayCode;
}



const Vector<String>& IniExperiment::getAntennas( void ) const
{
    return itsAntennas;
}

const Vector<String>& IniExperiment::getSources( void ) const
{
    return itsSources;
}




IniExperiment::~IniExperiment()
{
    //
    //  Delete our own databases
    //
    delete itsFrequencyDBptr;
    
    delete itsAntennaDBptr;

    delete itsSourceDBptr;
}
