//
//  IniExperiment - class that holds experiment-specific data, data is
//  read from inifiles!
//
//   Author:  Harro Verkouter, 8-12-1998
//
//   $Id: IniExperiment.h,v 1.1.1.1 2000/03/20 15:14:36 verkout Exp $
//
#ifndef INIEXPERIMENT_H
#define INIEXPERIMENT_H

#ifndef AIPS_AIPS_H
#include <aips/aips.h>
#endif

#ifndef AIPS_STRING_H
#include <aips/Utilities/String.h>
#endif

#ifndef AIPS_VECTOR_H
#include <aips/Arrays/Vector.h>
#endif

#ifndef VISIBILITY_H
#include <vlbi/j2ms/Visibility.h>
#endif


//
//  We need this; it's our baseclass
//
#ifndef EXPERIMENT_H
#include <vlbi/j2ms/Experiment.h>
#endif


//
//  Forward declarations...
//
class Scan;
class ostream;
class SourceDB;
class AntennaDB;
class FrequencyDB;


//
//  The class itself
//
class IniExperiment :
    public Experiment
{
public:
    //
    //  Create an IniExperiment. A number of settings are read from the
    //  inifile 'experiment.ini'
    //
    IniExperiment( const String& rootdir="" );

    //
    //  Provide access to the properties
    //
    virtual const String&          getCode( void ) const;

    virtual const String&          getMSname( void ) const;
    virtual void                   setMSname( const String& newmsname );

    virtual Visibility::domain     getDestinationDomain( void ) const;
    virtual Bool                   setDestinationDomain( Visibility::domain newoutputdomain );
    
    virtual const Vector<String>&  getAntennas( void ) const;
    virtual const Vector<String>&  getSources( void ) const;

    //
    // The experiment must have its own frequency DB!
    //
    virtual const FrequencyDB&     getFrequencyDB( void ) const;

    //
    // The user MAY have specified an antennadatabase.
    //
    // NOTE: attempt to get an AntennaDB whilst none was configure
    // will throw an AipsError!!!!
    //
    virtual Bool                   hasAntennaDB( void ) const;
    virtual const AntennaDB&       getAntennaDB( void ) const;

    //
    //  Also, the user MAY have specified a source DB
    //
    virtual Bool                   hasSourceDB( void ) const;
    virtual const SourceDB&        getSourceDB( void ) const;

    //
    //  VisibilityBuffer's are a part of the experiment, therefore we should ask
    //  the IniExperiment if it would be so kind as to create the
    //  indicated buffer. If it fails to create the buffer, a null-pointer
    //  is returned, otherwise a pointer to the newly created VisibilityBuffer is
    //  returned! 
    //
    //  NOTE: the caller is responsible for deleting the
    //  resources!!!!!!
    //
    virtual VisibilityBuffer*      getVisibilityBuffer( const String& visbufname );
    
    //
    //  return the arraycode
    //
    virtual const String&          getArrayCode( void ) const;

    //
    //  Tells us whether or not the Vernier delay was applied.
    //
    virtual Bool                   hasVernierDelay( void ) const;

    //
    //  Delete the resources used by this experiment
    //
    virtual ~IniExperiment();

private:
    //
    //  The experiment's private parts
    //

    //
    //  In what domain would the user like to have his/her data?
    //
    Visibility::domain  itsDestinationDomain;

    //
    // The experiment's code (e.g. 'fr002' or 'gl34' )
    //
    String     itsCode;

    //
    //  The name of the attached MS
    //
    String     itsMSname;

    //
    //  The list of antennas (two-letter code) used in this experiment
    //
    Vector<String>   itsAntennas;
    
    //
    //  The list of sources for this experiment (source-codes!)
    //
    Vector<String>   itsSources;

    //
    //  We NEED a pointer to a frequencyDB
    //
    FrequencyDB*     itsFrequencyDBptr;
    
    //
    //  Optionally, the user may have specified an antennaDb
    //  him/herself
    //
    AntennaDB*       itsAntennaDBptr;
    
    //
    //  Or a sourceDB
    //
    SourceDB*        itsSourceDBptr;

    //
    //  Was Vernier delay applied? Default = NO (False)
    //
    Bool             itsVernierDelayFlag;

    //
    //  ArrayCode defaults to EVN
    //
    String           itsArrayCode;
    
    //
    //  Prohibit these
    //
    IniExperiment( const IniExperiment& );
    IniExperiment& operator=( const IniExperiment& );
};

//
//  Print some diagnostics on a stream
//
ostream& operator<<( ostream& os, const IniExperiment& expref );


#endif
