//
//  Implementation of the IniFrequencyDB class
//
//
//  Author:  Harro Verkouter,  3-12-1998
//
//  $Id: IniFrequencyDB.cc,v 1.1.1.1 2000/03/20 15:14:35 verkout Exp $
//
//
#ifndef INIFREQUENCYDB_H
#include <vlbi/j2ms/IniFrequencyDB.h>
#endif

#ifndef INIFILE_H
#include <inifile/IniFile.h>
#endif

#ifndef FREQUENCYCONFIG_H
#include <vlbi/j2ms/FrequencyConfig.h>
#endif

#ifndef AIPS_REGEX_H
#include <aips/Utilities/Regex.h>
#endif

#ifndef AIPS_MEASURES_H
#include <aips/Measures.h>
#endif

#ifndef CORRELATIONCODE_H
#include <vlbi/DtdLUT/CorrelationCode.h>
#endif

#include <iostream.h>
#include <strstream.h>
#include <iomanip.h>


#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>


IniFrequencyDB::IniFrequencyDB( const String& inifilename ) :
    FrequencyDB(),
    itsInifilename( inifilename ),
    itsInifileptr( new IniFile(itsInifilename) )
{
}

FrequencyConfig IniFrequencyDB::operator[]( const String& key ) const
{
    FrequencyConfig   retval( key );
    
    this->readRecord( retval, key );
    return retval;
}


Bool IniFrequencyDB::searchFrequencyConfig( FrequencyConfig& fcref, const String& key ) const
{
    return this->readRecord( fcref, key );
}




IniFrequencyDB::~IniFrequencyDB()
{
    //
    //  Delete the inifile ptr!
    //
    delete itsInifileptr;
}


//
//  Private functions
//
Bool  IniFrequencyDB::readRecord( FrequencyConfig& fcref, const String& key ) const
{
    char                    buffer[128];
    uInt                    nrelements;
    Bool                    retval( False );
    ostrstream              entryname( buffer, 128 );
    FrequencyConfig::PolMap polcodes;
    //
    //  These two variables hold the properties for the buffer where
    //  the parsed subbanddefinition line goes into
    //
    const uInt              nrallocated( 5 );
    String                  elements[ nrallocated ];

    
    //
    //  Make sure fcref is empty
    //
    fcref = FrequencyConfig( key );

    //
    //  Init the polmap...
    //
    FrequencyConfig::initPolMap( polcodes );

    //
    //  Basically, we need to read the polarization map, we need to be
    //  able to translate code 0 and 1 into either x,y,l or r (or
    //  maybe i,q,u,v?)
    //
    String     polmap;
    
    itsInifileptr->getValue( key, "polmap", polmap, "<DEF>" );
    
    if( polmap=="<DEF>" )
    {
	cout << "No polmap!" << endl;
	return retval;
    }

    //
    //  Parse the polmap string, allowed format is:
    //
    //  <num>:<polcode>,<num>:<polcode>
    //
    //  where:
    //
    //   <num>     = [0|1]
    //   <polcode> = [l|L|r|R|x|X|y|Y]
    //
    uInt       nrvalid( 0 );
    String     num( "[01]" );
    String     polcode( "[lLrRxXyY]" );
    String     optionalwhite( "[ \n\r\t\v\f]*" );
    Regex      onepol( optionalwhite+num+":"+polcode+optionalwhite );
    
    nrelements = split( polmap, elements, 4, "," );
    
    for( uInt i=0; i<nrelements; i++ )
    {
	if( elements[i].matches(onepol) )
	{
	    Int     n;
	    Char    c;
	    
	    //
	    //  Get out the number
	    //
	    if( sscanf((const char*)elements[i], "%d:%c", &n, &c)==2 )
	    {
		if( n>=0 && n<2 && !polcodes[n] )
		{
		    polcodes[n]=c;
		    nrvalid++;
		}
	    }
	}
    }

    if( nrvalid!=2 )
    {
	cout << "Not 2 valid polarizationmap-entries were found for ";
	cout << "frequency specification " << key << " in file " << itsInifilename << endl;
	return retval;
    }

    //
    //  Copy the polmap into the frequency config!
    //
    fcref.setPolMap( polcodes );

    //
    //  From here, the retval is True! only errors can occur, but
    //  that's something else; the record has been found!
    //
    retval = True;

    //
    //  Attempt to fill the fcref object!
    //
    uInt     sbcnt( 0 );
    uInt     errcnt( 0 );
    Bool     done( False );
    
    while( !done )
    {
	String    sbtext;

	//
	//  Create the correct entryname:
	//
	//  [<config_label>]
	//  subband<nr>=<subband def>
	//
	entryname.seekp( 0, ios::beg );
	
	entryname << "subband" << sbcnt << ends;

	itsInifileptr->getValue( key, entryname.str(), sbtext, "<DEF>" );

	if( sbtext=="<DEF>" )
	{
	    done = True;
	    continue;
	}
	
	//
	//  Ok. parse the subband text to find out the settings for this subband
	//
	//  Expected format:
	//
	//  <double: startfreq>,<double: bandwidth>,<int: nrchannels>,<polarizations>,<sideband>
	//
	//  where <polarizations> is:
	//
	//  [xXyYrRlL:]* -> e.g.:   rr:lR:Lr or xr:rl:yy
	//
	//  and <sideband> is:
	//
	//  [USB|LSB]
	//
	nrelements = split( sbtext, elements, 5, "," );
	
	if( nrelements!=5 )
	{
	    cout << entryname.str() << " incorrectly specified in `" << key << "' in file `"
		 << itsInifilename << "'" << endl;
	    sbcnt++;
	    errcnt++;
	    continue;
	}

	//
	//  Ok. do the elements!
	//
	uInt               nrchan;
	Quantity           bandwidth;
	Quantity           startfreq;
	BaseBand::SideBand sideband;
	
	Quantity::read( startfreq, elements[0] );
	Quantity::read( bandwidth, elements[1] );
	nrchan = atoi( (const char*)elements[2] );
	sideband = BaseBand::stringToSideBand( elements[4] );
	

	//
	//  Create the baseband
	//
	BaseBand     bb( startfreq.get("Hz").getValue(), bandwidth.get("Hz").getValue(),
			 sideband );

	//
	//  Subband processed completely, add it to the frequency config!
	//
	fcref.addBaseBand( bb );

	//
	//  AND DONT FORGET TO INCREMENT THE SUBBANDCOUNT! D*&%head!
	//
	sbcnt++;
    }

    //
    //  If errcnt != 0 -> error(s) occurred. -> retval = False and
    //  Freq.Config becomes empty
    //
    if( errcnt )
    {
	retval = False;
	fcref = FrequencyConfig( key );

	cout << endl << "***********************************************" << endl;
	cout << "Error" << ((errcnt!=1)?("s"):("")) << " detected. Make sure format is like this:" << endl;
	cout << "subband<n>=<startfreq>,<bandwidth>,<nrchannels>,<polarizations>,<sideband>"
	     << endl << endl;
	cout << "where <polarizations> is: [xXyYrRlL:]* -> e.g.:   rr:lR:Lr or xr:rl:yy" << endl;
	cout << "and <sideband> is: [USB|LSB] " << endl << endl;
	cout << "***********************************************" << endl;
    }
    
    return retval;
}
