//
//  JIVEMSFiller - a MS filler, dedicated to deal with the situation @
//  JIVE
//
//  i.e. Must be able to cope with 'Bos-format (=raw binary dump)' +
//  COF (Correlator Output Format)
//
//  Author:   Harro Verkouter,   10-12-1998
//
//  $Id: JIVEMSFiller.h,v 1.2 2001/06/22 14:05:21 verkout Exp $
//
//
//
//
//  The JIVEMSFiller - object (attempts) to create global databases
//  for sources and antennas. It reads the 'J2MSDBDIR' environment
//  variable to decide where the 'global' databases live
//
#ifndef JIVEMSFILLER_H
#define JIVEMSFILLER_H

#ifndef AIPS_AIPS_H
#include <aips/aips.h>
#endif

#ifndef AIPS_STRING_H
#include <aips/Utilities/String.h>
#endif

#ifndef GLOBALDBS_H
#include <vlbi/j2ms/GlobalDBs.h>
#endif

#ifndef AIPS_MEASUREMENTSETS_H
#include <aips/MeasurementSets.h>
#endif

#ifndef AIPS_MATRIX_H
#include <aips/Arrays/Matrix.h>
#endif

#ifndef EXPERIMENTCACHE_H
#include <vlbi/j2ms/ExperimentCache.h>
#endif



//
//  Forward declarations
//
class Experiment;
class Visibility;
class FrequencyConfig;
class VisibilityBuffer;
class ClientConnection;

//
//
//  The Filler-class itself
//
//
class JIVEMSFiller
{
public:

    //
    //  Create a JIVEMSFiller. The string argument is optional and
    //  when not null, it specifies a client adress where the system
    //  will try to connect to. Error/diagnostic messages of the fill
    //  process are sent to this connection. For syntax of the
    //  clientconnection id, see libConnections
    //
    JIVEMSFiller( const char* logconnection=0 );

    //
    //  Do the job of adding the indicated scan to the MS of the
    //  indicated experiment. The boolean argument controls whether or
    //  not the user is asked to concatenate (if the MS already
    //  exists). true -> user is prompted; false -> data is
    //  concatenated blindly
    //
    Bool       fillDataToExperimentsMS( Experiment& expref, const String& visbufname, Bool ask=True );

    //
    //  Delete all the resources used by the JIVEMSFiller...
    //
    ~JIVEMSFiller();

private:
    //
    //  The filler's private parts
    //

    //
    //  The set of global databases:
    //
    static GlobalDBs          theirGlobalDBs;
    static ExperimentCache    theirCache;
    static ClientConnection*  myLC;

    //
    //  The private methods
    //

    //
    //  Attempt to create/open the MS for the indicated Experiment
    //
    MS*                     makeMS( const Experiment& expref, Bool ask );

    //
    //  This function makes sure the visibility data is in the domain
    //  requested by the experiment. 
    //  
    //  NOTE: this is ONLY for the datamatrix, nothing else!
    //
    //  NOTE: The contents of the matrix that is returned may or may
    //  not vary, since a reference to a static object is returned and
    //  the contents of the matrix may be changed after a call to
    //  makeData(). Therefore, copy the contents of the matrix to your
    //  own local variable to ensure proper operation.
    //
    static const Matrix<Complex>&  makeData( const Experiment& expref, const Visibility& vref );
    static const Matrix<Complex>&  makeDataOld( const Experiment& expref, const Visibility& vref );
    //
    //  Fill the MS with the data from the scan
    //
    Bool                           fillMSfromVisibilityBuffer( MeasurementSet& msref,
							       VisibilityBuffer& visbuf );

    //
    //  Write the indicated subband of the scan in the spectralwindow
    //  table of the given MS
    //
    Bool                           updateSpectralWindowTable( MeasurementSet& msref,
							      const j2ms::Scan& scanref,
							      const FrequencyConfig& fcref );

    //
    //  Write all antennas in the experiment to the antennatable in
    //  the indicated MS
    //
    Bool                           writeAntennaTable( MeasurementSet& msref,
						      const Experiment& expref );

    //
    //  Write entries in the FIELD and SOURCE table
    //
    Bool                           updateFieldTable( MeasurementSet& msref, const j2ms::Scan& scanref );
    
    //
    //  Calculate the UVW-value for the Scan/Visibility combination
    //  (NOTE: data is taken from the experiment/scan cache)
    //
    //  IMPORTANT NOTE: this function returns a reference to static
    //  Vector-object. The vector is modified everytime the 'calcUVW'
    //  method is invoked. Therefore, don't rely on the fact that the
    //  values found in the vector are always the same and if you want
    //  to keep the values, you should make a local copy (not the
    //  copy-c'tor, because that uses reference semantics. Great isn't
    //  it?! Marvellous!)
    //
    static const Vector<Double>&   calcUVW( const Visibility& vref );

    static void                    writeStatistics( MeasurementSet& ms,
						    const VisibilityBuffer& visbuf );

    //
    // For now, calculate the time of the visibility in the scan/frequencyconfig
    //
    /*
    const MEpoch&         calcVisibilityTime( const Scan& scanref, const Visibility& vref,
					      const FrequencyConfig& fcref );
					      */
    //
    //  Do not allow these object to be copied/assigned to!
    //
    JIVEMSFiller( const JIVEMSFiller& );
    JIVEMSFiller& operator=( const JIVEMSFiller& );
};



//
// In the interface we told the compiler that we have some static
// objects, now it's time to really declare them!
//
#ifdef JIVEMSFILLER_CC

GlobalDBs         JIVEMSFiller::theirGlobalDBs;
ExperimentCache   JIVEMSFiller::theirCache;
ClientConnection* JIVEMSFiller::myLC = 0;

#endif


#endif
