//
//  Scan - Abstract baseclass that defines the interface for accessing
//  the particular properties of a scan.
//
//  Make derived class as necessary...
//
//
//
//  NOTE:  Read through the class interface to find out which functions
//  *must* be implemented.
//
//  NOTE2: You can overload 'operator<<( ostream&, const <your class>&
//  )' -> merely showing info about the Scan in HRF, usually used for
//  displaying purposes.
//
//  NOTE3: There is no public c'tor -> this is to ensure that only
//  experiments can create valid Scan-objects.
//
//
//
//  Author:   Harro Verkouter,  7-12-1998
//
//  $Id: Scan.h,v 1.2 2001/06/22 14:05:23 verkout Exp $
//
#ifndef SCAN_H
#define SCAN_H


#ifndef AIPS_AIPS_H
#include <aips/aips.h>
#endif

#ifndef AIPS_STRING_H
#include <aips/Utilities/String.h>
#endif

#ifndef AIPS_MEPOCH_H
#include <aips/Measures/MEpoch.h>
#endif

#ifndef AIPS_MVEPOCH_H
#include <aips/Quanta/MVEpoch.h>
#endif

#ifndef CORRELATORSETUP_H
#include <vlbi/j2ms/CorrelatorSetup.h>
#endif


//
// Forward declarations -> we're just interested in the interface.
//
class ostream;
class Experiment;
class CorrelatorSetup;
class VisibilityBuffer;



//
//  This struct can be used as a utility; it holds all the basic
//  scanparameters. Use it to cache scan-information or pass it roud
//  or whatever you feel it fit for....
//
typedef struct _ScanInfo
{
    String          scanId;
    String          sourceCode;
    String          frequencyCode;
    MVEpoch         scanStart;
    MVEpoch         scanEnd;
    CorrelatorSetup correlatorSetup;
} ScanInfo;


//
// Let's put the scan-object in its own namespace....
//
namespace j2ms
{

//
//  The scan-class itself
//
class Scan
{
    //
    //  The Experiment class is our friend: only an experiment can
    //  create a fully valid scan; users should never be able to
    //  create a scan
    //
    //friend class Experiment;

public:
    //
    //  (RO)Access to properties of this scan.
    //
    //  For successfull operation, derived classes must implement
    //  these functions
    //

    //
    //  Textual representation of the scanId
    //
    virtual const String&     getScanId( void ) const = 0;

    //
    //  What source were we looking at?
    //
    //  NOTE: This is the code that the system uses for looking up the
    //  source in the database.
    //
    virtual const String&     getSourceCode( void ) const = 0;
    
    //
    //  What was the frequency setup?
    //
    //  NOTE: This is the code the system uses for looking up the
    //  frequency setup in the databases
    //
    virtual const String&     getFrequencyCode( void ) const = 0;

    //
    //  What was the Correlator setup used when correlating this scan?
    //  (info like: nr of lags, which polarizations were formed)
    //
    virtual const CorrelatorSetup& getCorrelatorSetup( void ) const = 0;

    //
    //  Return a fully defined Epoch, telling the system at what time
    //  the scan started...
    //
    virtual const MEpoch&     getStartTime( void ) const = 0;

    //
    //  The scan should have a visibility-buffer where the system can
    //  read visibilities from!
    //
    //  Create a visibility bufferclass, specific to one's own
    //  dataformat and translate your native visibilities into
    //  'standard' (i.e. within this system) Visibilities.
    //
    const VisibilityBuffer&   getVisibilityBuffer( void ) const;

    //
    //  Be able to retrieve experiment info via this:
    //
    //  NOTE: This is NOT a virtual function; you're not supposed to
    //  implement it. The Experiment will be stored in this class and
    //  is a required creator parameter!
    //
    const Experiment&         getExperiment( void ) const;
    
    //
    //  Delete the scan-resources
    //
    virtual ~Scan();

protected:
    //
    //  Create a Scan, telling the Scan what experiment it belongs to.
    //
    //  NOTE: If pyou pass a NULL-Experiment-ptr, an
    //  AipsError-exception is thrown!
    //
    Scan( const VisibilityBuffer* visbufptr );
    
    
private:
    //
    //  The scan's private parts:
    //

    //
    // A scan is connected to a 'parent' experiment. We keep a pointer
    // to a const Experiment, since we (for now) want to disable the
    // fact that a scan can influence its parent Experiment.
    //
    const VisibilityBuffer* itsVisibilityBufferptr;

    //
    //  Private methods (if any):
    //
    

    //
    //  Empty scans don't exist?
    //
    Scan();
    Scan( const Scan& );
    Scan& operator=( const Scan& );
};

//
// End of namespace
//
};

//
//  Show summary of the scan on a stream
//
ostream& operator<<( ostream& os, const j2ms::Scan& sref );


#endif
