//
//  Implementation of the VEXAntennaDB - class
//
//  Author: Harro Verkouter,   6-7-1999
//
//  $Id: VEXAntennaDB.cc,v 1.2 2000/03/29 09:57:43 verkout Exp $
//
//  $Log: VEXAntennaDB.cc,v $
//  Revision 1.2  2000/03/29 09:57:43  verkout
//  HV: Added extra antenna info like mount and axisoffset. Mounts are now passed on from VEX-file to FITSfile ok..(i hope..)
//
//  Revision 1.1.1.1  2000/03/20 15:14:36  verkout
//  HV: imported aips++ implement stuff
//
//
#ifndef VEXANTENNADB_H
#include <vlbi/j2ms/VEXAntennaDB.h>
#endif

#ifndef VEXPLUS_H
#include <prep_job/vexplus.h>
#endif

#ifndef AIPS_VECTOR_H
#include <aips/Arrays/Vector.h>
#endif

#include <iostream.h>
#include <iomanip.h>


VEXAntennaDB::VEXAntennaDB( const String& vexfilename ) :
    AntennaDB(),
    myVexFileName( vexfilename )
{
    VexPlus     vexfile( myVexFileName );
    static bool filenameShown( false );    

    filenameShown = false;

    //
    //  Attempt to open/parse and read VEX-file!
    //
    if( !vexfile.parseVex() )
    {
	char*          stationptr;
	char*          siteptr;
	char*          tlcptr;
	char*          mountptr;
	String         station;
	String         site;
	String         tlc;
	String         mount;
	MPosition      position;
	const char*    defsite( "<UNKNOWN>" );
	const char*    defmount( "none" );
	
	Vector<Double> pos( 3 );
	Vector<Double> off( 3 );

	for( int i=0; i<vexfile.N_Stations(); i++ )
	{
	    stationptr = vexfile.Station( i );
	    siteptr    = vexfile.Site( stationptr );
	    tlcptr     = vexfile.TwoLetter( stationptr );
	    mountptr   = vexfile.AxisMount( stationptr );
	    
	    if( !stationptr || !siteptr || !tlcptr || !mountptr )
	    {
		if( !filenameShown )
		{
		    cout << "Error in file " << myVexFileName << "\n";
		    filenameShown = true;
		}

		cout << "Antenna #" << setw(2) << i << ":\n";
		
		if( !stationptr )
		{
		    cout << "> Missing stationname\n";
		}
		if( !siteptr )
		{
		    siteptr = defsite;
		    cout << "> Missing sitename, defaulting to " << siteptr << "\n";
		}
		if( !tlcptr )
		{
		    tlcptr = stationptr;
		    cout << "> Missing two-letter-code, defaulting to "
			 << ((tlcptr!=0)?(tlcptr):("<NULL>")) << "\n";
		}
		if( !mountptr )
		{
		    mountptr = defmount;
		    cout << "> Missing mount, defaulting to " << mountptr << "\n";
		}
		
		cout << flush;
		
		//
		//  If the stationname is unknown, then we're in trouble!
		//
		if( !stationptr )
		{
		    continue;
		}
	    }
	    
	    //
	    //  Transform from char* -> String
	    //  because the automatic conversion does not seem to work...?!
	    //  when calling
	    //
	    //  AntennaInfo( stationptr, tlcptr, siteptr, position),
	    //  the compiler complains about not being able to convert
	    //  char*& to String? NOTE: the c'tor of AntennaInfo takes
	    //  three times 'const char*' + a MPosition&...
	    //
	    station = stationptr;
	    site    = siteptr;
	    tlc     = tlcptr;
	    mount   = mountptr;

	    //
	    //  Transfrom Vex-style mount defns to AIPS++ (and in the
	    //  end FITS)-style mount-strings
	    //
	    if( mount=="az_el" )
	    {
		mount = "alt-az";
	    }
	    else if( mount=="ha_dec" )
	    {
		mount = "equatorial";
	    }
	    else
	    {
		cout << "VEXAntennaDB: Mount type " << mount << " not recognized\n"
		     << "              It is likely it will show up incorrectly\n"
		     << "              in any exported dataformat (e.g. FITS)\n";
	    }
	    

	    pos( 0 ) = vexfile.SiteX( station );
	    pos( 1 ) = vexfile.SiteY( station );
	    pos( 2 ) = vexfile.SiteZ( station );	    

	    off = 0.0;
	    off( 0 ) = vexfile.AxisOffset( station );
	    
	    position = MPosition( pos, MPosition::ITRF );

	    myAntennaList.add( AntennaInfo(station, tlc, site, position, 0.0, mount, off) );
	}
    }
    //
    //  Done!
    //
}

//
//  Look up stuff!
//
AntennaInfo VEXAntennaDB::operator[]( const String& key ) const
{
    AntennaInfo   retval;
    
    for( uint32 i=0; i<myAntennaList.elements(); i++ )
    {
	if( myAntennaList(i).getName()==key )
	{
	    retval = myAntennaList(i);
	    break;
	}
    }
    return retval;
}

Bool VEXAntennaDB::searchAntenna( AntennaInfo& airef, const String& key ) const
{
    Bool   retval( True );
    
    airef = (*this)[ key ];
    
    if( airef.getName()=="" )
    {
	retval = False;
    }
    return retval;
}

//
//  Clean up...
//
VEXAntennaDB::~VEXAntennaDB()
{
}
