//
//  Visibility - a class that describes a visibility: it holds a label
//  (a VisLabel) that holds the logical indices telling us the origin
//  of the visibility, PLUS a matrix holding the visibility data (the
//  shape of the matrix is nr. correlations. times nr of spectral
//  channels)
//
//
//  Author:   Harro Verkouter, 18-11-1998
//
//
//  $Id: Visibility.h,v 1.1.1.1 2000/03/20 15:14:35 verkout Exp $
//
#ifndef VISIBILITY_H
#define VISIBILITY_H


#ifndef VISLABEL_H
#include <vlbi/DtdLUT/VisLabel.h>
#endif

#ifndef AIPS_MATRIX_H
#include <aips/Arrays/Matrix.h>
#endif

#ifndef AIPS_VECTOR_H
#include <aips/Arrays/Vector.h>
#endif

#ifndef AIPS_MEASURES_H
#include <aips/Measures.h>
#endif

#ifndef AIPS_MEPOCH_H
#include <aips/Measures/MEpoch.h>
#endif

#ifndef AIPS_MUVW_H
#include <aips/Measures/Muvw.h>
#endif

//
//  Forward declarations
//
class ostream;
class VisibilityBuffer;


class Visibility
{
public:
    //
    //  We can tell in what domain the visibility data is:
    //
    typedef enum _domain
    {
	undefined, frequency, time
    } domain;
    
    //
    //  An empty visibility
    //
    Visibility();

    //
    //  A defined visibility.
    //
    //  NOTE: a visibility MAY have a buffer attached, bust is not
    //  obliged to have one!
    //
    Visibility( const ExtVisLabel& label, const Matrix<Complex>& data,
		Visibility::domain datadomain=Visibility::undefined,
		Bool flag=False, const Vector<Double>& weight=Vector<Double>(),
		const MEpoch& time=MEpoch(), const Muvw& uvw=Muvw(),
		const VisibilityBuffer* buffer=0 );
    
    //
    //  Copy c'tor
    //
    Visibility( const Visibility& other );
    
    //
    //  assignment
    //
    Visibility& operator=( const Visibility& other );

    //
    //  Get the label/data
    //
    const ExtVisLabel&     getLabel( void ) const;
    const Matrix<Complex>& getData( void ) const;

    //
    //  Is this an autocorrelation?
    //
    Bool                   isAutoCorrelation( void ) const;

    Visibility::domain     getDatadomain( void ) const;

    //
    //  Get the weight for the whole visibility and if the visibility
    //  was flagged.
    //
    //  Defaults are:
    //
    //  weight = 1.0
    //  flag   = false (i.e. not flagged)
    //
    Bool                    getFlag( void ) const;
    const Vector<Double>&   getWeight( void ) const;

    
    //
    //  Return the time of the visibility
    //
    const MEpoch&           getTime( void ) const;
    
    //
    //  And an UVW-value
    //
    const Muvw&             getUVW( void ) const;
    
    //
    //  Tells us if we have a buffer where we came from + access to it
    //
    Bool                    hasVisibilityBuffer( void ) const;
    const VisibilityBuffer& getVisibilityBuffer( void ) const;

    //
    //  Clear all resources
    //
    ~Visibility();
    
private:
    //
    //  The visibilities private parts
    //
    Bool               itsFlag;
    Muvw               itsUVW;
    MEpoch             itsTime;
    ExtVisLabel        itsLabel;
    Vector<Double>     itsWeight;
    Matrix<Complex>    itsData;
    Visibility::domain itsDatadomain;
    
    //
    //  Optional pointer to buffer where we came from...
    //
    const VisibilityBuffer*  itsBufferptr;
    
    //
    // Disallow these:
    //

};

//
//  Output function!
//
ostream& operator<<( ostream& os, const Visibility& visref );
ostream& operator<<( ostream& os, const Visibility::domain& domainref );

#endif
