//
//  VisibilityBuffer - an abstract baseclass that defines the
//  interface for classes that would like to act as a visibility
//  buffer (i.e. deliverers of Visibilities!)
//
//
//
#ifndef VISIBILITYBUFFER_H
#define VISIBILITYBUFFER_H

#ifndef AIPS_AIPS_H
#include <aips/aips.h>
#endif

#ifndef AIPSEXCEPTIONOBJECT_H
#include <vlbi/Utilities/AIPSExceptionObject.h>
#endif

#ifndef EXPERIMENT_H
#include <vlbi/j2ms/Experiment.h>
#endif

//
//  Forward declarations
//
namespace j2ms
{
	class Scan;
}
class Visibility;

class VisibilityBuffer
{
public:

    //
    //  Visibility buffers must be attached to an Experiment
    //
    VisibilityBuffer( const Experiment* experiment ) :
	itsScanptr( 0 ),
	itsExperimentptr( experiment )
    {
	//
	//  If no experiment -> we give up!
	//
	if( !itsExperimentptr )
	{
	    throw( AipsError("VisibilityBuffer: No experiment given!") );
	}
    }
    
    //
    //  We MUST have an experiment
    //
    const Experiment&  getExperiment( void ) const
    {
	return *itsExperimentptr;
    }
    

    //
    //  Tell the current scan (if any)
    //
    const j2ms::Scan*         getScanptr( void ) const
    {
	return itsScanptr;
    }

    //
    //  Tells us about vernier delay
    //
    Bool               getVernierDelay( void ) const
    {
	return itsExperimentptr->hasVernierDelay();
    }
    
    //
    //  The VisibilityBuffer must at least implement these functions:
    //

    //
    //  retrieve the current visibility, do NOT increment/move to the
    //  next. NOTE: it changes the argument to be a COPY of the
    //  visibility so the user can mess around with it! Returns True
    //  if the argument 'vis' has been filled in.
    //
    //virtual Bool          getCurrentVisibility( Visibility& vis ) const = 0;
    virtual Bool          getCurrentVisibility( Visibility& vis ) = 0;
    
    //
    //  Tell buffer to prepare to get next visibility. NOTE: may also
    //  be used as a check to see if a next visibility is available.
    //
    virtual Bool          next( void ) = 0;

    //
    //  Tell the directory where the datafiles for this
    //  visibilitybuffer actually live!
    //
    virtual const String& getRootDir( void ) const = 0;

    //
    //  Destruct the buffer
    //
    virtual ~VisibilityBuffer()
    {
    }

protected:
    //
    //  Derived classes may set the current scanptr to point at the
    //  scan where the current visibility is attached to
    //
    //  NOTE: the pointer is NOT owned; the usr. is still responsible
    //  for deleting it!
    //
    void     setCurrentScanptr( const j2ms::Scan* scanptr )
    {
	itsScanptr = scanptr;
    }
    
    
private:

    //
    //  If we are attached to a scan, this ptr points at it. Otherwise
    //  we just don't know!
    //
    const j2ms::Scan*   itsScanptr;
    const Experiment*   itsExperimentptr;
    
    //
    //  Prohibit these
    //

    VisibilityBuffer( const VisibilityBuffer& );
    VisibilityBuffer& operator=( const VisibilityBuffer& );
};



#endif
