//
//  class:
//   
//          ms2uvfitsConverter
//
//  Purpose:
//
//          Convert Measurementsets to UVFITS format. This class
//          derives from FileFormatConverter, since the
//          ms2uvfitsConversion is a fileformat-conversion.
//
//  Use:
//
//          The only public interface functions are the default ctor,
//          dtor and the actual conversion function, 'convert()'.
//
//          The convert method:
//
//          Bool  convert( const String& fitsfilename, const String& MSfilename )
//
//          As can be seen, the function just needs the filename of
//          the fitsfile and the MS. Everything else will be covered
//          by the converter-object! Check the comments in the
//          convert-method, located in ms2uvfitsConverter.cc
//
//
//
//    Harro Verkouter, 11-3-1998
//
//  $Id: ms2uvfitsConverter.h,v 1.1.1.1 2000/03/20 15:16:48 verkout Exp $
//
#ifndef MS2UVFITSCONVERTER_H_INC
#define MS2UVFITSCONVERTER_H_INC


//
//  Aips++ includes
//
#ifndef AIPS_VECTOR_H
#include <aips/Arrays/Vector.h>
#endif

#ifndef AIPS_MEASUREMENTSET_H
#include <aips/MeasurementSets/MeasurementSet.h>
#endif

//
//  Need this for the base-class declaration
//
#include <vlbi/ms2uvfitsimpl/Converters/FileFormatConverter.h>


//
//  Forward declaration(s)
//
class hduGenerator;
class MSConversionInfo;
class ms2uvfitsConverter;
template <class T> class Block;
template <class T> class LinearContainer;
template <class K, class V> class SimpleOrderedMap;

//
//  Introduce this typedef as shorthand...
//
typedef ms2uvfitsConverter MS2UVF;

//
//  And this one as the type of functionptr we expect for fragmenting a MS
//
typedef MSConversionInfo** (*fragmentfn)( const MS& msref, unsigned int& npieces );



class ms2uvfitsConverter :
    public FileFormatConverter
{

public:
    //
    //  We keep an enumeration which describes all available
    //  hdugenerators, so the user can decide which one(s) he/she
    //  would like to use and in what order!
    //
    typedef enum _generator_t 
    {
	nogenerator, mainprimarygroup, antennabinary, antennaascii, freqbinary,
	    gaincurvebinary, tsysbinary, sourcebinary, idiheader, uvtable, arraygeometry,
	    idisource, idifrequency, idiantenna
    } generator_t;

    //
    //  Default ctor: this is the only thing needed.
    //
    ms2uvfitsConverter();
    
    //
    //  Do convert the MS into a UVFITS-file
    //
    Bool  convert( const String& fitsfilename, const String& MSfilename ) const;

    //
    //  Replace the current generators by those listed in the
    //  block. The conversion class will set up the engine in such a
    //  way that when the convert method is run, the (eventually)
    //  produced FITS-file will contain those HDU's U specified in the
    //  'hdus' parameter (and in the same order).
    //
    //Bool  setConversion( const Block<generator_t>& hdus );
    Bool   setConversion( const LinearContainer<generator_t>& hdus );

    Bool   setFragmenter( fragmentfn fnptr );

    //
    //  Destruct the conversionengine
    //
    virtual ~ms2uvfitsConverter();
    
private:
    //
    //  The objects private parts
    //

    //
    //  Here we keep a list of ptrs to generators, who generate a HDU
    //  from a MS. Since the translation from MS to UVFITS cannot be
    //  guaranteed to be lossless, we will have to find out some
    //  global info about the MS. The MSConversionInfo object
    //  generates this info (the MSConversionInfo object is passed to
    //  the generator).
    //
    Vector<hduGenerator*>   itshduGenerators;
    
    Vector<hduGenerator*>   itsAvailableGens;

    //
    //  This is the map that maps from enum to pointer to generator!
    //
    SimpleOrderedMap<ms2uvfitsConverter::_generator_t,hduGenerator*>*  itsEnumToGeneratorMapptr;

    //
    //  The current fragmentation function
    //
    fragmentfn              itsFragmentationFn;
    

    //*****************************************************************
    //              The object's private methods...
    //*****************************************************************
    

    //
    //  This static method takes a measurement set and fagments it
    //  into a number of pieces. It returns an array of pointers to
    //  MSConversionInfo objects and the number of elements is
    //  returned in the parameter npieces
    //
    //  NOTE: The caller is responsible for deleting the arrayelements
    //  AS well as the array itself. The array is allocated by a call
    //  to new[]!
    //
    static MSConversionInfo**  fragmentMS( const String& msfilename, uInt& npieces );

    //
    //  Prevent copying and assignment
    //
    ms2uvfitsConverter( const ms2uvfitsConverter& );
    ms2uvfitsConverter& operator=( const ms2uvfitsConverter& );
};

ostream& operator<<( ostream& os, const ms2uvfitsConverter::generator_t& genref );


#endif
