//
//  Implementation of the uvfits2msConverter-class. See
//  uvfits2msConverter.h for general details. Funcon specific comments
//  can be found in this file!
//
//
//  Author:
//
//    Harro Verkouter     23-3-1998
//
//    $Id: uvfits2msConverter.cc,v 1.1.1.1 2000/03/20 15:16:48 verkout Exp $
//
#ifndef UVFITS2MSCONVERTER_H_INC
#include <vlbi/ms2uvfitsimpl/Converters/uvfits2msConverter.h>
#endif

#ifndef AIPS_AIPS_H
#include <aips/aips.h>
#endif

#ifndef TABLEGENERATOR_H_INC
#include <vlbi/ms2uvfitsimpl/Generators/tableGenerator.h>
#endif

//#ifndef AIPS_EXCEPTIONS_H
//#include <aips/Exceptions.h>
//#endif
#ifndef EXCEPTIONOBJECT_H
#include <vlbi/Utilities/AIPSExceptionObject.h>
#endif

#ifndef AIPS_MEASUREMENTSET_H
#include <aips/MeasurementSets/MeasurementSet.h>
#endif

#ifndef FITSCONVERSIONINFO_H_INC
#include <vlbi/ms2uvfitsimpl/FITSConversionInfo.h>
#endif

#ifndef AIPS_TABLES_H
#include <aips/Tables.h>
#endif


//
//  Default c'tor -> nothing to do yet
//
uvfits2msConverter::uvfits2msConverter()
{
    //
    //  Ok. Make a list of tables we want to be able to extract from
    //  the FITSConverionInfo object!
    //
    itstableGenerators.resize(3);
    
    //
    //  Set all entries to null (to start  with)
    //
    itstableGenerators=(tableGenerator*)0;
    
    //
    //  Now fill in those tables we want to generate!
    //
    itstableGenerators(0)=new tableGenerator;
}

//
// The actual conversion function!
//
Bool uvfits2msConverter::convert( const String& msfilename, const String& fitsfilename ) const
{
    Bool   retval( False );
    
    cout << "Start converting fitsfile `" << fitsfilename << "' to MS `" << msfilename << "'" << endl;
 
    try
    {
        //
        //  Attempt to creaste the MS etc.
        //
        MeasurementSet*       themsptr( 0 );
        FITSConversionInfo    info( fitsfilename );
        
        //
        //  Step one: we need to create a MeasurementSet!  Return
        //  value is either a valid ptr or an exception is thrown!!!
        //
        themsptr=setupMS( msfilename );
        

        //
        // Loop over all generators!
        //
        Int    nrgenerators;
        
        itstableGenerators.shape( nrgenerators );
        
        for( Int i=0; i<nrgenerators; i++ )
        {
            //
            //  Check if we have a generator at all....
            //
            if( itstableGenerators(i) )
            {
                Table*   tableptr;

                tableptr=itstableGenerators(i)->generate( info );

                //
                //  Depending on what table-type we returned, we have
                //  to bind the columns of a table in our just created
                //  MS to the columns of the table that has been
                //  generated? The table object that is returned is
                //  already filled in! Maybe it is sufficient to copy
                //  everything across?
                //

                delete tableptr;
            }
            else
            {
                cout << "Darn! No tablegenerator for table #" << i+1 << ". Skipping it." << endl;
            }
        }
        retval=True;

        //
        //  We don't need these anymore!
        //
        delete themsptr;
        
    }
    catch( AIPSExceptionObject x )
    {
        cout << "Huh? " << x.getMesg() << endl;
    }
    end_try;
    
    return retval;
}



//
//  Destruct the conversion engine
//
uvfits2msConverter::~uvfits2msConverter()
{
    //
    //  That means: delete all generators!
    //
    Int    nrgenerators;
    
    itstableGenerators.shape( nrgenerators );
    
    for( Int i=0; i<nrgenerators; i++ )
    {
        delete itstableGenerators(i);
    }
    
    //
    //  And we're done!
    //
}


//********************************************************************************/
//
//              UVFITS2MS conversion private routines!
//
//********************************************************************************/


//
//  Like: setup a new MeasurementSet. This is the MeasurementSet that
//  will be written into!
//
//
//  Return: ptr to newly created MS. If this fails, an exception is
//  thrown.
//
MeasurementSet*  uvfits2msConverter::setupMS( const String& msfilename ) const
{
    MeasurementSet*    retval( 0 );
    
    if( !msfilename.length() )
    {
        throw( AipsError("No MS outputfilename defined...") );
        return retval;
    }
    
    //
    //  First we follow the method of TMS/NFRA, see MSDefiner{.cc,.h}
    //
    TableDesc   td=MS::requiredTableDesc();
    
    //
    //  We need to add a data column to the tabledescriptor
    //
    td.addColumn( ArrayColumnDesc<Complex>( MS::columnName(MS::DATA), 2) );

    //
    //  Set the unit of the column
    //
    td.rwColumnDesc(MS::columnName(MS::DATA)).rwKeywordSet().define("UNIT","Jy");
    
    //
    //  Replace the flag-column with a varibale shaped array
    //
    td.removeColumn(MS::columnName(MS::FLAG));
    MS::addColumnToDesc(td, MS::FLAG, 2);

    //
    //  Same goes for the sigma column
    //
    td.removeColumn(MS::columnName(MS::SIGMA));
    MS::addColumnToDesc(td, MS::SIGMA, 1);

    //
    // Add measure reference keyword to time column descriptor
    //
    String    key("MEASURE_REFERENCE");

    td.rwColumnDesc(MS::columnName(MS::TIME)).rwKeywordSet().define(key,"UTC");
    td.rwColumnDesc(MS::columnName(MS::TIME)).rwKeywordSet().define("MEASURE_VAL","124573.456");


    //
    //  Ok. Table descriptor set up.
    //
    SetupNewTable newtab( msfilename, td, Table::New );

    //
    //  Now we need some storage managers as well...
    //
    StManAipsIO      aipsStMan;
    IncrementalStMan incrStMan( "ISMData" );

    //
    // Default: bind all columns to the incremental storagemanager
    //
    newtab.bindAll( incrStMan );

    //
    //  Only the data column will be bound to the aipsIO storage manager
    //
    newtab.bindColumn( MS::columnName(MS::DATA), aipsStMan );
    newtab.bindColumn( MS::columnName(MS::ANTENNA1), aipsStMan );
    newtab.bindColumn( MS::columnName(MS::ANTENNA2), aipsStMan );

    
    //
    //  Finally: we may attempt to create the MS!!!
    //
    retval = new MeasurementSet( newtab );

    if( !retval )
    {
        //
        //  Somehow, the creation did not succeed and no exception was
        //  thrown?!
        //
        throw( AipsError("Failed to create MS `"+msfilename+"'") );
    }
    

    //
    // For now, all that's left is to create all the default
    // subtables!
    //
    retval->createDefaultSubtables( Table::New );
    
    return retval;
}

