//
//  Class:
//
//        FITSColumn
//
//
//  Purpose:
//
//        This class will describe the properties of a
//        column in a FITS table. These properties encompass the
//        datatype ( ..FORM fields in the FITSdescription), the
//        description of the column and the width and the offset of
//        the column within a row in the FITS table.
//
//        I choose not to have this column templated bute use the
//        AIPS++ DataType enumeration instead, to prevent us from
//        having to upcast to templated type.
//
//
//
//
//  Author:
//
//        Harro Verkouter     27-3-1998
//
//
//      $Id: FITSColumn.h,v 1.1.1.1 2000/03/20 15:16:48 verkout Exp $
//
#ifndef FITSCOLUMN_H_INC
#define FITSCOLUMN_H_INC


#ifndef BASICFITSUTILITY_H_INC
#include <vlbi/ms2uvfitsimpl/FITSUtils/BasicFITSUtility.h>
#endif

#ifndef AIPS_AIPS_H
#include <aips/aips.h>
#endif

#ifndef AIPS_STRING_H
#include <aips/Utilities/String.h>
#endif

#ifndef AIPS_FITS
#include <aips/FITS/fits.h>
#endif

#ifndef AIPS_SCALARCOLUMN_H
#include <aips/Tables/ScalarColumn.h>
#endif

#ifndef AIPS_ARRAYCOLUMN_H
#include <aips/Tables/ArrayColumn.h>
#endif

#ifndef AIPS_IPOSITION_H
#include <aips/Lattices/IPosition.h>
#endif

#ifndef AIPS_SLICER_H
#include <aips/Lattices/Slicer.h>
#endif

#ifndef AIPS_DATATYPE_H
#include <aips/Utilities/DataType.h>
#endif

#ifndef AIPS_ARRAY_H
#include <aips/Arrays/Array.h>
#endif

#ifndef AIPS_RECORDINTERFACE_H
#include <aips/Containers/RecordInterface.h>
#endif

#ifndef AIPS_RECORDDESC_H
#include <aips/Containers/RecordDesc.h>
#endif


//
// Forward declaration(s)
//
class FITSTableHDU;
class RecordFieldId;


class FITSColumn :
    public BasicFITSUtility
{
public:
    //
    // Return the AIPS++ datatype that describes the nature of this
    // column
    //
    DataType                getColumnDataType( void ) const;
    RecordDesc              getColumnDescription( void ) const;

    //
    //  Translate this Column to Fitskeywords
    //
    virtual FitsKeywordList asFITSKeys( void ) const;

    //
    //  Get or put the value in a cell. Derived columns MUST implement these methods!
    //
    virtual Bool            get( uInt rownr, RecordInterface& recordref, const RecordFieldId& fieldidx ) const =0;
    virtual Bool            put( uInt rownr, const RecordInterface& recordref, const RecordFieldId& fieldidx ) =0;

    
    //
    //  Let the user have access (mostly read-only) to some of the
    //  private parts!
    //

    //
    // The width of this particular column (in bytes)
    //
    virtual uLong           getWidthInBytes( void ) const;
    

    //
    //  Accessors, use FITS nomenclature:
    //
    const String&           ttype( void ) const;
    const String&           tform( void ) const;
    const String&           tunit( void ) const;

    //
    //  Some administrative functions, I don't think users will be
    //  much using these, but anyway.
    //
    Int                     columnnr( void ) const;
    uLong                   columnoffset( void ) const;

    //
    //  Do we happen to be a binary column, by any chance??
    //
    Bool                    isBinaryColumn( void ) const;

    //
    //  Destruct this Column
    //
    virtual ~FITSColumn();

protected:
    //
    //  Three constructors. We make all of them protected, so only
    //  derived classes (actual specialization of the column class)
    //  are able to construct a column object.
    //
    //  1) Create a complete column -> user has control over all
    //  Column properties (but the Column number and the offset)
    //
    //  2) Create a column from a FitsKeywordlist -> intended use:
    //  attempt to read the column's settings from the list of
    //  keywords just read from file! The parent HDU tells us what our
    //  ordinal number and offset are.
    //
    //  3) A kind of copy-constructor -> the column settings of
    //  'other' are copied across apart from the columnnumber and
    //  offset. This constructor finds its use when a HeaderDataUnit
    //  needs to add a column. The user specifies a column that needs
    //  to be added to the HeaderdataUnit. The user does not know the
    //  actual number of the column. The HeaderdataUnit does know and
    //  uses this c'tor to create the correct column.
    //
    //
    //  NOTE: the number parameter (as well as the private datamember
    //  'itsColumnnumber' are Ints, because the value '-1' is used to
    //  indicate that the column has not been assigned to a valid HDU!
    //
    FITSColumn( const String& type, const String& comment, const String& form, const String& unit );
    
    FITSColumn( Int number, uLong offset, FitsKeywordList& kwsource, FITSTableHDU* ptr2parent );
    
    FITSColumn( Int number, uLong offset, const FITSColumn& other, FITSTableHDU* ptr2parent );
    
    //
    //  The derived classes must have acces to these members: they
    //  first have to parse the TFORM field and hence they can decide
    //  what column type/shape datatype etc. it is!
    //
    Bool      setDataType( const DataType& newdatatype );
    Bool      setColumnDescription( const RecordDesc& newdescription );
    Bool      setForm( const String& newform );

    //
    //  Sometimes (e.g. for an ASCII-column) we may have to change the
    //  offset for this column...
    //
    void      setColumnOffset( uLong newoffset );
    

protected:

    //
    //  Yuck! Do I not like this.... derived classes (and friends for
    //  that matter) can use this method to tell this Column
    //  Object that is really is a BinaryColumn... but.. since we must
    //  make decisions for keywords in another object (the
    //  BinaryTable) we must be able to find out if there are Matrix
    //  columns present and if so, we must even know which one it is!
    //
    //  Wait... you haven't heard the best part yet... it's not even
    //  part of the standard, these so-called matrix columns! They're
    //  an invention of the the Classic AIPS team... and since we
    //  *really* need classic AIPS we'll need to find a way to get our
    //  data through into classic AIPS...
    // 
    void        setBinaryColumnFlag( void );

private:
    //
    //  These are the variables that describe the Column
    //

    //
    //  The Column type
    //
    String         itsColumnType;
    
    //
    //  The comment
    //
    String         itsComment;
    
    //
    // This string describes the form of the column, according to
    // form= element_count+BASCIC_TYPE_CHARACTER,
    //
    //  e.g.  a column of single precision floats would have the
    //  following form: "1E"
    //
    //  a column with an array of 10 characters would have the form:
    //  "10A"
    //
    String         itsForm;
    
    //
    //  The unit of the data in the column
    //
    String         itsUnit;
    

    //
    //  If appropriate: which Columnnumber is this?
    //
    Int            itsColumnnumber;

    //
    //  For that matter, we keep the current datatype in this variable
    //  (this is the AIPS++ DataType enum). This enum also holds
    //  arraytypes etc. just what we need!
    //
    DataType       columnDataType;

    //
    //  How many data-type elements does this column take (form=
    //  <nrelements><datatype>)
    //
    Int            itsNumberElements;
    
    //
    //  We also need to have the total column width and the column
    //  offset in the row of the table (if known of course: this is
    //  only valid if the column is attached to a HeaderDataUnit)!
    //

    //
    //  This is the column width in bytes
    //
    uInt           itsTotalWidth;
    
    //
    //  The offset of this column in the row (also in bytes)
    //
    uLong          itsOffsetInRow;

    //
    //  The record description that describes this column in AIPS++
    //  terminology (like a combination of DataType and IPosition). It
    //  will be filled in by the 'CheckFormatAndInit()' private member
    //  function.
    //
    RecordDesc     itsColumnDescription;

    //
    //  The flag that tells us if we are a matrix column or not..
    //
    Bool           myBinaryColumnFlag;


    //
    //  The Column's private methods
    //

    
    //
    //  We do not allow copying and assigning to Columns, nor do we
    //  allow default Columns!
    //
    FITSColumn();
    FITSColumn( const FITSColumn& );
    FITSColumn& operator=( const FITSColumn& );
    
};

//
//  Display the columns properties on a stream....
//
ostream& operator<<( ostream& osref, const FITSColumn& columnref );

#endif
