//
//  Implementation of the FITSMatrixAxis class...
// 
//  Author: Harro Verkouter,  13-09-1999
//
//  $Id: FITSMatrixAxis.cc,v 1.1.1.1 2000/03/20 15:16:48 verkout Exp $
//
//  $Log: FITSMatrixAxis.cc,v $
//  Revision 1.1.1.1  2000/03/20 15:16:48  verkout
//  HV: Imported aips++ implement-stuff
//
//
#ifndef FITSMATRIXAXIS_H
#include <vlbi/ms2uvfitsimpl/FITSUtils/FITSMatrixAxis.h>
#endif


#ifndef FITSMATRIXCOLUMN_H
#include <vlbi/ms2uvfitsimpl/FITSUtils/FITSMatrixColumn.h>
#endif

#include <iomanip.h>
#include <strstream.h>


ostream& operator<<( ostream& osref, const FITSMatrixAxis& axisref )
{
    osref << "MatrixAxis #" << setw(2) << axisref.axisnr() << " " 
          << axisref.ctype() << ": " 
          << axisref.naxis() << "\n"
          << "CRPIX=" << axisref.crpix() << "\n"
          << "CRVAL=" << axisref.crval() << "\n"
          << "CDELT=" << axisref.cdelt() << "\n"
          << "CROTA=" << axisref.crota() << endl;
    
    return osref;
}


//
//  First: the constructor with all the stuff in it!
//
//  Typically used by user's that want to add axes to a HDU.
//
FITSMatrixAxis::FITSMatrixAxis( const String& type, const String& comment, Int length,
				Double refpixel, Double refvalue, Double increment,
				Double rotangle ) :
    BasicFITSUtility(),
    itsAxistype( type ),
    itsComment( comment ),
    itsAxislength( length ),
    itsReferencepixel( refpixel ),
    itsReferencevalue( refvalue ),
    itsIncrement( increment ),
    itsRotationangle( rotangle ),
    itsAxisnumber( -1 )
{
    //
    //  Nothing left to do...
    //
}


FITSMatrixAxis::FITSMatrixAxis( Int axisnr, FitsKeywordList& kwsource,
				FITSMatrixColumn* const parent ) :
    BasicFITSUtility( (FITSBinaryColumn* const)parent ),
    itsAxistype( "" ),
    itsComment( "" ),
    itsAxislength( 1 ),
    itsReferencepixel( 1.0 ),
    itsReferencevalue( 0.0 ),
    itsIncrement( 1.0 ),
    itsRotationangle( 0.0 ),
    itsAxisnumber( axisnr )
{
    char           strmbuf[ 128 ];
    ostrstream     strm( strmbuf, sizeof(strmbuf) );
    FitsKeyword*   keywordptr( 0 );
    
    //
    //  Parse the keywordlist and attempt to extract all the info we
    //  need. Note: we assume the 'axisnr' that's being passed as an
    //  argument is zero based. In the FITSfile, it is one-based!
    //


    //
    //  Locate the NAXISn keyword!
    //
    strm.seekp( 0 );
    strm << "MAXIS" << itsAxisnumber+1 << ends;
    
    if( (keywordptr=kwsource(strm.str()))==0 )
    {
        //
        //  No MAXISn keyword!!!
        //
        ostrstream  error;
        
        error << "Rats! No MAXIS" << itsAxisnumber+1 << " keyword found. Aborting!";
        
        throw( AipsError(error.str()) );
    }

    //
    //  Fill in the length of the axis and the comment!
    //
    itsAxislength=keywordptr->asInt();
    itsComment=keywordptr->comm();

    //
    //  Delete the keyword!
    //
    kwsource.del();

    //
    //  Locate the axistype....
    //
    if( (keywordptr=kwsource(FITS::CTYPE, itsAxisnumber+1))!=0 )
    {
        itsAxistype=keywordptr->asString();

        //
        //  Delete!
        //
        kwsource.del();
    }

    
    //
    //  The axis-increment
    //
    if( (keywordptr=kwsource(FITS::CDELT, itsAxisnumber+1))!=0 )
    {
        itsIncrement=keywordptr->asDouble();
        
        kwsource.del();
    }

    //
    //  The reference pixel!
    //
    if( (keywordptr=kwsource(FITS::CRPIX, itsAxisnumber+1))!=0 )
    {
        itsReferencepixel=keywordptr->asDouble();
        
        kwsource.del();
    }
    
    //
    // The refernce value
    //
    if( (keywordptr=kwsource(FITS::CRVAL, itsAxisnumber+1))!=0 )
    {
        itsReferencevalue=keywordptr->asDouble();
        
        kwsource.del();
    }

    //
    // Rotation angle
    //
    if( (keywordptr=kwsource(FITS::CROTA, itsAxisnumber+1))!=0 )
    {
        itsRotationangle=keywordptr->asDouble();
        
        kwsource.del();
    }

    //
    //  e-thi.. e-thi.. e-that's all folks!
    //
}

//
//
//  The "copy"-c'tor
//
//
FITSMatrixAxis::FITSMatrixAxis( Int axisnr, const FITSMatrixAxis& other,
				FITSMatrixColumn* ptr2parent ) :
    BasicFITSUtility( (FITSBinaryColumn* const)ptr2parent ),
    itsAxistype( other.itsAxistype ),
    itsComment( other.itsComment ),
    itsAxislength( other.itsAxislength ),
    itsReferencepixel( other.itsReferencepixel ),
    itsReferencevalue( other.itsReferencevalue ),
    itsIncrement( other.itsIncrement ),
    itsRotationangle( other.itsRotationangle ),
    itsAxisnumber( axisnr )    
{
    //
    //  Nothing to do, all's done that needs to be done
    //
}


//
//  Translate this axis into a FitsKeywordList!
//
FitsKeywordList FITSMatrixAxis::asFITSKeys( void ) const
{
    char              buf[ 64 ];
    ostrstream        strm( buf, sizeof(buf) );
    FitsKeywordList   retval;
    
    //
    //  Ok. here we go....
    //

    //
    //  Only if the axis has a length other than 0 (which means a real
    //  degenerate), we need to add the description to the list of
    //  keywords
    //
    if( itsAxislength!=0 )
    {
        retval.mk( itsAxisnumber+1, FITS::CTYPE, itsAxistype );
        retval.mk( itsAxisnumber+1, FITS::CDELT, itsIncrement );
        retval.mk( itsAxisnumber+1, FITS::CRPIX, itsReferencepixel );
        retval.mk( itsAxisnumber+1, FITS::CRVAL, itsReferencevalue );
        retval.mk( itsAxisnumber+1, FITS::CROTA, itsRotationangle );
    }
    return retval;
}

Int FITSMatrixAxis::getAxisLength( void ) const
{
    return itsAxislength;
}

void FITSMatrixAxis::setAxisLength( Int newlength )
{
    itsAxislength=newlength;
}

//
//  Return internal values!
//
Int FITSMatrixAxis::axisnr( void ) const
{
    return itsAxisnumber;
}


Int FITSMatrixAxis::naxis( void ) const
{
    return itsAxislength;
}

Double FITSMatrixAxis::crpix( void ) const
{
    return itsReferencepixel;
}

Double FITSMatrixAxis::crval( void ) const
{
    return itsReferencevalue;
}

Double FITSMatrixAxis::cdelt( void ) const
{
    return itsIncrement;
}

Double FITSMatrixAxis::crota( void ) const
{
    return itsRotationangle;
}

const String& FITSMatrixAxis::ctype( void ) const
{
    return itsAxistype;
}


//
//  Destruct this axis!
//
FITSMatrixAxis::~FITSMatrixAxis()
{
}

