//
//  implementation of the matrix column class
//
//
//  Author:  Harro Verkouter, 13-09-1999
//
//  $Id: FITSMatrixColumn.cc,v 1.1.1.1 2000/03/20 15:16:48 verkout Exp $
//
//  $Log: FITSMatrixColumn.cc,v $
//  Revision 1.1.1.1  2000/03/20 15:16:48  verkout
//  HV: Imported aips++ implement-stuff
//
//
#ifndef FITSMATRIXCOLUMN_H
#include <vlbi/ms2uvfitsimpl/FITSUtils/FITSMatrixColumn.h>
#endif

#ifndef FITSMATRIXAXIS_H
#include <vlbi/ms2uvfitsimpl/FITSUtils/FITSMatrixAxis.h>
#endif

#ifndef FREEFUNCTIONS_H_INC
#include <vlbi/ms2uvfitsimpl/FITSUtils/freefunctions.h>
#endif

#ifndef AIPS_DATATYPE_H
#include <aips/Utilities/DataType.h>
#endif

#include <strstream.h>


FITSMatrixColumn::FITSMatrixColumn( const String& type, DataType tp, const String& unit,
				    const String& comment ) :
    FITSBinaryColumn( type, tp, IPosition(1,0), unit, comment ),
    myDataType( asScalar(tp) )
{
    this->setMatrixFlag();
}



FITSMatrixColumn::FITSMatrixColumn( Int number, uLong offset, FitsKeywordList& kwsource, 
                                    FITSBinaryTableHDU* ptr2parent ) :
    FITSBinaryColumn( number, offset, kwsource, ptr2parent ),
    myDataType( TpOther )
{
    //
    //  Check format and init acts upon the tform-keyword!
    //
    CheckFormatAndInit();

    //
    //  Tell the binary colum that it really is a matrix column
    //
    this->setMatrixFlag();

    //
    //  Since this is the matrix inary colum c'tor, we may have to filter
    //  out the TDIMn keyword!
    //
    FitsKeyword*   kwptr;
    
    kwptr=kwsource( FITS::TDIM, this->columnnr()+1 );
    if( kwptr )
    {
        //
        //  Read the IPosition from the keyword!
        //
        Char         locbuff[512];
        IPosition    shape;
        RecordDesc   coldesc;
        strstream    strm( locbuff, 512 );
        
        strm << kwptr->asString() << ends;
        
        strm >> shape;
        
        coldesc.addField( this->ttype(), this->getColumnDataType(), shape );
        
        setColumnDescription( coldesc );

        kwsource.del();
    }
    
    //
    //  And... read all the axes descriptions!
    //
    if( (kwptr=kwsource("MAXIS"))==0 )
    {
	throw( AipsError("Attempt to create MATRIX-column but the MAXIS keyword was not found!") );
    }
    
    Int        nraxes;

    nraxes = kwptr->asInt();
    
    kwsource.del();
    
    //
    //  Ok. Loop over all axes!
    //
    for( Int i=0; i<nraxes; i++ )
    {
	FITSMatrixAxis*   newaxis;
	
	newaxis = new FITSMatrixAxis( i, kwsource, this );
	myAxes.add( newaxis );
    }

    //
    // Ok. Done!
    //
}



FITSMatrixColumn::FITSMatrixColumn( Int number, uLong offset, const FITSMatrixColumn& other,
				    FITSBinaryTableHDU* ptr2parent ) :
    FITSBinaryColumn( number, offset, (const FITSBinaryColumn&)other, ptr2parent ),
    myDataType( other.myDataType )
{
    //
    //  Tell the BinaryColumn object within ourselves that we're a
    //  matrix column
    //
    this->setMatrixFlag();

    //
    //  Copy the axes across...
    //
    for( uInt i=0; i<other.myAxes.elements(); i++ )
    {
	myAxes.add( new FITSMatrixAxis(i, *(other.myAxes(i)), this) );
    }

    //
    //  Ok.. done!
    //
}


uInt FITSMatrixColumn::nrAxes( void ) const
{
    return myAxes.elements();
}

const FITSMatrixAxis* FITSMatrixColumn::getAxis( uInt axnr ) const
{
    const FITSMatrixAxis*   retval( 0 );
    
    if( axnr<myAxes.elements() )
    {
	retval = myAxes(axnr);
    }
    return retval;
}





Bool FITSMatrixColumn::addAxis( const FITSMatrixAxis& axis )
{
    if( this->hasParentHDU() )
    {
    	return False;
    }
    
    //
    //  Add the axis into ourselves, re-generate the tform string and
    //  re-set the stuff!
    //
    Int               curaxisnr;
    FITSMatrixAxis*   newaxisptr( 0 );
    
    //
    //  Step one: in order to determine the correct number for this
    //  axis, we first find out how many axes there already are.
    //
    curaxisnr = (Int)myAxes.elements();
    
    //
    //  Now we know that, we can safely create the new axisobject with
    //  the proper settings
    //
    newaxisptr= new FITSMatrixAxis( curaxisnr, axis, this );

    //
    //  Make room for this one...
    //
    myAxes.add( newaxisptr );
    
    //
    //  Update our internals...
    //
    char       formbuf[ 128 ];
    IPosition  curshape( this->shape() );
    ostrstream formstrm( formbuf, sizeof(formbuf) );
    
    formstrm.seekp( 0 );
    formstrm << curshape.product();

    switch( myDataType )
    {
	case TpChar:
	case TpUChar:
	    formstrm << "A";
	    break;
	    
        case TpBool:
            formstrm << "L";
            break;

        case TpShort:
        case TpUShort:
            formstrm << "I";
            break;
            
        case TpInt:
        case TpUInt:
            formstrm << "J";
            break;
            
        case TpFloat:
            formstrm << "E";
            break;
            
        case TpDouble:
            formstrm << "D";
            break;
            
        case TpComplex:
            formstrm << "C";
            break;
            
        case TpDComplex:
            formstrm << "M";
            break;
            
        default:
            throw( AipsError("FITSMatrixColumn: datatype not xpressible in FITS") );
            break;
    }
    formstrm << ends;

    //
    //  Update interna; structure of the column object
    //
    this->setForm( formstrm.str() );    
    this->CheckFormatAndInit();

    //
    //  Return success...
    //
    return True;
}



//
//  Return the shape of the matrix in this column
//
IPosition FITSMatrixColumn::shape( void ) const
{
    IPosition    retval( myAxes.elements() );

    for( uInt i=0; i<myAxes.elements(); i++ )
    {
	retval(i) = myAxes(i)->getAxisLength();
    }
    return retval;
}



//
//  Ok. Translate ourselves into keywords....
//
FitsKeywordList FITSMatrixColumn::asFITSKeys( void ) const
{
    FitsKeywordList     retval( this->FITSBinaryColumn::asFITSKeys() );
    
    //
    //  First, the ordinary column keywords, let the BinaryColumn
    //  object do that for us (already done by the time we end up
    //  here...)
    //
    //  Ok. Now add the TDIM keyword to indicate that this is a matrix
    //  column...
    //
    IPosition   ip( this->shape() );
    
    if( ip.nelements()>0 )
    {
        //
        //  Array with at least one domension
        //
	uInt        nel( ip.nelements() );
        char        buf[512];
        ostrstream  strm( buf, 512 );

        strm << "(";

	for( uInt i=0; i<nel; i++ )
	{
	    strm << ip(i);
	    if( i<(nel-1) )
	    {
		strm << ",";
	    }
	}
	strm << ")" << ends;
        
	cout << strm.str() << endl;
	
        retval.mk( this->columnnr()+1, FITS::TDIM, strm.str() );
    }

    //
    //  Add our own stuff: the axes descriptions...
    //
    for( uInt i=0; i<myAxes.elements(); i++ )
    {
	FitsKeyword*        kwptr;
	FitsKeyword*        newkw;
	FitsKeywordList     tmp( myAxes(i)->asFITSKeys() );
	
	tmp.first();
	tmp.next();
	kwptr = tmp.curr();
	
	while( kwptr!=0 )
	{
	    newkw = new FitsKeyword( *kwptr );
	    retval.insert( *newkw );
	    
	    kwptr = tmp.next();
	}	
    }
    return retval;
}
					      

FITSMatrixColumn::~FITSMatrixColumn()
{
    //
    //  Delete all the axes...
    //
    for( uInt i=0; i<myAxes.elements(); i++ )
    {
	delete myAxes(i);
    }
}

