//
//  Class:
//
//        FITSParameter
//
//
//  Purpose:
//
//        This class wil describe the basic properties of a parameter
//        in the description of a random group structure in a
//        FITS-file. The class is not templated, but when adding it to
//        a random group hdu, the datatypes must match.
//
//        
//
//
//  Author:
//
//        Harro Verkouter     24-3-1998
//
//
//      $Id: FITSParameter.h,v 1.1.1.1 2000/03/20 15:16:48 verkout Exp $
//
#ifndef FITSPARAMETER_H_INC
#define FITSPARAMETER_H_INC


#ifndef BASICFITSUTILITY_H_INC
#include <vlbi/ms2uvfitsimpl/FITSUtils/BasicFITSUtility.h>
#endif

#ifndef AIPS_AIPS_H
#include <aips/aips.h>
#endif

#ifndef AIPS_STRING_H
#include <aips/Utilities/String.h>
#endif

#ifndef AIPS_FITS
#include <aips/FITS/fits.h>
#endif

#ifndef AIPS_RECORDINTERFACE_H
#include <aips/Containers/RecordInterface.h>
#endif

#ifndef AIPS_RECORDFIELDID_H
#include <aips/Containers/RecordFiledId.h>
#endif


//
//  Forward declarations
//
class RecordFieldId;
class FITSPrimaryGroup;



class FITSParameter :
    public BasicFITSUtility
{
public:
    //
    //  Three constructors.
    //
    //  1) Create a complete parameter -> user has control over all
    //  parameter properties (but the parameter number) AND THE
    //  PARAMETER TYPE. So if you've just added a parameter to a
    //  PrimaryGroup and you want to get/put a/some values, get a
    //  reference to the parameter from the PrimaryGroup HDU rather
    //  than the object you used to add to the HDU! This is done to
    //  ensure that the parameter type of the parameters in the
    //  PrimaryGroup HDU match with the type of the HDU.
    //
    //  2) Create a parameter from a FitsKeywordlist -> intended use:
    //  attempt to read the parameters settings from the list of
    //  keywords just read from file!
    //
    //  3) A kind of copy-constructor -> the parameter settings of
    //  'other' are copied across apart from the parameternumber. This
    //  constructor finds its use when a HeaderDataUnit needs to add a
    //  parameter. The user specifies a parameter that needs to be
    //  added to the HeaderdataUnit. The user does not know the actual
    //  number of the parameter. The HeaderdataUnit does know and uses
    //  this c'tor to create the correct parameter.
    //
    FITSParameter( const String& type, const String& comment="", 
                   Double pscale=1.0, Double pzero=0.0 );
    
    //
    //  In the following two c'tors, the parameter data type is taken
    //  from the parent!
    //
    FITSParameter( Int number, FitsKeywordList& kwsource, FITSPrimaryGroup* const ptr2parent );
    
    FITSParameter( Int number, const FITSParameter& other, FITSPrimaryGroup* const ptr2parent );
    
    
    //
    //  Translate this parameter to Fitskeywords
    //
    virtual FitsKeywordList asFITSKeys( void ) const;

    //
    //  Get the parameter value for this parameter for random group
    //  #idx. 
    //
    //  NOTE: the result of the get-action will be stored in the first
    //  field of the record. In order to achieve that, the record will
    //  be first restructured to hold axactly one field of the
    //  internal datatype!
    //
    //  NOTE: when the record re-structuring does not succeed, the
    //  exception is caught within the function. A value of False is
    //  returned to the user.
    //
    Bool         get( uLong idx, RecordInterface& riref, const RecordFieldId& fieldidx=0 ) const;

    Bool         put( uLong idx, const RecordInterface& riref, const RecordFieldId& fieldidx=0 );


    //
    //  Get access to some of the properties of the parameter, READ
    //  ONLY and fits nomenclature
    //
    String       ptype( void ) const;
    Double       pscal( void ) const;
    Double       pzero( void ) const;
    

    //
    //  Destruct this parameter
    //
    virtual ~FITSParameter();

private:
    //
    //  These are the variables that describe the parameter
    //

    //
    //  The parameter type
    //
    String         itsParametertype;
    
    //
    //  The comment
    //
    String         itsComment;
    
    //
    //  If appropriate: which parameternumber is this?
    //
    Int            itsParameternumber;

    Double         itsScale;
    
    Double         itsZeropoint;

    //
    // Save the current basic fits type in here..., describes what
    // format the parameter should take.
    //
    DataType       itsDataType;


    //
    //  The parameter's private methods
    //

    //
    //  We do not allow copying and assigning to parameters, nor do we
    //  allow default parameters!
    //
    FITSParameter();
    FITSParameter( const FITSParameter& );
    FITSParameter& operator=( const FITSParameter& );
    
};


#endif
