//
//  Class:
//
//          FITSPrimaryGroup
//
//  Purpose:
//
//          This class describes a PrimaryGroup in a FITS file!
//
//
//  Author:
//
//          Harro Verkouter,   26-05-1998
//
//
//     $Id: FITSPrimaryGroup.h,v 1.1.1.1 2000/03/20 15:16:48 verkout Exp $
//
#ifndef FITSPRIMARYGROUP_H_INC
#define FITSPRIMARYGROUP_H_INC


//
//  We need this include since it is our baseclass!
//
#ifndef FITSARRAYHDU_H_INC
#include <vlbi/ms2uvfitsimpl/FITSUtils/FITSArrayHDU.h>
#endif

//
//  Also, vector is needed
//
#ifndef AIPS_VECTOR_H
#include <aips/Arrays/Vector.h>
#endif

//
//  Same goes for FITSParameter
//
#ifndef FITSPARAMETER_H_INC
#include <vlbi/ms2uvfitsimpl/FITSUtils/FITSParameter.h>
#endif

//
// Forward declarations
//
class RecordInterface;


//
//  The primary group hdu class. It isA FITSArrayHDU, therefore we
//  derive it from that class
//
class FITSPrimaryGroup :
    public FITSArrayHDU
{
    //
    //  The get/put functions of the FITSParameter class are friends
    //  of ours!
    //
    friend Bool FITSParameter::put( uLong idx, const RecordInterface& riref,
                                    const RecordFieldId& fieldidx );
    friend Bool FITSParameter::get( uLong idx, RecordInterface& riref,
                                    const RecordFieldId& fieldidx ) const;

public:
    //
    //  The public interface
    //

    //
    //  Create a primary group of a given type (basic fits type: the
    //  type of data items),
    //
    FITSPrimaryGroup( DataType basicfitstype );

    //
    //  Create the primary array from a fitskeywordlist, that's just
    //  been read from an existing fitsfile represented by
    //  'ptr2stream'
    //
    FITSPrimaryGroup( FitsKeywordList& kwlist, ByteIO* ptr2stream );

    //
    //  Primary groups have the possibility of adding parameters! The
    //  return value is the parameter number that was assigned to the
    //  parameter. If the return value is <0, the parameter was NOT
    //  added.
    //
    virtual Int       addParameter( const FITSParameter& parameter );

    //
    //  Set the number of groups (if possible)
    //
    Bool              setNumberOfGroups( uLong nrofgroups );

    //
    //  The type of this hdu
    //
    virtual HDUObject::hdu_t getHDUtype( void ) const;

    //
    // Define get/put actions on the primary group!
    //
    // Note: these functions are NOT defined on the baseclass, and
    // they are virtual in THIS class -> derived classes from this
    // object may overload the get/put actions.
    //
    virtual Bool      put( uLong group, const RecordInterface& recordref );
    
    virtual Bool      get( uLong group, RecordInterface& recordref ) const;

    //
    //  We have to implement our own version of the
    //  'getCurrentDataVolumeInBytes()' method, since there are two
    //  points we have to take into account:
    //
    //  1) axis 0 has length 0 -> product of all axes=0 which is
    //  usually quite wrong...
    //
    //  2) We have a number of groups, preceded by a number of
    //  parameters -> we have to take the parameters in account too.
    //
    virtual uLong     getCurrentDataVolumeInBytes( void ) const;

    //
    //  Users might be interested in the following info
    //

    //
    //  The shape of a group in the random groups structure
    //
    IPosition         groupShape( void ) const;
    uLong             getGroupSizeInBytes( void ) const;
    uLong             getNumberOfGroups( void ) const;
    RecordDesc        getGroupDescription( void ) const;

    //
    //  Check if the given parametr exists in this HDU. The number of
    //  occurrences is returned.
    //
    uInt              parameterExists( const String& parametertype ) const;
    uInt              getNumberOfParameters( void ) const;
    
    //
    //  Access parameters (by name or by number). Note: if the
    //  adressed parameter cannot be found, an aipsError exception is
    //  thrown!
    //
    FITSParameter&    getParameter( const String& parametername, uInt occurrence=0 ) const;
    FITSParameter&    getParameter( uInt parameternumber ) const;
    

    //
    //  Destruct the primary array
    //
    virtual ~FITSPrimaryGroup();
    
protected:
    //
    //  Protected functions
    //
    virtual FitsKeywordList   getDataDescriptionKeywords( void ) const;
    
private:
    //
    //  The primary groups private parts
    //
    
    //
    //  The number of groups represented by this HDU
    //
    uLong                   itsNumberOfGroups;
    
    //
    //  Primary groups have parameters! Keep a vector of pointers to
    //  the parameters
    //
    Vector<FITSParameter*>  itsParameterptrs;
    
    //
    //  The private methods
    //

    //
    //  These are undefined and inaccessible
    //
    FITSPrimaryGroup();
    FITSPrimaryGroup( const FITSPrimaryGroup& );
    FITSPrimaryGroup& operator=( const FITSPrimaryGroup& );
};



#endif
