// class:
//
//     MSConversionInfo
//
//
// purpose:
//
//
//     Set up conversion parameters for the transformation from MS to
//     UVFITS file. The UVFITS-'standard' tells us that there are a
//     number of 'global' FITS-keywords in the UVFITS file (like
//     e.g. OBSCODE, NO_STKD, NO_BAND...). This class will, upon
//     creation attempt to get this info from the MS. Other
//     information may also be nice to know, like time, units etc. All
//     this info has to be located once.
//
//     The MSConversionInfo object will attempt to open the MS for
//     you: you won't get a pointer to the MS! The object will delete
//     any resources itself! You will have access to the MS via a
//     reference: you can't delete a reference!
//
//
// Note:
//     If the MSConversionInfo object cannot open the MS, it will
//     throw an exception!
//
//
// Use:
//
//          MSConversionInfo     info( "DataSets/Measurementetname" );
//
//          cout << "Number of parizations: " << info.getNumPolarizations() << endl;
//
//
//
//
//
//      Author:   Harro Verkouter,   12-3-1998
//     
//
//       $Id: MSConversionInfo.h,v 1.3 2000/08/24 14:33:15 verkout Exp $
//
#ifndef MSCONVERSIONINFO_H_INC
#define MSCONVERSIONINFO_H_INC

//
//  Includes
//
#ifndef AIPS_AIPS_H
#include <aips/aips.h>
#endif

#ifndef AIPS_STRING_H
#include <aips/Utilities/String.h>
#endif

#ifndef AIPS_MVTIME_H
#include <aips/Quanta/MVTime.h>
#endif


//
//  Not exactly sure why I put this function here, but this header file is
//  included in all generator files so maybe that's the easiest way to get
//  this function everyewhere available
//
//  It rounds the double to the *nearest* integer:
//
//      -6.9999 will be rounded to -7.0
//      34.78   will be rounded to 35.0
//      -4.45   will be rounded to -4.0
//
Double round2int( Double r);



//
//  Forward declarations
//
class MeasurementSet;
class ROMSColumns;
class MSColumns;

//
//  The class declaration
//
class MSConversionInfo
{
public:
	//	
	// Enum for reference frequency specification: tell the system
	// if you want the EDGE or the CENTER of the band...
	//
	typedef enum _reference 
	{
		CenterBand, CenterChannel, Edge
	} reference;



    static uInt   lastVisibility;

    //
    //  Only allowed ctor is with name of MS. The MS will be opened,
    //  the ROMSColumns object will be initialized so as it will
    //  reference the opened MS. Further more, any 'global' info will
    //  be gathered.
    //
    MSConversionInfo( const String& msname );

    //
    //  Attaempt to attach to an already opened MS (used when
    //  splitting a MS -> usr. opens the MS, creates subtables, for
    //  each of which a MSConversionInfo object has to be created)
    //
    MSConversionInfo( MeasurementSet& msref, uInt thispiece, uInt maxpiece );
    
    //
    //  retrieve the original name of the MS
    //
    const String& getMSname( void ) const;

    //
    //  Grant access to the MS, so others can access the tables and
    //  whatsoever.
    //
    const MeasurementSet&  getMS( void ) const;

    //
    // Grant RO-access to the columns.
    //
    ROMSColumns&  getROColumns( void ) const;

    //
    //  At creation time, a lot of info can be gathered, like the nr
    //  of polarizations and/or the number of channels in this
    //  observation.
    //
    //  Provide acces to these cached items!
    //
    Int           getFirstStokes( void ) const;
    uInt          nrPolarizations( void ) const;
    uInt          nrChannels( void ) const;
    uInt          nrIFs( void ) const;
    
    //
    // Frequency info
    //
    Double        getReferenceFrequency( MSConversionInfo::reference r ) const;
    Double        getTotalBandwidth( void ) const;
    Double        getChannelwidth( void ) const;
    Double        getChannelincrement( void ) const;

    //
    //  The name of the DATA-column?
    //
    const String& getDataColumnName( void ) const;

    //
    //  Derived from value in first cell of time-column!
    //
    const MVTime& observingDate( void ) const;

    //
    //  The obscode=keyword values
    //
    const String& getObsCode( void ) const;
    
    //
    //  Multi-part fits file info
    //
    uInt         itsPieceNr( void ) const;
    uInt         maxPieceNr( void ) const;

    //
    //  Destruct the info object
    //
    ~MSConversionInfo();
    
private:
    //
    //  The private data section
    //
    String            itsMSname;

    //
    //  The pointer to the measurementset to which the info object
    //  applies
    //
    MeasurementSet*   itsMSptr;

    //
    //  The pointer to the RO Column object: the translators can only
    //  read from the columns!
    //
    ROMSColumns*      itsROColumnsptr;
    
    //
    // Number of polarizations/channels
    //
    uInt              itsNumberOfPolarizations;
    uInt              itsNumberOfChannels;
    uInt              itsNumberOfIFs;
    
    //
    //  Cached frequency info
    //
    Double            itsReferenceFrequency;
    Double            itsTotalBandwidth;
    Double            itsChannelincrement;

    //
    //  The observing date
    //
    MVTime            itsObservingDate;
    
    //
    //  The deduced columnname
    //
    String            itsDataColumnName;

    //
    //  Possibility to indicate which piece out of max nr of pieces
    //  tis set represents...
    //
    uInt              myPieceNr;
    uInt              nrPieces;

    String            myObsCode;
    Int               myFirstStokes;
    
    //
    //  The private method section
    //


    //
    //  Attempt to attach a MS. Throws excpetions if it fails...
    //
    Bool              attachMS( MeasurementSet& msref );

    
    //
    //  Forbid copying, assignment and default ctor!
    //
    MSConversionInfo();
    MSConversionInfo( const MSConversionInfo& );
    MSConversionInfo& operator=( const MSConversionInfo& );
};


#ifdef MSCONVERSIONINFO_CC

uInt MSConversionInfo::lastVisibility = 0;

#endif


#endif
