//
//  Implementation of the Container
//
#ifndef CONTAINER_H
#include <vlbi/tmplsrc/Container.h>
#endif

#ifndef AIPS_AIPS_H
#include <aips/aips.h>
#endif

#ifndef AIPS_EXCEPTIONS_H
#include <aips/Exceptions.h>
#endif


//
//  Global functions
//
template <class T>
ostream& operator<<( ostream& os, const Container<T>& vref )
{
    uint32    nr2show( vref.elements() );
    
    os << "[";
    
    for( uint32 i=0; i<nr2show; i++ )
    {
        os << vref(i);
        if( i<(nr2show-1) )
        {
            os << ",";
        }
    }
    os << "]";
    return os;
}

template <class T>
uint32 frequencyOfArg( const Container<T>& container, const T& arg )
{
    uint32    retval( 0 );
    
    for( uint32 cnt=0; cnt<container.elements(); cnt++ )
    {
	if( container(cnt)==arg )
	{
	    retval++;
	}
    }
    return retval;
}


//
//
//   The container class itself
//
//
template <class T>
Container<T>::Container() :
    itsStorage( 0 ),
    itsNrelements( 0 ),
    itsNrallocated( 0 )
{
    
}


template <class T>
Container<T>::Container( uint32 initiallength, const T& initialval ) :
    itsStorage( 0 ),
    itsNrelements( initiallength ),
    itsNrallocated( 0 )
{
    //
    //  Make storage and copy elements across!
    //
    itsStorage = allocateStorage( itsNrelements, itsNrallocated );
    
    //
    //  Set elements to initial value
    //
    setElements( itsStorage, initialval, itsNrelements );
}


template <class T>
Container<T>::Container( const Container<T>& other ) :
    itsStorage( 0 ),
    itsNrelements( other.itsNrelements ),
    itsNrallocated( 0 )
{
    //
    //  Make a copy of other. So 1) allocate storage b) copy values
    //  across
    //
    itsStorage = allocateStorage( itsNrelements, itsNrallocated );
    
    copyElements( itsStorage, other.itsStorage, itsNrelements );
}


template <class T>
void Container<T>::resize( uint32 newlength, const T& initval )
{
    //
    //  Now see what we have to do. If the new size is smaller, we
    //  don't have to do anything, apart from adjusting our internal
    //  length!
    //
    if( newlength>itsNrelements )
    {
        if( newlength>itsNrallocated )
        {
            //
            // Now we have something to do! Allocate a new block!
            //
            T*      newstorage;
            uint32  newnrallocated;
            
            newstorage = allocateStorage( newlength, newnrallocated );

            //
            //  Copy previous values across
            //
            copyElements( newstorage, itsStorage, itsNrelements );
            
            //
            //  Don't need the old storage anymore
            //
            delete [] itsStorage;
            
            itsStorage=newstorage;
            itsNrallocated=newnrallocated;
        }
        
        //
        //  Initialize newly added elements
        //
        setElements( itsStorage+itsNrelements, initval, newlength-itsNrelements );
    }
    //
    //  This instruction has to be done always!
    //
    itsNrelements=newlength;
}

template <class T>
void Container<T>::add( const T& newval )
{
    if( (itsNrelements+1)>=itsNrallocated )
    {
        //
        //  Allocate a new block!
        //
        T*     newstorage;
        uint32 newnrallocated;
        
        newstorage = allocateStorage( itsNrallocated, newnrallocated );
        copyElements( newstorage, itsStorage, itsNrelements );

        delete [] itsStorage;
        
        itsStorage=newstorage;
        itsNrallocated=newnrallocated;
    }

    itsStorage[itsNrelements]=newval;
    itsNrelements++;
}

//
//  Access elements
//
template <class T>
T& Container<T>::operator[]( uint32 idx )
{
    if( idx<itsNrelements )
    {
        return itsStorage[idx];
    }
    throw( AipsError("Container: Index out of range") );
    return *( new T());
}

template <class T>
T& Container<T>::operator()( uint32 idx )
{
    return itsStorage[idx];
}

template <class T>
const T& Container<T>::operator[]( uint32 idx ) const
{
    if( idx<itsNrelements )
    {
        return itsStorage[idx];
    }
    throw( AipsError("Container: Index out of range") );
    return *( new T());
}

template <class T>
const T& Container<T>::operator()( uint32 idx ) const
{
    return itsStorage[idx];
}

template <class T>
uint32 Container<T>::elements( void ) const
{
    return itsNrelements;
}

//
//  Access the storage
//
template <class T>
T* Container<T>::getstorage( void )
{
  return itsStorage;
}

template <class T>
const T* Container<T>::getstorage( void ) const
{
  return itsStorage;
}

//
//  Assignement!
//
template <class T>
Container<T>& Container<T>::operator=( const Container<T>& rval )
{
    if( this!=&rval )
    {
        //
        //  We must replace our contents with thos of rval!
        //
        delete [] itsStorage;
        
        itsNrelements=rval.itsNrelements;
        
        itsStorage = allocateStorage( itsNrelements, itsNrallocated );
        
        copyElements( itsStorage, rval.itsStorage, itsNrelements );
    }
    return *this;
}

template <class T>
Container<T>::~Container()
{
    delete [] itsStorage;
    itsNrelements=0;
    itsNrallocated=0;
}


//
//  Some private methods that we need to implement
//
template <class T>
void Container<T>::copyElements( T* dst, const T* src, uint32 nr2copy )
{
    if( !dst || !src || !nr2copy )
    {
        return;
    }
    
    for( uint32 i=0; i<nr2copy; i++ )
    {
        dst[i]=src[i];
    }
}

template <class T>
void Container<T>::setElements( T* dst, const T& val2set, uint32 nr2set )
{
    if( !dst || !nr2set )
    {
        return;
    }
    
    for( uint32 i=0; i<nr2set; i++ )
    {
        dst[i]=val2set;
    }
}

template <class T>
T* Container<T>::allocateStorage( const uint32 nr2allocate, uint32& reallyallocated )
{
    T*    retval;

    //
    //  Depending on the size of T, we allocate extra elements; if the
    //  type is very big, 10 elements extra may be too expensive to
    //  allocate! Here we lay the turning point at 1024 bytes.
    //
    if( sizeof(T)>1024 )
    {
	reallyallocated = nr2allocate+1;
    }
    else
    {
	reallyallocated = nr2allocate+10;
    }
    
    retval = new T[ reallyallocated ];
    return retval;
}

