//
//  class:
//
//         Container <>
//
//  purpose:
//
//         a templated lightweigth, slightly optimized storage container which
//         can be indexed linearly
//
//         NOTE:
//
//         When doing the template instantiations yourself, be sure to
//         instantiate the following templates to ensure smoot
//         operation:
//
//         template class Container<[your type]>;
//         template ostream& operator<<( ostream&, const Container<[your type]>& );
//
//
//
//  Author:
//
//         Harro Verkouter
//
//
//      $Id: Container.h,v 1.1.1.1 2000/03/20 15:15:58 verkout Exp $
//
#ifndef CONTAINER_H
#define CONTAINER_H

#ifndef HVTYPES_H
#include <vlbi/DtdLUT/hvtypes.h>
#endif

//
//  Forward declarations
//
class ostream;


//
// The obvious templated container-class: a lightweight one!
//
template <class T>
class Container
{

public:
    
    //
    //  Default c'tor: a container of zero length
    //
    Container();

    //
    //  Provide storage for 'initiallength' elements of type T and set
    //  them initially to initialval
    //
    Container( uint32 initiallength, const T& initialval=T() );
    
    //
    //  The copy c'tor
    //
    Container( const Container<T>& other );

    //
    //  Resize the Container; as much elements as possible are copied.
    //
    //  NOTE: the initval  parameter is only used when enlarging the
    //  container; newly allocated elements are initialized to initval!
    //
    void resize( uint32 newlength, const T& initval=T() );

    //
    //  Add a COPY of the argument to the container!
    //
    void add( const T& newval );
    

    //
    //  Gain access to the members.
    //   The [] forms perform boundschecking
    //   The () forms do not perform boundschecking!
    //
    T& operator[]( uint32 idx );
    T& operator()( uint32 idx );
    
    const T& operator[]( uint32 idx ) const;
    const T& operator()( uint32 idx ) const;
    
    //
    //  Use at own risk: get direct access to the storage
    //
    T*       getstorage( void );
    const T* getstorage( void ) const;

    //
    //  Some general interface functions
    //
    //  The number of elements in this container
    //
    uint32   elements( void ) const;
    
    //
    //  Assignment
    //
    Container<T>& operator=( const Container<T>& rval );

    //
    //  The destructor... for all object stored in the container, the
    //  desctructor is called!
    //
    ~Container();
    
private:
    //
    //  Private members are the array of elements and the num,ber of
    //  elements
    //
    T*       itsStorage;
    
    uint32   itsNrelements;
    uint32   itsNrallocated;

    //
    //  Private methods...
    //
    static void   copyElements( T* dst, const T* src, uint32 nr2copy );
    static void   setElements( T* dst, const T& val2set, uint32 nr2set );

    static T*     allocateStorage( const uint32 nr2allocate, uint32& reallyallocated );
};

//
//  Nice to have a function that dumps a container on a stream
//
template <class T>
ostream& operator<<( ostream& os, const Container<T>& vref ); 

//
//  Count how often the argument occurs in the container
//
template <class T>
uint32  frequencyOfArg( const Container<T>& container, const T& arg );

//
//  Declare any statics in here
//
#ifdef CONTAINER_CC

#endif


#endif
