//
//  Implementation of the LinearContainer
//
#ifndef LINEARCONTAINER_H_INC
#include <vlbi/tmplsrc/LinearContainer.h>
#endif

#ifndef AIPS_EXCEPTIONS_H
#include <aips/Exceptions.h>
#endif

template <class T>
ostream& operator<<( ostream& os, const LinearContainer<T>& vref )
{
    uInt    nr2show( vref.elements() );
    
    os << "[";
    
    for( uInt i=0; i<nr2show; i++ )
    {
        os << vref(i);
        if( i<(nr2show-1) )
        {
            os << ",";
        }
    }
    os << "]";
    return os;
}


template <class T>
LinearContainer<T>::LinearContainer() :
    storage( 0 ),
    nrelements( 0 ),
    nrallocated( 0 )
{
    
}


template <class T>
LinearContainer<T>::LinearContainer( uInt initiallength, const T& initialval ) :
    storage( 0 ),
    nrelements( initiallength ),
    nrallocated( 0 )
{
    //
    //  Make storage and copy elements across!
    //
    storage = allocateStorage( nrelements, nrallocated );
    
    //
    //  Set elements to initial value
    //
    setElements( storage, initialval, nrelements );
}


template <class T>
LinearContainer<T>::LinearContainer( const LinearContainer<T>& other ) :
    storage( 0 ),
    nrelements( other.nrelements ),
    nrallocated( 0 )
{
    //
    //  Make a copy of other. So 1) allocate storage b) copy values
    //  across
    //
    storage = allocateStorage( nrelements, nrallocated );
    
    copyElements( storage, other.storage, nrelements );
}


template <class T>
void LinearContainer<T>::resize( uInt newlength, const T& initval )
{
    //
    //  Now see what we have to do. If the new size is smaller, we
    //  don't have to do anything, apart from adjusting our internal
    //  length!
    //
    if( newlength>nrelements )
    {
        if( newlength>nrallocated )
        {
            //
            // Now we have something to do! Allocate a new block!
            //
            T*    newstorage;
            uInt  newnrallocated;
            
            newstorage = allocateStorage( newlength, newnrallocated );

            //
            //  Copy previous values across
            //
            copyElements( newstorage, storage, nrelements );
            
            //
            //  Don't need the old storage anymore
            //
            delete [] storage;
            
            storage=newstorage;
            nrallocated=newnrallocated;
        }
        
        //
        //  Initialize newly added elements
        //
        setElements( storage+nrelements, initval, newlength-nrelements );
    }
    //
    //  This instruction has to be done always!
    //
    nrelements=newlength;
}

template <class T>
void LinearContainer<T>::add( const T& newval )
{
    if( (nrelements+1)>=nrallocated )
    {
        //
        //  Allocate a new block!
        //
        T*   newstorage;
        uInt newnrallocated;
        
        newstorage = allocateStorage( nrallocated, newnrallocated );
        copyElements( newstorage, storage, nrelements );

        delete [] storage;
        
        storage=newstorage;
        nrallocated=newnrallocated;
    }

    storage[nrelements]=newval;
    nrelements++;
}

//
//  Access elements
//
template <class T>
T& LinearContainer<T>::operator[]( uInt idx )
{
    if( idx<nrelements )
    {
        return storage[idx];
    }
    throw( AipsError("conatiner: Index out of range") );
    return *( new T());
}

template <class T>
T& LinearContainer<T>::operator()( uInt idx )
{
    return storage[idx];
}

template <class T>
const T& LinearContainer<T>::operator[]( uInt idx ) const
{
    if( idx<nrelements )
    {
        return storage[idx];
    }
    throw( AipsError("conatiner: Index out of range") );
    return *( new T());
}

template <class T>
const T& LinearContainer<T>::operator()( uInt idx ) const
{
    return storage[idx];
}

template <class T>
uInt LinearContainer<T>::elements( void ) const
{
    return nrelements;
}

//
//  Access the storage
//
template <class T>
T* LinearContainer<T>::getstorage( void )
{
  return storage;
}

template <class T>
const T* LinearContainer<T>::getstorage( void ) const
{
  return storage;
}

//
//  Assignement!
//
template <class T>
LinearContainer<T>& LinearContainer<T>::operator=( const LinearContainer<T>& rval )
{
    if( this!=&rval )
    {
        //
        //  We must replace our contents with thos of rval!
        //
        delete [] storage;
        
        nrelements=rval.nrelements;
        
        storage = allocateStorage( nrelements, nrallocated );
        
        copyElements( storage, rval.storage, nrelements );
    }
    return *this;
}

template <class T>
LinearContainer<T>::~LinearContainer()
{
    delete [] storage;
    nrelements=0;
    nrallocated=0;
}


//
//  Some private methods that we need to implement
//
template <class T>
void LinearContainer<T>::copyElements( T* dst, const T* src, uInt nr2copy )
{
    if( !dst || !src || !nr2copy )
    {
        return;
    }
    
    for( uInt i=0; i<nr2copy; i++ )
    {
        dst[i]=src[i];
    }
}

template <class T>
void LinearContainer<T>::setElements( T* dst, const T& val2set, uInt nr2set )
{
    if( !dst || !nr2set )
    {
        return;
    }
    
    for( uInt i=0; i<nr2set; i++ )
    {
        dst[i]=val2set;
    }
}

template <class T>
T* LinearContainer<T>::allocateStorage( const uInt nr2allocate, uInt& reallyallocated )
{
    T*    retval;
    
    reallyallocated = nr2allocate+10;
    
    retval = new T[ reallyallocated ];
    return retval;
}

